/* Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without 
 * modifications, as long as this notice is preserved.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Author: Michael Schulz (see also AUTHORS and THANKS for more)
 * Birth: 12 Juli 1998
 */

/* -------------------------------------------------------------------------
 * @addtogroup HookBindings HookBindings - Adaptor to the callback system
 * @ingroup Hook
 * @{
 * \todo Please explain this module.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include "config.h"
#include "blah.h"
#include "hooker.h"
#include "netsearch.h"
#include "parse.h"
#include "swig_utils.h"

/* -- VARIABLES --------------------------------------------------------- */


/* --------------------------------------------------------------------------
 * pointer to the currently used tcl interpreter.
 * \todo This pointer isn't needed for newer swigs.
 */
Tcl_Interp *hkInterp;

/* --------------------------------------------------------------------------
 * the singelton Logger instance used in this module.
 */
static Logger logger;

/* --------------------------------------------------------------------------
 * typed hook result.
 * This pointer is used in our special typemaps.
 */
HookResult hkResult;

/* ---------------------------------------------------------------------------
 * to be called by every hook.
 */
static void initHookResult(Hook hook)
{
  hkResult->type = HTNone;
  hkResult->hook = hook;
  hkResult->data.intResult = 0;
  hkResult->data.stringResult = NULL;
}

/* ---------------------------------------------------------------------------
 * initialize the module when loading
 * this stores the tcl-hooks cdgHooks
 */
void hooker_init(Tcl_Interp *interp)
{
  Hook hook;
  int i;

  logger = (Logger) memMalloc(sizeof (LoggerStruct));
  logger->size = 0;
  logger->maxsize = 0;
  logger->buffer = NULL;
  hkInterp = interp;
  hkResult = (HookResult) memMalloc(sizeof (HookResultStruct));
  initHookResult(NULL);
  loggerSize(1024);

  for (i = 0; i < vectorSize(hkHooks); i++) {
    hook = (Hook) vectorElement(hkHooks, i);
    hook->cmd = NULL;
    hook->function = (HookFunction *) tclHookHandle;
  }

  hook = vectorElement(hkHooks, HOOK_PRINTF);
  hook->function = (HookFunction *) logPrintf;
  hook->cmd = strCopy("puts -nonewline");

  hook = vectorElement(hkHooks, HOOK_FLUSH);
  hook->function = (HookFunction *) logFlush;
  hook->cmd = NULL;

  hook = vectorElement(hkHooks, HOOK_EVAL);
  hook->function = (HookFunction *) evalHookHandle;
  hook->cmd = NULL;

  hook = vectorElement(hkHooks, HOOK_NSSEARCH);
  hook->function = (HookFunction *) netsearchHookHandle;
  hook->cmd = NULL;

  hook = vectorElement(hkHooks, HOOK_GLSINTERACTION);
  hook->function = (HookFunction *) glsInteractionHookHandle;
  hook->cmd = NULL;

  hook = vectorElement(hkHooks, HOOK_GETS);
  hook->function = (HookFunction *) getsHookHandle;
  hook->cmd = NULL;

  hook = vectorElement(hkHooks, HOOK_PROGRESS);
  hook->function = (HookFunction *) progressHookHandle;
  hook->cmd = NULL;

  hook = vectorElement(hkHooks, HOOK_PARTIALRESULT);
  hook->function = (HookFunction *) partialResultHookHandle;
  hook->cmd = NULL;

  hook = vectorElement(hkHooks, HOOK_ICINTERACTION);
  hook->function = (HookFunction *) ICinteractionHookHandle;
  hook->cmd = NULL;

  hook = vectorElement(hkHooks, HOOK_RESET);
  hook->function = (HookFunction *) tclHookHandle;
  hook->cmd = NULL;

}

/* ---------------------------------------------------------------------------
 * set the buffersize of the logger.
 * \param maxsize the size which to set the log buffer to.
 * \returns the old size of the log buffer.
 */
int loggerSize(int maxsize)
{
  int oldSize = logger->maxsize;

  if (maxsize < 2)
    maxsize = 2;

  logger->maxsize = maxsize;
  logger->buffer = (char *)memRealloc(logger->buffer, maxsize + 1);
  logger->buffer[logger->size] = 0;

  return oldSize;
}

/* ---------------------------------------------------------------------------
 * write a char to the logger.
 * This functions writes a single char and check if it is a specail char 
 * which should be escaped.
 */
TclResultType logWriteChar(char c)
{
#define DEBUG_WRITECHAR 0
  int n = 0;

  initHookResult(NULL);

  /* escape some substitution */
  if (c == '[' || c == ']' || c == '$' || c == '\\' || c == '"') {
    if (logger->size + 1 >= logger->maxsize) {
      if (logFlush() == TCL_ERROR) {
	/* logFlush generated the error message */
	return TCL_ERROR;
      }
    }
    logger->buffer[logger->size++] = '\\';
    n++;
  }
#if DEBUG_WRITECHAR
  putchar(c);
#endif
  logger->buffer[logger->size++] = c;
  logger->buffer[logger->size] = 0;

  if (logger->size >= logger->maxsize || c == '\n') {
    if (logFlush() == TCL_ERROR) {
      /* logFlush generated the error message */
      return TCL_ERROR;
    }
    n++;
  }

  hkResult->type = HTInt;
  hkResult->data.intResult = n;
  
  return TCL_OK;
}

/* ---------------------------------------------------------------------------
 * logWrite: writes some text into the logger
 */
TclResultType logWrite(char *text)
{
  int len = strlen(text);
  int n = 0;
  int i;

  for (i = 0; i < len; i++) {
    if (logWriteChar(text[i]) == TCL_ERROR) {
      /* logFlush generated the error message */
      return TCL_ERROR;
    }
    n += hkResult->data.intResult;
  }

  hkResult->type = HTInt;
  hkResult->data.intResult = n;
  
  return TCL_OK;
}

/* ---------------------------------------------------------------------------
 * setHookCmd: set the tcl-command of a named hook
 */
TclResultType setHookCmd(char *hookName, char *cmd)
{
  int hookNo;
  Hook hook;

  initHookResult(NULL);

  hookNo = hkFindNoOfHook(hookName);
  if (hookNo < 0) {
    Tcl_Obj *tclResult = Tcl_GetObjResult(hkInterp);
    Tcl_AppendStringsToObj(tclResult, "don't know hook `", hookName, "'", 
	(char *)NULL);
    return TCL_ERROR;
  }

  hook = (Hook) vectorElement(hkHooks, hookNo);

  if (strlen(cmd) == 0)
    hook->cmd = NULL;
  else
    hook->cmd = strCopy(cmd);

  return TCL_OK;
}

/* ---------------------------------------------------------------------------
 * getHookCmd: get the tcl-command of a named hook
 */
TclResultType getHookCmd(char *hookName)
{
  int hookNo;
  Hook hook;

  hookNo = hkFindNoOfHook(hookName);
  if (hookNo < 0) {
    Tcl_Obj *tclResult = Tcl_GetObjResult(hkInterp);
    Tcl_AppendStringsToObj(tclResult, "don't know hook `", hookName, "'", (char *)NULL);
    return TCL_ERROR;
  }

  hook = (Hook) vectorElement(hkHooks, hookNo);
  hkResult->hook = hook;
  hkResult->type = HTString;
  hkResult->data.stringResult = hook->cmd;

  return TCL_OK;
}

/* ---------------------------------------------------------------------------
 * logPrintf: print formated text into the logger
 */
TclResultType logPrintf(char *format, va_list ap)
{
  int result = logWrite(strVPrintf(format, ap));
  hkResult->hook = vectorElement(hkHooks, HOOK_PRINTF);
  return result;
}

/* ---------------------------------------------------------------------------
 * logFlush: flush the logwindow
 */
TclResultType logFlush()
{
  int result = TCL_OK;
  Hook hook;

  hook = vectorElement(hkHooks, HOOK_PRINTF);
  initHookResult(hook);
  if (hook && hook->cmd) {
    result = Tcl_VarEval(hkInterp, hook->cmd,
			 " \"", logger->buffer, "\" ", (char *)NULL);
  }

  if (result == TCL_OK) {
    hook = vectorElement(hkHooks, HOOK_FLUSH);
    initHookResult(hook);
    if (hook && hook->cmd) {
      result = Tcl_VarEval(hkInterp, hook->cmd, (char *)NULL);
    }
  }

  logger->buffer[0] = 0;
  logger->size = 0;

  if (result != TCL_OK) {
    hkResult->type = HTError;
  }

  return (result);
}

/* ---------------------------------------------------------------------------
 * netsearchHookHandle:
 *	arg1: NetSearchState
 *	arg3: mode 
 *	      1 - add rootnode
 *	      2 - solution node found
 *	      3 - add child
 *	      4 - add skipper
 *	      5 - widen searchspace
 *	      6 - closing open node
 */
TclResultType netsearchHookHandle(Hook hook, va_list ap)
{
  int result;
  int mode;
  Pointer p1;
  char b1[512], b2[512];

  if (!hook->cmd)
    return TCL_OK;

  initHookResult(hook);

  /* parse arguments */
  p1 = va_arg(ap, Pointer);
  mode = va_arg(ap, int);

  /* convert to ascii representation */
  SWIG_MakePtr(b1, p1, "_NetSearchState");
  sprintf(b2, "%d", mode);

  /* execute command */
  result = Tcl_VarEval(hkInterp, hook->cmd, " ", b1, " ", b2, (char *)NULL);

  /* exception handling */
  if (result != TCL_OK) {
    hkResult->type = HTError;
  }

  return result;
}

/* ---------------------------------------------------------------------------
 * evalHookHandle:
 *	called inside evalConstraint when a constraint fails
 *	the constraint can be unary or binary; levelvalues are still assigned
 *	arg1 : constraint
 */
TclResultType evalHookHandle(Hook hook, va_list ap)
{
  int result;
  Pointer p1;
  char b1[512];

  if (!hook->cmd)
    return TCL_OK;

  initHookResult(hook);

  /* parse arguments */
  p1 = va_arg(ap, Pointer);

  /* convert to ascii representation */
  SWIG_MakePtr(b1, p1, "_Constraint");

  /* execute command */
  result = Tcl_VarEval(hkInterp, hook->cmd, " ", b1, (char *)NULL);

  /* exception handling */
  if (result != TCL_OK) {
    hkResult->type = HTError;
  }

  return result;
}

/* ---------------------------------------------------------------------------
 * glsInteractionHookHandle:
 *	called inside the gls-module when a interaction with the algorithm
 *	is desired
 *	arg1 : GlsNet glsNet
 *	arg2 : String message
 */
TclResultType glsInteractionHookHandle(Hook hook, va_list ap)
{
  int result;
  Pointer glsNet;
  char *message;
  char b1[512];

  if (!hook->cmd)
    return TCL_OK;

  initHookResult(hook);

  /* parse arguments */
  glsNet = va_arg(ap, Pointer);
  message = va_arg(ap, char *);

  /* convert to ascii representation */
  SWIG_MakePtr(b1, glsNet, "_Pointer");

  /* execute command */
  result = Tcl_VarEval(hkInterp, hook->cmd, " ", b1, " {", message, "}",
		       (char *)NULL);

  /* exception handling */
  if (result != TCL_OK) {
    hkResult->type = HTError;
  }

  return result;
}

/* ---------------------------------------------------------------------------
 * getsHookHandle:
 *	handle gets from within tcl
 */
TclResultType getsHookHandle(Hook hook, String buffer, int size)
{
  int result;

  if (!hook->cmd)
    return TCL_OK;

  initHookResult(hook);

  /* execute command */
  result = Tcl_VarEval(hkInterp, hook->cmd, (char *)NULL);

  /* exception handling */
  if (result != TCL_OK) {
    hkResult->type = HTError;
  } else {
    strncpy(buffer, hkInterp->result, size);
  }

  return result;
}

/* ---------------------------------------------------------------------------
 * progressHookHandle:
 *	show progess messages in tcl
 */
TclResultType progressHookHandle(Hook hook, char *format, va_list ap)
{
  int result;
  char buffer[1024];

  if (!hook->cmd)
    return TCL_OK;

  initHookResult(hook);

  /* generate string */
  vsnprintf(buffer, 1024, format, ap);

  /* execute command */
  result = Tcl_VarEval(hkInterp, hook->cmd, " {", buffer, "}", (char *)NULL);

  /* exception handling */
  if (result != TCL_OK) {
    hkResult->type = HTError;
  }

  return result;
}



/* ---------------------------------------------------------------------------
 * partialResultHookHandle:
 * order displaying of another Parse
 */
TclResultType partialResultHookHandle(Hook hook, va_list ap)
{
  int result;
  char b1[512];
  Parse parse;

  if (!hook->cmd)
    return TCL_OK;

  initHookResult(hook);

  /* convert to ascii representation */
  parse = (Parse) va_arg(ap, Parse);
  SWIG_MakePtr(b1, parse, "_Parse");

  /* execute command */
  result = Tcl_VarEval(hkInterp, hook->cmd, " ", b1, (char *)NULL);

  /* exception handling */
  if (result != TCL_OK) {
    hkResult->type = HTError;
  }

  return result;
}

/* ---------------------------------------------------------------------------
 * ICinteractionHookHandle:
 *   get another word from the IC textbox.
 */
TclResultType ICinteractionHookHandle(Hook hook, va_list ap)
{
  int result;
  String buffer;

  buffer = va_arg(ap, String);

  if (!hook->cmd)
    return TCL_OK;

  initHookResult(hook);

  /* execute command */
  result = Tcl_VarEval(hkInterp, hook->cmd, (char *)NULL);

  /* exception handling */
  if (result != TCL_OK) {
    hkResult->type = HTError;
  } else {
    strcpy(buffer, hkInterp->result);
  }

  return result;
}

/* ---------------------------------------------------------------------------
   Call the Tcl-specified command, with no arguments.
*/
TclResultType resetHookHandle(Hook hook, va_list ap) {
  int result;
  char b1[512];
  Parse parse;

  if (!hook->cmd)
    return TCL_OK;

  initHookResult(hook);

  /* execute command */
  result = Tcl_VarEval(hkInterp, hook->cmd, " ", (char *)NULL);

  /* exception handling */
  if (result != TCL_OK) {
    hkResult->type = HTError;
  }

  return result;

}



/* ---------------------------------------------------------------------------
 * tclHookHandle: default tclHook-Handler
 *	arguments aren't used
 */
TclResultType tclHookHandle(Hook hook, va_list ap)
{
  int result;

  if (!hook->cmd)
    return TCL_OK;

  initHookResult(hook);

  /* execute command */
  result = Tcl_VarEval(hkInterp, hook->cmd, (char *)NULL);

  /* exception handling */
  if (result != TCL_OK) {
    hkResult->type = HTError;
  }

  return result;
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
/** @} */
