/* 
 * Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without 
 * modifications, as long as this notice is preserved.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/* -------------------------------------------------------------------------
 * @addtogroup Chunker Chunker - Interface to a Chunking Parser
 * @author Michael Daum (see also AUTHORS and THANKS for more)
 * @date 2002-09-09
 * 
 * $Id: chunker.c,v 1.41 2004/09/01 13:42:31 micha Exp $
 *
 * This module offers an interface for an external chunker.
 *
 * @{
 */

/* -- INCLUDES ------------------------------------------------------------- */
#include <ctype.h> 
#include <unistd.h> 
#include <signal.h> 
#include <stdio.h> 
#include <errno.h> 
#include <limits.h>
#include <string.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/wait.h>
#include "cdg.h" 
#include "chunker.h" 
#include "parse.h"
#include "hook.h"
#include "tagger.h"
#include "set.h"

/* -- MACROS --------------------------------------------------------------- */
// #define DEBUG_GETCHUNKSAT
// #define DEBUG_GETCHUNKS
// #define DEBUG_FINDCHUNK

/* -- TYPE DEFINITIONS ----------------------------------------------------- */

/* ------------------------------------------------------------------------- 
 * representation of the chunker.
 */
struct ChunkerStruct {
  ChunkerMode mode;	     /**< determines the mode of operation */
  LexemGraph lg;	     /**< the current lexemgraph that we are chunking */
  Parse parse;		     /**< parse generated from the annotation of the lattice */
  Level mainlevel;	     /**< mainlevels index in the current parse */
  int nrWords;		     /**< number of words in the current parse */
  int nrLevels;		     /**< number of levels in the current parse */
  List chunks;		     /**< found chunks */
  char **args;		     /**< command to start this chunker @see chunkerArgs */
  pid_t pid;		     /**< process id of a started chunker job */
  int pipe1[2];		     /**< pipe connected to the stdin of the chunker process */
  int pipe2[2];		     /**< pipe connected to the stdout of the chunker process */
};

/* -- VARIABLES ------------------------------------------------------------ */

/** indicates wether the chunker is used or not */
static Boolean chunkerUseChunker = FALSE; 

/** set the default chunker mode, @see DefaultChunker */
static ChunkerMode chunkerMode = RealChunker; 

/** string representation of the current command used for real chunking */
static String chunkerCommand = NULL;

/** NULL terminated array of command arguments used for real chunking */
static char **chunkerArgs = NULL; 

/* -- FUNCTIONS ------------------------------------------------------------ */

/* constructors, destructors & ko */
static Chunk newChunk(ChunkType type);
static Boolean initChunker(Chunker chunker);
static Boolean initFakeChunker(Chunker chunker);
static Boolean initRealChunker(Chunker chunker);
static void resetChunker(Chunker chunker);

/* chunker routines */
static List getChunks(Chunker chunker);
static List getFakeChunks(Chunker chunker);
static List getFakeChunksAt(Chunker chunker, Chunk parent, int index);
static ChunkType getFakeChunkType(Chunker chunker, int index);
static void postProcessChunks(Chunker chunker, List chunks);
static Chunk findChunk(List chunks, int from, int to);
static Chunk mergeChunk(Chunker chunker, Chunk target, Chunk source);
static Chunk embedChunk(Chunker chunker, Chunk target, Chunk source);
static int evalChunker(Chunker chunker, List annoChunks);
static int countChunks(List chunks);
static Boolean compareChunks(Chunk c1, Chunk c2);

/* TODO: these should go into the parse module as their names already say */
static int parseGetModifiee(Chunker chunker, int index);
static List parseGetRoots(Chunker chunker);
static String parseGetLabel(Chunker chunker, int index);
static LevelValue parseGetLevelValue(Chunker chunker, int index);
static String parseGetCategory(Chunker chunker, int index);
static GraphemNode parseGetGrapheme(Chunker chunker, int index);

/* misc */
static Boolean cmpChunks(Chunk c1, Chunk c2, Chunker chunker);
static Boolean cmpArcs(Arc arc1, Arc arc2);
static Boolean cmpGraphemes(GraphemNode g1, GraphemNode g2);
static void printChunk(unsigned long mode, Chunk chunk);
static GraphemNode findGrapheme(LexemGraph lg, GraphemNode gn);
static List getCategories(GraphemNode gn);
static String getCategory(GraphemNode gn);

/* -- IMPLEMENTATION ------------------------------------------------------- */

/* -------------------------------------------------------------------------
 * terminateChild
 *
 * This function waits for child with the specified pid to terminate.
 *
 * Return values:
 *
 * -1 error
 *  0 child died already
 *  1 child died after SIGTERM
 *  2 child died after SIGKILL
 *
 */
int terminateChild(pid_t pid)
{
  pid_t res;

  res = wait4(pid, (int *)NULL, WNOHANG, NULL);
  if (res == pid) {
    return 0;
  }
  if (res == 0) {
    /* child hasn't terminated yet */
    if (-1 == kill(pid, SIGTERM)) {
      cdgPrintf(CDG_WARNING, "WARNING: can't send SIGTERM to %d: %s\n", pid,
		strerror(errno));
    }
    usleep(100000);
    res = wait4(pid, (int *)NULL, WNOHANG, NULL);
    if (res == pid) {
      return 1;
    }
    if (res == 0) {
      if (-1 == kill(pid, SIGKILL)) {
	cdgPrintf(CDG_WARNING, "WARNING: can't send SIGKILL to %d: %s\n", pid,
		  strerror(errno));
      }
      usleep(100000);
      res = wait4(pid, (int *)NULL, WNOHANG, NULL);
      if (res == pid) {
	return 2;
      }
    }
  }
  /* error condition */
  cdgPrintf(CDG_ERROR, "ERROR: can't terminate %d: %s\n", pid,
	    strerror(errno));

  return -1;
}


/* ------------------------------------------------------------------------- 
 * construct a new chunker.
 *
 * @param mode one of the defined @ref ChunkerMode
 * @param lg data from which we're going to initialize 
 * @returns a new chunker
 *
 * In case of a @ref RealChunker a child process is forked in the background
 * specified by the @ref chunkerCommand. This function might return NULL
 * when chunking is switched off, no @ref chunkerCommand is defined  or the 
 * initialization of the chunker object fails.
 *
 * @see initChunker, initRealChunker, initFakeChunker
 */
Chunker chunkerNew(ChunkerMode mode, LexemGraph lg)
{
  Chunker chunker;
  int i;

  if (!chunkerUseChunker)
    return NULL;

  if (mode == DefaultChunker)
    mode = chunkerMode;

  if (mode != FakeChunker) {
    if(!chunkerCommand || strlen(chunkerCommand) == 0) {
      cdgPrintf(CDG_ERROR, "ERROR: no chunker command defined ... switching off the chunker\n");
      chunkerUseChunker = FALSE;
      return NULL;
    }
  }

  chunker = (Chunker) memMalloc(sizeof (ChunkerStruct));
  chunker->mode = mode;
  chunker->lg = lg;
  chunker->parse = NULL;
  chunker->nrWords = 0;
  chunker->nrLevels = 0;
  chunker->chunks = NULL;
  chunker->pid = -1;
  chunker->args = NULL;

  /* copy the chunkerArgs to our private structure */
  if(chunkerArgs) {
    for (i = 0; chunkerArgs[i]; i++) 
      chunker->args = (char **)memMalloc(sizeof (char *) * (i + 1));
    for (i = 0; chunkerArgs[i]; i++) {
      chunker->args[i] = strRegister(chunkerArgs[i]);
    }
    chunker->args[i] = NULL;
  }
  return (initChunker(chunker)?chunker:NULL);
}

/* ------------------------------------------------------------------------- 
 * initialize the chunker with the given data.
 *
 * @param chunker the current chunker
 * @returns true on success.
 *
 * This is called from @ref chunkerChunk and from chunkerNew. It calls 
 * @ref initFakeChunker or @ref initRealChunker depending on the chunker mode.
 */
Boolean initChunker(Chunker chunker)
{
  switch (chunker->mode) {
    case FakeChunker:
      return initFakeChunker(chunker);
      break;
    case RealChunker:
      return initRealChunker(chunker);
    case EvalChunker:
      return initFakeChunker(chunker) && initRealChunker(chunker);
      break;
    default: /* never reach */
      cdgPrintf(CDG_ERROR, "ERROR: unknown chunker mode\n");
      break;
  }

  return FALSE;
}

/* ------------------------------------------------------------------------- 
 * initialize a fake chunker with the given data.
 *
 * @param chunker the current chunker
 * @returns true on success.
 *
 * This function is called by initChunker() whenever the given Chunker
 * is in mode FakeChunker. In order to work properly the contained 
 * Lattice within the LexemGraph lg must have a corresponding 
 * Annotation from which the chunks can be faked.
 *
 * @see also initChunker, initFakeChunker, initRealChunker, chunkerChunk.
 */
Boolean initFakeChunker(Chunker chunker) 
{
  Level mainlevel = inputGetMainlevel(inputCurrentGrammar);
  AnnoEntry anno = findAnnoForLattice(chunker->lg->lattice, TRUE);

  /* check for available annotation */
  if (!anno) {
    cdgPrintf(CDG_ERROR,
	      "ERROR: no annotation for lattice %s found needed in fake mode\n",
              chunker->lg->lattice->id);
    return FALSE;
  }

  /* check for a defined mainlevel */
  if (!mainlevel) {
    cdgPrintf(CDG_ERROR, "ERROR: no mainlevel defined\n");
    return FALSE;
  }

  /* planed to work on linear lattices only right now */
  if (latticeBranches(chunker->lg->lattice)) {
    cdgPrintf(CDG_ERROR,
	      "ERROR: sorry, chunking only works for linear lattices right now\n");
    return FALSE;
  }

  /* get a parse from the annotation */
  chunker->parse = parseFromAnno(anno);
  if (!chunker->parse) {
    resetChunker(chunker);
    return FALSE;
  }

  /* decorate the parse */
  if (!parseDecorate(chunker->parse, chunker->lg, anno)) {
    cdgPrintf(CDG_ERROR,
	      "ERROR: could not decorate parse for annotation `%s'.\n",
	      anno->id);
    resetChunker(chunker);
    return FALSE;
  }

  chunker->nrWords = vectorSize(chunker->parse->words);
  chunker->nrLevels = listSize(chunker->parse->levels);
  chunker->mainlevel = mainlevel;
  
  return TRUE;
}

/* ------------------------------------------------------------------------- 
 * initialize a real chunker with the given data.
 * 
 * @param chunker the current chunker
 * @returns true on success.
 *
 * This function is called from @ref initChunker during the initialization
 * of the chunker, that is just before it starts chunking in @ref chunkerChunk.
 *
 * @see also initChunker, initFakeChunker, initRealChunker, chunkerChunk.
 */
Boolean initRealChunker(Chunker chunker)
{
  /* check for a defined mainlevel */
  chunker->mainlevel = inputGetMainlevel(inputCurrentGrammar);
  if (!chunker->mainlevel) {
    cdgPrintf(CDG_ERROR, "ERROR: no mainlevel defined\n");
    return FALSE;
  }

  /* create a fresh new chunker coprocess for each chunker object */

  /* create pipes */
  if (pipe(chunker->pipe1) < 0 || pipe(chunker->pipe2) < 0) {
    cdgPrintf(CDG_ERROR, "ERROR: can't create pipes\n");
    return FALSE;
  }

#if 0
  cdgPrintf(CDG_DEBUG, "DEBUG: forking %s ... \n", chunker->args[0]);
#endif

  /* fork starts a subprocess under pid 'pid' */
  if ((chunker->pid = fork()) < 0) {
    cdgPrintf(CDG_ERROR, "ERROR: can't fork: %s\n", strerror(errno));
    return FALSE;
  }

  if (chunker->pid) {
    /* parent thread */
    close(chunker->pipe1[0]);
    close(chunker->pipe2[1]);
#if 0
    cdgPrintf(CDG_DEBUG, "DEBUG: process %d started.\n", chunker->pid);
#endif
  } else {
    /* child thread */
    close(chunker->pipe1[1]);
    close(chunker->pipe2[0]);
    signal(SIGINT, SIG_DFL);
    signal(SIGTERM, SIG_DFL);
    signal(SIGXCPU, SIG_DFL);
    if (chunker->pipe1[0] != STDIN_FILENO) {
      if (dup2(chunker->pipe1[0], STDIN_FILENO) != STDIN_FILENO) {
	fprintf(stderr, "WARNING: child can't dup2 stdin: %s\n",
	    strerror(errno));
	close(chunker->pipe1[0]);
      }
    }
    if (chunker->pipe2[1] != STDOUT_FILENO) {
      if (dup2(chunker->pipe2[1], STDOUT_FILENO) != STDOUT_FILENO) {
	fprintf(stderr, "WARNING: child can't dup2 stdout: %s\n",
	    strerror(errno));
	close(chunker->pipe2[1]);
      }
    }
    execvp(chunker->args[0], chunker->args);

    /* never reach */
    fprintf(stderr, "ERROR: exec(%s) failed: %s\n", 
	chunker->args[0], strerror(errno));
    _exit(1);
  }

  return TRUE;
}

/* ------------------------------------------------------------------------- 
 * resetChunker: set the chunker in a state of innocence.
 *
 * parameters:
 *	chunker = the object of desire
 */
void resetChunker(Chunker chunker)
{
  if (chunker->parse) {		       /* not NULL aware */
    parseDelete(chunker->parse);
  }
  listForEachDelete(chunker->chunks, chunkerChunkDelete);
  chunker->lg = NULL;
  chunker->parse = NULL;
  chunker->nrWords = 0;
  chunker->nrLevels = 0;
  chunker->chunks = NULL;
  if (chunker->pid > 0) {
    close(chunker->pipe1[0]);
    close(chunker->pipe1[1]);
    close(chunker->pipe2[0]);
    close(chunker->pipe2[1]);
    terminateChild(chunker->pid);  /* in tagger.c */
  }

  if (chunker->args) {
    int i;
    for(i = 0; chunker->args[i]; i++) {
      cdgFreeString(chunker->args[i]);
    }
    memFree(chunker->args);
    chunker->args = NULL;
  }
}

/* ------------------------------------------------------------------------- 
 * chunkerDelete: destroy the chunker representation.
 *
 * parameters:
 *	chunker = the object to be destructed
 */
void chunkerDelete(Chunker chunker)
{
  if (!chunker)
    return;

  resetChunker(chunker);
  memFree(chunker);
}

/* ------------------------------------------------------------------------- 
 * construct a new chunk and initialize it.
 *
 * @param type one of the ChunkTypes NChunk, PChunk, ...
 * @returns a new empty Chunk.
 */
Chunk newChunk(ChunkType type)
{
  Chunk chunk;

  chunk = (Chunk)memMalloc(sizeof(ChunkStruct));
  chunk->type = type;
  chunk->nodes = NULL;
  chunk->from = NULL;
  chunk->to = NULL;
  chunk->subChunks = NULL;
  chunk->parent = NULL;
  chunk->head = NULL;

  return chunk;
}

/* ------------------------------------------------------------------------- 
 * chunkerCloneChunk: construct a copy of a given chunk including clones of subChunks.
 *
 * parameters:
 *	chunk = the original
 * returns: the copy.
 */
Chunk chunkerCloneChunk(Chunk chunk)
{
  Chunk clone;
  List l;
  
  if (!chunk)
    return NULL;

  clone = newChunk(chunk->type);
  clone->parent = chunk->parent;
  clone->from = chunk->from; /* these are references into the chunker->lg */
  clone->to = chunk->to;
  clone->head = chunk->head;
  clone->nodes = listClone(chunk->nodes);
  for (l = chunk->subChunks; l; l = listNext(l)) {
    clone->subChunks = listAppendElement(clone->subChunks, chunkerCloneChunk(listElement(l)));
  }

  return clone;
}

/* ------------------------------------------------------------------------- 
 * chunkerChunkDelete: destruct a chunk and all its subchunks.
 *
 * parameters:
 *	chunk = a chunk to be deallocated
 */
void chunkerChunkDelete(Chunk chunk)
{
  if (!chunk)
    return;

  listForEachDelete(chunk->subChunks, chunkerChunkDelete);
  listDelete(chunk->nodes);
  memFree(chunk);
}

/* ------------------------------------------------------------------------- 
 * cmpGraphemes: return true if g1 start before g2
 */
Boolean cmpGraphemes(GraphemNode g1, GraphemNode g2)
{
  return (g1->arc->from < g2->arc->from);
}

/* ------------------------------------------------------------------------- 
 * cmpArcs: return true if arc1 starts before arc2
 */
Boolean cmpArcs(Arc arc1, Arc arc2)
{
  return (arc1->from < arc2->from);
}

/* ------------------------------------------------------------------------- 
 * cmpChunks: return true if c1 starts before c2.
 */
Boolean cmpChunks(Chunk c1, Chunk c2, Chunker chunker)
{
  return (c1->from->arc->from < c2->from->arc->from);
}

/* ------------------------------------------------------------------------- 
 * parseGetRoots: get all unbound words (on the main level)
 *
 * parameters:
 *	chunker = the current chunker
 * returns: a list of word indices or NULL if there are no root bindings (?)
 *
 * Note: you become the owner of the returned
 *	list container, so deallocate it after you've consumed the result.
 */
List parseGetRoots(Chunker chunker)
{
  int i;
  List result = NULL;

  for (i = 0; i < chunker->nrWords; i ++) {
    if (parseGetModifiee(chunker, i) == -1) {;
      result = listAppendElement(result, (Pointer)i);
    }
  }

  return result;
}

/* ------------------------------------------------------------------------- 
 * parseGetModifiee: get the word this one is modifying (on the main level)
 *
 * parameters:
 *	chunker = the current chunker
 *	index  = the modifier index in the word vector of the 
 *		current parse
 * returns: the modifiee index
 */
int parseGetModifiee(Chunker chunker, int index)
{
  return (int)vectorElement(chunker->parse->verticesStructure, 
      chunker->nrLevels * index + chunker->mainlevel->no);
}

/* ------------------------------------------------------------------------- 
 * parseGetLabel: get the label of the dependency of a word (on the main level)
 *
 * parameters:
 *	chunker = the current chunker
 *	index = index of a word in the parse
 * returns: the label of that dependency
 */
String parseGetLabel(Chunker chunker, int index)
{
  return (String) vectorElement(chunker->parse->verticesLabels, 
      chunker->nrLevels * index + chunker->mainlevel->no);
}

/* ------------------------------------------------------------------------- 
 * parseGetLevelValue: get the dependency arc of a word.
 *
 * parameters:
 *	chunker = the current chunker
 *	index = index of a word in the parse.
 *
 * returns: the level value of this word
 */
LevelValue parseGetLevelValue(Chunker chunker, int index)
{
  if (index == -1 || index == INT_MAX) return NULL;
  return (LevelValue) vectorElement(chunker->parse->LVs, 
      inputCurrentGrammar->noOfLevels * index + chunker->mainlevel->no);
}

/* ------------------------------------------------------------------------- 
 * getCategory: get one POS-tag, warn if there are more than one
 *
 * parameter: 
 *	gn = a grapheme
 * returns: the first POS-tag available.
 */
String getCategory(GraphemNode gn)
{
  List cats;
  String cat;
  
  if (!gn)
    return NULL;

  cats = getCategories(gn);
  if (!cats) {
    return NULL;
  }

#if 0
  if (listSize(cats) > 1) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: got more than one POS-tag\n");
  }
#endif
  cat = listElement(cats);
  listDelete(cats);

  return cat;
}

/* ------------------------------------------------------------------------- 
 * getCategories: get all POS-tags of undeleted lexem nodes
 *
 * parameter: 
 *	gn = a lexem node
 * returns: a list of POS tags.
 */
List getCategories(GraphemNode gn)
{
  List cats = NULL;
  List l;
  Value value;
  List lns = NULL;

  Boolean lexemSorter(LexemNode lna, LexemNode lnb) 
  {
    return (lna->tagscore > lnb->tagscore);
  }

  for (l = gn->lexemes; l; l = listNext(l)) {
    LexemNode ln = listElement(l);
    
    if (bvElement(gn->lexemgraph->isDeletedNode, ln->no)) 
      continue;

    value = ln->lexem->values[taggerCategoryIndex];
    if (value->type == VTString) {
      lns = listInsertSorted(lns, ln, lexemSorter);
    }
  }

  for (l = lns; l; l = listNext(l)) {
    LexemNode ln = listElement(l);
    value = ln->lexem->values[taggerCategoryIndex];
    cats = listAddUniqueElement(cats, value->data.string);
#if 0
    cdgPrintf(CDG_DEBUG, "DEBUG: cat %s = score %g\n",
	value->data.string, ln->tagscore);
#endif
  }
  listDelete(lns);


  return cats;
}

/* ------------------------------------------------------------------------- 
 * parseGetCategory: get the POS-tag of a given word index
 *
 * parameters:
 *	chunker = the current chunker
 *	index = index of a word in the parse.
 *
 * returns: the POS-tag string or NULL if not defined
 */
String parseGetCategory(Chunker chunker, int index)
{
  if (index == -1 || index == INT_MAX) {
    return "";
  } else {
    LevelValue lv = parseGetLevelValue(chunker, index);
    LexemNode ln = lv->modifier;
    Value value = ln->lexem->values[taggerCategoryIndex];

    if (value->type != VTString) {
      return NULL;
    }
	    
    return value->data.string;
  }
}

/* ------------------------------------------------------------------------- 
 * parseGetGrapheme: get the grapheme node of a given word index
 *
 * parameters:
 *	chunker = the current chunker
 *	index = index of a word in the parse.
 *
 * returns: the arc.
 */
GraphemNode parseGetGrapheme(Chunker chunker, int index)
{
  LevelValue lv;

  if (index == -1 || index == INT_MAX) {
    return NULL;
  }

  lv = parseGetLevelValue(chunker, index);
  
  if (!lv) { /* never reach */
    cdgPrintf(CDG_ERROR, "ERROR: can't get grapheme at %d\n", index);
    abort();
  }
  return lv->modifier->grapheme;
}


/* ------------------------------------------------------------------------- 
 * getFakeChunkType
 */
ChunkType getFakeChunkType(Chunker chunker, int index)
{
  String cat = parseGetCategory(chunker, index);

  if (cat == strRegister("NN") ||
      cat == strRegister("NE") ||
      cat == strRegister("FM") ||
      cat == strRegister("CARD") ||
      cat == strRegister("PDS") ||
      cat == strRegister("PPER") ||
      cat == strRegister("PRF") ||
      cat == strRegister("PRELS") ||
      cat == strRegister("PPOSS") ||
      cat == strRegister("PWS") ||
      cat == strRegister("PIS") ||
      cat == strRegister("ADJA") ||
      cat == strRegister("KON") ||
      cat == strRegister("TRUNC")) {
    return NChunk;
  }

  if (cat == strRegister("VVINF") ||
      cat == strRegister("VVIZU") ||
      cat == strRegister("VVFIN") ||
      cat == strRegister("VVPP") ||
      cat == strRegister("VMINF") ||
      cat == strRegister("VMFIN") ||
      cat == strRegister("VMPP") ||
      cat == strRegister("VAINF") ||
      cat == strRegister("VAFIN") ||
      cat == strRegister("VAPP") ||
      cat == strRegister("PTKVZ") ||
      cat == strRegister("PTKZU")) {
    return VChunk;
  }

  if (cat == strRegister("APPR") ||
      cat == strRegister("APPRART") ||
      cat == strRegister("APPO")) {
    return PChunk;
  }

  if (cat == strRegister("KOKOM") ||
      cat == strRegister("PROAV") ||
      cat == strRegister("ADV") ||
      cat == strRegister("ADJD") ||
      cat == strRegister("KON")) {
    return NoChunk;
  }

  return UnknownChunk;
}

/* ------------------------------------------------------------------------- 
 * mergeChunk: add the source to the target chunk. the target chunk spans
 *	the words of both chunks.
 *	
 * parameters: 
 *	chunker = the current chunker
 *	target = the resulting chunk
 *	source = the chunk to be added to the target
 *
 * returns: the target chunk.
 */
Chunk mergeChunk(Chunker chunker, Chunk target, Chunk source)
{
  List l;

#ifdef DEBUG_GETCHUNKSAT
  cdgPrintf(CDG_DEBUG, "DEBUG: merging chunk %s <%s,%s> to target chunk %s <%s,%s>\n",
      chunkerStringOfChunkType(source), source->from->arc->word, source->to->arc->word,
      chunkerStringOfChunkType(target), target->from->arc->word, target->to->arc->word);
#endif
  
  /* enlarge the span */
  if (target->to->arc->to == source->from->arc->from) {
    target->to = source->to;
  }
  if (source->to->arc->to == target->from->arc->from) {
    target->from = source->from;
  }

  for (l = source->nodes; l; l = listNext(l)) {
    Arc arc = listElement(l);
    target->nodes = listInsertSorted(target->nodes, arc, cmpGraphemes);
  }
  
  /* add all sub chunks to the currently available subchunks */
  for (l = source->subChunks; l; l = listNext(l)) {
    Chunk clone = chunkerCloneChunk(listElement(l));
    target->subChunks = 
      listInsertSortedWithData(target->subChunks, clone, cmpChunks, chunker);
  }

  return target;
}

/* ------------------------------------------------------------------------- 
 * embedChunk: embed chunk as a subchunk into the target chunk
 *
 * parameters
 *	target = the resulting chunk
 *	source = the chunk to be embedded
 * returns: the target chunk.
 */
Chunk embedChunk(Chunker chunker, Chunk target, Chunk source)
{
  List l;

#ifdef DEBUG_GETCHUNKSAT
  cdgPrintf(CDG_DEBUG,
	    "DEBUG: embedding chunk %s <%s,%s> to target chunk %s <%s,%s>\n",
	    chunkerStringOfChunkType(source), 
	    source->from->arc->word, source->to->arc->word,
	    chunkerStringOfChunkType(target), 
	    target->from->arc->word, target->to->arc->word);
#endif
  if (target->to->arc->to == source->from->arc->from) {
    target->to = source->to;
  }
  if (source->to->arc->to == target->from->arc->from) {
    target->from = source->from;
  }
  for (l = source->nodes; l; l = listNext(l)) {
    Arc arc = listElement(l);
    target->nodes = listInsertSorted(target->nodes, arc, cmpArcs);
  }
  
  target->subChunks = 
    listInsertSortedWithData(target->subChunks, source, cmpChunks, chunker);

  return target;
}

/* ------------------------------------------------------------------------- 
 * get the chunks under the given root node.
 *
 * @param chunker the current chunker
 * @param parent dominating chunk
 * @param index index of the root node in the dependency tree
 */
List getFakeChunksAt(Chunker chunker, Chunk parent, int index)
{
  List chunks = NULL;
  List children, leftChildren, rightChildren;
  List l;
  Chunk thisChunk, chunk;
  GraphemNode thisNode;

  /* inline function ------------------------------------------------------- */
  void doit(List children) {
    List l, m;

    for (l = children; l; l = listNext(l)) {
      int childIndex = (int)listElement(l);
      String childLabel = parseGetLabel(chunker, childIndex);
      List childChunks = getFakeChunksAt(chunker, thisChunk, childIndex);

#ifdef DEBUG_GETCHUNKSAT
      cdgPrintf(CDG_DEBUG, "DEBUG: childChunks = ");
      for (m = childChunks; m; m = listNext(m)) {
	chunk = listElement(m);
	cdgPrintf(CDG_DEBUG, "%s <%s,%s> " ,
	    chunkerStringOfChunkType(chunk),
	    chunk->from->arc->word, chunk->to->arc->word);
      }
      cdgPrintf(CDG_DEBUG, "\n");
#endif

      /* integrate chunks to current chunk */
      for (m = childChunks; m; m = listNext(m)) {
	chunk = listElement(m);

	if ((thisChunk->to->arc->to == chunk->from->arc->from || /* check adjacency */
	    chunk->to->arc->to == thisChunk->from->arc->from) && 
	    chunk->parent == thisChunk && /* check directly domination */
	    childLabel != strRegister("GMOD") /* never integrate GMODs */ &&
	    childLabel != strRegister("ADV") /* never integrate ADVs */
	    ) { 

	  /* NChunks */
	  if (thisChunk->type == NChunk && 
	      (chunk->type == UnknownChunk || 
	       chunk->type == NChunk)) {
#ifdef DEBUG_GETCHUNKSAT
	    cdgPrintf(CDG_DEBUG, "DEBUG: appyling NC rule\n");
#endif
	    mergeChunk(chunker, thisChunk, chunk);
	    chunkerChunkDelete(chunk);
	    continue;
	  }

	  /* PChunks */
	  if (thisChunk->type == PChunk) {
	    /* adjectives */
	    if(chunk->type == UnknownChunk) {
#ifdef DEBUG_GETCHUNKSAT
  	      cdgPrintf(CDG_DEBUG, "DEBUG: appyling PC rule for XCs\n");
#endif

	      mergeChunk(chunker, thisChunk, chunk);
	      chunkerChunkDelete(chunk);
	      continue;
	    }

	    /* embedded NPs */
	    if (chunk->type == NChunk) {
#ifdef DEBUG_GETCHUNKSAT
  	      cdgPrintf(CDG_DEBUG, "DEBUG: appyling PC rule for NCs\n");
#endif
	      embedChunk(chunker, thisChunk, chunk);
	      continue;
	    }

	    /* embedded PPs, e.g. bis zum Frhjahr */
	    if (chunk->type == PChunk && thisChunk->from == thisChunk->to) {
#ifdef DEBUG_GETCHUNKSAT
  	      cdgPrintf(CDG_DEBUG, "DEBUG: appyling PC rule for PCs\n");
#endif
	      mergeChunk(chunker, thisChunk, chunk);
	      chunkerChunkDelete(chunk);
	      continue;
	    }
	  }

	  /* VChunks */
	  if (thisChunk->type == VChunk && chunk->type == VChunk) {
#ifdef DEBUG_GETCHUNKSAT
  	      cdgPrintf(CDG_DEBUG, "DEBUG: appyling VC rule\n");
#endif
	    mergeChunk(chunker, thisChunk, chunk);
	    chunkerChunkDelete(chunk);
	    continue;
	  }

#ifdef DEBUG_GETCHUNKSAT
	  cdgPrintf(CDG_DEBUG, "DEBUG: no rule applies\n");
#endif
	}
#ifdef DEBUG_GETCHUNKSAT
	else {
	  cdgPrintf(CDG_DEBUG, "DEBUG: not adjacent\n");
	}
#endif

	/* keep separate */
#ifdef DEBUG_GETCHUNKSAT
	cdgPrintf(CDG_DEBUG, "DEBUG: cannot add chunk %s <(%d,%d)%s,(%d,%d)%s> to target chunk %s <(%d,%d)%s,(%d,%d)%s>\n",
	      chunkerStringOfChunkType(chunk), 
	      chunk->from->arc->from, chunk->from->arc->to, 
	      chunk->from->arc->word, 
	      chunk->to->arc->from, chunk->to->arc->to, 
	      chunk->to->arc->word,
	      chunkerStringOfChunkType(thisChunk), 
	      thisChunk->from->arc->from, thisChunk->from->arc->to, 
	      thisChunk->from->arc->word, 
	      thisChunk->to->arc->from, thisChunk->to->arc->to, 
	      thisChunk->to->arc->word);
#endif
	chunks = listInsertSortedWithData(chunks, chunk, cmpChunks, chunker);
      }
      listDelete(childChunks);
    }
  }
  /* ------------------------------------------------------------------------ */
  /* ------------------------------------------------------------------------ */

  /* security first */
  if (index == -1 || index == INT_MAX) {
    return NULL;
  }

  /* construct a chunk for the current arc */
  thisNode = parseGetGrapheme(chunker, index);
  thisChunk = newChunk(getFakeChunkType(chunker, index));
  thisChunk->parent = parent;
  thisChunk->from = thisNode;
  thisChunk->to = thisNode;
  thisChunk->head = thisNode;
  thisChunk->nodes = listAppendElement(NULL, thisNode);
  children = parseGetModifiers(chunker->parse, chunker->mainlevel, index);

#ifdef DEBUG_GETCHUNKSAT
  cdgPrintf(CDG_DEBUG, "DEBUG: decending at %s\n",
      thisChunk->from->arc->word);
#endif


  /* end of recursion: each leaf node is its own chunk */
  if (!children) {
#ifdef DEBUG_GETCHUNKSAT
    cdgPrintf(CDG_DEBUG, "DEBUG: returning from %s (leaf reached)\n",
  	thisChunk->from->arc->word);
#endif
    return listAppendElement(NULL, thisChunk);
  }

  /* separate modifiers left and right of the current index */
  leftChildren = rightChildren = NULL;
  for (l = children; l; l = listNext(l)) {
    int childIndex = (int)listElement(l);
    GraphemNode gn = parseGetGrapheme(chunker, childIndex);

    /* left */
    if (thisChunk->from->arc->from >= gn->arc->to)  {
      leftChildren = listAppendElement(leftChildren, (Pointer)childIndex);
    }

    /* right */
    else {
      rightChildren = listAppendElement(rightChildren, (Pointer)childIndex);
    }
  }
  listDelete(children);

  /* process children */

  { /* descent at the left in reverse order */
    List reverseList = NULL; 
    /* TODO: rework on new ACL with listReverse */
#if 0 
    reverseList = listReverse(leftChildren);
#else
    for (l = leftChildren; l; l = listNext(l)) {
      reverseList = listPrependElement(reverseList, listElement(l));
    }
#endif
    listDelete(leftChildren);
    leftChildren = reverseList;
  }

  doit(leftChildren);
  doit(rightChildren);

  /* add this constructed chunk to the list if all chunks we have sofar */
  chunks = listInsertSortedWithData(chunks, thisChunk, cmpChunks, chunker);

#ifdef DEBUG_GETCHUNKSAT
  cdgPrintf(CDG_DEBUG, "DEBUG: returning from %s\n",
      thisChunk->from->arc->word);
#endif
  return chunks;
}

/* ------------------------------------------------------------------------- 
 * postProcessChunks: get rid of unwanted chunks.
 *
 * parameters:
 *	inputList = items to be filtered
 */
void postProcessChunks(Chunker chunker, List inputList)
{
  List l;
  Chunk chunk;
  String cat;

  for (l = inputList; l; l = listNext(l)) {
    chunk = listElement(l);
    cat = getCategory(chunk->from);
    if ((chunk->from == chunk->to
	 && (cat == strRegister("PTKZU") ||
	     cat == strRegister("KON")))
	|| chunk->type == UnknownChunk) {
#if 0
      cdgPrintf(CDG_DEBUG, "DEBUG: filtered chunk %s <%s, %s>\n",
		chunkerStringOfChunkType(chunk),
		chunk->from->arc->word, chunk->to->word);
#endif
      chunk->type = NoChunk;
    }
  }
}

/* ------------------------------------------------------------------------- 
 * this is the entry function to the fake chunker.
 *
 * @param chunker the current chunker to be used
 * @returns a list of chunks
 */
List getFakeChunks(Chunker chunker)
{
  List chunks = NULL;
  List result = NULL;
  List l, m, roots;
  int index;

  roots = parseGetRoots(chunker);
  for (l = roots; l; l = listNext(l)) {
    index = (int)listElement(l);
    chunks = getFakeChunksAt(chunker, NULL, index);
    postProcessChunks(chunker, chunks);
    for (m = chunks; m; m = listNext(m)) {
      result =
	listInsertSortedWithData(result, listElement(m), cmpChunks, chunker);
    }
    listDelete(chunks);
  }

  listDelete(roots);
  return result;
}

/* ------------------------------------------------------------------------- 
 * this is the entry function to the real chunker.
 *
 * @param chunker the current chunker to be used
 * @returns a list of chunks
 * 
 * Writes to the chunker, reads from the chunker and builds a list of 
 * chunks.
 */
List getChunks(Chunker chunker)
{
  int i, n;
  List l;
  List chunks = NULL;
  String str;
  char buffer[MAXBUFFER];
  List outputs = NULL;
  String bufptr;
  String token;
  Vector gns;

  /* write to chunker */
#ifdef DEBUG_GETCHUNKS
  cdgPrintf(CDG_DEBUG, "DEBUG: writing to chunker:\n");
#endif
  for (i = 0; i < vectorSize(chunker->lg->graphemnodes); i++) {
    GraphemNode gn = vectorElement(chunker->lg->graphemnodes, i);
    String cat = getCategory(gn);
    str = strPrintf("%s\t%s\n", gn->arc->word, cat);
    n = strlen(str);
#ifdef DEBUG_GETCHUNKS
    cdgPrintf(CDG_DEBUG, str);
#endif
    if(write(chunker->pipe1[1], str, n) != n) {
      cdgPrintf(CDG_ERROR, "ERROR: while writing %s(%d) to chunker: %s\n",
	  str, n, strerror(errno));
    }
    cdgFreeString(str);
  }
#ifdef DEBUG_GETCHUNKS
  cdgPrintf(CDG_DEBUG, "\n");
#endif

  /* finish with an empty line */
  str = strRegister("\n");
  if(write(chunker->pipe1[1], str, 1) != 1) {
    cdgPrintf(CDG_ERROR, "ERROR: while writing a newline to chunker: %s\n",
	strerror(errno));
  }
  cdgFreeString(str);

  close(chunker->pipe1[1]); /* is that a good idea ??? */

  /* read from chunker */
  for (i = 0; i < vectorSize(chunker->lg->graphemnodes); i++) {
    n = read(chunker->pipe2[0], buffer, MAXBUFFER);
    if(n < 0) {
      cdgPrintf(CDG_ERROR, "ERROR: while reading from chunker: %s\n",
	  strerror(errno));
      break;
    }
    if (n == 0)
      break;
    buffer[n] = '\0';
    outputs = listAppendElement(outputs, strRegister(buffer));
#if 0
    cdgPrintf(CDG_DEBUG, "DEBUG: reading\n%s\n", buffer);
#endif
  }
  close(chunker->pipe2[0]); /* is this a good idea ??? */

  if(!outputs) {
    return NULL;
  }
  
  /* concatenate the outputs */
  str = strCopy(strFromList(outputs)); /* allocate memory */
  listForEachDelete(outputs, cdgFreeString);
  
  /* split into lines */
  bufptr = str;
  outputs = NULL;
  token = strtok_r(str, "\n", &bufptr);
  while (token) {
    outputs = listAppendElement(outputs, strRegister(token));
    token = strtok_r(NULL, "\n", &bufptr);
  }
  memFree(str); /* allocated by strCopy */

#if 0
  for (l = outputs; l; l = listNext(l)) {
    cdgPrintf(CDG_DEBUG, "DEBUG: %s\n", listElement(l));
  }
#endif

  /* collect graphemnodes */
  gns = vectorNew(vectorSize(chunker->lg->graphemnodes));
  for (i = 0; i < vectorSize(chunker->lg->graphemnodes); i++) {
    GraphemNode gn = vectorElement(chunker->lg->graphemnodes, i);
    vectorSetElement(gns, gn, gn->arc->from);
  }

  /* 
   * convert output to chunks 
   * line format: <wordform>\t<chunk-type>\t<chunk-from>\t<chunk-to>\t<chunk-head>
   */
  for (l = outputs; l; l = listNext(l)) {
    int from = -1, to = -1, head = -1;
    char tag[3];
    Chunk chunk = NULL;

    /* get data */
    if (sscanf(listElement(l), "%s\t%2s\t%d\t%d\t%d", 
	  buffer, tag, &from, &to, &head) != 5) {
      cdgPrintf(CDG_WARNING, "WARNING: unexpected line format '%s'\n", 
	  listElement(l));
      continue;
    }

    from--;
    to--;
    head--;

#ifdef DEBUG_GETCHUNKS
    cdgPrintf(CDG_DEBUG, "DEBUG: %s\t%s\t%d\t%d\t%d\n", 
	buffer, tag, from, to, head);
#endif
    /* search matching chunk */
    chunk = findChunk(chunks, from, to);
    if (!chunk) {
      /* new chunk */
      chunk = newChunk(chunkerChunkTypeOfString(tag));
      chunks = listAppendElement(chunks, chunk);
      chunk->from = vectorElement(gns, from);
      chunk->to = vectorElement(gns, to);
      chunk->head = vectorElement(gns, head);
      for (i = from; i <= to; i++) {
	chunk->nodes = listAppendElement(chunk->nodes, vectorElement(gns, i));
      }
      continue;
    } 

    /* already contained */
    if (chunk->from->arc->from == from && chunk->to->arc->from == to) 
      continue;
    
    /* create a sub chunk inside chunk */
    if (chunk->from->arc->from <= from &&
	chunk->to->arc->from >= to) {
      Chunk subChunk = newChunk(chunkerChunkTypeOfString(tag));
      subChunk->from = vectorElement(gns, from);
      subChunk->to = vectorElement(gns, to);
      subChunk->head = vectorElement(gns, head);
      for (i = from; i <= to; i++) {
	subChunk->nodes = listAppendElement(subChunk->nodes, vectorElement(gns, i));
      }
#ifdef DEBUG_GETCHUNKS
      cdgPrintf(CDG_DEBUG, 
	  "DEBUG: creating a subchunk <%s:%d-%d> inside <%s:%d-%d>\n",
	  tag, from, to, 
	  chunkerStringOfChunkType(chunk),
	  chunk->from->arc->from, chunk->to->arc->from);
#endif
      chunk->subChunks = 
	listInsertSortedWithData(chunk->subChunks, subChunk, cmpChunks, chunker);
      continue;
    }

    /* never reach */
    cdgPrintf(CDG_ERROR, "ERROR: programming error\n");
    abort();
  }

  /* memory tinji */
  listForEachDelete(outputs, cdgFreeString);
  vectorDelete(gns);

  return chunks;
}

/* ------------------------------------------------------------------------- 
 * count the number of chunks.
 *
 * @param chunks list of chunks to be counted
 * @returns amount of chunks
 *
 * This not only counts the list length but also all sub chunks.
 */
int countChunks(List chunks)
{
  List l;
  int counter = 0;

  for (l = chunks; l; l = listNext(l)) {
    Chunk chunk = listElement(l);
    if (chunk->type == NChunk ||
	chunk->type == PChunk ||
	chunk->type == VChunk) {
      counter++;
    }
    counter += countChunks(chunk->subChunks);
  }

  return counter;
}

/* ------------------------------------------------------------------------- 
 * search the chunk that spans over the given indices.
 *
 * @param chunks a list of chunks
 * @param from the starting point of the span
 * @param to the ending point of the span
 * @returns the found Chunk or NULL if the specified span has not been
 * chunked so far
 *
 * This is a recursive function:
 *    - returns NULL if chunks is empty 
 *    - returns NULL if <from-to> is not spanned by any chunk
 *    - returns chunk X if it spans exactly <from-to>
 *    - if a chunk X spans more than <from-to>
 *       - return X if there's no exaclty spanning sub chunk Y
 *       - or the found sub chunk Y inside X
 */
Chunk findChunk(List chunks, int from, int to)
{
  List l;

  for (l = chunks; l; l = listNext(l)) {
    Chunk chunk = listElement(l);

    /* already spanned */
    if (chunk->from->arc->from == from && chunk->to->arc->from == to) {
#if DEBUG_FINDCHUNK
      cdgPrintf(CDG_DEBUG, "DEBUG: found in ");
      printChunk(CDG_DEBUG, chunk);
      cdgPrintf(CDG_DEBUG, "\n");
#endif
      return chunk;
    }

    /* check sub chunk */
    if (chunk->from->arc->from <= from && chunk->to->arc->from >= to) {
      Chunk subChunk;
      
#if DEBUG_FINDCHUNK
      cdgPrintf(CDG_DEBUG, "DEBUG: searching sub chunk\n");
#endif
      subChunk = findChunk(chunk->subChunks, from, to);
      if (subChunk) {
#if DEBUG_FINDCHUNK
	cdgPrintf(CDG_DEBUG, "DEBUG: found in sub chunk ");
	printChunk(CDG_DEBUG, subChunk);
	cdgPrintf(CDG_DEBUG, "\n");
#endif
	return subChunk;
      } else {
#if DEBUG_FINDCHUNK
	cdgPrintf(CDG_DEBUG, "DEBUG: found, no sub chunk, using ");
	printChunk(CDG_DEBUG, chunk);
	cdgPrintf(CDG_DEBUG, "\n");
#endif
	return chunk;
      }
    }
  }

#if DEBUG_FINDCHUNK
  cdgPrintf(CDG_DEBUG, "DEBUG: not found\n");
#endif

  return NULL;
}

/* ------------------------------------------------------------------------- 
 * printChunk: print a single chunk and all its subchunks
 * 
 * parameters:
 *	mode = print mode, e.g. CDG_INFO
 *	chunk = the chunk to be printed
 */
void printChunk(unsigned long mode, Chunk chunk)
{
  GraphemNode from, to;
  List l,ll;
  List subChunks = NULL;
  if (!chunk)
    return;

  if (chunk->type != NoChunk) 
    cdgPrintf(mode, "[%s ", chunkerStringOfChunkType(chunk));

  /* get span of sub chunks */
  if (chunk->subChunks) {
    subChunks = chunk->subChunks;
    from = ((Chunk)listElement(subChunks))->from;
    to = ((Chunk)listLastElement(subChunks))->to;
  } else {
    from = to = NULL;
  }

  /* print all words spanning the current chunk */
  l = chunk->nodes;
  while (l) {
    GraphemNode gn = listElement(l);

    /* ... which arent part of a subchunk */
    if (!subChunks || 
	gn->arc->to <= from->arc->from || gn->arc->from >= to->arc->to) {
      List cats = 
#if 0
	getCategories(gn);
#else
	listAppendElement(NULL, getCategory(gn));
#endif
      cdgPrintf(mode, "%s%s/", 
	  (gn == chunk->head && chunk->type != NoChunk)?"*":"", gn->arc->word);
      for (ll = cats; ll; ll = listNext(ll)) {
	String cat = listElement(ll);
	cdgPrintf(mode, "%s%s", cat, listNext(ll)?"/":"");
      }
      listDelete(cats);
      l = listNext(l);
    } else {
      Chunk subChunk = listElement(subChunks);
      subChunks = listNext(subChunks);
      if (subChunks) {
	from = ((Chunk)listElement(subChunks))->from;
      } else {
	from = to = NULL;
      }
      printChunk(mode, subChunk);
      for (; l; l = listNext(l)) {
	gn = listElement(l);
	if (gn->arc->from >= subChunk->to->arc->to)
	  break;
      }
    }

    if (l)
      cdgPrintf(mode, " ");
  }

  if (chunk->type != NoChunk) 
    cdgPrintf(mode, "]");
}

/* ------------------------------------------------------------------------- 
 * print the chunks of a lattice.
 * 
 * @param mode print mode, e.g. CDG_INFO
 * @param chunks the list of chunks to be printed
 */
void chunkerPrintChunks(unsigned long mode, List chunks)
{
  List l;
  
  if (!chunks) {
    return;
  }

  for (l = chunks; l; l = listNext(l)) {
    printChunk(mode, listElement(l));
    cdgPrintf(mode, " ");
  }
  cdgPrintf(mode, "\n");
}

/* ------------------------------------------------------------------------- 
 * return the string representation of a chunk type.
 */
ChunkType chunkerChunkTypeOfString(String tag)
{
  if (strcmp(tag, "NC") == 0) {
    return NChunk;
  }
  if (strcmp(tag, "VC") == 0) {
    return VChunk;
  }
  if (strcmp(tag, "PC") == 0) {
    return PChunk;
  }
  return NoChunk;
}

/* ------------------------------------------------------------------------- 
 * return the string representation of a chunk type.
 */
String chunkerStringOfChunkType(Chunk chunk)
{
  if (!chunk) {
    return strRegister("(null)");
  }
  switch(chunk->type) {
    case NChunk: return strRegister("NC"); 
    case VChunk: return strRegister("VC"); 
    case PChunk: return strRegister("PC"); 
    default: return strRegister("XC"); 
  }
}

/* ------------------------------------------------------------------------- 
 * compute the chunks.
 *
 * @param chunker the current chunker
 * @returns a list of chunks or NULL on failure. 
 *
 * The returned list is owned by the chunker and thus needs no extra 
 * deallocation besides chunker deletion. The list of computed chunks
 * is then used to assert the chunking information into the LexemGraph lg.
 */
List chunkerChunk(Chunker chunker) 
{
  List annoChunks = NULL;
  List l;

  void backlink(Chunk chunk) {
    List l;
    for (l = chunk->nodes; l; l = listNext(l)) {
      GraphemNode gn = listElement(l);
      gn->chunk = chunk;
    }
    for (l = chunk->subChunks; l; l = listNext(l)) {
      backlink(listElement(l));
    }
  }

  if (!chunker) {
    return NULL;
  }

  switch (chunker->mode) {
    case RealChunker:
      chunker->chunks = getChunks(chunker);
      break;
    case FakeChunker:
      chunker->chunks = getFakeChunks(chunker);
      break;
    case EvalChunker:
      chunker->chunks = getChunks(chunker);
      annoChunks = getFakeChunks(chunker);
      evalChunker(chunker, annoChunks);
      cdgPrintf(CDG_INFO, "\n");
      listForEachDelete(annoChunks, chunkerChunkDelete);
      break;
    default: /* never reach */
      cdgPrintf(CDG_ERROR, "ERROR: unknown chunker mode\n");
      return NULL;
  }

  if (!chunker->chunks)
    return NULL;

  /* enrich the lexemgraph with chunking information. */
  if (chunker->lg->chunks) {
    listForEachDelete(chunker->lg->chunks, chunkerChunkDelete);
    chunker->lg->chunks = NULL;
  }
  
  /* attach chunk graphemes */
  for (l = chunker->chunks; l; l = listNext(l)) {
    Chunk chunk = chunkerCloneChunk(listElement(l));
    chunker->lg->chunks = listAppendElement(chunker->lg->chunks, chunk);
    chunkerReplaceGraphemes(chunk, chunker->lg);
    backlink(chunk);
  }

  return chunker->chunks;
}

/* ------------------------------------------------------------------------- 
 * are two chunks isomorph.
 *
 * @param c1 the one chunk
 * @param c2 the other chunk
 * @returns TRUE if the two chunks are equal, else FALSE
 *
 * This is again a recursion where all sub chunks must match aswell.
 */
Boolean compareChunks(Chunk c1, Chunk c2)
{
  List l, ll;

  if ((c1 && !c2) || (!c1 && c2) ||
      c1->from->arc->from != c2->from->arc->from ||
      c1->from->arc->to != c2->from->arc->to ||
      c1->from->arc->word != c2->from->arc->word ||
      c1->to->arc->from != c2->to->arc->from ||
      c1->to->arc->to != c2->to->arc->to ||
      c1->to->arc->word != c2->to->arc->word || c1->type != c2->type) {
    return FALSE;
  }

  for (l = c1->subChunks, ll = c2->subChunks; l && ll;
       l = listNext(l), ll = listNext(ll)) {
    if (!compareChunks(listElement(l), listElement(ll))) {
      return FALSE;
    }
  }

  if (!l && !ll) 
    return TRUE;
  else
    return FALSE;
}

/* ------------------------------------------------------------------------- 
 * evaluate computed agains annotated chunks.
 *
 * @param chunker the current chunker
 * @param annoChunks the list of chunks extracted from the annotation
 * @returns the number of errors
 */
int evalChunker(Chunker chunker, List annoChunks)
{
  int noChunks = countChunks(chunker->chunks);
  int noAnnoChunks = countChunks(annoChunks);
  int noErrors = 0, noUnChunked = 0;
  List l;
  List errorChunks = NULL;
  List unChunked = NULL;

  if (noChunks != noAnnoChunks) {
    cdgPrintf(CDG_WARNING, "\nWARNING: got %d chunk(s) but was expecting %d\n",
	      noChunks, noAnnoChunks);
  }

  /* check computed chunks */
  for (l = chunker->chunks; l; l = listNext(l)) {
    Chunk chunk = listElement(l);
    Chunk annoChunk = findChunk(annoChunks, chunk->from->arc->from,
				chunk->to->arc->from);

    if (chunk->type == NoChunk)
      continue;

    if (!compareChunks(chunk, annoChunk))
      errorChunks = listAppendElement(errorChunks, chunk);
  }

  noErrors = listSize(errorChunks);
  if (noErrors) {
    cdgPrintf(CDG_WARNING, "\nWARNING: got %d erroneous chunk(s):\n", noErrors);
    for (l = errorChunks; l; l = listNext(l)) {
      Chunk chunk = listElement(l);

      cdgPrintf(CDG_WARNING, "      ");
      printChunk(CDG_WARNING, chunk);
      cdgPrintf(CDG_WARNING, "\n");
    }
  }

  /* check anno chunks */
  for (l = annoChunks; l; l = listNext(l)) {
    Chunk annoChunk = listElement(l);
    Chunk chunk = findChunk(chunker->chunks, annoChunk->from->arc->from,
			    annoChunk->to->arc->from);

    if (annoChunk->type == NoChunk)
      continue;

    if (annoChunk->type != NChunk &&
	annoChunk->type != VChunk && annoChunk->type != PChunk)
      continue;

    if (!compareChunks(chunk, annoChunk))
      unChunked = listAppendElement(unChunked, annoChunk);
  }

  noUnChunked = listSize(unChunked);
  if (noUnChunked) {
    cdgPrintf(CDG_WARNING, "\nWARNING: %d chunk(s) not found:\n", noUnChunked);
    for (l = unChunked; l; l = listNext(l)) {
      Chunk chunk = listElement(l);

      cdgPrintf(CDG_WARNING, "      ");
      printChunk(CDG_WARNING, chunk);
      cdgPrintf(CDG_WARNING, "\n");
    }
  }

  if (!noErrors && !noUnChunked) {
    cdgPrintf(CDG_INFO, "\nINFO: chunker agrees with annotations.\n",
	noChunks);
  }

  cdgPrintf(CDG_INFO, "\nINFO: got %d answers of which %d where correct, having %d keys\n",
      noChunks, noChunks - noErrors, noAnnoChunks);

  listDelete(errorChunks);
  listDelete(unChunked);

  return noErrors + noUnChunked;
}

/* ------------------------------------------------------------------------- 
 * find an equivalent grapheme in a given lexemgraph.
 *
 * parameters:
 *	lg = the lexemgraph
 *	arc = the arc (possibly not used in the lexemgraph)
 * returns: an equivalent arc.
 */
GraphemNode findGrapheme(LexemGraph lg, GraphemNode old)
{
  int i;

  if (!old)
    return NULL;

  for (i = 0; i < vectorSize(lg->graphemnodes); i++) {
    GraphemNode gn = vectorElement(lg->graphemnodes, i);
    if (gn->arc->from == old->arc->from && 
	gn->arc->to == old->arc->to) {
      return gn;
    }
  }

  return NULL;
}

/* ------------------------------------------------------------------------- 
 * chunkerReplaceGraphemes: replace all grapheme references in a chunk with those 
 *	given in a lexemgraph.
 *
 * parameters: 
 *	chunk = the structure using the arcs
 *	lg = the lexemgraph using equivalent arcs
 */
void chunkerReplaceGraphemes(Chunk chunk, LexemGraph lg)
{
  List l;

  if (!chunk)
    return;

  chunk->from = findGrapheme(lg, chunk->from);
  chunk->to = findGrapheme(lg, chunk->to);
  chunk->head = findGrapheme(lg, chunk->head);
  for (l = chunk->nodes; l; l = listNext(l)) {
    listSetElement(l, findGrapheme(lg, listElement(l)));
  }

  for (l = chunk->subChunks; l; l = listNext(l)) {
    chunkerReplaceGraphemes(listElement(l), lg);
  }
}

/* ------------------------------------------------------------------------- 
 * validation command for @ref chunkerCommand.
 *
 * @param name name of the variable (chunkerCommand in our case)
 * @param value the value to be set
 * @param var the address of a possibly converted value
 * @returns TRUE on success
 *
 * Basically the chunkerCommand value is converted into @ref chunkerArgs
 * which are then used when actually forking the chunker command.
 * 
 * @see chunkerInitialize, chunkerArgs
 */
Boolean chunkerCommandValidate(String name, String value, String *var)
{
  int i;

  /* check resposibility of this validator */
  if (name != strRegister("chunkerCommand")) {
    cdgPrintf(CDG_ERROR, "ERROR: chunkerCommand callback does not handle %s\n",
	name);
    return FALSE;
  }

  /* get rid of old stuff */
  chunkerArgs = NULL;
  if (chunkerArgs) {
    for(i = 0; chunkerArgs[i]; i++) {
      cdgFreeString(chunkerArgs[i]);
    }
    memFree(chunkerArgs);
    chunkerArgs = NULL;
  }

  /* set the chunker args */
  {
    List args, l;
    String myvalue = strCopy(value); /* allocate memory */
    String start, end;

    /* loop over the chunkerCommand and list the contained arguments in args */
    args = NULL;
    for (start = myvalue; *start; start = end) {

      /* skip over whitespace */
      for (;isspace((int)*start); start++);

      /* get quoted argument */
      if (*start == '"') {
	start++;

	/* skip to next quote */
	for (end = start; *end && *end != '"'; end++);

	if (!*end) {
	  cdgPrintf(CDG_ERROR, "ERROR: unbalanced quotes in taggerCommand\n");
	  return FALSE;
	}

	/* terminate the current argument */
	*end = '\0';
	end++;

      } 
      
      /* get unquoted argument */
      else {

	/* skip over whitespace */
	for (end = start; *end && !isspace((int)*end) && *end != '"'; end++);

	if (*end == '"') {
	  cdgPrintf(CDG_ERROR, "ERROR: unbalanced quotes in taggerCommand\n");
	  return FALSE;
	}

	
	/* terminate the current argument */
	if (*end) {
	  *end = '\0';
	  end++;
	}
      }
      if (start != end) {
	args = listAppendElement(args, start);
      }
    }

    /* put all args into the chunkerArgs */
    chunkerArgs = (char **)memMalloc(sizeof(char *)*(listSize(args)+1));
    for (l = args, i = 0; l; l = listNext(l), i++) {
      chunkerArgs[i] = strRegister(listElement(l));
    }
    chunkerArgs[i] = NULL;
    listDelete(args);
    memFree(myvalue); /* allocated by strCopy */
  }

#if 0
  fprintf(stderr, "chunkerArgs = \n");
  for (i = 0; chunkerArgs[i]; i++) {
    fprintf(stderr, "%d: <%s>\n", i, chunkerArgs[i]);
  }
  fprintf(stderr, "\n");
#endif


  return TRUE;
}

/* ------------------------------------------------------------------------- 
 * initialize the chunker module.
 *
 * This is called only once by @ref cdgInitialize when the application starts
 * up.
 *
 * @see chunkerFinalize
 */
void chunkerInitialize(void) 
{
  setRegister("chunker", SET_BOOL, &chunkerUseChunker, NULL, NULL, NULL, NULL);

  setRegister("chunkerMode", SET_ENUM, &chunkerMode,
	      NULL, NULL, NULL,
	      "fake", FakeChunker,
	      "real", RealChunker, "eval", EvalChunker, NULL);

  chunkerCommand = strRegister("");
  setRegister("chunkerCommand", SET_STRING, &chunkerCommand,
	      NULL, &chunkerCommandValidate, NULL, NULL);
}

/* ------------------------------------------------------------------------- 
 * finalize the chunker module.
 * 
 * This is called by @ref cdgFinalize. (No good module without a finalizer and
 * a initializer.)
 * 
 * @see chunkerInitialize
 */
void chunkerFinalize(void) 
{
}

/* ------------------------------------------------------------------------- */
/* -- ENDOFFILE ------------------------------------------------------------ */
/** @} */
