/*
 * Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without 
 * modifications, as long as this notice is preserved.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: eval.c,v 1.112 2004/09/27 17:07:03 micha Exp $ 
 */

/* -------------------------------------------------------------------------
  @addtogroup Eval Eval - routines to evaluate constraint formulas
  @author Ingo Schroeder
  @date 7/3/97
  
  This module provides functions to evaluate constraints (structures of
  type Constraint) on LVs (structures of type LevelValue). 
  The result of evaluating a constraint is always a Boolean, but
  since constraint formulas are defined compositionally, there are also
  functions for evaluating terms of other types.

  All evaluation functions use output parameters of type Value to
  compute and return their results. ErrorValue is returned if any
  error occurred during evaluation, such as an access to a feature of the
  root node. This typically does not cause a run-time error, but simply
  makes the containing formula evaluate to FALSE. Note that
  unpredictable results can be produced if an error occurs in a nested
  term. Therefore an evaluation error always produces a CDG_WARNING
  message.
  @{
 */

#include <config.h>

#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>
#include "cdg.h"
#include "input.h"
#include "lexemgraph.h"
#include "constraintnet.h"
#include "levelvalue.h"
#include "functions.h"
#include "predicates.h"
#include "eval.h"
#include "increment.h"
#include "hook.h"
#ifndef DARWIN
#include "compile.h"
#endif
#include "timer.h"
#include "set.h"
#include "chunker.h"

/* ----------------------------------------------------------------------
 * @{
 * @name Static Strings
 * These strings are used so often that we want to pre-compute
 * their addresses. This is the only location where it's really
 * important to have full speed.
 * They are initialized in evalInitialize()
 */
String static_string_id; /**< static string @a "id" */
String static_string_word; /**< static string @a "word" */
String static_string_from; /**< static string @a "from" */
String static_string_to; /**< static string @a "to" */
String static_string_info; /**< static string @a "info" */
String static_string_chunk_start; /**< static string @a "chunk_start" */
String static_string_chunk_end; /**< static string @a "chunk_end" */
String static_string_chunk_type; /**< static string @a "chunk_type" */
/* @} */

/** This variable is set to the current Constraint during an evaluation,
   NULL otherwise. */
Constraint evalCurrentConstraint = NULL;

/** Set to the current Formula during an evaluation,
   NULL otherwise.

   Actually this is not strictly true. During evaluation, it is set to the
   formula whose evaluation was most recently *begun*, which is not at all
   the same; if we evaluate the formula
   @code
       X@cat = N & X^cat != V
   @endcode
   then the first formula (&) will be evaluated first, then the second
   formula (=) and then the third one (!=), but after that,
   evalCurrentFormula is not reset to the first formula. To trace the
   progress of evaluation exactly, we would need either LISP-style dynamic
   binding, or an extra Formula argument across the entire chain of
   evaluation. But since evalCurrentFormula is only used within predHas()
   to distinguish different has() invocations from each other, and those do
   not nest, the current behaviour is sufficient.
*/
Formula evalCurrentFormula = NULL;

/** Implements subsumesWarnings.

   If a constraint evaluation tries to access lexical information of the
   root node, the corresponding subexpression unconditionally evaluates to
   FALSE, and a warning is printed. This flag can be used to turn off the
   warnings; however the interpretation of the erroneous expression as
   FALSE remains. It is therefore recommended that you write all
   constraints so that they will not access features of the root node in
   the first place.

*/
Boolean evalSloppySubsumesWarnings = FALSE;

/** Implements evalmethod. */
EvalMethodType evalEvaluationMethod = EMTInterpreted;

/** Implements peekvaluemethod. */
EvalMethodType evalPeekValueMethod = EMTCompiled;

/** Notes whether evaluation is locked.

   Since locks can be nested, this counter can rise above 1; any value
   higher than 0 means that cached results are still valid.
 */
int lock_counter = 0;

/** The number of time points in the lexeme graph that will
     be used in the evaluations
*/
int lock_width = 0;

/** At the moment, only the results of identical has() applications are
   cached because those are expected to be particularly expensive. In
   principle, this method could be extended to arbitrary identical
   subformulas.

   The result of the application of has() number X at time point T
   is stored in cell X * lock_width + T. 2 means TRUE, 1 means FALSE, and 0
   `unknown'.
*/
int *has_cache = NULL;

/* ----------------------------------------------------------------------
   Declare evaluation lock.

   When many constraints are evaluated on the same dependency tree, it can
   be worthwhile to cache the result of common subformulas. Therefore we
   offer the caller the possibility to declare that the dependency tree will
   not change during the next evaluations. The module eval is then free to
   reuse the results of previous evaluations until the tree is unlocked
   again.

   WIDTH specifies the number of time points in the lexeme graph that will
   be used in the evaluations.
*/
void lock_tree(int width) {

  /* do not even try to cache things if there are no has() formulas */
  if(!inputCurrentGrammar->is_context_sensitive) {
    return;
  }

  if(!lock_counter) {
    lock_width = width;
    has_cache = calloc(width*inputCurrentGrammar->no_has, sizeof(int));
  }

  lock_counter++;
}

/* ----------------------------------------------------------------------
   Release evaluation lock.

   If lock_tree() was called multiple times, evaluation remains locked onto
   the current tree until each call been cancelled individually.
*/
void unlock_tree(void) {

  if(!inputCurrentGrammar->is_context_sensitive) {
    return;
  }
  
  if(lock_counter <= 0) {
    cdgPrintf(CDG_ERROR, "ERROR: invalid tree unlock request!\n");
  }
  lock_counter--;
  if(!lock_counter) {
    lock_width = 0;
    free(has_cache);
    has_cache = NULL;
  }
}


/* ----------------------------------------------------------------------
   Compare Numbers with some margin for rounding error.

   Upon removing conflicts and introducing new ones, the score of a
   structure must frequently be multiplied and divided with constants.
   Repeated use of this practice may introduce rounding errors.
   Therefore we use a special function for comparison with some margin
   for error. The problem of distinguishing rounding errors from very
   subtle penalties remains unsolved. 
 */
Boolean significantlyGreater(Number a, Number b)
{

  if (b == 0.0) {
    return a > 0.0;
  }

  return (a / b > 1.00001);
}

/* ----------------------------------------------------------------------
   Allocate a Badness.
 */
Badness bNew(int no, int hard, Number soft)
{

  Badness result = memMalloc(sizeof (BadnessStruct));

  result->no = no;
  result->hard = hard;
  result->soft = soft;

  return result;
}

/* ----------------------------------------------------------------------
   Copy a Badness.

   src must have been allocated by the caller.
 */
void bCopy(Badness dest, Badness src)
{
  *dest = *src;
}

/* ----------------------------------------------------------------------
   Create deep copy of a Badness. 

   The result is a new Badness owned by the caller.
*/
Badness bClone(Badness b)
{
  return bNew(b->no, b->hard, b->soft);
}

/* ----------------------------------------------------------------------
   Add score to a Badness.

   This function adds a classical score to a generalized score, i.e. it
   either increments b->hard or multiplies b->soft by score.

*/
Badness bAdd(Badness b, Number score)
{

  b->no++;
  if (score == 0.0) {
    b->hard++;
  } else {
    b->soft *= score;
  }

  return b;
}

/* ----------------------------------------------------------------------
   Subtract score from a Badness.
 */
Badness bSubtract(Badness b, Number score)
{

  b->no--;
  if (score == 0.0) {
    b->hard--;
  } else {
    b->soft /= score;
  }

  return b;
}

/* ----------------------------------------------------------------------
   Add Badness b to a.
 */
Badness bAddBadness(Badness a, Badness b)
{
  a->no += b->no;
  a->hard += b->hard;
  a->soft *= b->soft;

  return a;
}

/* ----------------------------------------------------------------------
   Subtract Badness b from a.
 */
Badness bSubtractBadness(Badness a, Badness b)
{
  a->no -= b->no;
  a->hard -= b->hard;
  if (b->soft != 0.0)
    a->soft /= b->soft;
  else
    a->soft = 1.0;

  return a;
}

/* ----------------------------------------------------------------------
   De-allocate a Badness.
 */
void bDelete(Badness b)
{
  memFree(b);
}

/* ----------------------------------------------------------------------
   Print a badness in canonical form.
 */
void bPrint(unsigned long mode, Badness b)
{
  if (b->no == 0) {
    cdgPrintf(mode, "*** no conflicts! ***");
    return;
  }

  cdgPrintf(mode, "(%d/%4.3e)", b->hard, b->soft);
}

/* ----------------------------------------------------------------------
   Test if two badnesses are equal.
 */
Boolean bEqual(Badness a, Badness b)
{
  return (a->hard == b->hard && a->soft == b->soft);
}

/* ----------------------------------------------------------------------
   Compare two Badnesses.

   Returns TRUE if a is properly better than b, FALSE if they are equal.
 */
Boolean bCompare(Badness a, Badness b)
{
  /* fewer hard conflicts? */
  if (a->hard < b->hard)
    return TRUE;

  /* more hard conflicts? */
  if (a->hard > b->hard)
    return FALSE;

  /* better score? */
  if (significantlyGreater(a->soft, b->soft))
    return TRUE;
  
  return FALSE;
}

/* ----------------------------------------------------------------------
   Return new Badness better than any other.

   The result is owned by the caller.
 */
Badness bestBadness(void)
{
  return bNew(0, 0, 1.0);
}

/* ----------------------------------------------------------------------
   Return new Badness worse than any other.

   Strictly speaking, this is not the worst Badness that any structure
   could ever have, but I guess it will do.

   The result is owned by the caller.
 */
Badness worstBadness(void)
{
  return bNew(INT_MAX, INT_MAX, 0.0);
}

/* ----------------------------------------------------------------------
   Evaluate a formula.

   This function interprets the Boolean formulas defined in the cdg input
   language. It is basically a comprehensive listing of actions for each
   possible value of f->type.

   Each of the different types of formula is evaluated in the expected way.
   In particular, conjunctions, implications and disjunction are guaranteed
   to short-circuit in the same way as the corresponding C operators.

   Binary relations (FTEqual etc.) are evaluated by calling evalTerm() on
   both halves of the formula. If the two resulting types are not
   compatible, FALSE is returned with a warning. In general, every type is
   only compatible with itself. The ErrorValue is not compatible with any
   type, not even with itself.

   The various relations are defined as follows:

   - Two strings are equal if they are identical (since no strings
     are duplicated in cdg, this is equivalent to checking whether they 
     have the same reading).
   - Two numbers are equal if == holds between them.
   - Two lexeme nodes are equal if they were created from the same
     grapheme node or they are both NULL. Note that this may judge two
     simultaneous lexeme nodes as equal even though they are acoustically
     incompatible. This case should not arise since no constraint should
     ever be evaluated on two incompatible LVs; however, evalFormula() does
     not check this.
   - Two underspecified values or an underspecified and a specified 
     value are never equal.
   - FTGreater holds between two numbers that satisfy C's > operator.
   - FTLess holds between two numbers that satisfy C's < operator.

   If f->type is FTConnexion, subsumesConnexion() is called to check
   whether the specification f->data.connexion.c subsumes the actual
   configuration of the LVs found in the fields
   f->data.connexion.var1->levelvalue and
   f->data.connexion.var2->levelvalue.

   If f=>type is FTDirection, subsumesDirection() is called to check
   whether the specification f->data.direction.d subsumes the Direction of
   the LevelValue structure found in the field
   f->data.direction.var->levelvalue.

*/
Boolean evalFormula(Formula f, LexemGraph lg, Vector context)
{
  ValueStruct vas, vbs;
  Value a, b;

  evalCurrentFormula = f;
  
  switch(f->type) {
  
  case FTConstant:
    return (f->data.constant);
    /* -------------------------------------------------- */
  case FTNegation:
    return (!evalFormula(f->data.formula, lg, context));
    /* -------------------------------------------------- */
  case FTPredicate:
    return (f->data.predicate.function(f->data.predicate.args, lg, context));
    /* -------------------------------------------------- */
  case FTConjunction:
    return (evalFormula(f->data.junction.arg1, lg, context) &&
	    evalFormula(f->data.junction.arg2, lg, context));
    /* -------------------------------------------------- */
  case FTDisjunction:
    return (evalFormula(f->data.junction.arg1, lg, context) ||
	    evalFormula(f->data.junction.arg2, lg, context));
    /* -------------------------------------------------- */
  case FTImplication:
    return (!evalFormula(f->data.junction.arg1, lg, context) ||
	    evalFormula(f->data.junction.arg2, lg, context));
    /* -------------------------------------------------- */
  case FTBiimplication:
    {
      Boolean b1 = evalFormula(f->data.junction.arg1, lg, context);
      Boolean b2 = evalFormula(f->data.junction.arg2, lg, context);

      return b1 == b2;
    }
    /* -------------------------------------------------- */
  case FTEqual:
    a = evalTerm(f->data.relation.op1, &vas, lg, context);
    b = evalTerm(f->data.relation.op2, &vbs, lg, context);

    if (a->type == VTError || b->type == VTError)
      return (FALSE);

    if (a->type != b->type) {
      cdgPrintf(CDG_WARNING, "WARNING: type mismatch in formula:\n");
      printFormula(CDG_WARNING, f, FTConstant, 0);
      cdgPrintf(CDG_WARNING, "\n");
      return (FALSE);
    }

    switch (a->type) {
      case VTString:
	return (a->data.string == b->data.string);
      case VTNumber:
	return (a->data.number == b->data.number);
      case VTLexemNode:{
	if (a->data.lexemnode != NULL && b->data.lexemnode != NULL &&
	    a->data.lexemnode != NONSPEC && b->data.lexemnode != NONSPEC) {
	  return
	    (((LexemNode) a->data.lexemnode)->grapheme ==
	     ((LexemNode) b->data.lexemnode)->grapheme);
	} else if (a->data.lexemnode == NULL && b->data.lexemnode == NULL)
	  return (TRUE);
	else
	  return (FALSE);
      }
      default:
	return (FALSE);
    }
    break;
    /* -------------------------------------------------- */
  case FTGreater:
    a = evalTerm(f->data.relation.op1, &vas, lg, context);
    b = evalTerm(f->data.relation.op2, &vbs, lg, context);

    if (a->type == VTError || b->type == VTError)
      return (FALSE);

    if (a->type == VTNumber && b->type == VTNumber)
      return (a->data.number > b->data.number);

    cdgPrintf(CDG_WARNING, "WARNING: type mismatch in formula:\n");
    printFormula(CDG_WARNING, f, FTConstant, 0);
    cdgPrintf(CDG_WARNING, "\n"); 
    break;
   /* -------------------------------------------------- */
  case FTLess:
    a = evalTerm(f->data.relation.op1, &vas, lg, context);
    b = evalTerm(f->data.relation.op2, &vbs, lg, context);

    if (a->type == VTError || b->type == VTError)
      return (FALSE);

    if (a->type == VTNumber && b->type == VTNumber)
      return (a->data.number < b->data.number);

    cdgPrintf(CDG_WARNING, "WARNING: type mismatch in formula:\n");
    printFormula(CDG_WARNING, f, FTConstant, 0);
    cdgPrintf(CDG_WARNING, "\n");
    break;
    /* -------------------------------------------------- */
  case FTGreaterEqual:
    a = evalTerm(f->data.relation.op1, &vas, lg, context);
    b = evalTerm(f->data.relation.op2, &vbs, lg, context);

    if (a->type == VTError || b->type == VTError)
      return (FALSE);

    if (a->type == VTNumber && b->type == VTNumber)
      return (a->data.number >= b->data.number);

    cdgPrintf(CDG_WARNING, "WARNING: type mismatch in formula:\n");
    printFormula(CDG_WARNING, f, FTConstant, 0);
    cdgPrintf(CDG_WARNING, "\n");
    break;
    /* -------------------------------------------------- */
  case FTLessEqual:
    a = evalTerm(f->data.relation.op1, &vas, lg, context);
    b = evalTerm(f->data.relation.op2, &vbs, lg, context);

    if (a->type == VTError || b->type == VTError)
      return (FALSE);

    if (a->type == VTNumber && b->type == VTNumber)
      return (a->data.number <= b->data.number);

    cdgPrintf(CDG_WARNING, "WARNING: type mismatch in formula:\n");
    printFormula(CDG_WARNING, f, FTConstant, 0);
    cdgPrintf(CDG_WARNING, "\n");
    break;
    /* -------------------------------------------------- */
  case FTNotEqual:
    a = evalTerm(f->data.relation.op1, &vas, lg, context);
    b = evalTerm(f->data.relation.op2, &vbs, lg, context);

    if (a->type == VTError || b->type == VTError)
      return (FALSE);

    if (a->type == VTNumber && b->type == VTNumber)
      return (a->data.number != b->data.number);
    else if (a->type == VTString && b->type == VTString)
      return (a->data.string != b->data.string);
    else if (a->type == VTLexemNode && b->type == VTLexemNode) {
      if (a->data.lexemnode == NULL && b->data.lexemnode != NULL)
	return (TRUE);
      if (a->data.lexemnode != NULL && b->data.lexemnode == NULL)
	return (TRUE);
      if (a->data.lexemnode == NULL && b->data.lexemnode == NULL)
	return (FALSE);
      if (a->data.lexemnode == NONSPEC || b->data.lexemnode == NONSPEC)
	return (TRUE);
      return (((LexemNode) a->data.lexemnode)->grapheme !=
	      ((LexemNode) b->data.lexemnode)->grapheme);
    }

    cdgPrintf(CDG_WARNING, "WARNING: type mismatch in formula:\n");
    printFormula(CDG_WARNING, f, FTConstant, 0);
    cdgPrintf(CDG_WARNING, "\n");
    break;
    /* -------------------------------------------------- */
  case FTConnexion:
    return (subsumesConnexion(f->data.connexion.c,
			      getCon(f->data.connexion.var1->levelvalue,
				     f->data.connexion.var2->levelvalue)));
    /* -------------------------------------------------- */
  case FTDirection:
    return (subsumesDirection(f->data.direction.d,
			      getDir(f->data.direction.var->levelvalue)));
  }
  return (FALSE);
}

/* ----------------------------------------------------------------------
 * returns value for a path in a feature matrix
 */
Value peekValue(Value val, List path)
{

  Value v;

  /* anchor */
  if (path == NULL) {
    return val;
  } else {

    /* recursion */
    if (val->type == VTConjunction) {
      String att = (String) listElement(path);
      List l;

      for (l = val->data.list; l != NULL; l = listNext(l)) {
	v = (Value) listElement(l);
	if (v->type != VTAVNode) {
	  cdgPrintf(CDG_WARNING, "WARNING: value ");
	  printValue(CDG_WARNING, v, 0);
	  cdgPrintf(CDG_WARNING, " in value ");
	  printValue(CDG_WARNING, val, 0);
	  cdgPrintf(CDG_WARNING,
		    " should be attribute/value pair, but isn't!\n\n");
	  return ErrorValue;
	}

	if (((AVNode) v->data.avnode)->attribute == att) {
	  return peekValue(((AVNode) v->data.avnode)->value, listNext(path));
	}
      }
      return ErrorValue;
    } else {
      cdgPrintf(CDG_WARNING,
		"WARNING: can't find any path in something not a conjunction\n");
      return ErrorValue;
    }
  }
}

/* ----------------------------------------------------------------------------
   evaluate a CDG Term.

   @param t the term to be evaluated 
   @param val temporary storage where the result is stored in
   @param lg the current lexemgraph
   @param context the current parse tree (or NULL)
   @returns a pointer to @c val or the ErrorValue

   This function evaluates a term on a value of the cdg formula language,
   computing both its type and its value. Both are returned as the result of
   evalTerm(). The function performs an exhaustive case distinction over the
   possible values of @c t->type:

   - TTTopPeek: The field @c t->data.peek.varinfo->levelvalue->modifiee is
     checked. If the modifiee is underspecified, ErrorValue is returned,
     otherwise peekValue() is applied to the path @c t->data.peek.path in the
     value of the modifiee.\n
     \n
     Exceptions occur if @c t->data.peek.path is @c id, @c word, @c from or @c to. 
     These special attributes are notated like features of a lexeme node, but
     are actually attributes of the lexeme node, and not the lexeme itself.
     These attributes are computed as follows:
     - the attribute id is computed by setting @c val->lexemnode
       to the modifiee itself. If the modifiee is underspecified,
       ErrorValue is returned, VTLexemNode otherwise.
     - the attribute word is computed by setting
       @c val->data.string to the field word of the lexical entry of the
       modifier. (Strictly speaking, this *is* an attribute of the lexeme
       rather than the lexeme node; but as it does not appear in the
       feature matrix, it cannot be computed by peekValue().) If the
       modifiee is underspecified, ErrorValue is returned, @c val otherwise.
     - the attribute from is computed by setting @c val->data.number 
       to the field from of the modifiee. If the modifiee is underspecified,
       @c lg->max+1 is assigned instead. In any case, @c val is returned.
     - the attribute to is computed by setting
       @c val->data.number to the field to of the modifiee. If the
       modifiee is underspecified, the number @c lg->max+2 is assigned
       instead. In any case, @c val is returned. \n
     \n
   - TTBottomPeek: The same algorithm is applied to the
     modifier of the LevelValue -- simplified by the fact that modifiers cannot
     be underspecified. 
   - TTLabel: @c val->data.string is set to the field label of the LV and @c val is returned.
   - TTLevel: @c val->data.string is set to the identifier of the Level of LV,
     and @c val is returned.
   - TTAdd, TTSubtract, TTMultiply, TTDivide: The two fields
     @c t->data.operation.op1 and @c t->data.operation.op1 are evaluated by
     evalTerm(). ErrorValue is returned unless both results are numbers. 
     Dividing by @c 0 also returns ErrorValue. Otherwise the results are
     conjoined by the corresponding C operator (+, -, * oder /),
     @c val->data.number is set to the result of this computation, and VTNumber
     is returned.
     applied to t->data.function.args (and lg) and both the resulting type
     and the computed are passed to the caller of evalTerm().
   - TTString: @c val->data.string is set to @c t->data.string, and @c val is returned.
   - TTNumber: @c val->data.number is set to @c t->data.number, and @c val is returned.

*/
Value evalTerm(Term t, Value val, LexemGraph lg, Vector context)
{
  /* Variables for add, sub, ... */
  ValueStruct a, b;
  Value va, vb;

  /* Variables for TTTopPeek, BottomPeek */
  LexemNode ln;
  LevelValue lv;
  List l;

  switch (t->type) {
    case TTTopPeek:{
      lv = (LevelValue) t->data.peek.varinfo->levelvalue;
      l = t->data.peek.path;
      ln = (LexemNode) lv->modifiee;

      if (l) {
	/* special handling of special attribute */
	String s = (String) listElement(l);

	if (!listNext(l)) {

	  /* special handling of `id' attribute */
	  if (s == static_string_id) {
	    val->data.lexemnode = (struct LexemNodeStruct *)ln;
	    val->type = VTLexemNode;
	    return val;
	  }

	  if (ln == NONSPEC) {
	    /* Modifiee not specified */
	    if (s == static_string_from) {
	      val->data.number = lg->max + 1;
	      val->type = VTNumber;
	      return val;
	    }
	    if (s == static_string_to) {
	      val->data.number = lg->max + 2;
	      val->type = VTNumber;
	      return val;
	    }
            return ErrorValue;
	  }

	  if (!ln) {
	    return ErrorValue;
	  }
	  /* special handling of `word' attribute */
	  if (s == static_string_word) {
	    val->data.string = ln->lexem->word;
	    val->type = VTString;
	    return val;
	  }
	  /* special handling of `from' attribute */
	  if (s == static_string_from) {
	    val->data.number = (Number) ln->arc->from;
	    val->type = VTNumber;
	    return val;
	  }
	  /* special handling of `to' attribute */
	  if (s == static_string_to) {
	    val->data.number = (Number) ln->arc->to;
	    val->type = VTNumber;
	    return val;
	  }
	  /* special handling of `chunk_start' attribute */
	  if (s == static_string_chunk_start) {
	    Chunk chunk = ln->grapheme->chunk;
	    if (!chunk) 
	      return ErrorValue;
	    val->data.number = chunk->from->arc->from;
	    val->type = VTNumber;
	    return val;
	  }
	  /* special handling of `chunk_end' attribute */
	  if (s == static_string_chunk_end) {
	    Chunk chunk = ln->grapheme->chunk;
	    if (!chunk) 
	      return ErrorValue;
	    val->data.number = chunk->to->arc->to;
	    val->type = VTNumber;
	    return val;
	  }
	  /* special handling of `chunk_type' attribute */
	  if (s == static_string_chunk_type) {
	    Chunk chunk = ln->grapheme->chunk;
	    if (!chunk) 
	      return ErrorValue;
	    val->data.string = chunkerStringOfChunkType(chunk);
	    val->type = VTString;
	    return val;
	  }
	} else {
	  /* `info' attribute must have additional attributes */
	  if (s == static_string_info) {
	    /* peek value from infostring */
	    if (!ln || !ln->arc->info) {
	      return ErrorValue;
	    }
	    return peekValue(ln->arc->info, listNext(l));
	  }
	}
      }

      if (!ln || ln == NONSPEC) {
	return ErrorValue;
      }

      switch (evalPeekValueMethod) {
	case EMTCompiled:
	  return (ln->lexem->values[t->data.peek.pathindex]);
	case EMTInterpreted:
	  return peekValue(ln->lexem->value, l);
	default:
	  cdgPrintf(CDG_WARNING, "WARNING: invalid evalpeekvalue method\n");
	  return ErrorValue;
      }
    }
    case TTBottomPeek:{
      lv = (LevelValue) t->data.peek.varinfo->levelvalue;
      l = t->data.peek.path;
      ln = lv->modifier;

      /* special handling of special attribute */
      if (l) {
	String s = (String) listElement(l);

	if (!listNext(l)) {
	  /* special handling of `id' attribute */
	  if (s == static_string_id) {
	    val->data.lexemnode = (struct LexemNodeStruct *)ln;
	    val->type = VTLexemNode;
	    return val;
	  }

	  /* special handling of `word' attribute */
	  if (s == static_string_word) {
	    val->data.string = ln->lexem->word;
	    val->type = VTString;
	    return val;
	  }
	  /* special handling of `from' attribute */
	  if (s == static_string_from) {
	    val->data.number = (Number) ln->arc->from;
	    val->type = VTNumber;
	    return val;
	  }
	  /* special handling of `to' attribute */
	  if (s == static_string_to) {
	    val->data.number = (Number) ln->arc->to;
	    val->type = VTNumber;
	    return val;
	  }
	  /* special handling of `chunk_start' attribute */
	  if (s == static_string_chunk_start) {
	    Chunk chunk = ln->grapheme->chunk;
	    if (!chunk) {
	      return ErrorValue;
	    } else {
	      val->data.number = (Number) chunk->from->arc->from;
	      val->type = VTNumber;
	      return val;
	    }
	  }
	  /* special handling of `chunk_end' attribute */
	  if (s == static_string_chunk_end) {
	    Chunk chunk = ln->grapheme->chunk;
	    if (!chunk) {
	      return ErrorValue;
	    } else {
	      val->data.number = (Number) chunk->to->arc->to;
	      val->type = VTNumber;
	      return val;
	    }
	  }
	  /* special handling of `chunk_type' attribute */
	  if (s == static_string_chunk_type) {
	    Chunk chunk = ln->grapheme->chunk;
	    if (!chunk) 
	      return ErrorValue;
	    val->data.string = chunkerStringOfChunkType(chunk);
	    val->type = VTString;
	    return val;
	  }
	} else {
	  /* `info' attribute must have additional attributes */
	  if (s == static_string_info) {
	    /* peek value from infostring */
	    if (!ln->arc->info) {
	      return ErrorValue;
	    }
	    return peekValue(ln->arc->info, listNext(l));
	  }
	}
      }

      switch (evalPeekValueMethod) {
	case EMTCompiled:
	  return (ln->lexem->values[t->data.peek.pathindex]);
	case EMTInterpreted:
	  return peekValue(ln->lexem->value, l);
	default:
	  cdgPrintf(CDG_WARNING, "WARNING: invalid evalpeekvalue method\n");
	  return ErrorValue;
      }
    }
      /* -------------------------------------------------- */
    case TTLabel:
      val->data.string = ((LevelValue) t->data.label->levelvalue)->label;
      val->type = VTString;
      return val;
      /* -------------------------------------------------- */
    case TTLevel:
      val->data.string = ((LevelValue) t->data.level->levelvalue)->level->id;
      val->type = VTString;
      return val;
      /* -------------------------------------------------- */
    case TTAdd:
      if ((va = evalTerm(t->data.operation.op1, &a, lg, context))->type != VTNumber ||
	  (vb = evalTerm(t->data.operation.op2, &b, lg, context))->type != VTNumber) {
	cdgPrintf(CDG_WARNING, "WARNING: type mismatch in term:\n");
	printTerm(CDG_WARNING, t, FALSE);
	return ErrorValue;
      } else {
	val->data.number = va->data.number + vb->data.number;
	val->type = VTNumber;
      }
      return val;
      /* -------------------------------------------------- */
    case TTSubtract:
      if ((va = evalTerm(t->data.operation.op1, &a, lg, context))->type != VTNumber ||
	  (vb = evalTerm(t->data.operation.op2, &b, lg, context))->type != VTNumber) {
	cdgPrintf(CDG_WARNING, "WARNING: type mismatch in term:\n");
	printTerm(CDG_WARNING, t, FALSE);
	return ErrorValue;
      } else {
	val->data.number = va->data.number - vb->data.number;
	val->type = VTNumber;
      }
      return val;
      /* -------------------------------------------------- */
    case TTMultiply:
      if ((va = evalTerm(t->data.operation.op1, &a, lg, context))->type != VTNumber ||
	  (vb = evalTerm(t->data.operation.op2, &b, lg, context))->type != VTNumber) {
	cdgPrintf(CDG_WARNING, "WARNING: type mismatch in term:\n");
	printTerm(CDG_WARNING, t, FALSE);
	return ErrorValue;
      } else {
	val->data.number = va->data.number * vb->data.number;
	val->type = VTNumber;
      }
      return val;
      /* -------------------------------------------------- */
    case TTDivide:
      if ((va = evalTerm(t->data.operation.op1, &a, lg, context))->type != VTNumber ||
	  (vb = evalTerm(t->data.operation.op2, &b, lg, context))->type != VTNumber) {
	cdgPrintf(CDG_WARNING, "WARNING: type mismatch in term:\n");
	printTerm(CDG_WARNING, t, FALSE);
	return ErrorValue;
      } else if (vb->data.number == 0.0) {
	cdgPrintf(CDG_WARNING, "WARNING: division by zero in term:\n");
	printTerm(CDG_WARNING, t, FALSE);
	return ErrorValue;
      } else {
	val->data.number = va->data.number / vb->data.number;
	val->type = VTNumber;
      }
      return val;
      /* -------------------------------------------------- */
    case TTFunction:
      return (Value) t->data.function.function(t->data.function.args, val, lg, context);
      /* -------------------------------------------------- */
    case TTString:
      val->data.string = t->data.string;
      val->type = VTString;
      return val;
      /* -------------------------------------------------- */
    case TTNumber:
      val->data.number = t->data.number;
      val->type = VTNumber;
      return val;
    default:
      return ErrorValue;
  }
}

/* ----------------------------------------------------------------------
   Evaluate a constraint w.r.t. the specified LVs.

   Here's the detailed algorithm:
   
   - check the number of arguments and returns TRUE (with a warning) if
     it does not match the arity of c
   - instantiate the variables in c->vars with the LVs passed 
     as arguments
   - do *not* check whether c is applicable to these LVs. Applying
     constraints to arbitrary LVs can produce spurious constraint
     violations without warning! Hence evalConstraint() should only be
     called on constraints whose applicability was previously checked by
     match{Unary,Binary}Signature().
   - set evalCurrentConstraint and evaluates c->formula using evalFormula()
   - reset evalCurrentConstraint
   - if the result was FALSE and c has a variable penalty
     (if c->penaltyTerm is non-NULL), set c->penalty to evalTerm(c->penaltyTerm)
   - execute the hook HOOK_EVAL
   - de-instantiate c->vars
   - increase either evalUnary or evalBinary
   - return the result of evalFormula()

  The net is actually only used to record statistics (the fields evalUnary
  and evalBinary) and can be NULL.

  The context may be NULL. If it is not NULL it should be a Vector of LVs,
  each of which must be located at its index as calculated by lvIndex(). 
  This is then passed on to predicates and functions that need to examine
  the context of an LV to operate.

  If the constraint c uses such a function or predicate (that is, if it has
  its is_context_sensitive flag set) and there is no context, then
  evalConstraint() returns TRUE. This is obviously not always correct, so
  if you do not provide context to your evaluations and still use
  context-sensitive constraints, you can miss some conflicts.

  This function can evaluate constraints of arbitrary arity. The number of
  LVs passed must equal the arity. Unary and binary constraints can also be
  evaluated by evalUnaryConstraint() and evalBinaryConstraint(). Those
  functions are slightly faster than using the general vararg mechanism.

*/
Boolean evalConstraint(Constraint c, ConstraintNet net, Vector context, ...)
{
  va_list ap;
  LevelValue lv;
  VarInfo vi;
  List l;
  Boolean result = TRUE;

  Value val;
  ValueStruct vas;

  LexemGraph lg;
  int counter = 0;
  LevelValue lva = NULL, lvb = NULL;

  /* if we have no context, don't even try to find a conflict. */
  if(c->is_context_sensitive && !context) {
    return TRUE;
  }
  
  /* instantiate constraint variables */
  l = c->vars;
  va_start(ap, context);
  while ((lv = va_arg(ap, LevelValue)) != NULL && l != NULL) {
    vi = (VarInfo) listElement(l);
    vi->levelvalue = lv;

    counter++;
    if (counter == 1)
      lva = lv;
    if (counter == 2)
      lvb = lv;
    l = listNext(l);
  }
  va_end(ap);

  lg = lva->modifier->lexemgraph;
  result = evalFormula(c->formula, lg, context);

  if (result == FALSE && c->penaltyTerm != NULL) {
    /* determine actual penalty */
    val = evalTerm(c->penaltyTerm, &vas, lg, context);
    if (val->type != VTNumber) {
      cdgPrintf(CDG_WARNING,
		"WARNING: penalty term of `%s' doesn't evaluate to a number\n",
		c->id);
      c->penalty = 0.0;
    } else {
      c->penalty = val->data.number;
    }
  }
  evalCurrentConstraint = NULL;

  if ((hkVerbosity & CDG_HOOK) && !result)
    cdgExecHook(HOOK_EVAL, c);

#if 0
  /* de-instantiate constraint variables */
  for (l = c->vars; l != NULL; l = listNext(l))
    ((VarInfo) listElement(l))->levelvalue = NULL;
#endif

  /* count */
  if (net) {
    if (counter == 1)
      net->evalUnary++;
    else
      net->evalBinary++;
  }

  return (result);
}

/* ----------------------------------------------------------------------
 * Evaluate a unary constraint on LV.

   This function evaluates a unary constraint. It has the same effect as
   evalConstraint() used with four arguments, but saves the overhead of
   accessing the variable argument list. Also, it will call a compiled
   constraint function rather than evalConstraint() if possible.

 */
Boolean evalUnaryConstraint(Constraint c, ConstraintNet net,
                            Vector context, LevelValue lv)
{
  VarInfo vi;
  Boolean result;

  Value val;
  ValueStruct vas;

  LexemGraph lg;

  /* if we have no context, don't even try to find a conflict. */
  if(c->is_context_sensitive && !context) {
    return TRUE;
  }

  /* instantiate constraint variable */
  vi = (VarInfo) listElement(c->vars);
  vi->levelvalue = lv;

  lg = lv->modifier->lexemgraph;
  evalCurrentConstraint = c;

  /* eval compiled */
  if (evalEvaluationMethod == EMTCompiled) {
    result = (*c->cfunction) (lg, context, lv);
  }

  /* eval interpreted */
  else {

    result = evalFormula(c->formula, lg, context);
    if (c->penaltyTerm && !result) {
      /* determine actual penalty */
      val = evalTerm(c->penaltyTerm, &vas, lg, context);
      if (val->type != VTNumber) {
	cdgPrintf(CDG_WARNING,
		  "WARNING: penalty term of `%s' doesn't evaluate to a number\n",
		  c->id);
	c->penalty = 0.0;
      } else {
	c->penalty = val->data.number;
      }
    }
  }
  evalCurrentConstraint = NULL;

  if ((hkVerbosity & CDG_HOOK) && !result) {
    cdgExecHook(HOOK_EVAL, c);
  }

  if (net) {
    net->evalUnary++;
  }

  return (result);
}

/* ------------------------------------------------------------
   Evaluate one binary constraint on two LVs.

   This function evaluates a binary constraint. It has the same effect as
   evalConstraint() used with five arguments, but saves the overhead of
   accessing the variable argument list. Also, it will call a compiled
   constraint function rather than evalConstraint() if possible.
*/
Boolean evalBinaryConstraint(Constraint c, ConstraintNet net,
                             Vector context,                      
			     LevelValue lva, LevelValue lvb)
{
  VarInfo vi;
  Boolean result;

  Value val;
  ValueStruct vas;

  LexemGraph lg;

  /* if we have no context, don't even try to find a conflict. */
  if(c->is_context_sensitive && !context) {
    return TRUE;
  }

  /* instantiate constraint variables */
  vi = (VarInfo) listElement(c->vars);
  vi->levelvalue = lva;

  vi = (VarInfo) listElement(listNext(c->vars));
  vi->levelvalue = lvb;

  evalCurrentConstraint = c;
  lg = lva->modifier->lexemgraph;

  /* eval compiled */
  if (evalEvaluationMethod == EMTCompiled) {
    result = (*c->cfunction) (lg, context, lva, lvb);
  }

  /* eval interpreted */
  else {
    result = evalFormula(c->formula, lg, context);
    if (c->penaltyTerm && !result) {
      /* determine actual penalty */
      val = evalTerm(c->penaltyTerm, &vas, lg, context);
      if (val->type != VTNumber) {
	cdgPrintf(CDG_WARNING,
		  "WARNING: penalty term of `%s' doesn't evaluate to a number\n",
		  c->id);
	c->penalty = 0.0;
      } else {
	c->penalty = val->data.number;
      }
    }
  }
  evalCurrentConstraint = NULL;

  if ((hkVerbosity & CDG_HOOK) && !result) {
    cdgExecHook(HOOK_EVAL, c);
  }

  if (net) {
    net->evalBinary++;
  }

  return (result);
}

/* ------------------------------------------------------------
   Eval unary constraints on LV.

   This function applies all unary constraints to lv. It uses the Vector
   cdgConstraintVector in much the same way as evalBinary() uses
   cdgConstraintMatrix. The same exceptions for deactivated constraints
   apply as in that function. In addition, evalUnary() stores all violated
   local constraints in the List lv->constraints.

*/
Number evalUnary(LevelValue lv, ConstraintNet net,
                 Vector context, Boolean use_cs_only,
                 Badness b, List *conflicts)
{
  List l;
  Number score = 1.0;
  Constraint c;
  int index;

  /* evalUnary() can operate in fast mode, calculating only the cumulative
     score, or in thorough mode, calculating the score, the Badness and the
     List of all conflicts found. The thorough mode is requested by passing
     a valid pointer as the third argument. */
  Boolean thorough = (b != NULL);

  /* grant a cleared violations-list */
  if (!context) {
    /* the lv->constraints only contain local unary constraints,
       no context-sensitive ones */
    listDelete(lv->constraints);
    lv->constraints = NULL;
  }

  index = inputUCIndex(lv->level->no, getDir(lv));
  for (l = inputCurrentGrammar->hashedUCs[index]; l != NULL; l = listNext(l)) {
    c = (Constraint) listElement(l);

    /* correct arity and activated */
    if (listSize(c->vars) != 1 || !c->section->active || !c->active)
      continue;

    if(use_cs_only && !c->is_context_sensitive) {
      continue;
    }
    if (!evalUnaryConstraint(c, net, context, lv)) {
#ifdef DEBUGEVALUNARY
      cdgPrintf(CDG_DEBUG, "\nDEBUG: constraint `%s' failed.\n", c->id);
#endif
      if (!context) {
	/* add a local unary violation to the list of violated constraints */
	lv->constraints = listPrependElement(lv->constraints, c);
      }
      if (!thorough && c->penalty == 0.0) {
	return (0.0);
      }
      score *= c->penalty;
      if (thorough) {
	bAdd(b, c->penalty);
	if (conflicts) {
	  *conflicts =
	    listPrependElement(*conflicts, cvNew(c, lv, NULL));
	}
      }
    }
  }
#ifdef DEBUGEVALUNARY
  cdgPrintf(CDG_DEBUG, "OK.\n");
#endif
  return (score);
}


/* ----------------------------------------------------------------------
   Evaluate all constraints on two LVs.

   This function performs a conceptually simple task: it computes the
   combined score of two LVs in a constraint net and returns the computed
   score. This is done by applying evalConstraint() to all known
   constraints and tallying the scores of all violated constraints.

   For efficiency reasons, the actual implementation of this function is
   much more complicated:

   - The entire function can be run in fast mode or in thorough mode. In
     fast mode, the function will stop applying constraints if a constraint
     with score 0 is violated, and return 0 immediately. In thorough mode,
     it will keep applying all possible constraints and bAdd() the results
     to the Badness b. Also, ConstraintViolation structures will be
     allocated and inserted into the List conflicts. Thorough mode is
     requested by passing a non-null value for b. For fast mode, both b and
     conflicts should be NULL.
   - In fast mode, if the LVs lva and lvb have already been
     evaluated by this function and scUseCache was set during both
     invocations of evalBinary(), the result is read from net->cache and
     not computed again.
   - To make this case more probable, the LVs lva and lvb are always sorted by
     the number of their respective levels. (That is, only half of the
     entire cache is ever used.)
   - Even if no precomputed result exists, not all constraints may be
     evaluated:

     Unary constraints are ignored completely.

     Inactive constraints and constraints from inactive sections are
     skipped likewise.

     Of the remaining constraints, only those that may actually fail are
     evaluated. Hence not all constraints are evaluated, but only those
     whose signature matches the actual configuration existing between
     lva and lvb. The List of these constraints is found in the
     matrix cdgConstraintMatrix[].
     
     In fast mode, if a hard constraint is violated, the result 0.0 is
     returned immediately.
   
     If the parameter context is NULL, context-sensitive constraints will
     not be evaluated.

     If the parameter use_cs_only is set, ONLY context-sensitive constraints
     will be evaluated. See evalInContext() for why this is occasionally
     necessary.
     
     - Theoretically, a binary constraint may be violated twice by two
     LVs. Think of a constraint that forbids the label X to appear more
     than once in an analysis. If two LVs with label X are evaluated
     jointly, both (a,b) and (b,a) violate this condition. 
     But since the two violations do not really indicate different errors,
     this kind of double fault is usually not desired.

     A double fault can only occur if the LVs match the signature of a
     constraint in both variants of instantiation. Since every binary
     constraint signature must explicitly state the two levels it refers
     to, this is only possible if both LVs belong to the same level and
     either their configuration is symmetrical, or it is is asymmetrical,
     but the signature of the constraint subsumes both configurations.

     To eliminate all double faults, evalBinary() checks for all
     of these possibilities explicitly.

   - If a value was computed rather than looked up, and
     scUseCache is set, it is registered in net->cache.

   The parameter net can be NULL, so that LVs can be evaluated even in the
   absence of a constraint net. In this case no caching is possible.

 */
Number evalBinary(LevelValue lva, LevelValue lvb,
                  ConstraintNet net, Vector context,
                  Boolean use_cs_only,
		  Badness b, List *conflicts)
{
  List l;
  Number score = 1.0;
  Constraint c;
  int index;
  int connexion;
  Direction dir1, dir2;
  List violated = NULL;
  Boolean thorough = (b != NULL);
  Boolean foundContextSensitive = FALSE;

  /* force lv order */
  if (lva->indexWRTNet > lvb->indexWRTNet) {
    LevelValue lv = lvb;

    lvb = lva;
    lva = lv;
  }

  /* lookup scorecache */
  if (!thorough && scUseCache && net) {
    score = scGetScore(net->cache, lva, lvb);
    if (score != -1.0)
      return (score);
    else
      score = 1.0;
  }

  dir1 = getDir(lva);
  connexion = getCon(lva, lvb);
  dir2 = getDir(lvb);

  index = inputBCIndex(lva->level->no, lvb->level->no, dir1, connexion, dir2);
#ifdef DEBUG_EVALBINARY
  cdgPrintf(CDG_DEBUG, "DEBUG: looking up cell %d {X", index);
  printDirection(CDG_DEBUG, dir1);
  cdgPrintf(CDG_DEBUG, "%s", lva->level->id);
  printConnexion(CDG_DEBUG, connexion);
  cdgPrintf(CDG_DEBUG, "Y");
  printDirection(CDG_DEBUG, dir2);
  cdgPrintf(CDG_DEBUG, "%s}", lvb->level->id);
  cdgPrintf(CDG_DEBUG, " (%d constraints)\n",
	    listSize(inputCurrentGrammar->hashedBCs[index]));
#endif

  lock_tree(lva->modifier->lexemgraph->max);

  /* simplest case: lva and lvb belong to the same level,
     and their configuration is symmetrical. That means
     that there is only one class of constraints that they can violate. */

  if (lva->level->no == lvb->level->no &&
      connexion == inverseConnexion(connexion) && dir1 == dir2) {

    /* give each constraint in the list two chances to fail */
    for (l = inputCurrentGrammar->hashedBCs[index]; l != NULL; l = listNext(l)) {
      c = (Constraint) listElement(l);
      /* activated? */
      if (!c->section->active || !c->active) {
#ifdef DEBUG_EVALBINARY
	cdgPrintf(CDG_DEBUG, "DEBUG: skipping constraint %s\n", c->id);
#endif
	continue;
      }
      if (c->is_context_sensitive)
	foundContextSensitive = TRUE;
      else if(use_cs_only)
        continue;
      if (!evalBinaryConstraint(c, net, context, lva, lvb)
	  || !evalBinaryConstraint(c, net, context, lvb, lva)) {
	if (!thorough && c->penalty == 0.0) {
#ifdef DEBUG_EVALBINARY
	  cdgPrintf(CDG_DEBUG, "DEBUG: bailing out!\n");
#endif
	  score = 0.0;
	  goto end;
	}
	score *= c->penalty;
	if (thorough) {
	  bAdd(b, c->penalty);
	  if (conflicts) {
	    *conflicts =
	      listPrependElement(*conflicts, cvNew(c, lva, lvb));
	  }
	}
      }
    }
  }

  /* general case: there are two classes of constraints that lva and lvb
     may violate, e.g. {X\SYN/Y\SEM} and {X\SEM\Y\SYN}. */
  else {

    /* evaluate each constraint from either of two lists,
       but without double faults */
    violated = NULL;

    /* first list */

    for (l = inputCurrentGrammar->hashedBCs[index]; l != NULL; l = listNext(l)) {
      c = (Constraint) listElement(l);
      /* activated? */
      if (!c->section->active || !c->active) {
#ifdef DEBUG_EVALBINARY
	cdgPrintf(CDG_DEBUG, "DEBUG: skipping constraint %s\n", c->id);
#endif
	continue;
      }
      if (c->is_context_sensitive)
	foundContextSensitive = TRUE;
      else if(use_cs_only)
        continue;
      if (!evalBinaryConstraint(c, net, context, lva, lvb)) {
	violated = listPrependElement(violated, c);
	c->active = FALSE;
	if (!thorough && c->penalty == 0.0) {
#ifdef DEBUG_EVALBINARY
	  cdgPrintf(CDG_DEBUG, "DEBUG: bailing out!\n");
#endif
	  score = 0.0;
	  goto end;
	}
	score *= c->penalty;
	if (thorough) {
	  bAdd(b, c->penalty);
	  if (conflicts) {
	    *conflicts =
	      listPrependElement(*conflicts, cvNew(c, lva, lvb));
	  }
	}
      }
    }

    index = inputBCIndex(lvb->level->no,
			 lva->level->no,
			 dir2, inverseConnexion(connexion), dir1);
#ifdef DEBUG_EVALBINARY
    cdgPrintf(CDG_DEBUG, "DEBUG: looking up cell %d {X", index);
    printDirection(CDG_DEBUG, dir2);
    cdgPrintf(CDG_DEBUG, "%s", lvb->level->id);
    printConnexion(CDG_DEBUG, inverseConnexion(connexion));
    cdgPrintf(CDG_DEBUG, "Y");
    printDirection(CDG_DEBUG, dir1);
    cdgPrintf(CDG_DEBUG, "%s}", lva->level->id);
    cdgPrintf(CDG_DEBUG, " (%d constraints)\n",
	      listSize(inputCurrentGrammar->hashedBCs[index]));
#endif

    /* second list */
    for (l = inputCurrentGrammar->hashedBCs[index]; l != NULL; l = listNext(l)) {
      c = (Constraint) listElement(l);

      /* activated? */
      if (!c->section->active || !c->active) {
#ifdef DEBUG_EVALBINARY
	cdgPrintf(CDG_DEBUG, "DEBUG: skipping constraint %s\n", c->id);
#endif
	continue;
      }
      if (c->is_context_sensitive)
	foundContextSensitive = TRUE;
      else if(use_cs_only)
        continue;
      if (!evalBinaryConstraint(c, net, context, lvb, lva)) {
	if (!thorough && c->penalty == 0.0) {
#ifdef DEBUG_EVALBINARY
	  cdgPrintf(CDG_DEBUG, "DEBUG: bailing out!\n");
#endif
	  score = 0.0;
	  goto end;
	}
	score *= c->penalty;
	if (thorough) {
	  bAdd(b, c->penalty);
	  if (conflicts) {
	    *conflicts =
	      listPrependElement(*conflicts, cvNew(c, lva, lvb));
	  }
	}
      }
    }
  }

end:
  unlock_tree();
  for (l = violated; l; l = listNext(l)) {
    c = listElement(l);
    c->active = TRUE;
  }
  listDelete(violated);

  /* store computed score in scorecache */
  if (scUseCache && net && !foundContextSensitive) {
    scSetScore(net->cache, lva, lvb, score);
  }

  return (score);
}


/* ----------------------------------------------------------------------
   Eval context-sensitive constraints only.

   This function evaluates all LVs in the Vector LVs against each other,
   using only context-sensitive constraints, and assuming that the context
   is context.

   This is useful for methods that cannot normally provide context, e.g. 
   because they operate incrementally. Such methods should use normal
   evalUnary and evalBinary calls with no context while they are building a
   structure, and then this function once the structure is complete.

   Note that LVs and context can be different; one frobbing method actually
   needs to eval a subset of a structure, but in the global context.
*/
Number evalInContext(Vector LVs, Vector context, Badness b, List* conflicts) {
  int i,j;
  LevelValue lva, lvb;
  Number result = 1.0;
  
  /* Don't pay for what you don't use */
  if(!inputCurrentGrammar->is_context_sensitive) {
    return 1.0;
  }
  
  for(i = 0; i < vectorSize(LVs); i++) {
    lva = vectorElement(LVs, i);
    if(!lva) continue;
    result *= evalUnary(lva, NULL, context, TRUE, b, conflicts);
    for(j = i+1; j < vectorSize(LVs); j++) {
      lvb = vectorElement(LVs, j);
      if(!lvb) continue;
      result *= evalBinary(lva, lvb, NULL, context, TRUE, b, conflicts);
    }
  }

  return result;
}


/* ----------------------------------------------------------------------
 * Initialize the eval module
 */
void evalInitialize(void)
{
  static_string_id = strRegister("id");
  static_string_word = strRegister("word");
  static_string_from = strRegister("from");
  static_string_to = strRegister("to");
  static_string_info = strRegister("info");
  static_string_chunk_start = strRegister("chunk_start");
  static_string_chunk_end = strRegister("chunk_end");
  static_string_chunk_type = strRegister("chunk_type");

  setRegister("subsumesWarnings", SET_BOOL, &evalSloppySubsumesWarnings, NULL,
	      NULL, NULL, "sloppy", TRUE, "full", FALSE, NULL);
  setRegister("evalmethod", SET_ENUM, &evalEvaluationMethod, NULL,
	      (BooleanFunction *) evalValidateEvalMethod, NULL, "compiled",
	      EMTCompiled, "interpreted", EMTInterpreted, NULL);
  setRegister("peekvaluemethod", SET_ENUM, &evalPeekValueMethod, NULL, NULL,
	      NULL, "compiled", EMTCompiled, "interpreted", EMTInterpreted,
	      NULL);
}

/* ----------------------------------------------------------------------
 * finalize the eval module
 */
void evalFinalize(void)
{
  cdgFreeString(static_string_id);
  cdgFreeString(static_string_word);
  cdgFreeString(static_string_from);
  cdgFreeString(static_string_to);
  cdgFreeString(static_string_info);
  cdgFreeString(static_string_chunk_start);
  cdgFreeString(static_string_chunk_end);
  cdgFreeString(static_string_chunk_type);
}

/* ----------------------------------------------------------------------
   This is the validation function for the CDG variable evalmethod.
 */
Boolean evalValidateEvalMethod(String name, String value, int *var)
{
  Constraint c;
  HashIterator hi;
  if (strcmp(name, "evalmethod") == 0) {
    if (strcmp(value, "compiled") == 0) {

      if (*var != EMTCompiled) {
	return FALSE;
      }

      /* Check whether there are constraints at all. */
      if (!inputCurrentGrammar->constraints) {
	cdgPrintf(CDG_WARNING, "WARNING: no constraints loaded, "
		  "refusing to switch to compiled evaluation\n");
	return FALSE;
      }
      /* Check whether all constraints have a valid cfunction. */
      hi = hashIteratorNew(inputCurrentGrammar->constraints);
      while((c = hashIteratorNextValue(hi))) {
	if (!c->cfunction) {
	  cdgPrintf(CDG_WARNING, "WARNING: invalid function for `%s', "
		    "refusing to switch to compiled evaluation\n", c->id);
	  return FALSE;
	}
      }
      hashIteratorDelete(hi);
      return TRUE;
    } else if (strcmp(value, "interpreted") == 0) {
      return *var == EMTInterpreted;
    }
  }
  return FALSE;
}

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/*@}*/
