/*
 * File: functions.c
 * Author: Ingo Schroeder
 * Purpose: CDG functions
 * Birth: 17/3/97
 *
 * $Id: functions.c,v 1.50 2004/09/27 17:07:04 micha Exp $
 */

/* -- INCLUDES ---------------------------------------------------------- */
#include <config.h>

#include <stdio.h>
#include <string.h>
#include <blah.h>
#include <math.h>
#include "cdg.h"
#include "lexemgraph.h"
#include "functions.h"
#include "increment.h"
#include "hook.h"

/* -- MACROS ------------------------------------------------------------ */

/* -- FORWARD DECLARATIONS ---------------------------------------------- */

/* -- VARIABLES --------------------------------------------------------- */
CDGFunctionStruct eval_functions[] = {
  {"abs", funcAbs, FALSE, "1",
   "funcAbs", "returns the absolute value"},
  {"acoustics", funcAcoustics, FALSE, "1",
   "funcAcoustics", "returns normalized acoustic score"},
  {"exp", funcExp, FALSE, "1",
    "funcExp", "returns the value of e raised to the power of x"},
  {"distance", funcDistance, FALSE, "2",
   "funcDistance", "returns the distance between the lexem nodes"},
  {"match", funcMatch, FALSE, "3", "funcMatch", "read the manual"},
  {"max", funcMax, FALSE, ">0", "funcMax", "returns the maximum"},
  {"min", funcMin, FALSE, ">0", "funcMin", "returns the minimum"},
  {"pts", funcPts, FALSE, "1", "funcPts", "returns tag score"},
  {"realcat", funcCat, FALSE, "1", "peeks at the annotated tag"},
  {"lookup", funcLookup, FALSE, ">0", "funcLookup" "holds user-defined data"},
  {"quotes", funcQuotes, FALSE, "1", "funcQuotes", 
   "return level of quotation of a word"},
  {"parens", funcParens, FALSE, "1", "funcParens", 
   "return level of parentheses of a word"},
  {"phrasequotes", funcPhrasequotes, FALSE, "1", "funcPhrasequotes", 
   "return level of phrase quotation of a word"},
  {"parent", funcParent, TRUE, "1", "funcParent", 
   "return parent of a word"},
  {"height", funcHeight, TRUE, "1", "funcHeight", 
   "return height of subtree under a word"},



  {(String) NULL, (VTFunction *) NULL, FALSE, (String) NULL}
};

/* -- FUNCTIONS --------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * return a pointer to the function implementing a functor
 */
CDGFunction funcGetFunction(String functor)
{
  int i;

  for (i = 0; eval_functions[i].functor != NULL; i++) {
    if (strcmp(functor, eval_functions[i].functor) == 0) {
      return &eval_functions[i];
      break;
    }
  }

  return (CDGFunction)NULL;
}

/* ----------------------------------------------------------------------
 * return the name of the function implementing a functor
 */
String funcGetFunctionName(String functor)
{
  int i;

  for (i = 0; eval_functions[i].functor != NULL; i++) {
    if (strcmp(functor, eval_functions[i].functor) == 0) {
      return eval_functions[i].functionName;
      break;
    }
  }

  return NULL;
}


/* ----------------------------------------------------------------------
 * returns distance between lexem nodes
 */
Value funcDistance(List args, Value val, LexemGraph lg, Vector context)
{
   Value a, b;
   ValueStruct vsa, vsb;


   a = evalTerm((Term) listElement(args), &vsa, lg, context);
   b = evalTerm((Term) listElement(listNext(args)), &vsb, lg, context);

   if (a->type != VTLexemNode || b->type != VTLexemNode) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: arguments of function `distance' not all ids in `%s'\n",
	 evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
      return ErrorValue;
   }
   if (a->data.lexemnode == NULL || b->data.lexemnode == NULL) {
      val->data.number = 0.0;
   } else {
     val->data.number = (Number)lgDistanceOfNodes(lg, 
                                                  (LexemNode)a->data.lexemnode, 
                                                  (LexemNode)b->data.lexemnode);
   }
   val->type = VTNumber;
   return val;
}

/* ----------------------------------------------------------------------
 * returns absolute value of its argument
 */
Value funcAbs(List args, Value val, LexemGraph lg, Vector context)
{
  Value v;

   v = evalTerm((Term) listElement(args), val, lg, context);
   switch (v->type) {
      case VTNumber:
        if (v->data.number < 0.0) {
	    v->data.number = -v->data.number;
        }
        break;
      default:
	 cdgPrintf(CDG_WARNING,
	    "WARNING: argument of function `abs' not a number in `%s'\n",
	    evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
	 return ErrorValue;
   }

   return v;
}
/* ----------------------------------------------------------------------
 * returns the value of e raised to the power of x
 */
Value funcExp(List args, Value val, LexemGraph lg, Vector context)
{
  Value v;

  v = evalTerm((Term) listElement(args), val, lg, context);
  switch (v->type) {
    case VTNumber:
      v->data.number = exp(v->data.number);
      break;
    default:
      cdgPrintf(CDG_WARNING,
		"WARNING: argument of function `exp' not a number in `%s'\n",
		evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return ErrorValue;
  }

  return v;
}

/* ----------------------------------------------------------------------
 * returns the minimum
 */
Value funcMin(List args, Value val, LexemGraph lg, Vector context)
{
  ValueStruct vsa;
   Value v;
   Number minValue=0.0;
   List l;


   l = args;
   if (l != NULL && listElement(l) != NULL) {
     v = evalTerm((Term) listElement(l), &vsa, lg, context);
     if (v->type == VTNumber) {
       minValue = v->data.number;
     } else {
       cdgPrintf(CDG_WARNING,
                 "WARNING: argument of function `min' not a number in `%s'\n",
		 evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
       return ErrorValue;
     }    
     l = listNext(l);
   }

   for (;
      l != NULL && listElement(l) != NULL;
      l = listNext(l)) {

      v = evalTerm((Term) listElement(l), &vsa, lg, context);
      if (v->type == VTNumber) {
	       minValue = min(minValue, v->data.number);
      } else {
        cdgPrintf(CDG_WARNING,
                  "WARNING: argument of function `min' not a number in `%s'\n",
	 	  evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
        return ErrorValue;
      }
   }
   val->data.number = minValue;
   val->type = VTNumber;

   return val;
}

/* ----------------------------------------------------------------------
 * returns the maximum
 */
Value funcMax(List args, Value val, LexemGraph lg, Vector context)
{
  ValueStruct vsa;
   Value v;
   Number maxValue=0.0;
   List l;

   l = args;
   if (l != NULL && listElement(l) != NULL) {
     v = evalTerm((Term) listElement(l), &vsa, lg, context);
     if (v->type == VTNumber) {
       maxValue = v->data.number;
     } else {
       cdgPrintf(CDG_WARNING,
                 "WARNING: argument of function `max' not a number in `%s'\n",
		 evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
       return ErrorValue;
     }    
     l = listNext(l);
   }

   for (;
      l != NULL && listElement(l) != NULL;
      l = listNext(l)) {

      v = evalTerm((Term) listElement(l), &vsa, lg, context);
      if (v->type == VTNumber) {
	       maxValue = max(maxValue, v->data.number);
      } else {
        cdgPrintf(CDG_WARNING,
                  "WARNING: argument of function `max' not a number in `%s'\n",
		  evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
        return ErrorValue;
      }
   }
   val->data.number = maxValue;
   val->type = VTNumber;

   return val;
}

/* ----------------------------------------------------------------------
 * Returns the number corresponding to the first sort that subsumes 
 * the additional argument.
 *
 * Function assumes well-formedness of VTList argument 
 * (it's enforced in parser.y):
 *
 *   VTString VTNumber VYString VTNumber ...
 */
Value funcMatch( List args, Value val, LexemGraph lg, Vector context)
{
  ValueStruct vsa, vsb, vsh;
  Value va, vb, vh;
  Hierarchy h;
  int a, b;
  List l;
  

  vh = evalTerm((Term) listNthElement(args, 1), &vsh, lg, context);
  if ( vh->type != VTString ) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type mismatch (1st arg) for `match' in `%s'\n",
	      evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return ErrorValue;
  }
  
  va = evalTerm((Term) listNthElement(args, 2), &vsa, lg, context);
  if ( va->type != VTString && va->type != VTList ) {
    if (!evalSloppySubsumesWarnings)
      cdgPrintf(CDG_WARNING,
		"WARNING: type mismatch (2nd arg) for `match' in `%s'\n",
		evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return ErrorValue;
  }

  vb = evalTerm((Term) listNthElement(args, 3), &vsb, lg, context);
  if (vb->type != VTString) {
    if (!evalSloppySubsumesWarnings)
      cdgPrintf(CDG_WARNING,
		"WARNING: type mismatch (3rd arg) for `match' in `%s'\n",
		evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return ErrorValue;
  }
  
  /* get hierarchy */
  h = findHierarchy(vh->data.string);
  if (!h) {
    cdgPrintf(CDG_WARNING,
              "WARNING: hierarchy `%s' not found in `%s'\n", vh->data.string,
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    return ErrorValue;
  }
  
   /* find second sort */
   if ( 0 > ( b = (int)hashGet( h->types, vb->data.string ) - 1 ) ) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: type `%s' not found in hierarchy `%s' in `%s'\n",
	 vb->data.string, vh->data.string,
	 evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
      return ErrorValue;
   }

   if ( va->type == VTString ) {
     /* This is a MATCH( STRING, STRING, STRING) call. */
     if ( 0 > ( a = (int)hashGet( h->types, va->data.string ) - 1) ) {
       cdgPrintf(CDG_WARNING,
		 "WARNING: type `%s' not found in hierarchy `%s' in `%s'\n",
		 va->data.string, vh->data.string,
		 evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
       return ErrorValue;
     }
     if ( (int)arrayElement( h->distance, a, b ) >= 0 ) {
       val->data.number = 1.0;
       val->type = VTNumber;
       return val;
     }     
   } else {
     /* This is a MATCH( STRING, LIST, STRING) call. */

     /* Loop through list */
     for ( l = va->data.list; l != NULL; l = listNext(listNext(l)) ) {
       /* find first sort */
       if ( 0 > ( a = (int)hashGet( h->types, ((Value)listElement(l))->data.string ) - 1 ) ) {
	 cdgPrintf(CDG_WARNING,
		   "WARNING: type `%s' not found in hierarchy `%s' in `%s'\n",
		   ((Value)listElement(l))->data.string, vh->data.string,
		   evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
	 return ErrorValue;
       }
       if ( (int)arrayElement( h->distance, a, b ) >= 0 ) {
	 return (Value)listElement(listNext(l));
       }     
     }
   }
   val->data.number = 0.0;
   val->type = VTNumber;
   return val;     
}

/* ----------------------------------------------------------------------
 * Returns the normalized acoustic score.
 */
Value funcAcoustics(List args, Value val, LexemGraph lg, Vector context)
{
   Value a;
   ValueStruct vsa;
   LexemNode ln;


   a=evalTerm((Term)listElement(args), &vsa, lg, context);

   if (a->type!=VTLexemNode) {
     cdgPrintf(CDG_WARNING,
	       "WARNING: argument of function `acoustics' not an id in `%s'\n",
	       evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
     return ErrorValue;
   }

   ln=(LexemNode)a->data.lexemnode;
   
   if (!ln || !ln->arc) {
     cdgPrintf(CDG_WARNING,
	       "WARNING: argument of function `acoustics' not complete in `%s'\n",
	       evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
     return ErrorValue;
   }

   val->data.number=ln->arc->penalty;
   val->type=VTNumber;
   return val;
}

/* ----------------------------------------------------------------------
 * Returns annotated POS tag.
 */
Value funcCat(List args, Value val, LexemGraph lg, Vector context)
{
   Value a;
   ValueStruct vsa;
   LexemNode node;


   a = evalTerm((Term) listElement(args), &vsa, lg, context);

   if (a->type != VTLexemNode) {
      cdgPrintf(CDG_WARNING,
	 "WARNING: argument of function `realcat' not an id in `%s'\n",
	 evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
      return ErrorValue;
   }

   node = (LexemNode) a->data.lexemnode;
   if (node == NULL || node == NONSPEC) {
     cdgPrintf(CDG_WARNING, 
	 "WARNING: argument of function `realcat' is %s\n",
	 (node)?"nil": "nonspecific");
     return ErrorValue;
   }

   val->type = VTString;
   val->data.string = (String) vectorElement(lg->tags, node->arc->from);

   return val;
}

/* ----------------------------------------------------------------------
 * Returns the tag score.
 */
Value funcPts(List args, Value val, LexemGraph lg, Vector context)
{
   Value a;
   ValueStruct vsa;
   LexemNode ln;


   a=evalTerm((Term)listElement(args), &vsa, lg, context);

   if (a->type!=VTLexemNode) {
     cdgPrintf(CDG_WARNING,
	       "WARNING: argument of function `pts' not an id in `%s'\n",
	       evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
     return ErrorValue;
   }

   ln=(LexemNode)a->data.lexemnode;
   
   if (!ln) {
     cdgPrintf(CDG_WARNING,
	       "WARNING: argument of function `pts' not complete in `%s'\n",
	       evalCurrentConstraint ? evalCurrentConstraint->id: "*UNKNOWN*");
     return ErrorValue;
   }

   val->data.number=ln->tagscore;
   val->type=VTNumber;
   return val;
}

/* ----------------------------------------------------------------------
   Returns user-defined data.

   The first argument must be the name of a defined map. The second
   argument is a key into this map, and the corresponding Value is
   returned. 
*/
Value funcLookup(List args, Value val, LexemGraph lg, Vector context) {
  ValueStruct vsm;
  Value vm;
  String key = NULL;
  Value result;
  Hashtable h;
  List l;
  int counter;
  String constraint =
    evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*";
  String sep = strRegister(":");
  

  vm = evalTerm(listElement(args), &vsm, lg, context);
  if ( vm->type != VTString ) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: type mismatch (1st arg) for `lookup' in `%s'\n",
              constraint);
    return ErrorValue;
  }
  
  /* find map */
  h = hashGet(inputCurrentGrammar->maps, vm->data.string);

  if (h) {
    /* construct key */
    counter = 1;
    for(l = listNext(args); l != NULL; l = listNext(l)) {
      Term t = listElement(l);
      ValueStruct vs;
      Value v = evalTerm(t, &vs, lg, context);
      counter++;
      if (v->type != VTString) {
	cdgPrintf(CDG_WARNING,
		  "WARNING: type mismatch (arg %d) for `lookup' in `%s'\n",
		  counter, constraint);
	return ErrorValue;
      }    

      key = key ? strAppend(key, sep, v->data.string, NULL) : v->data.string;
    }

    if((result = hashGet(h, key))) {
      return result;
    }
  } else {
    cdgPrintf(CDG_WARNING,
              "WARNING: map `%s' not found in `%s'\n",
              vm->data.string, constraint);
  }
    
  /* return the empty string */
  val->type = VTString;
  val->data.string = strRegister("");
  return val;
  
}


/* ----------------------------------------------------------------------
   How many levels of sentence quotation does the word occupy?
 */
Value funcQuotes(List l, Value val, LexemGraph lg, Vector context) {
  ValueStruct vas;
  Value a;
  LexemNode ln;
  a = evalTerm((Term) listElement(l), &vas, lg, context);
  if (a->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `quotes' not a lexeme node\n");
    return ErrorValue;
  }
  ln = (LexemNode)a->data.lexemnode;
  if (!ln || NONSPEC == ln) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `quotes' not a lexeme node\n");
    return ErrorValue;
  }
  val->data.number = ln->arc->quotes;
  val->type = VTNumber;
  return val;  
}

/* ----------------------------------------------------------------------
   How many levels of parentheses does the word occupy?
 */
Value funcParens(List l, Value val, LexemGraph lg, Vector context) {
  ValueStruct vas;
  Value a;
  LexemNode ln;
  a = evalTerm((Term) listElement(l), &vas, lg, context);
  if (a->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `parens' not a lexeme node\n");
    return ErrorValue;
  }
  ln = (LexemNode)a->data.lexemnode;
  if (!ln || NONSPEC == ln) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `parens' not a lexeme node\n");
    return ErrorValue;
  }
  val->data.number = ln->arc->parens;
  val->type = VTNumber;
  return val;  
}

/* ----------------------------------------------------------------------
   How many levels of phrase quotation does the word occupy?
 */
Value funcPhrasequotes(List l, Value val, LexemGraph lg, Vector context) {
  ValueStruct vas;
  Value a;
  LexemNode ln;
  a = evalTerm((Term) listElement(l), &vas, lg, context);
  if (a->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `phrasequotes' not a lexeme node in constraint `%s'\n", evalCurrentConstraint->id);
    return ErrorValue;
  }
  ln = (LexemNode)a->data.lexemnode;
  if (!ln || NONSPEC == ln) {
    cdgPrintf(CDG_WARNING,
	      "WARNING: argument of predicate `phrasequotes' not a lexeme node in constraint `%s'\n", evalCurrentConstraint->id);
    return ErrorValue;
  }
  val->data.number = ln->arc->phraseQuotes;
  val->type = VTNumber;
  return val;  
}


/* ----------------------------------------------------------------------
   Return the word above a given word.
*/
Value funcParent(List args, Value val, LexemGraph lg, Vector context) {
  Value v;
  int a;
  LexemNode ln;
  LevelValue lv;
  VarInfo vi = listElement(evalCurrentConstraint->vars);
  int nrlevels = inputCurrentGrammar->noOfLevels;
  
  v = evalTerm((Term) listElement(args), val, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (1st arg) for `parent' in `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return ErrorValue;
  }
  ln = (LexemNode) val->data.lexemnode;
  if (!ln || ln == NONSPEC) {
    return ErrorValue;
  }
  a = ln->arc->from;
  
#ifdef DEBUGPARENT
  cdgPrintf(CDG_DEBUG, "parent(%s) = ", ln->arc->word);
#endif

  lv = vectorElement(context, nrlevels * a + vi->level->no);

#ifdef DEBUGPARENT
  if(!lv || !spec(lv->modifiee)) {
    cdgPrintf(CDG_DEBUG, "NIL.\n");
  } else {
  cdgPrintf(CDG_DEBUG, "%s.\n", lv->modifiee->arc->word);
  }
#endif
    
  v->data.lexemnode = (struct LexemNodeStruct *)lv->modifiee;
  v->type = VTLexemNode;
  return v;  

}

/* ----------------------------------------------------------------------
   Return the maximal length of dependency chains under a given word.
*/
Value funcHeight(List args, Value val, LexemGraph lg, Vector context) {
  Value v;
  LexemNode ln;
  Term t = listElement(args);
  Level level;
  
  v = evalTerm(t , val, lg, context);
  if (val->type != VTLexemNode) {
    cdgPrintf(CDG_WARNING,
              "WARNING: type mismatch (1st arg) for `height' in `%s'\n",
              evalCurrentConstraint ? evalCurrentConstraint->id : "*UNKNOWN*");
    printTypeError(CDG_WARNING, val, VTLexemNode);
    return ErrorValue;
  }
  ln = (LexemNode) val->data.lexemnode;
  if (!ln || ln == NONSPEC) {
    return ErrorValue;
  }

  /* deduce level to search */
  if(NULL == (level = findLevel(t->data.peek.varinfo->levelname))) {
    cdgPrintf(CDG_ERROR, "ERROR: level `%s' not found!\n");
    return ErrorValue;
  }
  
  v->data.number = funcHeightImpl(ln, level->no, lg, context);
  v->type = VTNumber;
  return v;  
}

/* ----------------------------------------------------------------------
   Perform the work of funcHeight().
*/
int funcHeightImpl(LexemNode ln, int levelno, LexemGraph lg, Vector context) {
  int i;
  LevelValue lv;
  int nrlevels = inputCurrentGrammar->noOfLevels;
  int result = 0;
  
  for(i = lg->min; i <= lg->max; i++) {
    int index = nrlevels * i + levelno;
    int height = 0;
    if(index >= vectorSize(context)) break;
    
    while(1) {
      height++;
      if(height > lg->max) break;
      lv = vectorElement(context, index);
      if(!lv || !spec(lv->modifiee)) break;
      if(lv->modifiee == ln) {
        if(height > result) {
          result = height;
        }
        break;
      }
      index = nrlevels * lv->modifiee->arc->from + levelno;
    }
  }
  return result;
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
