/* 
 * Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without 
 * modifications, as long as this notice is preserved.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Birth: 1.9.98
 *
 */

/* -------------------------------------------------------------------------
 * @addtogroup ISearch ISearch - An incremental best first search
 * \author Michael Schulz (see also AUTHORS and THANKS for more)
 *
 * $Id: increment.c,v 1.72 2004/09/27 17:07:04 micha Exp $
 * 
 * @{
 */

/* -- INCLUDES ---------------------------------------------------------- */
#include <config.h>

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <blah.h>
#include "cdg.h"
#include "command.h"
#include "input.h"
#include "eval.h"
#include "statistics.h"
#include "levelvalue.h"
#include "constraintnet.h"
#include "lexemgraph.h"
#include "netsearch.h"
#include "increment.h"
#include "timer.h"
#include "hook.h"
#include "parse.h"
#include "set.h"

/* -- MACROS ------------------------------------------------------------ */
#define MAX_SOLUTIONS 50
#define MAX_SEARCHRESULTS 7

/* -- FUNCTIONS --------------------------------------------------------- */
static Boolean incIsNonSpecCompatibleFormula(Formula, Boolean *);
static Boolean incIsNonSpecCompatibleTerm(Term);
static int incDeleteNonSpecValuesFromNet(ConstraintNet);
static Boolean incHasNonSpecParses(ConstraintNet);
static int incEndOfSentenceComputing(ConstraintNet);
static void incPrintParses(List);
static Boolean incAddArc(ConstraintNet, Arc, Number);
static void incRegisterParses(ConstraintNet, List);

/* ----------------------------------------------------------------------
 * check all parses in the net.
 * This returns true if there's a parse using nonspec values.
 */
Boolean incHasNonSpecParses(ConstraintNet net)
{
  List l;

  for (l = net->parses; l; l = listNext(l)) {
    if(parseIsUnderspecified(listElement(l))) {
      return TRUE;
    }
  }

  return FALSE;
}

/* ----------------------------------------------------------------------
 * recompute scores of nonspec-levelvalues.
 * This function recomputes the score of nonspec levelvalues as they
 * might be judged differently in every incremental step. If the new
 * score drops below a \c threshold then the levelvalue is
 * marked deleted.
 * @param net the constraint net whose nodes are examined
 * @returns the number of deleted levelvalues
 */
int incRecomputeNonspecs(ConstraintNet net, Number threshold)
{
  ConstraintNode node;
  int i, j, n;
  LevelValue lv;

  n = 0;
  /* loop thru all nodes of the net */
  for (i = 0; i < vectorSize(net->nodes); i++) {
    node = (ConstraintNode) vectorElement(net->nodes, i);

    /* loop thru all old levelvalues */
    for (j = 0; j < vectorSize(node->values); j++) {
      lv = (LevelValue) vectorElement(node->values, j);

      /* normal and deleted values are skipped */
      if (lv->modifiee != NONSPEC || lv->isDeleted)
	continue;
      
      lv->score = evalUnary(lv, net, NULL, FALSE, NULL, NULL);
      if (lv->score <= threshold) {
	lv->isDeleted = TRUE;
	node->noValidValues--;
	n++;
      }
    }
  }

  return n;
}

/* ----------------------------------------------------------------------
 * delete nonspec-levelvalues.
 * This function iterates over all domains of all nodes in the \c net
 * and marks levelvalues using nonspec modifiees as deleted.
 * @returns the number of deleted levelvalues
 */
int incDeleteNonSpecValuesFromNet(ConstraintNet net)
{
  int i,j, n;
  LevelValue lv;
  ConstraintNode node;

  n = 0;
  /* loop thru all nodes ... */
  for (i = 0; i < vectorSize(net->nodes); i++) {
    node = (ConstraintNode)vectorElement(net->nodes, i);

    /* ... and look at their domains */
    for (j = 0; j < vectorSize(node->values); j++) {
      lv = (LevelValue) (vectorElement(node->values, j));
      if (lv->modifiee == NONSPEC && !lv->isDeleted) {
	lv->score = 0.0;
	lv->isDeleted = TRUE;
	n++;
	node->noValidValues--;
      }
    }
  }

  return n;
}

/* ----------------------------------------------------------------------
 * incEndOfSentenceComputing:
 */
int incEndOfSentenceComputing(ConstraintNet net)
{
  int n, i, j;
  ConstraintNode node;
  List l;
  LevelValue lv;
  Parse p;
  int noValues, deletedValues;

  cdgPrintf(CDG_DEBUG, "DEBUG: end-of-sentence-computing\n");

  /* 
   * delete values which are unwanted in a final solution 
   */
  n = incDeleteNonSpecValuesFromNet(net);
  cdgPrintf(CDG_DEBUG, "DEBUG: deleted %d nonspec-value(s)\n", n);

  /* 
   * freeze good bindings
   */
  deletedValues = noValues = 0;
  if (!incHasNonSpecParses(net)) {

    /* delete all levelvalues of the net */
    for (i = 0; i < vectorSize(net->nodes); i++) {
      node = (ConstraintNode) vectorElement(net->nodes, i);
      node->noValidValues = 0;

      for (j = 0; j < vectorSize(node->values); j++) {
	lv = (LevelValue) vectorElement(node->values, j);
	if (!lv->isDeleted) {
	  lv->isDeleted = TRUE;
	  deletedValues++;
	}
      }
    }

    /* undelete levelvalues which are used in solutions */
    for (l = net->parses; l != NULL; l = listNext(l)) {
      p = listElement(l);
      for(i = 0; i < vectorSize(p->LVs); i++) {
        lv = vectorElement(p->LVs, i);
        lv->isDeleted = FALSE;
        cnFindNode(net, lv)->noValidValues++;
	noValues++;
      }
    }
  }

  if (deletedValues != noValues) {
    cdgPrintf(CDG_DEBUG, "DEBUG: deleted %d unused values, %d remaining values\n",
	      deletedValues, noValues);
  }

  return n;
}


/* ----------------------------------------------------------------------
 * adds an arc to an existing constraintnet.
 */
Boolean incAddArc(ConstraintNet net, Arc arc, Number threshold)
{
#define DEBIG_INCADDARC
  GraphemNode new_gn;
  int oldSize = vectorSize(net->lexemgraph->graphemnodes);

  /* build new graphemnode from arc and insert into lexemgraph */
  cdgPrintf(CDG_DEBUG, "DEBUG: building graphemnode\n");
  if (!lgNewIter(net->lexemgraph, arc)) {
    cdgPrintf(CDG_WARNING, "WARNING: no lexical entry for `%s'\n", arc->word);
    return FALSE;
  }

  /* set to NULL to force lgComputeDistances() to reallocate */
  memFree(net->lexemgraph->distance);
  net->lexemgraph->distance = NULL;
  if (!lgNewFinal(net->lexemgraph)) {
    return FALSE;
  }
  new_gn = vectorElement(net->lexemgraph->graphemnodes, oldSize);

  /* build new lexemnodes from graphemnode and inset into lexemgraph */
  cdgPrintf(CDG_DEBUG, "DEBUG: updating lexemgraph\n");

  /* build new constraint-nodes from graphemnode */
  cdgPrintf(CDG_DEBUG, "DEBUG: building constraintnodes \n");
  cnBuildIter(net, new_gn, TRUE);

#if 1
  {
  }
#endif

  incRecomputeNonspecs(net, threshold);
  cnBuildFinal(net, TRUE);

  return TRUE;
}

/* ----------------------------------------------------------------------------
 * print list of parses.
 */
static void incPrintParses(List parses)
{
  if (!parses) {
    cdgPrintf(CDG_WARNING, "WARNING: no parses found\n");
    return;
  }

  cdgPrintf(CDG_INFO, "INFO: %d parse(s) with score %4.3e:\n",
	    listSize(parses), ((Parse) listElement(parses))->score);

  if (hkVerbosity & CDG_SEARCHRESULT) {
    List l;
    int no = 0;

    for (l = parses; l && no < MAX_SEARCHRESULTS; l = listNext(l)) {
      no++;
      parsePrint(listElement(parses));
      cdgPrintf(CDG_INFO, "\n\n");
    }
    if (l) {
      cdgPrintf(CDG_INFO, "INFO: %d more parse(s) available\n", listSize(l));
    }
  }
}

/* ----------------------------------------------------------------------------
 * register list of parses to the constraint net.
 * see nsRegisterParses()
 */
static void incRegisterParses(ConstraintNet net, List parses)
{
  List l;
  Parse p;
  int i = 0;
  
  /* get rid of all old parses */
  listDelete(net->parses); /* the parses themselves are registered globally */
  net->parses = NULL;

  /* add parses */
  for (l = parses; l != NULL; l = listNext(l)) {
    p = (Parse)listElement(l);
    if (i > MAX_SEARCHRESULTS) {
      cdgPrintf(CDG_WARNING, "WARNING: suppressing registration of parse %s\n", p->id);
      parses = listDeleteElement(parses, p);
      parseDelete(p);
      continue;
    }
      
    p->solutionNo = ++i;
    p->noOfSolutions = listSize(parses);
    p->optimal = FALSE;
      
    /* can't fail here, the grammar is definitely present */
    parseEval(p);
    net->parses = listPrependElement(net->parses, p);

    /* add to global List */
    parseRegister(p);
    cdgPrintf(CDG_INFO, "INFO: parse %s added to the global parses list.\n", p->id);
  }      
}

/* ----------------------------------------------------------------------------
 * incremental parse a wordgraph for solutions.
 * This is the entry function to the increment module. provided a wordgraph `lattice',
 * an agenda size 'agSize', the agenda threshold `agThreshold' and a threshold
 * `nonSpecThreshold' a constraint net is built and solved incrementally.
 *
 * nonSpecThreshold determines the minimum score a levelvalue using a NONSPEC 
 * lexem can have before being deleted.
 *
 * returns a new constraint net containing the final solution 
 * (without any NONSPECs)
 * 
 */
ConstraintNet incSearch(Lattice lattice, int agSize, Number agThreshold, Number nonSpecThreshold)
{
   ConstraintNet net;
   int noParses;
   Timer profileTime = timerNew();
   Timer loopTime = timerNew();
   int evalUnaries = 0, statUnaries = 0, evalBinaries = 0;
   List l, parses;
   Arc arc;
   Boolean oldCnUseNonspec = cnUseNonSpec;

   /* switch to nonspec mode */
   cnUseNonSpec = TRUE;
   
   /* build constraint net in steps */
   net = cnBuildInit(); 
   net->lexemgraph = lgNewInit();
   net->lexemgraph->lattice = lattice;
   net->nodes = vectorNew(10);
   net->values = vectorNew(10);
   
   /* MAIN-LOOP ------------------------------------------------------------ */
   for (l = lattice->arcs; l; l = listNext(l)) {
     arc = (Arc) listElement(l);
           
      evalUnaries = net->evalUnary;
      evalBinaries = net->evalBinary;
      statUnaries = net->statUnary;

      /* reset cdgCtrlCTrapped for every segment */
      cdgCtrlCTrapped = FALSE;

      /* read chunk of input */
      timerStart(loopTime);
      timerStart(profileTime);
      cdgPrintf(CDG_DEBUG, "DEBUG: *** reading word '%s' (%02d-%02d)\n", 
	  arc->word, arc->from, arc->to);
      incAddArc(net, arc, nonSpecThreshold);
      
      cdgPrintf(CDG_PROFILE, "PROFILE: updated network in %lums\n", 
	  timerElapsed(profileTime));

      /* compute bestStatesSoFar of expanded net */
      cdgPrintf(CDG_DEBUG, "DEBUG: searching\n");
      parses = nsSearch(net, STBranchBound, agSize, agThreshold);

      /* test for a sentence end (no nonspec parses) */
#if 0
      if (!incHasNonSpecParses(net)) {
	cdgPrintf(CDG_INFO, "INFO: possible end of sentence !!!\n");
	incEndOfSentenceComputing(net);
      }
#endif

      /* end of segment ---------------------------------------------------*/

      /* print information about ongoing analysis */
      cdgPrintf(CDG_PROFILE, 
	 "PROFILE: looptime: %lums\n", timerElapsed(loopTime));
      noParses = listSize(parses);
      incRegisterParses(net, parses);
      incPrintParses(parses);
      listDelete(parses);

      /* protection from unwanted solution cascade -> emergency exit */
      if (noParses > MAX_SOLUTIONS) {
	 cdgPrintf(CDG_ERROR, 
		   "ERROR: too many solutions, analysis aborted\n");
	 cnDelete(net);
	 timerFree(loopTime);
	 timerFree(profileTime);
	 return NULL;
      }
   }
   /* end of MAIN-LOOP */

   /* 
    * now that we know the sentence is over 
    * delete unused bindings and
    * search once again the final solutions if needed
    */
   cdgPrintf(CDG_INFO, "INFO: end of sentence!!!\n");
   incEndOfSentenceComputing(net);

   /* last not least once again in case of previous NONSPEC-solutions */
   if (incHasNonSpecParses(net)) {
      cdgPrintf(CDG_DEBUG, "INFO: searching\n");
      parses = nsSearch(net, STBranchBound, agSize, agThreshold);

      /* lets see what we have */
      incPrintParses(parses);
      incRegisterParses(net, parses);
   }

   timerFree(loopTime);
   timerFree(profileTime);

   /* switch to the mode we had before */
   cnUseNonSpec = oldCnUseNonspec;
   
   return (net);
}

/* ----------------------------------------------------------------------
  nonspec analysis of the current constraint grammar.
  This function determines a list of constraints which
  do not guard a possible nonspec reference and thus might cause
  a logical error. This is done by descending the constraint formula
  and the constraint penalty term recursively the way the constraint
  interpreter does.
 
  @returns list of observed constraints.
 */
List incGetNonSpecIncompatibles(void)
{
   List result;
   Constraint c;
   HashIterator hi;
   
   result = NULL;
   hi = hashIteratorNew(inputCurrentGrammar->constraints);
   while((c = hashIteratorNextValue(hi))) {
     if (!incIsNonSpecCompatibleFormula(c->formula, NULL) ||
         (c->penaltyTerm && !incIsNonSpecCompatibleTerm(c->penaltyTerm))) {
       result = listAppendElement(result, c);
     }
   }
   hashIteratorDelete(hi);
   return result;
}

/* ----------------------------------------------------------------------
 * incIsNonSpecCompatibleFormula:
 *	scan if formula is NONSPEC-compatible
 */
Boolean incIsNonSpecCompatibleFormula(Formula formula, Boolean *found)
{
   List l;
   Boolean result;
   static Boolean thisFoundNonSpec;
   Boolean *foundNonSpec;

   if (found == NULL) {
      thisFoundNonSpec = FALSE;
      foundNonSpec = &thisFoundNonSpec;
   } else {
      foundNonSpec = found;
   }

   if (*foundNonSpec || formula->type == FTConstant)
      return TRUE;

   if (formula->type == FTNegation)
      return incIsNonSpecCompatibleFormula(formula->data.formula, foundNonSpec);

   if (formula->type == FTPredicate) {
      if (strcmp(formula->data.predicate.functor, "nonspec") == 0 ||
	 strcmp(formula->data.predicate.functor, "spec") == 0) {
	 *foundNonSpec = TRUE;	/* if these predicates are found then
				 * the whole formula is considered compatible 
				 */
	 return TRUE;
      }
      result = TRUE;
      for (l = formula->data.predicate.args; l && result; l = listNext(l))
	 result = incIsNonSpecCompatibleTerm((Term) listElement(l));
      return result;
   }
   if (formula->type == FTConjunction ||
      formula->type == FTDisjunction ||
      formula->type == FTImplication ||
      formula->type == FTBiimplication) {
      return ((incIsNonSpecCompatibleFormula(formula->data.junction.arg1,
	       foundNonSpec) &&
	    incIsNonSpecCompatibleFormula(formula->data.junction.arg2,
	       foundNonSpec))
	 || *foundNonSpec);
   }
   if (formula->type == FTEqual ||
      formula->type == FTGreater ||
      formula->type == FTLess ||
      formula->type == FTGreaterEqual ||
      formula->type == FTLessEqual ||
      formula->type == FTNotEqual) {
      return ((incIsNonSpecCompatibleTerm(formula->data.relation.op1) &&
	    incIsNonSpecCompatibleTerm(formula->data.relation.op2)));
   }
   return TRUE;
}

/* ----------------------------------------------------------------------
 * incIsNonSpecCompatibleTerm:
 *	scan if term is NONSPEC-compatible
 */
Boolean incIsNonSpecCompatibleTerm(Term term)
{
   List feature, l;
   String featureId;
   Boolean result;

   if(term->type == TTBottomPeek ||
      term->type == TTLabel ||
      term->type == TTLevel ||
      term->type == TTString ||
      term->type == TTNumber)
      return TRUE;

   if (term->type == TTTopPeek) {
      feature = term->data.peek.path;
      if (feature != NULL && 
	  listElement(feature) != NULL) {

	 featureId = (String)(listElement(feature));
	 
	 /* legal top-peeks */
	 if (strcmp(featureId, "id") == 0 || /* ^id could be considered
						incompatible */
	     strcmp(featureId, "from") == 0 ||
	     strcmp(featureId, "to") == 0) 
	    return TRUE;
	 else {
	    return FALSE;
	}
      } 
   }

   if (term->type == TTAdd || 
       term->type == TTSubtract ||
       term->type == TTMultiply || 
       term->type == TTDivide) 
      return (incIsNonSpecCompatibleTerm(term->data.operation.op1) &&
	      incIsNonSpecCompatibleTerm(term->data.operation.op2));

   if (term->type == TTFunction) {
      if (strcmp(term->data.function.functor, "distance") == 0)
	 return TRUE;

      result = TRUE;
      for (l = term->data.function.args; l && result; l=listNext(l))
	 result = incIsNonSpecCompatibleTerm((Term) listElement(l));
      return result;
   }

   cdgPrintf(CDG_DEBUG, "DEBUG:   invalid term\n");
   return FALSE;
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
/** @} */
