/* 
 * Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without 
 * modifications, as long as this notice is preserved.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Birth: 3/3/97
 */

/* -------------------------------------------------------------------------
 * @addtogroup Input Input - reading in dependency constraint grammars
 * \author Ingo Schroeder (see also AUTHORS and THANKS for more)
 *
 * $Id: input.c,v 1.307 2004/10/19 13:05:09 foth Exp $
 * 
 * \todo move docu from libcdg.tex
 * @{
 */

/* ------------------------------------------------------------------------- */
#include <config.h>
#include <unistd.h>
#include <errno.h>
#include <fcntl.h>
#include <ctype.h>
#include <dlfcn.h>
#include <values.h>
#include <stdio.h>
#include <blah.h>
#include <string.h>
#include <float.h>
#include <math.h>     /* exp() */
#include <sys/stat.h> /* stat() */
#include <sys/wait.h> /* wait() */
#include <signal.h>   /* signal() */
#include <strings.h>   /* strtok_r() */
#include <libgen.h>
#include <limits.h>
#include <locale.h>
#include <dirent.h>
#include <sys/types.h>
#include "cdg.h"
#include "hook.h"
#include "eval.h"
#include "input.h"
#include "parse.h"
#include "parsing.h"
#include "set.h"
#include "tagger.h"
#include "command.h"
#include "functions.h"
#include "predicates.h"
#include "cdgdb.h"
#ifndef DARWIN
#include "compile.h"
#endif


/* -- EXTERNAL VARIABLES ------------------------------------------------ */
Lattice inputMostRecentlyCreatedLattice = NULL;
Section inputDefaultSection = NULL;
List inputSections = NULL;
Input inputCurrentGrammar = NULL;
String inputPreprocessor = NULL;
String inputPrepSuffix = NULL;
List inputAnnoCats = NULL;
String inputAnnoCatString = NULL;
List inputCompoundCats = NULL;
String inputCompoundString = NULL;
List inputSpontaneousUpcase = NULL;
String inputSpontaneousUpcaseString = NULL;
String inputLocale = NULL;
String inputFeatureHierarchy = NULL;
Boolean inputDeduceCompounds = FALSE;
String inputAnnodirs = NULL;
int templateMode = 2;

Value ErrorValue = NULL; /**< the global error value, see inputInitialize() */

/* how to integrate acoustic scores with CDG scores */
VmAcousticsType vmAcoustics=vmAcOff;

/* -- INTERNAL VARIABLES ------------------------------------------------ */
String *prepArgv = NULL;
int prepArgNo = 0;

/* -- FORWARD DECLARATIONS ---------------------------------------------- */

/* -- FUNCTIONS --------------------------------------------------------- */

/* ----------------------------------------------------------------------------
 * free memory for a path
 */
void inputFreePath(List l)
{
  listForEachDelete(l, cdgFreeString);
}

/* ----------------------------------------------------------------------------
 * free memory for section
 */
void freeSection(Section s)
{
  cdgFreeString(s->id);
  memFree(s);
}

/* ----------------------------------------------------------------------------
 * free memory for constraint
 */
void freeConstraint(Constraint c)
{
  List l;
  VarInfo vi;

  for (l = c->vars; l != NULL; l = listNext(l)) {
    vi = (VarInfo) listElement(l);
    cdgFreeString(vi->varname);
    if (vi->levelname != NULL)
      cdgFreeString(vi->levelname);
    memFree(vi);
  }
  listDelete(c->vars);

  c->section->counter--;
  if (c->section->counter == 0 && c->section != inputDefaultSection) {
    inputSections = listDeleteElement(inputSections, c->section);
    freeSection(c->section);
  }
  cdgFreeString(c->id);
  memFree(c->sig);
  if (c->penaltyTerm != NULL)
    freeTerm(c->penaltyTerm);
  freeFormula(c->formula);
  memFree(c);
}

/* ----------------------------------------------------------------------------
 * free memory for formula
 */
void freeFormula(Formula f)
{
  if (f->type == FTNegation) {
    freeFormula(f->data.formula);
  } else if (f->type == FTPredicate) {
    cdgFreeString(f->data.predicate.functor);
    listForEachDelete(f->data.predicate.args, freeTerm);
  } else if (f->type == FTConjunction || f->type == FTDisjunction ||
	     f->type == FTImplication || f->type == FTBiimplication) {
    freeFormula(f->data.junction.arg1);
    freeFormula(f->data.junction.arg2);
  } else if (f->type == FTEqual || f->type == FTGreater ||
	     f->type == FTLess || f->type == FTGreaterEqual ||
	     f->type == FTLessEqual || f->type == FTNotEqual) {
    freeTerm(f->data.relation.op1);
    freeTerm(f->data.relation.op2);
  }
  memFree(f);
}

/* ----------------------------------------------------------------------------
 * free memory for term
 */
void freeTerm( Term t )
{
  if (t->type == TTTopPeek || t->type == TTBottomPeek) {
    inputFreePath(t->data.peek.path);
  } else if (t->type == TTAdd ||
	     t->type == TTSubtract ||
	     t->type == TTMultiply || t->type == TTDivide) {
    freeTerm(t->data.operation.op1);
    freeTerm(t->data.operation.op2);
  } else if (t->type == TTFunction) {
    cdgFreeString(t->data.function.functor);
    listForEachDelete(t->data.function.args, freeTerm);
  } else if (t->type == TTString) {
    cdgFreeString(t->data.string);
  }
  memFree(t);
}

/* ----------------------------------------------------------------------------
 * free memory for value
 */
void freeValue( Value v )
{
  if (v->type == VTString) {
    cdgFreeString(v->data.string);
  } else if (v->type == VTAVNode) {
    cdgFreeString(((AVNode) v->data.avnode)->attribute);
    freeValue(((AVNode) v->data.avnode)->value);
    memFree(((AVNode) v->data.avnode));
  } else if (v->type == VTConjunction || v->type == VTDisjunction ||
	     v->type == VTList) {
    listForEachDelete(v->data.list, freeValue);
  }
  memFree(v);
}



/* ---------------------------------------------------------------------------
 * free memory for level declaration
 * filename is shared among all input structures and must not be freed
 */
void freeLevel( Level d )
{
  /* up- and downFeatures aren't really paths but lists of lists of strings. 
   */
  listForEachDelete(d->properties, freeLevelProp);
  listForEachDelete(d->labels, cdgFreeString);
  cdgFreeString(d->id);
  memFree(d);
}


/* ----------------------------------------------------------------------------
 * free memory for a flag in a level declaration
 */
void freeLevelProp(Property df)
{
  cdgFreeString(df->name);
  cdgFreeString(df->value);
  memFree(df);
}

/* ----------------------------------------------------------------------------
 * free memory for lexical entry
 * filename is shared among all input structures and must not be freed
 */
void freeLexiconItem(LexiconItem l)
{
  cdgFreeString(l->description);
  cdgFreeString(l->word);
  freeValue(l->value);
  if (l->values) {
    memFree(l->values);
  }
  listDelete(l->distinctive_features);
  memFree(l);
}

/* ----------------------------------------------------------------------------
   Free list of LexiconItem
*/
void freeLexiconItems(List l)
{
  listForEachDelete(l, freeLexiconItem);
}

/* ----------------------------------------------------------------------------
 * free memory for lattice
 * filename is shared among all input structures and must not be freed
 */
void freeLattice( Lattice lat )
{
  List l;
  Value v;

  for (l = lat->arcs; l != NULL; l = listNext(l)) {
    cdgFreeString(((Arc) listElement(l))->word);
    v = ((Arc) listElement(l))->info;
    if (v) {
      freeValue(v);
    }
    memFree(listElement(l));
  }
  listDelete(lat->arcs);

  cdgFreeString(lat->id);
  memFree(lat);
}

/* ----------------------------------------------------------------------------
 * free memory for annotations
 */
void freeAnnotation(Annotation a)
{
  List l;

  for (l = a->specs; l != NULL; l = listNext(l)) {
    cdgFreeString(((Specification) listElement(l))->kind);
    cdgFreeString(((Specification) listElement(l))->name);
    memFree((Specification) listElement(l));
  }
  listDelete(a->specs);

  cdgFreeString(a->word);
  memFree(a);
}

/* ----------------------------------------------------------------------------
 * free memory for annotation entry
 * filename is shared among all input structures and must not be freed
 */
void freeAnnoEntry(AnnoEntry a)
{
  listForEachDelete(a->annos, freeAnnotation);
  cdgFreeString(a->id);
  cdgFreeString(a->lattice);

  memFree(a);
}

/* ----------------------------------------------------------------------------
 * free memory for subsumption
 */
void freeSubsumption(Subsumption s)
{
  cdgFreeString(s->type);
  listForEachDelete(s->types, cdgFreeString);

  memFree(s);
}

/* ----------------------------------------------------------------------------
 * freeSort
 */
void freeSort(Sort s)
{
  cdgFreeString(s->id);
  listForEachDelete(s->subsorts, freeSort);

  memFree(s);
}

/* ----------------------------------------------------------------------------
 * free memory for hierarchy
 * filename is shared among all input structures and must not be freed
 */
void freeHierarchy(Hierarchy h)
{
  cdgFreeString(h->id);
  if (h->types) {
    HashIterator hi;
    String s;

    hi = hashIteratorNew(h->types);
    while (NULL != (s = (String) hashIteratorNextKey(hi))) {
      cdgFreeString(s);
    }
    hashIteratorDelete(hi);
    hashDelete(h->types);
  }

  if (h->distance) {
    arrayDelete(h->distance);
  }

  memFree(h);
}

/* ----------------------------------------------------------------------------
 * free memory for parameter
 */
void freeParameter(Constraint c)
{
  List l;

  for (l = c->vars; l != NULL; l = listNext(l)) {
    cdgFreeString(((VarInfo) listElement(l))->varname);
    memFree(listElement(l));
  }
  listDelete(c->vars);

  freeFormula(c->formula);
  memFree(c);
}

/* ----------------------------------------------------------------------------
   Compare two Specifications.

   The aim here is to establish a consistent order between Specifications
   so that writeAnnotation()'s output will be as consistent as possible.
   Then, when making changes to an annotation, the automatic diffs will
   always be relevant and never spurious.
*/
Boolean inputSpecCompare(Specification a, Specification b) {

  Level la, lb;
  
  /* sort morphology specs before dependency specs */
  if(STTag == a->type && STDep == b->type) {
    return TRUE;
  }
  if(STDep == a->type && STTag == b->type) {
    return FALSE;
  }

  /* sort morphology specs alphabetically */
  if(STTag == a->type && STTag == b->type) {
    return strcmp(a->kind, b->kind) < 0;
  }

  /* sort level specs by level order */
  la = findLevel(a->kind);
  lb = findLevel(b->kind);
  if(la && lb) {
    return la->no < lb->no;
  }

  return FALSE;
  
}


/* ----------------------------------------------------------------------------
 * Detect strings that look like CDG keywords.
 */
Boolean isKeyword(String s) {
  if(!strcmp(s,"true") ||
     !strcmp(s,"false") ||
     !strcmp(s,"level") ||
     !strcmp(s,"label")) {
    return TRUE;
  }
  return FALSE;
}

/* ----------------------------------------------------------------------------
 * prints an identifier, taking care to quote if necessary.
 */
String formatIdentifier(String s) {
  static char buffer[1024];
  int in, out;

  /* not empty strings with no special characters */
  if(strlen(s) > 0 &&
     !isdigit(((unsigned char *)s)[0]) &&
     !isKeyword(s) &&
     strcspn(s, " @#$%^&*()-=+\\|`~\"';:/!?.>,<[]{}") >= strlen(s)) {
    return strRegister(s);
  }
  
  strcpy(buffer, "'");
  in = 0, out = 1;
  while(s[in]) {
    if(out>1021) {
      break;
    }
    switch(s[in]) {
    case '"':
    case '\'':
    case '\\':
      buffer[out++] = '\\';
      /* fall through */
    default:
      buffer[out++] = s[in++];
    }
  }
  buffer[out++] = '\'';
  buffer[out] = '\0';
  return strRegister(buffer);
}

/* ---------------------------------------------------------------------------
 * convert a time in milliseconds to a human readable string
 * TODO : this only displays time up to 24 d 20 h 31 m 23 s 647 ms
 */
String formatTime(unsigned long time)
{
  int days, hours, minutes, seconds, milliSeconds;
  Boolean isFirst;
  String result;

  if(time == 0) {
    return strRegister("0 ms");
  }

  /* calculate the units */
  days = time / 86400000;
  time = time - days * 86400000;
  hours = time / 3600000;
  time = time - hours * 3600000;
  minutes = time / 60000;
  time = time - minutes * 60000;
  seconds = time / 1000;
  milliSeconds = time - seconds * 1000;

  isFirst = TRUE;
  result = "";
  
  if(days) {
    result = strPrintf("%d d", days);
    isFirst = FALSE;
  }
  
  if(hours) {
    result = strPrintf("%s%s%d h", result, isFirst?"":" ", hours);
    isFirst = FALSE;
  }
   
  if(minutes) {
    result = strPrintf("%s%s%d m", result, isFirst?"":" ", minutes);
    isFirst = FALSE;
  }

  if(seconds) {
    result = strPrintf("%s%s%d s", result, isFirst?"":" ", seconds);
    isFirst = FALSE;
  }

  if(milliSeconds) {
    result = strPrintf("%s%s%d ms", result, isFirst?"":" ", milliSeconds);
  }

  return result;
}

/* ----------------------------------------------------------------------------
 * force a lattice to use the simplest possible time points
 */
Boolean trivialise(Lattice lat)
{
  List l;
  Arc a;
  int i, next;

  Boolean used[10000];
  int map[10000];

  /* initialise table */  
  for(i = 0; i < 10000; i++) {
    used[i] = FALSE;
  }

  /* find all used timepoints */
  for(l = lat->arcs; l != NULL; l = listNext(l)) {
    a = listElement(l);
    if(a->from < 0) return FALSE;
    if(a->to < 0) return FALSE;
    if(a->from > 9999) return FALSE;
    if(a->to > 9999) return FALSE;
    used[a->from] = used[a->to] = TRUE;
  }

  /* build translation table */
  next = 0;
  for(i = 0; i < 10000; i++) {
    if(used[i]) {
      map[i] = next;
      next++;
    }
  }

  /* translate all arcs */
  for(l = lat->arcs; l != NULL; l = listNext(l)) {
    a = listElement(l);
    a->from = map[a->from];
    a->to = map[a->to];
  }

  return TRUE;

}

/* ---------------------------------------------------------------------------
 * Does a wordgraph g contain word w?
 */
Boolean graphContainsWord(Lattice g, String w) {

  List l;

  for(l = g->arcs; l != NULL; l = listNext(l)) {
    if(!strcmp(((Arc) listElement(l))->word, w)) return TRUE;
  }

  return FALSE;


}


/* ---------------------------------------------------------------------------
 * Return the Arc that precedes A in LAT, or NULL.
 * Finds only the first predecessor if several exist. 
 */
Arc inputLatticePrevArc(Lattice lat, Arc a) {
  Arc result;
  List l;
  for(l = lat->arcs; l != NULL; l = listNext(l)) {
    result = listElement(l);
    if(result->to == a->from) {
      return result;
    }
  }
  return NULL;
}

/* ----------------------------------------------------------------------------
 * check constraints and delete some if necessary
 */
Hashtable checkedConstraints(Input input, Hashtable constraints)
{
  Constraint c;
  List k;
  VarInfo vi;
  HashIterator hi;
  Hashtable ht =
    hashNew( 500, 0.8, hashStringHashFunction, hashStringEqualFunction );

  hi = hashIteratorNew(constraints);
  while((c = hashIteratorNextValue(hi))) {
    Boolean ok = TRUE;

    /* check c->vars[..]->level */
    for (k = c->vars; k != NULL; k = listNext(k)) {
      vi = (VarInfo) listElement(k);
      if(!vi->levelname) continue;
      if (!(vi->level = findLevel(vi->levelname))) {

        /* the level is gone so we delete this constraint */
        cdgPrintf(CDG_WARNING,
                  "WARNING: deleting constraint `%s' with invalid level `%s'\n",
                  c->id, vi->levelname);
        freeConstraint(c);
        ok = FALSE;
        break;
      }
    }
    if(ok) {
      hashSet(ht, c->id, c);
    }
  }
  hashIteratorDelete(hi);
  hashDelete(constraints);

  return ht;
}

/* ----------------------------------------------------------------------------
 * do two paths read the same?
 */
Boolean comparePath(List a, List b)
{
  if (a == NULL) {
    if (b == NULL)
      return TRUE;
    return FALSE;
  }

  if (b == NULL)
    return FALSE;

  if (strcmp(listElement(a), listElement(b)))
    return FALSE;
  
  return comparePath(listNext(a), listNext(b));
}

/* ----------------------------------------------------------------------
   Does UP subsume DOWN in H?
*/
Boolean inputSubsumes(Hierarchy h, String up, String down) {
  int a, b;
  
  /* find both sorts */
  if (0 > (a = (int)hashGet(h->types, up) - 1)) {
    return FALSE;
  }
  if (0 > (b = (int)hashGet(h->types, down) - 1)) {
    return FALSE;
  }
  
  /* return canned answer */
  return (int)arrayElement(h->distance, a, b) >= 0; 
}


/* ----------------------------------------------------------------------------
 * Detect slots that are written like lexical features but aren't:
 * `word', `id', etc. 
 */
Boolean inputPseudoFeature(List path) {
  String s;
  if(listSize(path) != 1) {
    return FALSE;
  }
  s = listElement(path);
  if(s == static_string_id ||
     s == static_string_word ||
     s == static_string_from ||
     s == static_string_to ||
     s == static_string_info ||
     s == static_string_chunk_start ||
     s == static_string_chunk_end ||
     s == static_string_chunk_type) {
    return TRUE;
  }
  return FALSE;
}

/* ----------------------------------------------------------------------------
   Compute index of edge type
*/
EdgeType etEncode(Level level, String labelname, Direction d)
{
  int labelno = listIndex(level->labels, labelname) - 1;

  if (-1 == labelno) {
    return -1;
  }
#ifdef DEBUGETENCODE
  cdgPrintf(CDG_DEBUG, "etEncode(%s", level->id);
  printDirection(CDG_DEBUG, d);
  cdgPrintf(CDG_DEBUG, "%s) == ", labelname);
  cdgPrintf(CDG_DEBUG, "%d\n",
	    d + labelno * nrDirections +
	    level->no * nrDirections * inputCurrentGrammar->maxNrLabels);
#endif
  return (d +
	  labelno * nrDirections +
	  level->no * nrDirections * inputCurrentGrammar->maxNrLabels);
}

/* ----------------------------------------------------------------------------
   Return corresponding Direction
*/
Direction etDecodeDirection(EdgeType edgetype) {
  return edgetype % nrDirections;
}

/* ----------------------------------------------------------------------------
 * FIXME: no comment
 */
Boolean etValid(EdgeType edgetype) {
  Level level = etDecodeLevel(edgetype);
  int labelno = (edgetype / nrDirections) % inputCurrentGrammar->maxNrLabels;
  if(labelno+1 > listSize(level->labels)) {
    return FALSE;
  }
  return TRUE;
}

/* ----------------------------------------------------------------------------
   Return corresponding Label
*/
String etDecodeLabel(EdgeType edgetype)
{
  Level level;
  int labelno;

  if (etValid(edgetype)) {
    level = etDecodeLevel(edgetype);
    labelno = (edgetype / nrDirections) % inputCurrentGrammar->maxNrLabels;
    return listNthElement(level->labels, labelno + 1);
  }
  return NULL;
}

/* ----------------------------------------------------------------------------
  Return corresponding Level
*/
Level etDecodeLevel(EdgeType edgetype)
{
  return listNthElement(inputCurrentGrammar->levels,
		 ((edgetype / nrDirections / inputCurrentGrammar->maxNrLabels) +
		  1));
}


/* ----------------------------------------------------------------------------
   Print edge type symbolically
*/
void etPrint(unsigned long int mode, EdgeType et)
{
  String label = etDecodeLabel(et);

  cdgPrintf(mode, "%s(%s)",
	    etDecodeLevel(et)->id,
	    formatIdentifier(label ? label : "<dummmy>"));
  printDirection(mode, etDecodeDirection(et));
}


/* ----------------------------------------------------------------------------
   Return number of theoretically possible edge types
*/
EdgeType etMax(void)
{
  return listSize(inputCurrentGrammar->levels) *
    inputCurrentGrammar->maxNrLabels * nrDirections;
}


/* ----------------------------------------------------------------------------
   Compute INPUT->upFeatures and INPUT->downFeatures.

   This function scans all constraints in INPUT and determines what types
   of LV access which lexical feature. To keep track of what edge types are
   affected at any moment, it maintains data structures called Context
   (really just Hashtables of BitStrings).

   A Context stores a Bitstring under the name of each constraint variable. 
   Each bit indicates whether the corresponding edge type can occur as an
   instantiation of that variable if evaluation reaches this point.

   The function then follows the control flow in the constraint and
   effectively qualifies each feature access with the edge types that
   actually perform it. The result is that when building LVs for a
   constraint net, more of them can be proved to be exactly equivalent and
   combined into one disjunctive LV.

*/
void inputCollectFeatures(Input input) 
{

  Constraint c = NULL;
  int i;
  int nrEdgeTypes = etMax();
  int nrFeatures = listSize(inputCurrentGrammar->allAttributes);
  HashIterator hi;
  
  typedef Hashtable Context;

  Context context =
    hashNew(10, 0.7, hashStringHashFunction, hashStringEqualFunction);

  /* invert bits in a BitString, should be in ACL */
  void bitInvert(BitString bs) {
    int i;
    for(i = 0; i < bitSize(bs); i++) {
      if(bitGet(bs,i)) {
        bitClear(bs, i);
      } else {
        bitSet(bs,i);
      }
    }
  }
  
  /* apply bitInvert() to all members of HT */
  void contextInvert(Context ht) {
    HashIterator hi = hashIteratorNew(ht);
    BitString bs;
    while (NULL != (bs = hashIteratorNextValue(hi))) {
      bitInvert(bs);
    }
    hashIteratorDelete(hi);
  }

  /* apply bitAnd() to corresponding members of A and B */
  void contextAnd(Context a, Context b) {
    HashIterator hi = hashIteratorNew(a);
    String key;
    BitString bsa, bsb;
    while (NULL != (key = hashIteratorNextKey(hi))) {
      bsa = hashGet(a, key);
      bsb = hashGet(b, key);
      bitAnd(bsa,bsb);
    }
    hashIteratorDelete(hi);
  }

  /* apply bitOr() to corresponding members of A and B */
  void contextOr(Context a, Context b) {
    HashIterator hi = hashIteratorNew(a);
    String key;
    BitString bsa, bsb;
    while (NULL != (key = hashIteratorNextKey(hi))) {
      bsa = hashGet(a, key);
      bsb = hashGet(b, key);
      bitOr(bsa,bsb);
    }
    hashIteratorDelete(hi);
  }

  /* clone a Context */
  Context context_copy(Context ht) {
    Context result =
      hashNew(10, 0.7, hashStringHashFunction, hashStringEqualFunction);
    HashIterator hi = hashIteratorNew(ht);
    String key;
    while (NULL != (key = hashIteratorNextKey(hi))) {
      hashSet(result, key, bitClone(hashGet(ht, key)));
    }
    hashIteratorDelete(hi);
    return result;
  }

  /* deallocate a Context */
  void context_free(Context ht) {
    HashIterator hi = hashIteratorNew(ht);
    String key;
    while (NULL != (key = hashIteratorNextKey(hi))) {
      bitDelete(hashGet(ht, key));
    }
    hashIteratorDelete(hi);
    hashDelete(ht);
  }
  
  void debug_context(Context ht) {
    HashIterator hi = hashIteratorNew(ht);
    String key;
    while (NULL != (key = hashIteratorNextKey(hi))) {
      BitString bs = hashGet(ht, key);
      cdgPrintf(CDG_DEBUG, " Edge %s can be of type: ", key);
      for(i = 0; i < bitSize(bs); i++) {
        if(etValid(i) && bitGet(bs, i)) {
          etPrint(CDG_DEBUG, i);
          cdgPrintf(CDG_DEBUG, " ");
        }
      }
      cdgPrintf(CDG_DEBUG, "\n");
    }
    hashIteratorDelete(hi);
  }

  /* Really register a feature
     as being significant to particular edge types. */
  void register_access(Context context, Term t) {
    int index;
    BitString bs;
    String varname;
    
    if(inputPseudoFeature(t->data.peek.path)) {
      return;
    }

    /* get numeric index of lexical feature */
    index = t->data.peek.pathindex;

    /* what edge types can the variable be bound to? */
    varname = t->data.peek.varinfo->varname;
    bs = hashGet(context, varname);

    for(i = 0; i < nrEdgeTypes; i++) {
      if(bitGet(bs, i)) {
        BitString features;

        /* distinguish modifier and modifiee accesses */
        if(t->type == TTTopPeek) {
          features = vectorElement(inputCurrentGrammar->upFeatures, i);
        }
        else {
          features = vectorElement(inputCurrentGrammar->downFeatures, i);
        }

        /* set appropriate relevance bit */
        bitSet(features, index);

#ifdef DEBUGINPUTCOLLECTFEATURES
        if(etValid(i)) {
          etPrint(CDG_DEBUG, i);
          cdgPrintf(CDG_DEBUG, " edges access X%s", t->type == TTTopPeek ? "^" : "@");
          printPath(CDG_DEBUG, t->data.peek.path);
          cdgPrintf(CDG_DEBUG, " in constraint `%s'.\n", c->id);
        }
#endif
      }
    }
  }

  /* unmark all edgetypes that do not have one of these labels */
  void restrict_labels(BitString bs, Hierarchy h, Level level, String node) {
    List l;
    for(l = level->labels; l != NULL; l = listNext(l)) {
      String label = listElement(l);
      if(!inputSubsumes(h, node, label)) {
        bitClear(bs, etEncode(level, label, Left));
        bitClear(bs, etEncode(level, label, Right));
        bitClear(bs, etEncode(level, label, Nil));
      }
    }
  }

  /* unmark all edgetypes that do not have this level */
  void restrict_level(BitString bs, String levelname) {
    int i;
    Level level;
    if(!levelname) {
      return;
    }
    level = findLevel(levelname);
    if(!level) {
      return;
    }
    for(i = 0; i < nrEdgeTypes; i++) {
      if(etDecodeLevel(i) != level) {
        bitClear(bs, i);
      }
    }
  }

  /* unmark all edgetypes that have this level */
  void restrict_not_level(BitString bs, String levelname) {
    int i;
    Level level;
    if(!levelname) {
      return;
    }
    level = findLevel(levelname);
    if(!level) {
      return;
    }
    for(i = 0; i < nrEdgeTypes; i++) {
      if(etDecodeLevel(i) == level) {
        bitClear(bs, i);
      }
    }
  }
  
  /* unmark all edgetypes that do not have this label */
  void restrict_label(BitString bs, String label) {
    int i;
    for(i = 0; i < nrEdgeTypes; i++) {
      if(etDecodeLabel(i) != label) {
        bitClear(bs, i);
      }
    }
  }

  /* unmark all edgetypes that have this label */
  void restrict_not_label(BitString bs, String label) {
    int i;
    for(i = 0; i < nrEdgeTypes; i++) {
      if(etDecodeLabel(i) == label) {
        bitClear(bs, i);
      }
    }
  }

  /* unmark all edgetypes that do not have this direction */
  void restrict_dir(BitString bs, Direction d) {
    int i;
    if(d == AnyDir) {
      return;
    }
    for(i = 0; i < nrEdgeTypes; i++) {
      if(!subsumesDirection(d, etDecodeDirection(i))) {
        bitClear(bs, i);
      }
    }
  }
  
  /* propagate feature access information through a Term */
  void analyzeTerm(Context context, Term t) {
    List l;
    
    switch(t->type) {
    
    case TTTopPeek:
    case TTBottomPeek:
      register_access(context, t);
      break;

    case TTAdd:
    case TTSubtract:
    case TTMultiply:
    case TTDivide:
      analyzeTerm(context, t->data.operation.op1);
      analyzeTerm(context, t->data.operation.op2);
      break;

    case TTFunction:
      l = t->data.function.args;
      while (l != NULL) {
        analyzeTerm(context, listElement(l));
        l = listNext(l);
      }
      break;
    default:
      /* do nothing */
      break;
    }
  }

  /* Propagate feature access information through a Formula.

     OLD is the context that was valid when the formula was entered.
     This is used to qualify feature accesses.
     
     NEW is a context with all bits set. It is used to propagate up
     restrictions introduced by the formula itself. */
  void analyzeFormula(Context context, Formula f) {

    /* (A) Detect restrictions
           that affect further evaluation of the formula */
    
    /* detect "X/" */
    if(f->type == FTDirection) {
      BitString bs = hashGet(context, f->data.direction.var->varname);
      restrict_dir(bs, f->data.direction.d);
    }
    
    /* detect "subsumes(Labels, X.label, FOO)" */
    if(f->type == FTPredicate &&
       f->data.predicate.function == predSubsumes) {
      if(listSize(f->data.predicate.args) > 2) {
        Term t1 = listElement(f->data.predicate.args);
        Term t2 = listElement(listNext(f->data.predicate.args));
        Term t3 = listElement(listNext(listNext(f->data.predicate.args)));
        if(t2->type == TTString && t3->type == TTLabel) {
          Hierarchy h = findHierarchy(t1->data.string);
          if(h) {
            String node = t2->data.string;
            VarInfo vi = t3->data.peek.varinfo;
            String varname = vi->varname;
            BitString bs = hashGet(context, varname);
            Level level = vi->level;
            restrict_labels(bs, h, level, node);
          }
        }
      }
    }
    
    if (f->type == FTEqual) {
      Term t1 = f->data.relation.op1;
      Term t2 = f->data.relation.op2;
      
      /* detect "X.label = FOO" */
      if(t1->type == TTLabel && t2->type == TTString) {
        String varname = t1->data.label->varname;
        BitString bs = hashGet(context, varname);
        String label = t2->data.string;
        restrict_label(bs, label);
      } else if(t2->type == TTLabel && t1->type == TTString) {
        String varname = t2->data.label->varname;
        BitString bs = hashGet(context, varname);
        String label = t1->data.string;
        restrict_label(bs, label);
      }
      
      /* detect "X.level = BAR" */
      else if(t1->type == TTLevel && t2->type == TTString) {
        String varname = t1->data.level->varname;
        BitString bs = hashGet(context, varname);
        restrict_level(bs, t2->data.string);
      } else if(t2->type == TTLevel && t1->type == TTString) {
        String varname = t2->data.level->varname;
        BitString bs = hashGet(context, varname);
        restrict_level(bs, t1->data.string);
      } 
    }

    if (f->type == FTNotEqual) {
      Term t1 = f->data.relation.op1;
      Term t2 = f->data.relation.op2;
      
      /* detect "X.label != FOO" */
      if(t1->type == TTLabel && t2->type == TTString) {
        String varname = t1->data.label->varname;
        BitString bs = hashGet(context, varname);
        String label = t2->data.string;
        restrict_not_label(bs, label);
      } else if(t2->type == TTLabel && t1->type == TTString) {
        String varname = t2->data.label->varname;
        BitString bs = hashGet(context, varname);
        String label = t1->data.string;
        restrict_not_label(bs, label);
      }
      
      /* detect "X.level != BAR" */
      else if(t1->type == TTLevel && t2->type == TTString) {
        String varname = t1->data.level->varname;
        BitString bs = hashGet(context, varname);
        restrict_not_level(bs, t2->data.string);
      } else if(t2->type == TTLevel && t1->type == TTString) {
        String varname = t2->data.level->varname;
        BitString bs = hashGet(context, varname);
        restrict_not_level(bs, t1->data.string);
      } 
    }    

    /* (B) analyze Terms in the Formula
           to actually find feature accesses */

    /* predicates mark the end of the control flow, so the Context
       does not change anymore, and we can pass it by reference without
       fear. But they may contain feature accesses, so we have to scan them
       as well. */
    if (f->type == FTPredicate) {
      List l = f->data.predicate.args;
      while (l != NULL) {
        analyzeTerm(context, listElement(l));
        l = listNext(l);
      }
    }

    /* exactly the same goes for Boolean operators */
    if (f->type == FTEqual ||
        f->type == FTGreater ||
        f->type == FTLess ||
        f->type == FTGreaterEqual ||
        f->type == FTLessEqual ||
        f->type == FTNotEqual) {
      analyzeTerm(context, f->data.relation.op1);
      analyzeTerm(context, f->data.relation.op2);
    }
    
    /* (C) analyze subformulas and update the global Context */

    /* conjunctions etc. are evaluated left-to-right and on the
       need-to-know principle. This means that any constraints imposed on
       the constraint variables in the premise are still valid in the
       conclusion. Therefore we can simply pass the current Context along,
       allowing it to be modified by the subformula. */
    if (f->type == FTConjunction ||
        f->type == FTImplication ||
        f->type == FTBiimplication) {
      analyzeFormula(context, f->data.junction.arg1);
      analyzeFormula(context, f->data.junction.arg2);
    }

    /* But the constraints imposed by disjunctions are valid only within
       the subformula, so we have to separate what we already know about
       the context and what the subformula contributed itself. We can,
       however, utilize the combined Context of all subformulas to
       constrain our own Context further. */
    if (f->type == FTDisjunction) {
      Context new1 = context_copy(context);
      Context new2 = context_copy(context);
      analyzeFormula(new1, f->data.junction.arg1);
      analyzeFormula(new2, f->data.junction.arg2);
      contextOr(new1,new2);
      contextAnd(context, new1);
      context_free(new1);
      context_free(new2);
    }

    /* The information inside a negation could also be used to affect the
       current context, but that would require at least one more Hashtable
       to distinguish what edge types were already forbidden and which were
       forbidden by the negated formula. At the moment we simply pretend
       that negated formulas do not impose any edge type restrictions.
       This loses some optimization but makes no errors. */
    if (f->type == FTNegation) {
      Context new = context_copy(context);
      analyzeFormula(new, f->data.formula);
      context_free(new);
    }

  }

  /* ----------------------------------------------------------------------------
     inputCollectFeatures() begins
     ----------------------------------------------------------------------*/
  
  /* allocate Vectors of BitStrings */
  inputCurrentGrammar->upFeatures = vectorNew(nrEdgeTypes);
  inputCurrentGrammar->downFeatures = vectorNew(nrEdgeTypes);
  for(i = 0; i < nrEdgeTypes; i++) {
    BitString bs = bitNew(nrFeatures);
    bitClearAll(bs);
    vectorSetElement(inputCurrentGrammar->upFeatures, bs, i);
    bs = bitNew(nrFeatures);
    bitClearAll(bs);
    vectorSetElement(inputCurrentGrammar->downFeatures, bs, i);
  }

  /* analyze constraints to fill the maps */
  hi = hashIteratorNew(inputCurrentGrammar->constraints);
  while((c = hashIteratorNextValue(hi))) {
    List m;

    /* first of all, find out
       what variables this constraints talks about */
    for(m = c->vars; m != NULL; m = listNext(m)) {
      VarInfo vi = listElement(m);
      BitString bs = hashGet(context, vi->varname);
      if(!bs) {
        bs = bitNew(nrEdgeTypes);
      }
      bitSetAll(bs);      

      /* exploit information in the signature */
      restrict_level(bs, vi->levelname);
      restrict_dir(bs, vi->dir);
      hashSet(context, vi->varname, bs);
    }
    
    /* now scan the entire formula */
    analyzeFormula(context, c->formula);
  }
  hashIteratorDelete(hi);
  
#ifdef DEBUGINPUTCOLLECTFEATURES
  for(i = 0; i < nrFeatures; i++) {
    int j;
    cdgPrintf(CDG_DEBUG, "Feature `");
    printPath(CDG_DEBUG, listNthElement(inputCurrentGrammar->allAttributes, i+1));
    cdgPrintf(CDG_DEBUG, "' is used by:\n");

    for(j = 0; j < nrEdgeTypes; j++) {
      BitString up = vectorElement(inputCurrentGrammar->upFeatures, j);
      BitString down = vectorElement(inputCurrentGrammar->downFeatures, j);
      if(!etValid(j)) continue;
      if(bitGet(down, i)) {
        etPrint(CDG_DEBUG, j);
        cdgPrintf(CDG_DEBUG, "@\n");
      }
      if(bitGet(up, i)) {
        etPrint(CDG_DEBUG, j);
        cdgPrintf(CDG_DEBUG, "^\n");
      }
    }
  }
#endif

  context_free(context);

}


Boolean inputCheckTerm(Term t, Constraint c);
Boolean inputCheckFormula(Formula f, Constraint c);


/* ----------------------------------------------------------------------
   Check that the argument list ARGS conforms to the specification ARITY.

   This saves the actual functions from having to check the
   argument number at runtime. The type of arguments cannot be checked
   since it might be specified indirectly, by a feature access.
*/
Boolean inputCheckArity(String arity, List args,
                        String functor, Constraint c) {
  int actual = listSize(args);
  int formal, min, max;
  List l;
  if(1 == sscanf(arity, ">%d", &formal)) {
    if(actual <= formal) {
      cdgPrintf(CDG_WARNING,
                "WARNING: wrong arity %d for `%s()'"
                " in constraint `%s', deleting.\n",
                actual, functor, c->id);
      return FALSE;
    }
  }
  else if(2 == sscanf(arity, "%d-%d", &min, &max)) {
    if(actual < min || actual > max) {
      cdgPrintf(CDG_WARNING,
                "WARNING: wrong arity %d for `%s()'"
                " in constraint `%s', deleting.\n",
                actual, functor, c->id);
      return FALSE;
    }
  }
  else if(1 == sscanf(arity, "%d", &formal)) {
    if(actual != formal) {
      cdgPrintf(CDG_WARNING,
                "WARNING: wrong arity %d for `%s()'"
                " in constraint `%s', deleting.\n",
                actual, functor, c->id);
      return FALSE;
    }
  }
  for(l = args; l != NULL; l = listNext(l)) {
    if(!inputCheckTerm(listElement(l), c)) {
      return FALSE;
    }
  }
  return TRUE;
}

Boolean inputCheckTerm(Term t, Constraint c) {
    
  switch(t->type) {
  case TTAdd:
  case TTSubtract:
  case TTMultiply:
  case TTDivide:
    return
      inputCheckTerm(t->data.operation.op1, c) &&
      inputCheckTerm(t->data.operation.op2, c);
      
  case TTFunction:
    {
      /* check that the argument list has correct length */
      CDGFunction func = funcGetFunction(t->data.function.functor);
      return inputCheckArity(func->arity,
                             t->data.function.args,
                             t->data.function.functor,
                             c);
    }
  case TTTopPeek:
  case TTBottomPeek:
  case TTLabel:
  case TTLevel:
  case TTString:
  case TTNumber:
    return TRUE;
  }
  return TRUE;
}

Boolean inputCheckFormula(Formula f, Constraint c) {
    
  switch(f->type) {
  case FTConjunction:
  case FTDisjunction:
  case FTImplication:
  case FTBiimplication:
    return
      inputCheckFormula(f->data.junction.arg1, c) &&
      inputCheckFormula(f->data.junction.arg2, c);

  case FTNegation:
    return inputCheckFormula(f->data.junction.arg1, c);

  case FTPredicate:
    {
      /* check that the argument list has correct length */
      CDGPredicate pred = predGetFunction(f->data.predicate.functor);
      return inputCheckArity(pred->arity,
                             f->data.predicate.args,
                             f->data.predicate.functor,
                             c);
    }
  case FTConnexion:
  case FTConstant:
  case FTDirection:
    return TRUE;

  case FTEqual:
  case FTGreater:
  case FTGreaterEqual:
  case FTLess:
  case FTLessEqual:
  case FTNotEqual:
    return
      inputCheckTerm(f->data.relation.op1, c) &&
      inputCheckTerm(f->data.relation.op2, c);
  }
  return TRUE;
}

/* ----------------------------------------------------------------------
   Checks the number of arguments of function and predicate invocations.

   If a constraint contains a detectably wrong call, it is removed from
   the grammar. 
*/
void inputCheckArities(Input input) {
  List l, m;
  List bad_constraints = NULL;
  Constraint c;
 
  /* scan all constraint formulas */
  for(l = m = hashListOfKeys(input->constraints); l; l = listNext(l)) {
    c = hashGet(input->constraints, listElement(l));
    if(!inputCheckFormula(c->formula, c)) {
      bad_constraints = listPrependElement(bad_constraints, c->id);
    }
  }
  listDelete(m);

  for(l = bad_constraints; l != NULL; l = listNext(l)) {
    String id = listElement(l);
    c = hashGet(input->constraints, id);
    hashRemove(input->constraints, id);
    cdgPrintf(CDG_WARNING, "WARNING: deleting `%s'\n", id);
    freeConstraint(c);
  }
  
  /* clean up */
  listDelete(bad_constraints);
}

/* ----------------------------------------------------------------------
   Are two Values different?

   Returns FALSE only if the Values have the same type and exactly the same
   content. Even two conjunctions that have the same constituents but in
   different order are considered different.
*/
Boolean inputCompareValue(Value v1, Value v2) {

  List l, m;

  if(v1->type != v2->type) {
    return TRUE;
  }
  switch(v1->type) {
  case VTNumber:
    return v1->data.number != v2->data.number;
  case VTString:
    return v1->data.string != v2->data.string;
  case VTAVNode:
      return
      ((AVNode)v1->data.avnode)->attribute !=
      ((AVNode)v2->data.avnode)->attribute
      ||
      inputCompareValue(((AVNode)v1->data.avnode)->value,
                        ((AVNode)v2->data.avnode)->value);
  case VTConjunction:
  case VTDisjunction:
  case VTList:
    if(listSize(v1->data.list) != listSize(v2->data.list)) {
      return TRUE;
    }
    for(l = v1->data.list,m = v2->data.list;
        l;
        l = listNext(l), m = listNext(m)) {
      if(inputCompareValue(listElement(l), listElement(m))) {
        return TRUE;
      }
    }
    return FALSE;
    
  case VTError:
    return FALSE;
  case VTLexemNode:
  case VTGraphemNode:
    return v1->data.lexemnode != v2->data.lexemnode;
  default:
    return TRUE;
  }
  return TRUE;
}

/* ----------------------------------------------------------------------
   Are the two terms different?

   Returns FALSE only if the Terms have the same type and exactly the same
   content. Even two conjunctions that have the same constituents but in
   different order are considered different.
*/
Boolean inputCompareTerm(Term t1, Term t2) {

  List l, m;
  VarInfo vi1, vi2;
  
  if(t1->type != t2->type) {
    return TRUE;
  }

  switch(t1->type) {
  case TTTopPeek:
  case TTBottomPeek:
    if(evalPeekValueMethod == EMTCompiled) {
      if(t1->data.peek.pathindex != t2->data.peek.pathindex) {
        return TRUE;
      }
    } else {
      if(!comparePath(t1->data.peek.path, t2->data.peek.path)) {
        return TRUE;
      }
    }
    vi1 = t1->data.peek.varinfo;
    vi2 = t2->data.peek.varinfo;
    if(vi1->varname != vi2->varname) {
      return TRUE;
    }
    return FALSE;
  case TTLabel:
  case TTLevel:
    return t1->data.level->varname != t2->data.level->varname;
  case TTAdd:
  case TTSubtract:
  case TTMultiply:
  case TTDivide:
    return
      inputCompareTerm(t1->data.operation.op1, t2->data.operation.op1)
      ||
      inputCompareTerm(t1->data.operation.op2, t2->data.operation.op2);
  case TTFunction:
    if(t1->data.function.function != t2->data.function.function) {
      return TRUE;
    }
    if(listSize(t1->data.function.args) !=
       listSize(t2->data.function.args)) {
      return TRUE;
    }
    for(l = t1->data.function.args,m = t2->data.function.args;
        l;
        l = listNext(l), m = listNext(m)) {
      if(inputCompareTerm(listElement(l), listElement(m))) {
        return TRUE;
      }
    }
    return FALSE;
  case TTString:
    return t1->data.string != t2->data.string;
  case TTNumber:
    return t1->data.number != t2->data.number;
  }

  return TRUE;
}

/* ----------------------------------------------------------------------
   Are the two formulas different?

   This function does much less than the name suggests. It returns FALSE
   only if f1 and f2 are two `has' invocations with the same second
   argument. At the moment we don't need any more from it.
*/
Boolean inputCompareFormula(Formula f1, Formula f2) {
  if(f1->type != f2->type) {
    return TRUE;
  }
  if(f1->type != FTPredicate) {
    return TRUE;
  }
  if(f1->data.predicate.function != predHas) {
    return TRUE;
  }
  if(listSize(f1->data.predicate.args) != 2 ||
     listSize(f2->data.predicate.args) != 2) {
    return TRUE;
  }
  /* We compare only the second argument, not the first. This means that
     `has(X@id, OBJA)' and `has(Y^id, OBJA)' are considered identical. 

     predHas() is aware of this aggressive definition. */
  return inputCompareTerm(listNthElement(f1->data.predicate.args, 2),
                          listNthElement(f2->data.predicate.args, 2));
}


/* ----------------------------------------------------------------------
   Looks for common subformulas in a grammar and assigns them the same
   index. At the moment this only detects identical applications of has().
*/
void inputOptimizeGrammar(Input input) {
  Vector v = vectorNew(100);
  List l, m;
  int counter = 0;
  int i,j;
  
  void descendFormula(Formula f) {
    switch(f->type) {
    case FTConjunction:
    case FTDisjunction:
    case FTImplication:
    case FTBiimplication:
      descendFormula(f->data.junction.arg1);
      descendFormula(f->data.junction.arg2);
      break;

    case FTNegation:
      descendFormula(f->data.junction.arg1);
      break;

    case FTPredicate:
      if(f->data.predicate.function == predHas) {
        vectorAddElement(v, f);
      }
      break;
      
    case FTConnexion:
    case FTConstant:
    case FTDirection:
    case FTEqual:
    case FTGreater:
    case FTGreaterEqual:
    case FTLess:
    case FTLessEqual:
    case FTNotEqual:
      return;
    }
  }
  
  /* find all has() applications */
  for(l = m = hashListOfKeys(input->constraints); l; l = listNext(l)) {
    descendFormula(((Constraint)
                    hashGet(input->constraints, listElement(l)))->formula);
  }
  listDelete(m);
  
  /* co-index equal formulas */
  for(i = 0; i < vectorSize(v); i++) {
    Formula f1 = vectorElement(v, i);
    f1->index = -1;
    for(j = 0; j < i; j++) {
      Formula f2 = vectorElement(v, j);
      if(!inputCompareFormula(f1, f2)) {
        f1->index = f2->index;
        break;
      }
    }
    if(f1->index == -1) {
      f1->index = counter++;
    }
  }
#ifdef DEBUGOPTIMIZEGRAMMAR
  for(i = 0; i < vectorSize(v); i++) {
    Formula f = vectorElement(v, i);
    cdgPrintf(CDG_INFO, "<%d> ", f->index);
    printFormula(CDG_INFO, f, f->type, 0);
    cdgPrintf(CDG_INFO, "\n");
  }
#endif
  input->no_has = counter;
  
  /* clean up */
  vectorDelete(v);
}

/* ----------------------------------------------------------------------
   Calculate levelMatrixCounter, hashedUCs, hashedBCs etc.
*/
void inputCacheConstraints(Input input) {
  int i, j, n, o, p;
  HashIterator hi;
  Constraint c;
  VarInfo vi;
  int levelno1 = -1, levelno2 = -1;
  int index, counter;
  
  input->is_context_sensitive = FALSE;
  
  /* check function calls for correct arity */
  inputCheckArities(input);
  
  /* calculate hashedUCs */
  if (input->hashedUCs != NULL) {
    for (i = 0; i < input->hashedUCsize; i++)
      listDelete((List) input->hashedUCs[i]);
    memFree(input->hashedUCs);
    input->hashedUCs = NULL;
  }

  input->hashedUCsize = input->noOfLevels * nrDirections;
  input->hashedUCs =
    (List *) memMalloc(sizeof(List *) * input->hashedUCsize);
  for (i = 0; i < input->noOfLevels * nrDirections; i++)
    input->hashedUCs[i] = NULL;

  index = input->noOfLevels * input->noOfLevels;

  if(input->levelMatrixCounter) {
    vectorDelete(input->levelMatrixCounter);
  }
  input->levelMatrixCounter = vectorNew(index);
  vectorSetAllElements(input->levelMatrixCounter, (void *)0);

  /* calculate hashedBCs */
  if (input->hashedBCs != NULL) {
    for (i = 0; i < input->hashedBCsize; i++)
      listDelete((List) input->hashedBCs[i]);
    memFree(input->hashedBCs);
  }

  input->hashedBCsize = index * nrDirections * nrConnexions * nrDirections;
  input->hashedBCs = (List *) memMalloc(sizeof(List *) * input->hashedBCsize);
  for (i = 0; i < input->hashedBCsize; i++)
    input->hashedBCs[i] = NULL;
  input->constraints = checkedConstraints(input, input->constraints);

  /* count and classify constraints */
  counter = 0;
  hi = hashIteratorNew(input->constraints);
  while((c = hashIteratorNextValue(hi))) {
    c->no = counter++;
    if(c->is_context_sensitive) {
      input->is_context_sensitive = TRUE;
    }
    
    /* collect info about the constraint's signature */
    vi = (VarInfo) listNthElement(c->vars, 1);
    c->sig->l1 = vi->level;
    c->sig->d1 = vi->dir;
    levelno1 = vi->level ? vi->level->no : -1;

    if(listSize(c->vars) > 1) {
      vi = (VarInfo) listNthElement(c->vars, 2);
      c->sig->l2 = vi->level;
      c->sig->d2 = vi->dir;
      levelno2 = vi->level ? vi->level->no : -1;
    }

    if (listSize(c->vars) == 1) {
      c->sig->l1 = vi->level;
      c->sig->d1 = vi->dir;

      for (n = minDirection; n <= maxDirection; n++) {
        if (!subsumesDirection(c->sig->d1, n))
          continue;
        for(i = 0; i < input->noOfLevels; i++) {
          /* if level specified, skip all others */
          if(levelno1 != -1 && levelno1 != i) {
            continue;
          }
          index = inputUCIndex(i, n);
          if (c->penalty == 0 && !c->penaltyTerm)
            input->hashedUCs[index] = listPrependElement(input->hashedUCs[index], c);
          else
            input->hashedUCs[index] = listAppendElement(input->hashedUCs[index], c);
        }
      }

      /* that's all for unary constraints */
      continue;
    }

#ifdef DEBUGINITTABLES
    cdgPrintf(CDG_DEBUG,
              "DEBUG: constraint %s is type {%s}\n",
              c->id, c->signature);
#endif


    /* binary constraints are trickier because they may fall into
       many categories simultaneously: a {X/SYN/\Y/SYN} constraint
       goes into only one cell, but a {X:SYN,Y:SYN} goes into 112
       cells, and a {X,Y} constraint goes into 112n cells. */
    for (i = 0; i < input->noOfLevels; i++) {
      if(levelno1 != -1 && levelno1 != i) continue; 

      for (j = 0; j < input->noOfLevels; j++) {
        if(levelno2 != -1 && levelno2 != j) continue; 

        index = i * input->noOfLevels + j;

        /* count the constraint */
        vectorSetElement(input->levelMatrixCounter,
                           (void *)((int)vectorElement(input->levelMatrixCounter, index) + 1),
                           index);

        /* classify the constraint */
        for (n = minDirection; n <= maxDirection; n++) {
          if (!subsumesDirection(c->sig->d1, n))
            continue;
          for (o = minConnexion; o <= maxConnexion; o++) {
            if (!subsumesConnexion(c->sig->conn, o))
              continue;
            for (p = minDirection; p <= maxDirection; p++) {
              if (!subsumesDirection(c->sig->d2, p))
                continue;

              /* place hard constraints at start of list for early failing */

              index = inputBCIndex(i, j, n, o, p);
#ifdef DEBUGINITTABLES
              cdgPrintf(CDG_DEBUG, "DEBUG:  adding %s to cell %d: {X", c->id, index);
              printDirection(CDG_DEBUG, n);
              cdgPrintf(CDG_DEBUG, "%s", ((Level)listNthElement(input->levels, i+1))->id);
              printConnexion(CDG_DEBUG, o);
              cdgPrintf(CDG_DEBUG, "Y");
              printDirection(CDG_DEBUG, p);
              cdgPrintf(CDG_DEBUG, "%s}\n", ((Level)listNthElement(input->levels, j+1))->id);
#endif
              if (c->penalty == 0)
                input->hashedBCs[index] =
                  listPrependElement(input->hashedBCs[index], c);
              else
                input->hashedBCs[index] =
                  listAppendElement(input->hashedBCs[index], c);
            }
          }
        }
      }
    }
  }
  hashIteratorDelete(hi);


  /* Scan constraint formulas for common expression optimization */
  inputOptimizeGrammar(input);

}

/* ----------------------------------------------------------------------------
 * Gather features from a Term into input->allAttributes,
 * and set t->data.path.pathindex.
 */
void traverseTerm(Input input, Term t) {

  List l;
  
  if(!t) {
    return;
  }
  switch (t->type) {
  case TTTopPeek:
  case TTBottomPeek:
    t->data.peek.pathindex = inputRegisterPath(t->data.peek.path, input);
    break;
  case TTLabel:
  case TTLevel:
    break;
  case TTAdd: 
  case TTSubtract:   
  case TTMultiply:    
  case TTDivide:   
    traverseTerm(input, t->data.operation.op1);
    traverseTerm(input, t->data.operation.op2);
    break;
  case TTFunction:
    for(l = t->data.function.args; l; l = listNext(l)) {
      traverseTerm(input, listElement(l));
    }
    break;
  case TTString:
  case TTNumber:
    break;

  default:
    cdgPrintf(CDG_ERROR, "ERROR: invalid term type %d\n", t->type);
    abort();
  }

  
}

/* ----------------------------------------------------------------------------
 * Gather features from a formula into input->allAttributes.
 */
void traverseFormula(Input input, Formula f) {

  List l;

  switch(f->type) {
  case FTConstant:
    break;
  case FTNegation:
    traverseFormula(input, f->data.formula);
    break;
  case FTPredicate:
    for(l = f->data.predicate.args; l; l = listNext(l)) {
      traverseTerm(input, listElement(l));
    }
    break;
  case FTConjunction:
  case FTDisjunction:
  case FTImplication:
  case FTBiimplication:
    traverseFormula(input, f->data.junction.arg1);
    traverseFormula(input, f->data.junction.arg2);
    break;
  case FTEqual:
  case FTGreater:
  case FTLess:
  case FTGreaterEqual:
  case FTLessEqual:
  case FTNotEqual:
    traverseTerm(input, f->data.relation.op1);
    traverseTerm(input, f->data.relation.op2);
    break;
  case FTConnexion:
  case FTDirection:
    break;
  }
}


/* ----------------------------------------------------------------------------
 * Build the field input->allAttributes for the grammar.
 */
void inputGatherAttributes(Input input) {

  Constraint c;
  AnnoEntry ae;
  HashIterator hi;
  LexiconItem le;
  List l, m, n;
  Boolean annowarning = FALSE;
  
  /* The user's favorite attributes,
     as declared in anno-categories,
     get the best seats. */
  for(l = inputAnnoCats; l != NULL; l = listNext(l)) {
    inputRegisterPath(listElement(l), input);
  }
  
  if(!input->lexiconCurrent) {
    /* (a) find features in lexical entries */
    hi = hashIteratorNew(input->lexicon);
    while(NULL != (l = hashIteratorNextValue(hi))) {
      for(m = l; m != NULL; m = listNext(m)) {
        le = listElement(m);
        traverseItem(1, le, le->value, NULL, input);
      }
    }
    hashIteratorDelete(hi);
  }
  
  if(!input->constraintsCurrent) {
    /* (b) find features in constraints */
    hi = hashIteratorNew(input->constraints);
    while((c = hashIteratorNextValue(hi))) {
      traverseFormula(input, c->formula);
      traverseTerm(input, c->penaltyTerm);
    }
    hashIteratorDelete(hi);
  }

  if(!input->annosCurrent) {
    /* (c) find features in annotations */
    hi = hashIteratorNew(input->annotations);
    while((ae = hashIteratorNextValue(hi))) {
      for (m = ae->annos; m; m = listNext(m)) {
        Annotation an = (Annotation) listElement(m);
        for (n = an->specs; n; n = listNext(n)) {
          Specification s = (Specification) listElement(n);

          List path;
          /* Find and criticize instances of "id/<something>" */
          if (s->type != STTag) {
            continue;
          }
          if (strcmp(s->kind, "id") == 0) {
            if(!annowarning) {
              cdgPrintf(CDG_WARNING,
                        "WARNING: annotation specifies a lexical ID, ignoring it... done.\n");
            }
            annowarning = TRUE;
            continue;
          }
          path = scanPath(s->kind);
          inputRegisterPath(path, input);
          inputFreePath(path);
        }
      }
    }
    hashIteratorDelete(hi);
  }
#ifdef DEBUGINPUTGATHERATTRIBUTES
  {
    List list, path;
    int featureNo = 0;
    
    for (list = input->allAttributes; list; list = listNext(list)) {
      path = listElement(list);
      cdgPrintf(CDG_WARNING, "Feature #%d = ", featureNo++);
      printPath(CDG_WARNING, path);
      cdgPrintf(CDG_WARNING, "\n");
    }
  }
#endif
  
}


/* ----------------------------------------------------------------------------
 * Remove all but one of the list from the lexicon.
 */
void inputRemoveDuplicates(Hashtable lexicon, List class) {
  List l;
  String form = ((LexiconItem)listElement(class))->word;
  List oldlist = hashGet(lexicon, form);
  List newlist = NULL;

  for(l = oldlist; l != NULL; l = listNext(l)) {
    LexiconItem le = listElement(l);
    if(listContains(listNext(class), le)) {
      cdgPrintf(CDG_WARNING, "WARNING: deleting `%s'\n", le->description);
    } else {
      newlist = listPrependElement(newlist, le);
    }
  }
  listDelete(oldlist);
  if(newlist) {
    hashSet(lexicon, form, newlist);
  }
}


/* ----------------------------------------------------------------------------
 * Note PATH as a distinguishing feature for all ITEMS.
 * Also, append the value for path to the description of all items.
 */
void noteDistinction(Input input, List items, List path)
{
  char *p;
  char buffer[1024], feature[255];
  List l;
  LexiconItem le;

  for (l = items; l != NULL; l = listNext(l)) {
    Value v;
    le = listElement(l);
    le->distinctive_features = listAddUniqueElement(le->distinctive_features, path);

    v = le->values[inputLookupPath(path, input)];
    if(v->type != VTError) {
      strcpy(buffer, "");
      strcpy(feature, sprintValue(v));
      p = strcat(buffer, le->description);
      p = strcat(p, "_");
      p = strcat(p, feature);
      
#ifdef DEBUGEXTENDIDS
      cdgPrintf(CDG_DEBUG, "%s ===> %s\n", le->description, buffer);
#endif
      le->description = strRegister(buffer);
    }
  }
}

/* ----------------------------------------------------------------------------
   Annotate lexical items with distinctive features.

   ITEMS is a list of lexical items, presumably with the same lexical
   reading, otherwise they wouldn't need disambiguation.

   FEATURES is a list of paths (i.e., lists of strings) that may be used to
   find differences between the lexical items. We do not use a BitString
   here because the user can specify which features to use preferentially,
   i.e., "name lexicon entries by their case rather than by their gender",
   therefore we need to be able to represent an ordering between the
   features. 

   When a difference is found, each item is renamed so that its name
   reflects the value it has for that path.

   If no difference at all can be found between two items, a warning is
   given and inputRemoveDuplicates() is called to remove one of them.
   
*/
void resolvePartition(Input input, List items, List features)
{

  List path;
  List l, m, class, partitions = NULL;
  LexiconItem le1, le2;

  /* disambiguation complete */
  if (listSize(items) <= 1) {
    return;
  }

  /* duplicate entries? */
  if(!features) {
    cdgPrintf(CDG_WARNING, "WARNING: duplicate lexical definition for `%s'\n",
              ((LexiconItem)listElement(items))->word);
    inputRemoveDuplicates(input->lexicon, items);
    return;
  }

  path = listElement(features);

#ifdef DEBUGRESOLVEPARTITION
  cdgPrintf(CDG_DEBUG, "resolving class `%s' ",
            ((LexiconItem)listElement(items))->description);
  cdgPrintf(CDG_DEBUG, "by feature ");
  printPath(CDG_DEBUG, path);
  cdgPrintf(CDG_DEBUG, ".\n%d members:\n", listSize(items));
                            
  for(m = items; m != NULL; m = listNext(m)) {
    printLexiconItem(CDG_DEBUG, listElement(m));
  }
#endif

  /* classify the items into partitions
     according to their value for the first feature */
  for (l = items; l != NULL; l = listNext(l)) {
    le1 = listElement(l);
    for (m = partitions; m != NULL; m = listNext(m)) {
      class = listElement(m);
      le2 = listElement(class);

      /* insert into correct class */
      if (!inputCompareLeByAtt(le1, le2, path, FALSE)) {
	class = listAppendElement(class, le1);
	break;
      }
    }

    /* open new equivalence class */
    if (m == NULL) {
      class = listPrependElement(NULL, le1);
      partitions = listPrependElement(partitions, class);
    }
  }

  /* if at least some items have different values,
     note this attribute as a distinction */
  if(listSize(partitions) > 1) {
#ifdef DEBUGRESOLVEPARTITION
    cdgPrintf(CDG_DEBUG, "Result of partitioning:");
    for(l = partitions; l != NULL; l = listNext(l)) {
      cdgPrintf(CDG_DEBUG, " %d", listSize(listElement(l)));
    }
    cdgPrintf(CDG_DEBUG, "\n");
#endif
    noteDistinction(input, items, path);
  }

  /* resolve each partition in turn */
  for (l = partitions; l != NULL; l = listNext(l)) {
    resolvePartition(input, listElement(l), listNext(features));
  }

  listForEachDelete(partitions, listDelete);

}


/* ----------------------------------------------------------------------
   Find identically named lexicon items
   and rename them according to their content.
*/
void inputResolveLexiconNames(Input input) {

  HashIterator hi = hashIteratorNew(input->lexicon);
  List l, m;
  while (NULL != (l=hashIteratorNextValue(hi))) {
    for(m = l; m != NULL; m = listNext(m)) {
      LexiconItem le = listElement(m);
      cdgFreeString(le->description);
      le->description = strRegister(le->word);
    }
    resolvePartition(input, l, inputCurrentGrammar->allAttributes);
  }
  hashIteratorDelete(hi);
}

/* ----------------------------------------------------------------------
   Calculate levelPrio, levelMatrixCounter and maxLabelLength.
*/
void inputCacheLevels(Input input) {
  int i;
  List l, m;

  /* find out how long the longest label is
     so we can display solutions more nicely */
  input->maxLabelLength = 0;
  input->maxNrLabels = 0;
  for (l = input->levels; l != NULL; l = listNext(l)) {
    Level level = listElement(l);
    int length = listSize(level->labels);
    for(m = level->labels; m != NULL; m = listNext(m)) {
      i = strlen(listElement(m));
      if(input->maxLabelLength < i) {
        input->maxLabelLength = i;
      }
    }
    input->maxNrLabels = max(input->maxNrLabels, length);
  }
}

/* ----------------------------------------------------------------------------
   Calculate derived fields of InputStruct.

   This function computes all those fields of an Input that are not
   explicitly loaded grammar items. For instance, it classifies all
   existing constraints into the matrix ->hashedBCs for faster access. 
   Therefore it must always be called between adding grammar items to a
   grammar and using it.
*/
void inputCacheInput(Input input)
{
  Level level;
  List l;
  Hierarchy h;
  
  input->attsCurrent =
    input->lexiconCurrent &&
    input->annosCurrent &&
    input->constraintsCurrent;

  if(!input->levelsCurrent) {

    /* count number of levels
       
    This belongs into inputCacheLevels, but there is a cyclical dependency
    between caching levels and constraints, so we do this part first. */
    input->noOfLevels = 0;
    for (l = input->levels; l != NULL; l = listNext(l)) {
      level = (Level) (listElement(l));
      input->noOfLevels = max(input->noOfLevels, level->no + 1);
    }
  }

  /* calculate various stuff to do with constraints */
  if(!input->constraintsCurrent) {
    inputCacheConstraints(input);
  }

  /* calculate various stuff to do with levels */
  if(!input->levelsCurrent) {
    inputCacheLevels(input);
  }
  
  if(!input->attsCurrent) {
    int i = listSize(input->allAttributes);

    /* gather all features used in the grammar */
    inputGatherAttributes(input);

    /* If this step did not turn up any new features, we can leave all
       existing feature tables as they are and just allocate them in the
       new lexicon items. If it did, we have to destroy all existing tables
       and compute them anew. To make this distinction, we set attsCurrent.
    */
    if(i == listSize(input->allAttributes) ){
      input->attsCurrent = TRUE;
    }
    
    /* pre-compute all feature accesses */
    inputCacheLexicon(input);
    inputCacheParses(input, cdgParses);
  }

  /* resolve name clashes in the lexicon */
  if(!input->lexiconCurrent) {
    inputResolveLexiconNames(input);
  }
  
  if(!input->constraintsCurrent) {
    /* associate features with edge types that use them */
    inputCollectFeatures(input);
  }

  input->annosCurrent = TRUE;
  input->lexiconCurrent = TRUE;
  input->constraintsCurrent = TRUE;
  input->levelsCurrent = TRUE;
  input->attsCurrent = TRUE;
  
  /* looks up index of category path */
  (void)taggerLookupCategoryIndex();

  /* look up feature hierarchy */
  if(inputFeatureHierarchy != strRegister("")) {
    h = findHierarchy(inputFeatureHierarchy);
    if (!h) {
      cdgPrintf(CDG_ERROR, 
                "ERROR: no hierarchy `%s' defined.\n",
                inputFeatureHierarchy);
      inputFeatureHierarchy = strRegister("");
    }
  }
}

/* ----------------------------------------------------------------------------
 * frees a complete Input structure
 */
void inputFreeInput( Input ip )
{

  /* get rid of a dll grammar */
#ifndef DARWIN
  comFinitGrammar(ip);
#endif
  
  hashForEachFreeValue(ip->constraints,freeConstraint);
  listForEachDelete(ip->levels, freeLevel);
  hashForEachFreeValue(ip->lexicon, freeLexiconItems);
  listForEachDelete(ip->lattices, freeLattice);
  hashForEachFreeValue(ip->annotations, freeAnnoEntry);
  hashForEachFreeValue(ip->hierarchies, freeHierarchy);
  listForEachDelete(ip->parameters, freeParameter);

  hashForEachFreeValue(ip->maps, hashDelete);
  listDelete(ip->pragmas);
  
  if (ip->hashedUCs) {
    int i;
    for (i=0; i < ip->hashedUCsize; i++) {
      listDelete( (List)ip->hashedUCs[i]);
    }
    memFree(ip->hashedUCs);
  }

  if (ip->hashedBCs) {
    int i;
    for (i = 0; i < ip->hashedBCsize; i++) {
      listDelete( (List)ip->hashedBCs[i]);
    }
    memFree(ip->hashedBCs);
  }

  if (ip->levelMatrixCounter) { vectorDelete(ip->levelMatrixCounter); }

  listForEachDelete(ip->allAttributes, listDelete);
  ip->allAttributes = NULL;

  listForEachDelete(ip->files, cdgFreeString);

  if(ip->upFeatures) {
    int i;
    for(i = 0; i < vectorSize(ip->upFeatures); i++) {
      bitDelete(vectorElement(ip->upFeatures, i));
    }
    vectorDelete(ip->upFeatures);
  }
  if(ip->downFeatures) {
    int i;
    for(i = 0; i < vectorSize(ip->downFeatures); i++) {
      bitDelete(vectorElement(ip->downFeatures, i));
    }
    vectorDelete(ip->downFeatures);
  }
  
  memFree(ip);
}


/* ----------------------------------------------------------------------------
 * build a new input structure
 */
Input inputNewInput(void) 
{
  Input input = (Input) memMalloc( sizeof(InputStruct) );

  input->constraints = hashNew( 500, 0.8, hashStringHashFunction, hashStringEqualFunction );
  ;
  input->levels = NULL;
  input->lexicon = 
    hashNew( 500, 0.8, hashStringHashFunction, hashStringEqualFunction );
  input->lattices = NULL;
  input->annotations = hashNew( 500, 0.8, hashStringHashFunction,
                                hashStringEqualFunction );
  input->hierarchies = hashNew( 500, 0.8, hashStringHashFunction,
                                hashStringEqualFunction );
  input->maps = hashNew( 500, 0.8, hashStringHashFunction,
                         hashStringEqualFunction );
    
  input->parameters = NULL;
  input->pragmas = NULL;
  input->xcdgwordgraphsToDate = FALSE;
  input->xcdglevelsToDate = FALSE;
  input->xcdgconstraintsToDate = FALSE;
  input->xcdglexiconToDate = FALSE;
  input->xcdghierarchiesToDate = FALSE;

  input->annosCurrent = TRUE;
  input->lexiconCurrent = TRUE;
  input->constraintsCurrent = TRUE;
  input->levelsCurrent = TRUE;
  input->attsCurrent = FALSE;

  input->files = NULL;
  input->dllHandle = NULL;
  input->noOfLevels = 0;
  input->maxLabelLength = 0;
  input->hashedUCs = NULL;
  input->hashedBCs = NULL;
  input->hashedUCsize = 0;
  input->hashedBCsize = 0;
  input->levelMatrixCounter = NULL;
  input->allAttributes = NULL;

  input->upFeatures = NULL;
  input->downFeatures = NULL;
  return input;
}

/* ----------------------------------------------------------------------------
 * looks up filename and stores it if necessary in input,
 * returns the canonical string for it
 * maybe extended to look only at the basename
 * TODO: should be a hash table
 */
String inputRegisterFilename( Input input, String s )
{
  char t[PATH_MAX];
  String u;
  
  realpath(s, t);
  u = strRegister(t);
  
  if (!listContains(input->files, u)) {
    /* Ok, it's a new one... */
    input->files = listAppendElement(input->files, u);

    /* Since it's new, inform the user. */
    cdgPrintf(CDG_INFO, "INFO: loading file `%s'\n", u);
  } else {
    cdgFreeString(u);
  }
  return u;
}

/* ----------------------------------------------------------------------------
 * - inserts old items into target->lexicon
 * - should be a local function to mergeInput()
 */
void mergeLexiconItemForMergeInput(List l, Input target)
{
  List m;
  target->lexiconCurrent = FALSE;
  for(m = l; m != NULL; m = listNext(m)) {
    inputLexiconPut(listElement(m), target->lexicon);
  }
  target->xcdglexiconToDate = FALSE;
}

/* ----------------------------------------------------------------------------
 * merges constraint grammar `source' with that already in `target'
 */
void mergeInput(Input target, Input source)
{

  int i;
  HashIterator hi;
  List l, k, kk, tmp;
  Constraint c1, c2;
  AnnoEntry a1, a2;
  Lattice l1;
  Hierarchy h1, h2;
  Hashtable m1, m2;
  String s1;
  
  /* Constraints are uniquely identified by their identifier. */
  hi = hashIteratorNew(source->constraints);
  while((c1 = hashIteratorNextValue(hi))) {
    target->constraintsCurrent = FALSE;
    target->xcdgconstraintsToDate = FALSE;
    if ((c2 = hashGet(target->constraints, c1->id))) {
      /* older version present */
      cdgPrintf(CDG_WARNING,
                "WARNING: Constraint `%s' multiply defined\n",
                c1->id);
      freeConstraint(c2);
    }
    hashSet(target->constraints, c1->id, c1);
  }
  hashIteratorDelete(hi);
  hashDelete(source->constraints);
    

  /* 
   * Level declarations are uniquely identified by their name. Old level
   * declarations are replaced by the new ones with the same name.
   */
  l = source->levels;
  while (l != NULL) {
    target->levelsCurrent = FALSE;
    target->xcdglevelsToDate = FALSE;
    kk = k = target->levels;
    while (k != NULL &&
           strcmp(((Level) listElement(k))->id, ((Level) listElement(l))->id) != 0) {
      kk = k;
      k = listNext(k);
    }

    if (k == NULL) {
      /* no older version */
      if (kk == NULL) {
        target->levels = l;
        l = listNext(l);
        listSetNext(target->levels, NULL);
      } else {
        listSetNext(kk, l);
        l = listNext(l);
        listSetNext(listNext(kk), NULL);
      }
    } else {
      /* k holds older version of l */
      if (kk != k) {
        listSetNext(kk, l);
        l = listNext(l);
        listSetNext(listNext(kk), listNext(k));
        freeLevel((Level) listElement(k));
        memFree(k);
      } else {
        freeLevel((Level) listElement(target->levels));
        tmp = listNext(k);
        memFree(target->levels);
        target->levels = l;
        l = listNext(l);
        listSetNext(target->levels, tmp);
      }
    }
  }
  source->levels = NULL;
  /* Assign unique level numbers. */
  for (i = 0, l = target->levels; l != NULL; i++, l = listNext(l)) {
    ((Level) listElement(l))->no = i;
  }

  /* Lexical entries are uniquely identified by their identifier. */
  {
    Pointer p;

    hi=hashIteratorNew(source->lexicon);

    for (p=hashIteratorNextValue(hi);
         p;
         p=hashIteratorNextValue(hi))
      {
        mergeLexiconItemForMergeInput(p, target);
      }
    hashIteratorDelete(hi);
  }
  hashDelete(source->lexicon);
  source->lexicon = NULL;

  /* 
   * Lattices are uniquely identified by their identifier. 
   */
  for(l = source->lattices; l != NULL; l = listNext(l)) {
    l1 = listElement(l);
    target->xcdgwordgraphsToDate = FALSE;
#if 0
    /* Checking lattices for duplicates is a nice service, but it takes
       quadratic time for a large N, which is annoying. Duplicate lattices
       are no big deal even if they do occur, so I'm disabling it until we
       get a data structure that allows O(1) lookup *and* maintains
       order. */
    if ((l2 = findLatticeInInput(target, l1->id))) {
      /* older version present */
      cdgPrintf(CDG_WARNING,
                "WARNING: Lattice `%s' multiply defined\n",
                l1->id);
      freeLattice(l2);
    }
#endif
    target->lattices = listPrependElement(target->lattices, l1);
  }
  listDelete(source->lattices);
  source->lattices = NULL;

  /* 
   * Annotations are uniquely identified by their identifier. Old annotations
   * are replaced by the new ones with the same identifier.
   */
  hi = hashIteratorNew(source->annotations);
  while((a1 = hashIteratorNextValue(hi))) {
    target->xcdgannotationsToDate = FALSE;
    target->annosCurrent = FALSE;
    if ((a2 = hashGet(target->annotations, a1->id))) {
      /* older version present */
      cdgPrintf(CDG_WARNING,
                "WARNING: Annotation `%s' multiply defined\n",
                a1->id);
      freeAnnoEntry(a2);
    }
    hashSet(target->annotations, a1->id, a1);
  }
  hashIteratorDelete(hi);
  hashDelete(source->annotations);
  
  /* 
   * Hierarchies are uniquely identified by their identifier. Old hierarchies
   * are replaced by the new ones with the same identifier.
   */
  hi = hashIteratorNew(source->hierarchies);
  while((h1 = hashIteratorNextValue(hi))) {
    target->xcdghierarchiesToDate = FALSE;
    if ((h2 = hashGet(target->hierarchies, h1->id))) {
      /* older version present */
      cdgPrintf(CDG_WARNING,
                "WARNING: Hierarchy `%s' multiply defined\n",
                h1->id);
      freeHierarchy(h2);
    }
    hashSet(target->hierarchies, h1->id, h1);
  }
  hashIteratorDelete(hi);
  hashDelete(source->hierarchies);
  source->hierarchies = NULL;


  hi = hashIteratorNew(source->maps);
  while((s1 = hashIteratorNextKey(hi))) {
    m1 = hashGet(source->maps, s1);
    if ((m2 = hashGet(target->maps, s1))) {
      /* older version present */
      cdgPrintf(CDG_WARNING, "WARNING: Map `%s' multiply defined\n", s1);
      hashDelete(m2);
    }
    hashSet(target->maps, s1, m1);
  }
  hashIteratorDelete(hi);
  hashDelete(source->maps);
  source->maps = NULL;

  /* 
   * Parameters are special because only one set of parameters may exist
   * at a given point of time.
   * All parameters are deleted if there are new parameters available.
   */
  if (source->parameters != NULL) {
    listForEachDelete(target->parameters, freeConstraint);
    target->parameters = source->parameters;
    source->parameters = NULL;
  }

  for(l = source->pragmas; l != NULL; l = listNext(l)) {
    if (!listContains(target->pragmas, listElement(l))) {
      target->pragmas = listAppendElement(target->pragmas, listElement(l));
    }
  }

  for (l=source->files; l; l=listNext(l)) {
    if (!listContains(target->files, listElement(l))) {
      /* Ok, it's a new one... */
      target->files = listAppendElement(target->files, listElement(l));
    }
  }
  listDelete(source->files);
  source->files = NULL;

#ifndef DARWIN
  /* hand over the dll grammar */
  if (source->dllHandle) {
    if (target->dllHandle && 
	target->dllHandle != source->dllHandle) {
      cdgPrintf(CDG_WARNING, 
		"WARNING: there's already a dll grammar loaded, unloading old one.\n");
      comFinitGrammar(target);
    }
    /* initialize the dll grammar */
    target->dllHandle = source->dllHandle;
    source->dllHandle = NULL;
    comInitGrammar(target); 
  }
#endif
}

/* ----------------------------------------------------------------------------
 * fillTypesHashFromSortTree
 */
Boolean fillTypesHashFromSortTree( Hierarchy h, Sort s )
{
  Hashtable ht = h->types;
  List l, m;
  Sort t;
  String str;

  if ( hashGet( ht, s->id ) != NULL ) {
    cdgPrintf(CDG_WARNING,
              "WARNING: multiple introduction of sort `%s' in tree hierarchy `%s', ignored\n",
              s->id, h->id );
    return( FALSE );
  }

  /* 
   * Could also be str = s->id since we know that s->id is registered 
   * Maybe we want to add a usage counter to the strings, so we explicitly
   * tell the system that we want to use that string.
   */
  str = strRegister( s->id );

  /* value is index+1 in order to avoid zero (==NULL) */
  (void)hashSet( ht, str, (Pointer)(hashSize( ht )+1) ); 

  /* check whether `bot' subsumes something */
  if ( strcmp( s->id, "bot" ) == 0 && s->subsorts != NULL ) {
    cdgPrintf( CDG_WARNING, 
               "WARNING: special type `bot' must not subsume any type (e.g. `%s') in `%s', ignored\n",
               ((Sort)listElement(s->subsorts))->id, h->id );
    listForEachDelete( s->subsorts, freeSort );
    s->subsorts = NULL;
    /* should this be FALSE? */
    return( TRUE );
  }

  l = m = s->subsorts;
  while ( l != NULL ) {
    t = (Sort)listElement(l);
    if ( strcmp( t->id, "top" ) == 0 ) {
      cdgPrintf( CDG_WARNING,
                 "WARNING: special type `top' must not be subsumed by type `%s' in `%s', ignored\n",
                 s->id, h->id );
    }
    /* calling strcmp twice :-( */
    if ( strcmp( t->id, "top" ) != 0 && fillTypesHashFromSortTree( h, t ) ) {
      m = l;
      l = listNext(l);
    } else {
      if ( m == l ) {
        s->subsorts = listNext(s->subsorts);
        memFree( l );
        l = m = s->subsorts;
      } else {
        listSetNext(m, listNext(l));
        memFree( l );
        l = listNext(m);
      }
      freeSort( t );
    }
  }
  return (TRUE);
}

/* ----------------------------------------------------------------------------
 * markDirectSucessorsInHierarchy
 */
void markDirectSucessorsInHierarchy( Hierarchy h, Sort s )
{
  Array a = h->distance;
  Hashtable ht = h->types;
  List l;
  Sort t;
  int x, y;

  if ( 0 > ( x = (int)hashGet( ht, s->id ) - 1 ) ) {
    cdgPrintf(CDG_ERROR,
              "ERROR: error in markDirectSucessorsInHierarchy with `%s'\n",
              s->id);
    return;
  }

  for ( l = s->subsorts; l != NULL; l = listNext(l) ) {
    t = (Sort)listElement(l);
    if ( 0 > ( y = (int)hashGet( ht, t->id ) - 1 ) ) {
      cdgPrintf(CDG_ERROR,
                "ERROR: error in markDirectSucessorsInHierarchy with `%s'\n",
                t->id);
      return;
    }
    arraySetElement( a, (Pointer)1, x, y );    
    markDirectSucessorsInHierarchy( h, t );
  }
}

/* ----------------------------------------------------------------------------
 * transitiveClosure
 */
void transitiveClosure( Array a, int size )
{
  int i, j, k;

  /* Warshall's algorithm for transitive closure */
  for ( i = 0; i < size; i++ ) {
    for ( j = 0; j < size; j++ ) {
      int ji = (int)arrayElement( a, j, i );

      if ( ji > 0 ) {
        for ( k = 0; k < size; k++ ) {
          int ik = (int)arrayElement( a, i, k );
          int jk = (int)arrayElement( a, j, k );

          if ( ik > 0 && ( jk == -1 || ji+ik < jk ) ) {
            arraySetElement( a, (Pointer)(ji+ik), j, k );
          }
        }
      }
    }
  } 
}

/* ----------------------------------------------------------------------------
 * inputSortTreeToHierarchy
 */
Hierarchy inputSortTreeToHierarchy( String id, Sort s )
{
  int size, i, j;
  int top, bot;
  Hierarchy h = (Hierarchy)memMalloc( sizeof(HierarchyStruct) );

  h->id = strRegister(id);
  h->types = hashNew( 20, 0.7, hashStringHashFunction, hashStringEqualFunction );
  /* collects types in hashtable */
  fillTypesHashFromSortTree( h, s );
  /* if `top' not already there, insert it */
  if ( 0 > (top = (int)hashGet( h->types, "top" ) - 1 ) ) {
    top = hashSize( h->types );
    (void)hashSet( h->types, strRegister( "top" ), (Pointer)(top+1) );     
  }
  /* if `bot' not already there, insert it */
  if ( 0 > (bot = (int)hashGet( h->types, "bot" ) - 1 ) ) {
    bot = hashSize( h->types );
    (void)hashSet( h->types, strRegister( "bot" ), (Pointer)(bot+1) );
  }

  size = hashSize( h->types );

  /* register direct subsumption */
  h->distance = arrayNew( size, size, 0 );
  arraySetAllElements( h->distance, (Pointer)-1 );
  markDirectSucessorsInHierarchy( h, s );

  for ( i = 0; i < size; i++ ) {
    Boolean iSubsumesSomething = FALSE;
    Boolean iIsSubsumed = FALSE;

    for ( j = 0; j < size; j++ ) {
      if ( (int)arrayElement( h->distance, i, j ) > 0 ) {
        iSubsumesSomething = TRUE;
      }
      if ( (int)arrayElement( h->distance, j, i ) > 0 ) {
        iIsSubsumed = TRUE;
      }
    }
    if ( iSubsumesSomething == FALSE ) {
      arraySetElement( h->distance, (Pointer)1, i, bot );
    }
    if ( iIsSubsumed == FALSE ) {
      arraySetElement( h->distance, (Pointer)1, top, i );
    }
    /* reflexivity */
    arraySetElement( h->distance, (Pointer)0, i, i );
  }

  transitiveClosure( h->distance, size );
  /* save because we copied everything in fillTypesHashFromSortTree */
  freeSort( s );

  return( h );
}

/* ----------------------------------------------------------------------------
 * inputSubsumptionListToHierarchy
 */
Hierarchy inputSubsumptionListToHierarchy( String id, List l )
{
  int size, x, y;
  List m, n;
  Subsumption s;
  char *str;
  Hierarchy h = (Hierarchy)memMalloc( sizeof(HierarchyStruct) );
  int top, bot;
  int i, j;
  HashIterator hi;

  h->id = strRegister(id);
  h->types = hashNew( 20, 0.7, hashStringHashFunction, hashStringEqualFunction );
  for ( n = l; n != NULL; n = listNext(n) ) {
    s = (Subsumption)listElement(n);
    if ( NULL == hashGet( h->types, s->type ) ) {
      str = strRegister( s->type );      
      /* value is index+1 in order to avoid zero (==NULL) */
      (void)hashSet( h->types, str, (Pointer)(hashSize( h->types )+1) ); 
    }
    for ( m = s->types; m != NULL; m = listNext(m) ) {
      if ( NULL == hashGet( h->types, listElement(m) ) ) {
        str = strRegister( (String)listElement(m) );
        /* value is index+1 in order to avoid zero (==NULL) */
        (void)hashSet( h->types, str, (Pointer)(hashSize( h->types )+1) ); 
      }
    }
  }

  /* if `top' not already there, insert it */
  if ( 0 > (top = (int)hashGet( h->types, "top" ) - 1 ) ) {
    top = hashSize( h->types );
    (void)hashSet( h->types, strRegister( "top" ), (Pointer)(top+1) );     
  }
  /* if `bot' not already there, insert it */
  if ( 0 > (bot = (int)hashGet( h->types, "bot" ) - 1 ) ) {
    bot = hashSize( h->types );
    (void)hashSet( h->types, strRegister( "bot" ), (Pointer)(bot+1) );
  }

  size = hashSize( h->types );

  h->distance = arrayNew( size, size, 0 );
  arraySetAllElements( h->distance, (Pointer)-1 );
  for ( n = l; n != NULL; n = listNext(n) ) {
    s = (Subsumption)listElement(n);
    if ( strcmp( s->type, "top" ) == 0 && s->subsumesFlag == FALSE ) {
      cdgPrintf( CDG_WARNING, 
                 "WARNING: special type `top' must not be subsumed by any type in `%s', ignored\n",
                 h->id );
      continue;
    }
    if ( strcmp( s->type, "bot" ) == 0 && s->subsumesFlag == TRUE ) {
      cdgPrintf( CDG_WARNING, 
                 "WARNING: special type `bot' must not subsume any other type in `%s', ignored\n",
                 h->id );
      continue;
    }
    x = (int)hashGet( h->types, s->type ) - 1;

    /* register direct subsumption */
    for ( m = s->types; m != NULL; m = listNext(m) ) {
      y = (int)hashGet( h->types, (char *)listElement(m) ) - 1;
      if ( s->subsumesFlag == TRUE ) {
        if ( y == top ) {
          cdgPrintf( CDG_WARNING, 
                     "WARNING: special type `top' must not be subsumed by type `%s' in `%s', ignored\n",
                     s->type, h->id );
          continue;
        }
        arraySetElement( h->distance, (Pointer)1, x, y );
      } else {
        if ( y == bot ) {
          cdgPrintf( CDG_WARNING, 
                     "WARNING: special type `bot' must not subsume type `%s' in `%s', ignored\n",
                     s->type, h->id );
          continue;
        }
        arraySetElement( h->distance, (Pointer)1, y, x );
      }
    }
  }

  for ( i = 0; i < size; i++ ) {
    Boolean iSubsumesSomething = FALSE;
    Boolean iIsSubsumed = FALSE;

    for ( j = 0; j < size; j++ ) {
      if ( (int)arrayElement( h->distance, i, j ) > 0 ) {
        iSubsumesSomething = TRUE;
      }
      if ( (int)arrayElement( h->distance, j, i ) > 0 ) {
        iIsSubsumed = TRUE;
      }
    }
    if ( iSubsumesSomething == FALSE && i != bot ) {
      arraySetElement( h->distance, (Pointer)1, i, bot );
    }
    if ( iIsSubsumed == FALSE && i != top ) {
      arraySetElement( h->distance, (Pointer)1, top, i );
    }
  }

  transitiveClosure( h->distance, size );

  /* check for cycles */
  hi = hashIteratorNew( h->types );
  while ( NULL != ( str = (char *)hashIteratorNextKey( hi ) ) ) {
    x = (int)hashGet( h->types, str ) - 1;
    if ( (int)arrayElement( h->distance, x, x ) > 0 ) {
      cdgPrintf( CDG_WARNING,
                 "WARNING: cycle in hierarchy `%s' for type `%s'\n",
                 id, str );
    }
    /* reflexivity, can't be done before */
    arraySetElement(h->distance, (Pointer)0, x, x );
  }
  hashIteratorDelete( hi );


  /* TODO: delete complete subsumption list l */
  listForEachDelete(l, freeSubsumption);

  return ( h );
}

/* ----------------------------------------------------------------------
   Transform list of lists to a proper data map.

   The input parser reads data maps as lists of lists for simplicity.
   Here the list structure is converted to a proper Hashtable that maps
   strings to Values. If there are multiple keys in one entry declaration,
   they are first concatenated with `:'.   
*/
Hashtable inputHashFromList(List items) {
  List l, m;
  Hashtable h = hashNew( 500, 0.8, hashStringHashFunction,
                         hashStringEqualFunction );
  String sep = strRegister(":");

  for(l = items; l != NULL; l = listNext(l)) {
    String key, value;
    m = listElement(l);
    key = strCopy(listElement(m));
    m = listNext(m);
    while(m && listNext(m)) {
      String s =listElement(m);
      key = strAppend(key, sep, s, NULL);
      m = listNext(m);
    }
    key = strRegister(key);
    value = listElement(m);
    hashSet(h, key, value);
  }
  return h;
}

/* ----------------------------------------------------------------------------
 * print constraint
 */
void printConstraint(long unsigned int mode, Constraint c) {

  cdgPrintf(mode, "#%d: %s  %s\n\n",
            c->no,
            c->active ? "(ACTIVE) " : "(PASSIVE)",
            formatConstraint(c));

  if(c->is_context_sensitive) {
    cdgPrintf(mode, "\nThis constraint is context-sensitive.\n");
  }

}

/* ----------------------------------------------------------------------
   Return formatted representation of a constraint's penalty.
*/
String formatPenalty(Constraint c) {
  return c->penaltyTerm ? formatTerm(c->penaltyTerm, TRUE) : strPrintf("%g", c->penalty);
}

/* ----------------------------------------------------------------------------
 * Return formatted representation of a constraint.
 */

String formatConstraint(Constraint c) {

  return strPrintf("{%s} : %s : %s : %s :\n%s;", 
                   formatSignature(c->sig),
                   c->id,
                   c->section->id,
                   formatPenalty(c),
                   formatFormula(c->formula, FTConstant, 0));
}


/* ----------------------------------------------------------------------------
 * break a formula 
 */
String formatIndent(int indent)
{
  return strPrintf("\n%*s", indent, "");
}

/* ----------------------------------------------------------------------
 * Print a formula.
 */
void printFormula(unsigned long mode, Formula f, FormulaType enclosing, int indent) {
  cdgPrintf(mode, "%s", formatFormula(f, enclosing, indent));
}

/* ----------------------------------------------------------------------
   Return formatted representation of a Formula.
*/
String formatFormula(Formula f, FormulaType enclosing, int indent) {

  if (f->type == FTConstant) {
    return strPrintf("%s", f->data.constant ? "true" : "false");
  } 

  if (f->type == FTNegation) {

    /* An indentation is necessary because the unary ~ 
       takes up a column that might cause following lines
       to be misaligned. */
    /* But we do not print brackets here because some simple
       types of formulas might not want them, as in `~true'. */

    return strPrintf("~%s", formatFormula(f->data.formula, f->type, indent+1));
  }

  if (f->type == FTPredicate) {
    String args = NULL;
    List l = f->data.predicate.args;
    while (l != NULL) {
      args = strCat(args, formatTerm((Term) listElement(l), FALSE));
      if(listNext(l)) {
        args = strCat(args, ",");
      }
      l = listNext(l);
    }
    return strPrintf("%s(%s)",
                     f->data.predicate.functor,
                     args);
  }

  /* & and | are treated specially to avoid redundant brackets. */
  if (f->type == FTConjunction || f->type == FTDisjunction) {

    String result = NULL;

    /* suppress brackets if the enclosing Formula
       is of the same type */
    if(enclosing != f->type) {
      indent++;
      result = strCat(result, "(");
    }

    result =
      strCat(result, formatFormula(f->data.junction.arg1, f->type, indent));
    result = strCat(result,
                    strPrintf(" %s", f->type == FTConjunction ? "&" : "|"));

    /* always break at & or | */
    result = strCat(result, formatIndent(indent));

    result =
      strCat(result, formatFormula(f->data.junction.arg2, f->type, indent));

    /* suppress brackets if the enclosing Formula
       is of the same type */
    if(enclosing != f->type) {
      result = strCat(result, ")");
      indent--;
    }
    return result;
  }

  else if (f->type == FTImplication || f->type == FTBiimplication) {
    FormulaType left = f->data.junction.arg1->type;
    FormulaType right = f->data.junction.arg2->type;
    Boolean flat = 
      left != FTNegation &&
      left != FTConjunction &&
      left != FTDisjunction &&
      left != FTImplication &&
      left != FTBiimplication &&
      right != FTNegation &&
      right != FTConjunction &&
      right != FTDisjunction &&
      right != FTImplication &&
      right != FTBiimplication;

    return strPrintf("(%s%s%s%s%s)", 
                     
                     formatFormula(f->data.junction.arg1, f->type, indent+1),

                     /* these operators get a line of their own 
                        unless both arguments are flat */
                     flat ? " " : formatIndent(indent+1),
                     f->type == FTImplication ? "->" : "<->",
                     flat ? " " : formatIndent(indent+1),
                     formatFormula(f->data.junction.arg2, f->type, indent+1));
  }

  else if (f->type == FTEqual || f->type == FTGreater ||
           f->type == FTLess || f->type == FTGreaterEqual ||
           f->type == FTLessEqual || f->type == FTNotEqual) {

    return strPrintf("%s%s%s%s%s",
                     
                     /* for clarity */
                     enclosing == FTNegation ? "(" : "",
                     formatTerm(f->data.relation.op1, FALSE),
                     f->type == FTEqual ? " = " :
                     f->type == FTGreater ? " > " :
                     f->type == FTLess ? " < " :
                     f->type == FTGreaterEqual ? " >= " :
                     f->type == FTLessEqual ? " <= " : " != ",
                     formatTerm(f->data.relation.op2, FALSE),
                     /* for clarity */
                     enclosing == FTNegation ? ")" : "");
    
  }

  else if (f->type == FTConnexion) {
    return strPrintf("%s%s%s",
                     f->data.connexion.var1->varname,
                     formatConnexion(f->data.connexion.c),
                     f->data.connexion.var2->varname);
  } else if (f->type == FTDirection) {
    return strPrintf("%s%s",
                     f->data.direction.var->varname,
                     formatDirection(f->data.direction.d));
  }
  else {
    return NULL;
  }
}

/* ----------------------------------------------------------------------------
   Print the Term T in mode MODE.

   If termContext is set, directly enclosing terms are enclosed in square
   brackets. This is so that the output of printTerm() is always valid CDG
   input, since Terms may not be nested directly into each other according
   to the input grammar.
*/
void printTerm(long unsigned int mode, Term t, Boolean termContext) {
  cdgPrintf(mode, "%s", formatTerm(t, termContext));
}

   
/* ----------------------------------------------------------------------
   Return formatted representation of a Term.
   
   See printTerm() for the meaning of termContext.
*/
String formatTerm(Term t, Boolean termContext)
{
  if (t->type == TTTopPeek) {
    return strPrintf("%s^%s",
                     ((VarInfo) t->data.peek.varinfo)->varname,
                     formatPath(t->data.peek.path));
  } else if (t->type == TTBottomPeek) {
    return strPrintf("%s@%s",
                     ((VarInfo) t->data.peek.varinfo)->varname,
                     formatPath(t->data.peek.path));
  } else if (t->type == TTLabel) {
    return strPrintf("%s.label", ((VarInfo) t->data.label)->varname);
  } else if (t->type == TTLevel) {
    return strPrintf("%s.level", ((VarInfo) t->data.level)->varname);
  } else if (t->type == TTAdd) {
    return strPrintf("%s%s+%s%s",
                     termContext ? "[" : "",
                     formatTerm(t->data.operation.op1, TRUE),
                     formatTerm(t->data.operation.op2, TRUE),
                     termContext ? "]" : "");
  } else if (t->type == TTSubtract) {
    return strPrintf("%s%s-%s%s",
                     termContext ? "[" : "",
                     formatTerm(t->data.operation.op1, TRUE),
                     formatTerm(t->data.operation.op2, TRUE),
                     termContext ? "]" : "");
  } else if (t->type == TTMultiply) {
    return strPrintf("%s%s*%s%s",
                     termContext ? "[" : "",
                     formatTerm(t->data.operation.op1, TRUE),
                     formatTerm(t->data.operation.op2, TRUE),
                     termContext ? "]" : "");
  } else if (t->type == TTDivide) {
    return strPrintf("%s%s/%s%s",
                     termContext ? "[" : "",
                     formatTerm(t->data.operation.op1, TRUE),
                     formatTerm(t->data.operation.op2, TRUE),
                     termContext ? "]" : "");
  } else if (t->type == TTFunction) {
    String args = NULL;
    List l = t->data.function.args;
    while (l != NULL) {
      args = strCat(args, formatTerm((Term) listElement(l), FALSE));
      if(listNext(l)) {
        args = strCat(args, ",");
      }
      l = listNext(l);
    }
    return strPrintf("%s(%s)",
                     t->data.function.functor,
                     args);
  } else if (t->type == TTString) {
    return formatIdentifier(t->data.string);
  } else if (t->type == TTNumber) {
    if ( t->data.number - (int)t->data.number == 0.0 ) {
      return strPrintf("%d", (int)t->data.number);
    } else {
      return strPrintf("%g", t->data.number);
    }
  }
  else return NULL;
}


/* ----------------------------------------------------------------------------
 * print level declaration
 */
void printLevel(long unsigned int mode, Level l)
{
  List m;

  cdgPrintf(mode, "%s", l->id);
  if ( l->properties != NULL ) {
    cdgPrintf( mode, " [" );
    for (m = l->properties; m != NULL; m = listNext(m)) {
      cdgPrintf( mode, " %s=%s", ((Property)listElement(m))->name,  ((Property)listElement(m))->value );
    }
    cdgPrintf( mode, " ]" );
  }
  cdgPrintf( mode, " #" );
  for (m = l->labels; m != NULL; m = listNext(m)) {
    cdgPrintf(mode, " ");
    cdgPrintf(mode, formatIdentifier(listElement(m)));
  }
  cdgPrintf(mode, ";");

  cdgPrintf(mode, "  (%s) (%s)\n", 
            l->showflag ? "shown" : "hidden",
            l->useflag ? "used" : "unused");


}

/* ----------------------------------------------------------------------
   Complain in a legible way that V is not of type VT.
*/
void printTypeError(unsigned long mode, Value v, ValueType vt) {
  cdgPrintf(mode, "  Value `");
  if(v->type == VTError) {
    cdgPrintf(mode, "<ERROR>");
  } else {
    printValue(mode, v, 0);
  }
  cdgPrintf(CDG_WARNING, "' is not ");
  switch (vt) {
  case VTString:
    cdgPrintf(mode, "a string\n");
    break;
  case VTNumber:
    cdgPrintf(mode, "a number\n");
    break;
  case VTAVNode:
    cdgPrintf(mode, "an AV node\n");
    break;
  case VTList:
    cdgPrintf(mode, "a list\n");
    break;
  case VTDisjunction:
    cdgPrintf(mode, "a disjunction\n");
    break;
  case VTConjunction:
    cdgPrintf(mode, "a conjunction\n");
    break;
  case VTLexemNode:
    cdgPrintf(mode, "a lexeme node\n");
    break;
  default:
    break;
  }  
}

/* ----------------------------------------------------------------------------
 * print value
 */
void printValue(unsigned long mode, Value v, int indent)
{
  cdgPrintf(mode, formatValue(v, indent));
}

/* ----------------------------------------------------------------------------
 * print value
 */
String formatValue(Value v, int indent)
{
  List l;
  Boolean flat;
  String str = NULL;
  Value nextValue;

  switch (v->type) {
    /* format a string */
  case VTString:
    str = formatIdentifier(v->data.string);
    break;
    /* format a number */
  case VTNumber:
    if (v->data.number - (int)v->data.number == 0.0) {
      str = strPrintf("%d", (int)v->data.number);
    } else {
      str = strPrintf("%g", v->data.number);
    }
    break;
    /* format an avnode */
  case VTAVNode:
    nextValue = ((AVNode) v->data.avnode)->value;
    if (nextValue->type == VTConjunction) {
      str = strPrintf("\n%s%s:%s", 
                      formatIndent(indent+1),
                      ((AVNode) v->data.avnode)->attribute,
                      formatValue(nextValue, indent+1));
    } else {
      str = strPrintf("%s:%s", ((AVNode) v->data.avnode)->attribute,
                      formatValue(nextValue, indent));
    }
    break;
    /* format a list of values */
  case VTList:
    for (l = v->data.list; l; l = listNext(l)) {
      if (!str) {
        str = formatValue(listElement(l), indent);
      } else {
        str = strPrintf("%s, %s", str, formatValue(listElement(l), indent));
      }
    }
    str = strPrintf("<%s>", str);
    break;
    /* format a disjunction of values */
  case VTDisjunction:
    flat = ((Value) listElement((List) v->data.list))->type != VTConjunction;
    if (!flat) {
      str = formatIndent(indent);
    }
      
    if (v->index) {
      str = strPrintf("\n%s#%d(", str, v->index);
    } else {
      str = strPrintf("\n%s(", str);
    }

    for (l = v->data.list; l != NULL; l = listNext(l)) {
      str = strPrintf("%s%s", str, formatValue(listElement(l), indent+1));
      if (listNext(l)) {
        if (flat) {
          str = strPrintf("%s | ", str);
        } else {
          str = strPrintf("%s\n%s|", str, formatIndent(indent));
        }
      }
    }

    str = strPrintf("%s\n%s)", str, flat?"":formatIndent(indent));
    break;
    /* format a conjunction of values */
  case VTConjunction:
    if(!v->data.list) {
      str = strPrintf("[]");
      break;
    }
    nextValue = listElement(v->data.list);
    if (nextValue->type == VTAVNode) {
      nextValue = ((AVNode)nextValue->data.avnode)->value;
    } 
    if (nextValue->type == VTConjunction) {
      str = strPrintf("[");
    } else {
      str = strPrintf("[\n%s", formatIndent(indent+1));
    }
    for (l = v->data.list; l != NULL; l = listNext(l)) {
      str = strPrintf("%s%s%s", str, formatValue(listElement(l), indent+1),
                      listNext(l)?", ":"");
    }
    str = strPrintf("%s\n%s]", str, formatIndent(indent));
    break;
    /* never reach */
  default:
    cdgPrintf(CDG_ERROR,
              "ERROR: formatValue: illegal value type (%d)\n", v->type);
    abort();
  }

  return str;
}

/* ----------------------------------------------------------------------------
 * print lexical entry
 */
void printLexiconItem(long unsigned int mode, LexiconItem l)
{
  cdgPrintf(mode, formatLexiconItem(l));
}

/* ----------------------------------------------------------------------------
 * format lexical entry
 */
String formatLexiconItem(LexiconItem l)
{
  return strPrintf("%s := %s : %s;\n", l->description, l->word, 
                   formatValue(l->value, 0));
}

/* ----------------------------------------------------------------------------
 * print annotation entry
 */
void printAnnoEntry(long unsigned int mode, AnnoEntry ae)
{
  List l, m;
  Annotation a;

  cdgPrintf(mode, "%s : %s : <-> ", ae->id, ae->lattice);
  for (l = ae->annos; l != NULL; l = listNext(l)) {
    a = (Annotation) listElement(l);
    cdgPrintf(mode, "\n%3.3d %3.3d %s:", a->from, a->to, a->word);
    for (m = a->specs; m != NULL; m = listNext(m)) {
      if (((Specification) listElement(m))->type == STTag) {
        cdgPrintf(mode, "\n  %s/%s",
                  ((Specification) listElement(m))->kind,
                  ((Specification) listElement(m))->name);
      } else {
        cdgPrintf(mode, "\n  %s->%s->%d",
                  ((Specification) listElement(m))->kind,
                  ((Specification) listElement(m))->name,
                  ((Specification) listElement(m))->position);
      }
    }
  }
  cdgPrintf(mode, "\n");
}

/* ----------------------------------------------------------------------------
 * print sort
 */
void printSort(long unsigned int mode, Sort s)
{
  List l;
  int i;

  for (i = 0; i < s->depth; i++)
    cdgPrintf(mode, "  ");
  cdgPrintf(mode, "%s", s->id);
  if (s->subsorts != NULL) {
    cdgPrintf(mode, "(\n");
    for (l = s->subsorts; l != NULL; l = listNext(l))
      printSort(mode, (Sort) listElement(l));
    for (i = 0; i < s->depth; i++)
      cdgPrintf(mode, "  ");
    cdgPrintf(mode, ")");
  }
  cdgPrintf(mode, "\n");
}


/* ----------------------------------------------------------------------------
 * print hierarchy
 */
void printHierarchy(unsigned long mode, Hierarchy h)
{
  Hashtable ht = h->types;
  List keys = hashListOfKeys(ht);
  List done, nodes, checkdist, temp;
  Array a = h->distance;
  int i, j, x, y, k;
  Pointer item;

  /* remove "bot" */
  keys = listDeleteElement(keys, strRegister( "bot" ));

  nodes = NULL;                        /* initialize lists "done" and "nodes" */
  done = NULL;                         /* list "done" will later be used for pruning */
  cdgPrintf(mode, "\n\n");
  i = 0;                               /* level in hierarchy and number of tabulators to print 
                                        */

  k = 1;                               /* find "top" */
  item = listNthElement(keys, k);
  while (item != strRegister( "top" )) {
    if (listNthElement(keys, k) == NULL) {
      cdgPrintf(mode, "Oops ! Top was not found");
      exit(1);}
    k = k + 1;
    item = listNthElement(keys, k);
  }                                   


  nodes = listPrependElement(nodes, item); /* append "top" to list */
  nodes = listPrependElement(nodes, (Pointer) i);  /* cast is necessary because lists expect
                                                 * pointers */
  /* the list "nodes" now contains "0" and "top" */
  while (nodes != NULL) {
    i = (int)listElement(nodes) + 1; /* determine level in hierarchy for */
    /* the next entries in "nodes" */

    for (j = (int)listElement(nodes); j > 0; j = j - 1) {  /* print tabulators */
      cdgPrintf(mode, "\t");
    }
    temp = nodes;                      /* list "temp" only needed to free elements */
    nodes = listNext(nodes);               /* first entry is no longer needed */
    memFree(temp);                     /* free element */

    item = listElement(nodes);

    temp = nodes;                      /* as above */
    nodes = listNext(nodes);
    memFree(temp);

    if (listContains(done, item))
      cdgPrintf(mode, "%-25.25s (look above)\n", (char *)item); /* this part of hierarchy has
                                                                 * already been printed */
    else {
      cdgPrintf(mode, "%-25.25s\n", (char *)item);      /* this part of the hierarchy is new */
      done = listPrependElement(done, item);
      x = (int)hashGet(ht, item) - 1;  /* determine x-coordinate for distance-table */

      for (checkdist = keys; checkdist != NULL; checkdist = listNext(checkdist)) {
        y = (int)hashGet(ht, listElement(checkdist)) - 1;
        if ((int)arrayElement(a, x, y) == 1) {       /* direct successor ? */
          nodes = listPrependElement(nodes, listElement(checkdist));
          nodes = listPrependElement(nodes, (Pointer) i);
        }
      }
    }
  }

  cdgPrintf(mode, "\n\n");
  listDelete(nodes);
  listDelete(keys);
  listDelete(done);

}

/* ----------------------------------------------------------------------------
 * print lattice
 */
void printLattice(long unsigned int mode, Lattice lattice)
{
  List l;

  /* trivial lattices are printed inline... */
  if (!(latticeBranches(lattice) || latticeHasInfo(lattice))) 
    {
      cdgPrintf(mode, "%s : ", lattice->id);
      for (l = lattice->arcs; l != NULL; l = listNext(l)) {
        Arc arc=listElement(l);

        cdgPrintf(mode, "%s%s", arc->word, listNext(l) ? ", " : ";");
      }
    } else { 
      /* ...but nontrivial ones are enumerated line by line */
      cdgPrintf(mode, "\n%s : ", lattice->id);
      cdgPrintf(mode, "\n");

      for (l = lattice->arcs; l != NULL; l = listNext(l)) {
        Arc arc=listElement(l);

        cdgPrintf(mode, "\t%2d %2d %15s %f ",
                  arc->from, arc->to,
                  arc->word, arc->penalty);
        if (arc->info) {
          printValue(mode, arc->info, 0);
        }      
        cdgPrintf(mode, "\n");
      }
    }

  cdgPrintf(mode, "\n");
}

/* ----------------------------------------------------------------------------
 * print path
 */
void printPath(unsigned long mode, List p) {
  cdgPrintf(mode, formatPath(p));
}

/* ----------------------------------------------------------------------
   Return formatted representation of a Path.
*/
String formatPath(List p) {
  String result = NULL;
  while (p != NULL) {
    result = strCat(result, formatIdentifier(listElement(p)));
    if (listNext(p) != NULL) {
      result = strCat(result, ":");
    }
    p = listNext(p);
  }
  return result;
}


/* ----------------------------------------------------------------------------
 * print path to String
 */
String sprintPath(List p) {
  char buffer[1024];
  List l;
  Boolean first = TRUE;
  Boolean error;
  int index;
  
  strcpy(buffer, "");
  index = 0;
  for(l = p; l != NULL; l = listNext(l)) {
    String item = listElement(l);
    error = FALSE;

    if(!first) {
      index++;
      if(index>1023) {
        error = TRUE;
      } else {
        strcat(buffer, ":");
      }
    }
    index += strlen(item);
    if(index > 1023) {
      error = TRUE;
    } else {
      strcat(buffer, item);
    }
    if(error) {
      cdgPrintf(CDG_ERROR, "PANIC: path longer than 1024 bytes!\n");
      cdgPrintf(CDG_ERROR, "PANIC: Who's supposed to understand that, anyway?\n");
      abort();
    }
    first = FALSE;
  }

  return strRegister(buffer);
}

/* ----------------------------------------------------------------------------
 * Reconstruct path from its printed representation.
 * The returned list contains a list of shared strings.
 */
List scanPath(String s) {
  List result = NULL;
  char buffer[strlen(s)+1];
  char *bufptr = buffer;
  String token;
  
  strcpy(buffer, s);

  /* get :-separated list of components */
  token = strtok_r(buffer, ":", &bufptr);
  while (token) {
    result = listAppendElement(result, strRegister(token));
    token = strtok_r(NULL, ":", &bufptr);
  }

  return result;
}


/* ----------------------------------------------------------------------------
 * remove nested conjunctions from a Value
 */

void cleanup(Value v)
{

  List p, q = NULL;
  List add = NULL;

  switch (v->type) {
  case VTAVNode:
    cleanup(((AVNode) v->data.avnode)->value);
    break;
  case VTList:
  case VTDisjunction:
    for (p = v->data.list; p != NULL; p = listNext(p))
      cleanup((Value) listElement(p));
    break;
  case VTConjunction:
    for (p = v->data.list; p != NULL; p = listNext(p)) {
      cleanup((Value) listElement(p));
      if (((Value) listElement(p))->type == VTConjunction) {
        add = listAppendList(add, ((Value) listElement(p))->data.list);
        ((Value) listElement(p))->data.list = NULL;
      }
    }
    /* remove conjunctions... */
    for (p = v->data.list; p != NULL; p = listNext(p)) {
      if (((Value) (listElement(p)))->type != VTConjunction)
        q = listAppendElement(q, listElement(p));
    }
    listDelete(v->data.list);
    v->data.list = q;

    /* ... and add their contents */
    v->data.list = listAppendList(v->data.list, add);
    
  default:
    break;
  }

}


/* ----------------------------------------------------------------------------
 * create a deep copy of a Value
 */

Value cloneValue(Value v)
{

  /* allocate new vv;
     vv->type = v->type;

     IF v is numeric
     vv->data.number = v->data.number;

     ELSIF v is a string
     vv->data.string = strdup( v->data.string ); 

     ELSIF v is an AVNode
     allocate new av;
     av->attribute = strdup( v->data.avnode->attribute );
     av->value = cloneValue( v->data.avnode->value );
     vv->data.avnode = av;

     ELSIF v is a conjunction or disjunction
     l = listClone( v->data.list );
     FOR all p in l:
     p = cloneValue( p );
     ROF

     FI

     return vv.  

  */

  Value vv;
  AVNode av;
  List l, p;

  if(!v) {
    return NULL;
  }
  
  vv = (Value) memMalloc(sizeof(ValueStruct));
  vv->type = v->type;
  vv->index = v->index;

  switch (v->type) {
  case VTNumber:
    vv->data.number = v->data.number;
    break;
  case VTString:
    vv->data.string = strRegister( v->data.string );
    break;
  case VTAVNode:
    av = (AVNode) memMalloc(sizeof(AVNodeStruct));
    av->attribute = strRegister( ((AVNode) v->data.avnode)->attribute );
    av->value = cloneValue(((AVNode) v->data.avnode)->value);
    vv->data.avnode = (struct AVNodeStruct *)av;
    break;
  case VTList:
  case VTConjunction:
  case VTDisjunction:
    l = listClone(v->data.list);
    for (p = l; p != NULL; p = listNext(p)) {
      listSetElement(p, cloneValue((Value) listElement(p)));
    }
    vv->data.list = l;
    break;
  default:
    cdgPrintf(CDG_ERROR, "ERROR: cloneValue: illegal value type (%d)\n", 
              v->type);
    abort();
  }

  return (vv);

}

/* ----------------------------------------------------------------------------
 * return a new expansion state with no decisions made
 */
Expansion esNew(Value v)
{

  Expansion result;
  int i;

  result = (Expansion) memMalloc(sizeof(ExpansionStruct));
  result->v = cloneValue(v);
  for (i = 0; i < 9; i++)
    result->decisions[i] = 0;

  return (result);

}

/* ----------------------------------------------------------------------------
 * return a copy of the expansion state
 */
Expansion esClone(Expansion es)
{

  Expansion result;
  int i;

  result = (Expansion) memMalloc(sizeof(ExpansionStruct));
  result->v = cloneValue(es->v);
  for (i = 0; i < 9; i++)
    result->decisions[i] = es->decisions[i];

  return (result);

}

/* ----------------------------------------------------------------------------
 * delete expansion state
 */
void freeExpansionState(Expansion es)
{

  freeValue(es->v);
  memFree(es);

}

/* ----------------------------------------------------------------------------
 * Finds the first disjunctive node in a Value
 */
Value disjunction( Value v )
{
  List l;
  Value vv;

  switch(v->type) {
  case VTAVNode:
    return( disjunction( ((AVNode)v->data.avnode)->value) );
  case VTDisjunction:
    return(v);
  case VTList:
  case VTConjunction:
    for(l=v->data.list;l!=NULL;l=listNext(l)) {
      if((vv = disjunction(listElement(l)))!=NULL) return(vv);   
    }
  default:
    break;
  }

  return(NULL);
}


/* ----------------------------------------------------------------------------
 * calculate all disambiguations for a Value 
 */
List expandValue(Value v)
{

  /* 
     result = [];
     list = [ clone(v) ];

     WHILE list contains items:

     take Expansion x from list;

     IF x->v is not disjunctive
     result .= x;

     ELSE
     find a disjunction d in x;

     IF d is not empty
     FOR all alternatives a of d:
     xx = clone(x);
     dd = corresponding disjunction in xx;
     replace dd by a;

     IF  xx id not disjunctive
     result .= xx;
     ELSE
     list .= xx;
     ROF        
     FI

     FI

     ELIHW

     delete list;
     cleanup result.

  */

  List worklist, p, result, l;
  Value d;
  Expansion x;

#ifdef DEBUGEXPANDVALUE
  cdgPrintf(CDG_DEBUG, "\nEXPANDVALUE: expanding value\n");
  printValue(CDG_DEBUG, v, 0);
  cdgPrintf(CDG_DEBUG, "\n");
#endif
  result = (List) NULL;
  worklist = listPrependElement(NULL, esNew(v));

  /* step through list */
  p = worklist;
  while (p != NULL) {

    /* read item from list */
    x = (Expansion) listElement(p);

    /* if it's not disjunctive, add the value to the result */
    if ((d = disjunction(x->v)) == NULL) {
      result = listPrependElement(result, cloneValue(x->v));
    } else {

      Value xx, dd, a;
      int i;
      Expansion copy;

      /* index out of range? */
      if (d->index > 9 || d->index < 0) {
        cdgPrintf(CDG_WARNING,
                  "WARNING: ignoring distribution index %d\n", d->index);
        d->index = 0;
      }
#ifdef DEBUGEXPANDVALUE
      cdgPrintf(CDG_DEBUG, "\nEXPANDVALUE: processing value\n");
      printValue(CDG_DEBUG, x->v, 0);
      cdgPrintf(CDG_DEBUG, "\n(I already made these decisions: [");
      for (i = 1; i < 9; i++)
        cdgPrintf(CDG_DEBUG, "%d", x->decisions[i]);
      cdgPrintf(CDG_DEBUG, "]\n\n");
#endif

      /* make n new Values from this one */
      for (i = 1, l = d->data.list; l != NULL; i++, l = listNext(l)) {

        /* clone new Value from x */
        copy = esClone(x);
        xx = copy->v;

        /* find place to modify in xx */
        dd = disjunction(xx);

        /* find nth alternative from x */
        a = cloneValue((Value) listElement(l));

#ifdef DEBUGEXPANDVALUE
        cdgPrintf(CDG_DEBUG, "EXPANDVALUE: alternative #%d=", i);
        printValue(CDG_DEBUG, a, 0);
        cdgPrintf(CDG_DEBUG, "\n");
#endif

        /* delete the list of alternatives from the clone... */
        {
          List q;

          for (q = dd->data.list; q != NULL; q = listNext(q))
            freeValue((Value) listElement(q));
          listDelete(dd->data.list);
        }

        /* ...and replace it by the nth alternative. */
        dd->type = a->type;
        dd->index = a->index;
        dd->data = a->data;
        /* this has implicitly modified xx! */

        /* note the decision we made */
        if (d->index)
          copy->decisions[d->index] = i;

#ifdef DEBUGEXPANDVALUE
        cdgPrintf(CDG_DEBUG, "EXPANDVALUE: after substituting the clone looks like this:\n");
        printValue(CDG_DEBUG, xx, 0);
        cdgPrintf(CDG_DEBUG, "\n");
#endif

        /* put the simplified Value on the list if allowed */

        /* if this is a distributed disjunction, */
        if (d->index > 0

            /* and a decision has already been made, */
            && x->decisions[d->index] != 0

            /* and this is not the correct followup, */
            && x->decisions[d->index] != i) {

          /* forget about this value. */
          freeValue( xx );
          memFree( copy );
#ifdef DEBUGEXPANDVALUE
          cdgPrintf(CDG_DEBUG, "EXPANDVALUE: this value violates previously held obligations.\n\n");
#endif
        } else {
#ifdef DEBUGEXPANDVALUE
          cdgPrintf(CDG_DEBUG, "EXPANDVALUE: this value is ok.\n\n");
#endif
          /* accept and store the expansion */
          if (disjunction(xx))
            worklist = listAppendElement(worklist, copy);
          else {
            result = listAppendElement(result, xx);
	    memFree(copy);
	  }
        }
      }
    }

    p = listNext(p);

  }

  /* delete working list */
  listForEachDelete(worklist, freeExpansionState);

  /* remove nested conjunctions from list of results */
  for (p = result; p != NULL; p = listNext(p))
    cleanup((Value) listElement(p));

#ifdef DEBUGEXPANDVALUE
  cdgPrintf(CDG_DEBUG, "EXPANDVALUE: List of results =\n");
  {
    List p;

    if (listSize(l) > 1) {
      cdgPrintf(CDG_DEBUG, "[\n");
      for (p = l; p != NULL; p = listNext(p)) {
	printValue(CDG_DEBUG, (Value) listElement(p), 0);
	cdgPrintf(CDG_DEBUG, "\n");
      }
      cdgPrintf(CDG_DEBUG, "]\n");
    }
  }
#endif

  /* list of expansions is now complete */
  return (result);
}



/* ----------------------------------------------------------------------------
   Print atomic Value to String.
 */
String sprintValue(Value v)
{
  char buffer[1024] = "";
  List l;
  int index;
  String s;
  Boolean error = FALSE;
  
  switch ( v->type ) {
  case VTError:
    /* empty set symbol denotes unspecified features */
    return(strRegister(""));
  case VTString:
    return v->data.string;
  case VTNumber:
    if ( v->data.number - (int)v->data.number == 0.0 ) {
      snprintf(buffer, 1024, "%d", (int)v->data.number);
    } else {
      snprintf(buffer, 1024, "%g", v->data.number);
    }
    return strRegister(buffer);
    
  case VTList:
    index = 1;
    strcat(buffer, "<");
    for ( l = v->data.list; l != NULL; l = listNext(l) ) {
      s = sprintValue(listElement(l));
      index += strlen(s);
      if(index > 1023) {
        error = TRUE;
      } else {
        strcat(buffer, s);
      }
      if (listNext(l) != NULL) {
        index +=2;
        if(index > 1023) {
          error = TRUE;
        } else {
          strcat(buffer, ", ");
        }
      }
    }

    if(index + 1 > 1023) {
      error = TRUE;
    }
    if(error) {
      cdgPrintf(CDG_ERROR, "PANIC: Value list longer than 1024 bytes!\n");
      cdgPrintf(CDG_ERROR, "PANIC: Who's supposed to understand that, anyway?\n");
      abort();
    } 
    
    strcat(buffer, ">");
    return strRegister(buffer);
  default:
    cdgPrintf(CDG_ERROR, 
              "ERROR: sprintValue: illegal value type (%d)\n", 
              v->type);
    abort();
  }
  return NULL;
}

/* ----------------------------------------------------------------------------
 * Does FEATURE distinguish A and B?
 */
Boolean inputCompareLeByAtt(LexiconItem a, LexiconItem b,
                            List feature, Boolean unify)
{
  Value va = peekValue(a->value, feature);
  Value vb = peekValue(b->value, feature);
  ValueType vtA = va->type;
  ValueType vtB = vb->type;
  List l, m;


#ifdef DEBUGCOMPARELE
  cdgPrintf(CDG_DEBUG, "DEBUG: checking attribute ");
  printPath(CDG_DEBUG, feature);
  cdgPrintf(CDG_DEBUG, "... ");
#endif

  /* if neither item has the feature, it doesn't. */
  if (vtA == VTError && vtB == VTError) {
#ifdef DEBUGCOMPARELE
    cdgPrintf(CDG_DEBUG, "equal.\n");
#endif
    return FALSE;
  }

  /* What if one item has the feature and the other doesn't? Well, for
   * disambiguating lexical entries, an absent feature should not be
   * different from a specified feature because there is no way of
   * unambiguously specifying the form with the feature absent. Such a
   * situation is really an error in the lexicon and should be marked as a
   * duplicate lexical entry. On the other hand, when building LVs it makes
   * a big difference -- imagine a verb which demands a certain case for
   * its object, and one which demands none because it doesn't even take an
   * object. Clearly the two should not be merged into one LV. */

  if (vtA == VTError || vtB == VTError) {
    if (unify) {
#ifdef DEBUGCOMPARELE
      cdgPrintf(CDG_DEBUG, "equal.\n");
#endif
      return FALSE;
    } else {
#ifdef DEBUGCOMPARELE
      cdgPrintf(CDG_DEBUG, "different.\n");
#endif
      return TRUE;
    }
  }

  /* different types are a difference. */
  if (vtA != vtB) {
#ifdef DEBUGCOMPARELE
    cdgPrintf(CDG_DEBUG, "different types!\n");
#endif
    return TRUE;
  }

  /* different strings are a difference. */
  if (vtA == VTString) {
    if (strcmp(va->data.string, vb->data.string) != 0) {
#ifdef DEBUGCOMPARELE
      cdgPrintf(CDG_DEBUG, "%s != %s\n", va->data.string, vb->data.string);
#endif
      return TRUE;
    } else {
      return FALSE;
    }
  }

  /* different numbers are a difference. */
  if (vtA == VTNumber) {
    if (va->data.number != vb->data.number) {
#ifdef DEBUGCOMPARELE
      cdgPrintf(CDG_DEBUG, "%f != %f\n", va->data.number, vb->data.number);
#endif
      return TRUE;
    } else {
      return FALSE;
    }
  }

  /* different lists are a difference. */
  if (vtA == VTList) {
    if (listSize(va->data.list) != listSize(vb->data.list)) {
#ifdef DEBUGCOMPARELE
      cdgPrintf(CDG_DEBUG, "list length differs.\n");
#endif
      return TRUE;
    }
    for (l = va->data.list, m = vb->data.list;
	 l != NULL; l = listNext(l), m = listNext(m)) {
      Value v1 = listElement(l);
      Value v2 = listElement(m);

      if (v1->data.string != v2->data.string) {
#ifdef DEBUGCOMPARELE
	cdgPrintf(CDG_DEBUG, "list key differs.\n");
#endif
	return TRUE;
      }
      l = listNext(l), m = listNext(m);
      v1 = listElement(l);
      v2 = listElement(m);
      if (v1->data.number != v2->data.number) {
#ifdef DEBUGCOMPARELE
	cdgPrintf(CDG_DEBUG, "list value differs.\n");
#endif
	return TRUE;
      }
    }
  }

  return FALSE;
}

/* ----------------------------------------------------------------------------
 * Check whether one of FEATURES distinguishes A from B.
 */
Boolean inputCompareLeByAtts(LexiconItem a, LexiconItem b, BitString features)
{
  int i;
  List l;
  if (!lgCompactLVs) {
    return TRUE;
  }
#ifdef DEBUGCOMPARELE
  cdgPrintf(CDG_DEBUG, "\nDEBUG: equivalence requested between\n");
  printValue(CDG_DEBUG, a->value, 0);
  cdgPrintf(CDG_DEBUG, "\n and\n");
  printValue(CDG_DEBUG, b->value, 0);
  cdgPrintf(CDG_DEBUG, "\n");
#endif

  for(i = 0, l = inputCurrentGrammar->allAttributes;
      i < bitSize(features);
      i++, l = listNext(l)) {
    if(!bitGet(features, i)) {
      continue;
    }
    if (inputCompareLeByAtt(a, b, listElement(l), FALSE)) {
      return TRUE;
    }
  }
#ifdef DEBUGCOMPARELE
  cdgPrintf(CDG_DEBUG, "DISTINCTION: equivalence confirmed.\n");
#endif

  return FALSE;
}


/* ----------------------------------------------------------------------------
   Look up path in list of features.

   PATH is a Feature path, i.e. a List of Strings that interpreted as a
   path into a Value. It this path is already registered in the field
   allAttributes of the specified input, its 0-based index in that List is
   returned. Otherwise -1 is returned.
*/
int inputLookupPath(List path, Input input)
{
  int i;
  List l;

  for (i=0, l=input->allAttributes; l; i++, l=listNext(l))
    { if (comparePath(path, listElement(l))) { return i; } }
  return -1;
}

/* ----------------------------------------------------------------------------
   Insert a path into the list of features.

   PATH is a Feature path, i.e. a List of Strings that interpreted as a
   path into a Value. Its 0-based index in the List allAttributes of the
   specified input is returned. If PATH was not present there, it is first
   appended to the list.

   If PATH is a pseudo feature, -1 is returned.
*/
int inputRegisterPath(List path, Input input)
{
  int i=inputLookupPath(path, input);

  if(inputPseudoFeature(path)) {
    return -1;
  }
  
  if (i>=0) { return i; }
  i=listSize(input->allAttributes);
#ifdef DEBUGINPUTREGISTERPATH
  cdgPrintf(CDG_WARNING, "Feature #%d = ", i);
  printPath(CDG_WARNING, path);
  cdgPrintf(CDG_WARNING, "\n");
#endif
  input->allAttributes=listAppendElement(input->allAttributes, listClone(path));

  /* note that the set of known features has changed */
  input->attsCurrent = FALSE;
  return i;
}

/* ----------------------------------------------------------------------------
   Scan lexical entry for features (pass 1)
   or cache all present features (pass 2).
*/
void traverseItem(int pass, LexiconItem lex, Value v, List p, Input i)
{
  int index;

  switch (v->type) {
  case VTNumber:
  case VTString:
  case VTList: 
    index = inputRegisterPath(p, i);
    if (pass == 2) {
      if(index >= 0) {
        lex->values[index] = v;
      } else {
        cdgPrintf(CDG_ERROR, "ERROR: the feature `");
        printPath(CDG_ERROR, p);
        cdgPrintf(CDG_ERROR,
                  "' is special to CDG, you cannot use it explicitly.\n");
      }
    }
    break;
  case VTConjunction:
    {
      List l;
      for (l = v->data.list; l; l = listNext(l)) {
        /* Recursion */
        traverseItem(pass, lex, listElement(l), p, i);
      }
      break;
    }
  case VTAVNode:
    {
      AVNode av=(AVNode)v->data.avnode;

      /* Extend the path */
      p=listAppendElement(p, av->attribute);
      /* Recursion */
      traverseItem(pass, lex, av->value, p, i);
      /* Deallocate memory */
      listDeleteLastElement(p);
      break;
    }
  case VTDisjunction:
  default:
    /* This shouldn't happen */
    cdgPrintf(CDG_ERROR, "ERROR: Unknown ValueType\n");
  }
}

/* ----------------------------------------------------------------------------
 * Build the field ->values for one lexical item.
 */
void inputCacheLI(Input input, LexiconItem le) {
  int i, nrFeatures = listSize(input->allAttributes);
  
  /* if no new attribute was introduced since we last computed the feature
     table, we don't have to do anything. */
  if(le->values && input->attsCurrent) {
    return;
  }
  
  /* Otherwise we have to delete the old one... */
  if(le->values) {
    memFree(le->values);
  }

  /* ...and compute a new one. */
  le->values=(Value*)memMalloc(nrFeatures*sizeof(Value));
  for(i = 0; i < nrFeatures; i++) {
    le->values[i] = ErrorValue;
  }
  traverseItem(2, le, le->value, NULL, input);
}

/* ----------------------------------------------------------------------------
   Pre-compute all possible feature accesses and update the
   field ->values of each item in the lexicon.
*/
void inputCacheLexicon(Input input) {

  HashIterator hi;
  List l, m;

  /* pre-access all paths in each lexical item */
  hi = hashIteratorNew(input->lexicon);
  while((l=hashIteratorNextValue(hi))) {
    for(m = l; m != NULL; m = listNext(m)) {
      inputCacheLI(input, listElement(m));
    }
  }
  hashIteratorDelete(hi);
}

/* ----------------------------------------------------------------------------
   Update the ->values table items living in decorated Parses.

   This does exactly the same as inputCacheLexicon(), but it has to be done
   separately for decorated Parses, because the lexicon items in their
   lexeme graphs are deep copies of those in the lexicon.
*/
void inputCacheParses(Input input, Hashtable ht) {
  HashIterator hi;
  Parse p;
  int i;

  hi = hashIteratorNew(ht);
  while((p=hashIteratorNextValue(hi))) {
    if(parseIsAbstract(p)) continue;
    for(i = 0; i < vectorSize(p->lg->nodes); i++) {
      LexemNode ln = vectorElement(p->lg->nodes, i);
      inputCacheLI(input, ln->lexem);
    }
  }
  hashIteratorDelete(hi);
}



/* ----------------------------------------------------------------------------
 * calculate index for use with input->hashedUCs
 */
int inputUCIndex(int level, int dir)
{
  return (dir + level * nrDirections);
}

/* ----------------------------------------------------------------------------
 * calculate index for use with input->hashedBCs
 */
int inputBCIndex(int levela, int levelb, int dira, int conn, int dirb)
{
  return (dirb +
          conn * nrDirections +
          dira * nrDirections * nrConnexions +
          levelb * nrDirections * nrConnexions * nrDirections +
          levela * nrDirections * nrConnexions * nrDirections * inputCurrentGrammar->noOfLevels);
}

/* ----------------------------------------------------------------------------
 * return the inverse of a Connexion
 */
Connexion inverseConnexion(Connexion x)
{

  if (x == Over)
    return (Under);
  else if (x == Under)
    return (Over);
  else
    return (x);

}

/* ----------------------------------------------------------------------------
 * Does Connexion a subsume Connexion b?
 */
Boolean subsumesConnexion(Connexion a, Connexion b)
{

  if (a == b)
    return TRUE;
  else if (a == AnyCon)
    return TRUE;
  else if (b == Identical && (a == Top || a == Bottom))
    return TRUE;
  else if (b == Inverse && (a == Over || a == Under))
    return TRUE;
  else
    return FALSE;

}


/* ----------------------------------------------------------------------------
 * print out a Connexion
 */
void printConnexion(unsigned long mode, Connexion x) {
  cdgPrintf(mode, formatConnexion(x));
}

/* ----------------------------------------------------------------------
   Return formatted representation of a Connexion.
*/
String formatConnexion(Connexion c) {
  switch (c) {
  case Top       : return "/\\";
  case Bottom    : return "\\/";
  case Over      : return "/";
  case Under     : return "\\";
  case None      : return "||";
  case Inverse   : return "~=";
  case Identical : return "==";
  case AnyCon    : return ",";
  }
  return NULL;
}

/* ----------------------------------------------------------------------------
 * print out a Direction
 */
void printDirection(unsigned long mode, Direction x) {
  cdgPrintf(mode, "%s", formatDirection(x));
}

/* ----------------------------------------------------------------------
   Return formatted representation of a Direction.
*/
String formatDirection(Direction d) {

  switch (d) {
  case Left:   return "\\";
  case Right:  return "/";
  case Nil:    return "|";
  case NonNil: return "!";
  case AnyDir: return ":";
  }
  return NULL;
}


/* ----------------------------------------------------------------------------
 * Does Direction a subsume Direction b?
 */
Boolean subsumesDirection(Direction a, Direction b)
{

  if (a == b)
    return TRUE;
  else if (a == AnyDir)
    return TRUE;
  else if (a == NonNil && (b == Left || b == Right))
    return TRUE;
  else
    return FALSE;

}


/* ----------------------------------------------------------------------------
 * Does level specification a match the level b?
 */
Boolean matchLevel(Level a, Level b) {
  return(!a || (a == b));
}

/* ----------------------------------------------------------------------------
 * destructively invert a Signature
 */
void invertSignature(Signature x)
{

  Direction d;
  Level l;

  x->conn = inverseConnexion(x->conn);

  d = x->d1;
  l = x->l1;
  x->d1 = x->d2;
  x->d2 = d;
  x->l1 = x->l2;
  x->l2 = l;

}

/* ----------------------------------------------------------------------------
 * Return formatted representation of a Signature
 */
String formatSignature(Signature x)
{

  return strPrintf("X%s%s%s",
                   (x->l1 || x->conn != AnyCon) ?
                   formatDirection(x->d1) : "",
                   x->l1 ? x->l1->id : "",
                   x->arity == 1 ? "" :

                   strPrintf("%sY%s%s",
                             formatConnexion(x->conn),
                             (x->l2 || x->conn != AnyCon) ? 
                             formatDirection(x->d2) : "", 
                             x->l2 ? x->l2->id : ""));
}

/* ----------------------------------------------------------------------------
   Set the order of the levels.

   The arguments must be a permutation of the set of names of all defined
   levels of analysis. The List inputCurrentGrammar->levels is sorted
   accordingly, and each level's no is reset.
 */
Boolean setLevelSort(String levelIds[])
{
  int i;
  List l, neworder = NULL;

  for (i = 0; levelIds[i] != NULL; i++) {
    Level level;
    if(NULL != (level = findLevel(levelIds[i]))) {
      neworder = listAppendElement(neworder, level);
    } else {
      cdgPrintf(CDG_ERROR, "ERROR: unknown level `%s'\n", levelIds[i]);
      listDelete(neworder);
      return FALSE;
    }
  }

  /* reset the field ->no */
  i = 0;
  for(l = neworder; l != NULL; l = listNext(l)) {
    Level lvl = listElement(l);
    lvl->no = i++;
  }
  
  listDelete(inputCurrentGrammar->levels);
  inputCurrentGrammar->levels = neworder;
  return TRUE;
}

/* ----------------------------------------------------------------------------
   Return a string listing all defined Levels in their current order,
   separated by spaces.
 */
String getLevelSort(void) {
  String result = NULL;
  List l;
  for (l = inputCurrentGrammar->levels; l!=NULL; l=listNext(l)) {
    Level level = listElement(l);
    if(result) {
      result = strAppend(result, level->id, " ", NULL);
    } else {
      result = strAppend(level->id, " ", NULL);
    }
  }
  return strRegister(result);
}

/* ----------------------------------------------------------------------------
 * getLevelProperty:
 *      return the property of a given name or NULL
 */
String getLevelProperty( Level level, String name ) 
{
  List l;
  Property prop;
  String thisName = strRegister(name);

  for (l = level->properties; l != NULL; l = listNext(l)) {
    prop = (Property) (listElement(l));
    if (prop->name == thisName)
      return prop->value;
  }

  return NULL;
}

/* ----------------------------------------------------------------------
   Find files whose name matches the pattern *NAME.cda and load them.
*/
void loadAnnotations(String name) {
  char dirname[PATH_MAX];
  struct dirent *de;
  String pattern = strAppend(name, ".cda", NULL);

  /* read and maybe load items from directory DIR */
  void try(String dir) {
    DIR *handle;

    handle = opendir(dir);
    if(!handle) {
      return;
    }

    while(NULL != (de = readdir(handle))) {
      unsigned int mode;
      if(strstr(de->d_name, pattern)) {
        String fullname = strAppend(dir, "/", de->d_name, NULL);
        cdgPrintf(CDG_INFO, "INFO: autoloading `%s'\n", de->d_name);
        
        mode = hkVerbosity;
        hkVerbosity = hkVerbosity & (~CDG_INFO);
        inputLoad(fullname);
        hkVerbosity = mode;
        if(!parseResult) continue;
        
        /* if that contains anything but annotations, refuse to accept it */
        if(hashSize(parseResult->constraints) ||
           listSize(parseResult->levels) ||
           hashSize(parseResult->lexicon) ||
           listSize(parseResult->lattices) ||
           hashSize(parseResult->hierarchies) ||
           listSize(parseResult->parameters) ||
           hashSize(parseResult->maps)) {
          cdgPrintf(CDG_WARNING,
                    "WARNING: `%s' contains items other than annotations!\n",
                    de->d_name);
          inputFreeInput(parseResult);
          continue;
        }
        
        /* promote new data to current data */
        mergeInput(inputCurrentGrammar, parseResult);
        inputCacheInput(inputCurrentGrammar);
        memFree(parseResult);
        parseResult = NULL;
      }
    }
    closedir(handle);
  }

  realpath(inputAnnodirs, dirname);
  try(dirname);

  /* If there are huge numbers of annotations, it is also allowed to keep
     them in subdirs named 000, 001 etc., which respectively hold the
     first, second etc. 10,000 files. For this to work, the name must
     contain a decimal number. */
  {
    int i = strcspn(name, "0123456789");
    int nr, batch;
    String s;
    if(1 == sscanf(name+i, "%d", &nr)) {
      batch = nr / 10000;
      s = strPrintf("%s/%03d", dirname, batch);
      try(s);
      strDelete(s);
    }
  }
}

/* ----------------------------------------------------------------------------
   Return the annotation for LAT, or NULL.

   Basically this function returns any existing AnnoEntry whose
   field ->lattice is the same as lat->id.

   If there are several, one of them is returned arbitrarily. However,
   annotations whose name is the same as the lattice name are preferred
   against others.

   If there is none, and AUTO is set, the file system is queried for files
   whose name indicates that they might hold an annotation for LAT, and
   they are then loaded.
   
 */
AnnoEntry findAnnoForLattice(Lattice lat, Boolean autoload)
{
  HashIterator hi;
  AnnoEntry result;

  if (!lat)
    return NULL;

  /* find annotation with the same name and the correct lattice id */
  if(NULL != (result = hashGet(inputCurrentGrammar->annotations, lat->id)) &&
     result->lattice == lat->id) {
    return result;
  }

  /* find any annotation with the correct lattice id */
  hi = hashIteratorNew(inputCurrentGrammar->annotations);
  while((result = hashIteratorNextValue(hi))) {
    if(result->lattice == lat->id) {
      hashIteratorDelete(hi);
      return result;
    }
  }
  hashIteratorDelete(hi);

  /* maybe autoload  */
  if(autoload) {
    loadAnnotations(lat->id);
    return findAnnoForLattice(lat, FALSE);
  }
  
  return NULL;
}

/* ----------------------------------------------------------------------------
   Return the annotation with the name ID, or NULL.

   If there is none, and AUTO is set, the file system is queried for files
   whose name indicates that they might hold that annotation, and they are
   then loaded.
   
 */
AnnoEntry findAnnoByName(String id, Boolean autoload)
{
  AnnoEntry result;

  result = hashGet(inputCurrentGrammar->annotations, id);
  if(result) {
    return result;
  }

  /* maybe autoload  */
  if(autoload) {
    loadAnnotations(id);
    return findAnnoByName(id, FALSE);
  }
  
  return NULL;
}


/* ----------------------------------------------------------------------------
 * returns the annotations for an utterance
 */
List findAnnotations(Lattice lat, Boolean autoload)
{
  HashIterator hi;
  AnnoEntry ae;
  List result = NULL;

  if (!lat)
    return NULL;

  hi = hashIteratorNew(inputCurrentGrammar->annotations);
  while((ae = hashIteratorNextValue(hi))) {
    if(ae->lattice == lat->id) {
      result = listPrependElement(result, ae);
    }
  }
  hashIteratorDelete(hi);
  if(result) {
    return result;
  }

  /* maybe autoload  */
  if(autoload) {
    loadAnnotations(lat->id);
    return findAnnotations(lat, FALSE);
  }

  return NULL;
}

/* ----------------------------------------------------------------------------
 * Find a Level in the current input by ID.
 */
Level findLevel(String id)
{
  Level result;
  List l;
  String regId = strRegister(id);

  for (l = inputCurrentGrammar->levels; l != NULL; l = listNext(l)) {
    result = listElement(l);
    if (result->id == regId) {
      return result;
    }
  }

  return NULL;
}

/* ----------------------------------------------------------------------------
 * Find a Constraint in the current input by ID.
 */
Constraint findConstraint(String id) {
  return hashGet(inputCurrentGrammar->constraints, id);
}

/* ----------------------------------------------------------------------------
   Returns lattice with the name ID.
 */
Lattice findLattice(String id) {
  return findLatticeInInput(inputCurrentGrammar, id);
}

/* ----------------------------------------------------------------------------
   Returns lattice with the name ID, looking in IP.
 */
Lattice findLatticeInInput(Input ip, String id) {
  List l;
  for(l = ip->lattices; l != NULL; l = listNext(l)) {
    Lattice lat = listElement(l);
    if(!strcmp(lat->id, id)) {
      return lat;
    }
  }
  return NULL;
}

/* ----------------------------------------------------------------------------
 * returns constraint section by name
 */
Section findSection(String id) {
  List l;
  for (l = inputSections; l; l = listNext(l)) {
    Section s = listElement(l);
    if (strcmp(s->id, id) == 0) {
      return s;
    }
  }
  return NULL;
}


/* ----------------------------------------------------------------------------
 * Returns hierarchy by name
 */
Hierarchy findHierarchy(String id) {
  return hashGet(inputCurrentGrammar->hierarchies, id);
}

/* ----------------------------------------------------------------------------
   Checks whether lat is anything more complicated than a linear succession 
   of words.
*/
Boolean latticeBranches(Lattice lat) {

  Arc a,b;
  List l;

  a = NULL;
  for(l = lat->arcs; l != NULL; l = listNext(l)) {
    b = listElement(l);
    if(a && a->to != b->from) {
      return TRUE;
    }
    a = b;
  }
  return FALSE;
}

/* ----------------------------------------------------------------------------
   Checks whether there is an infostring in lattice
*/
Boolean latticeHasInfo(Lattice lat) {
  Arc arc;
  List l;

  for(l = lat->arcs; l != NULL; l = listNext(l)) {
    arc = listElement(l);
    if(arc->info != NULL) {
      return TRUE;
    }
  }
  return FALSE;
}



/* ----------------------------------------------------------------------------
   Clones a Lattice
*/
Lattice inputCloneLattice(Lattice lat) {

  Lattice result = memMalloc(sizeof(LatticeStruct));
  List l;
  Arc arc;

  result->id = strRegister( lat->id );
  result->filename = strRegister( lat->filename );
  result->lineNo = lat->lineNo;

  result->arcs = NULL;

  for(l = lat->arcs; l != NULL; l = listNext(l)) {
    Arc old=(Arc)listElement(l);
    arc = newArc(old->from, old->to, old->no, strRegister(old->word),
                 old->penalty, cloneValue(old->info),
		 old->fstart, old->fstop);
    arc->quotes = old->quotes;
    arc->parens = old->parens;
    arc->phraseQuotes = old->phraseQuotes;
    result->arcs = listAppendElement(result->arcs, arc);
  }

  return result;
}


/* ----------------------------------------------------------------------------
   Parse inputCompoundCatString into inputCompoundCats.
*/
void inputParseCompoundCats(String name, String *var)
{
  char buffer[strlen(*var)+1];
  char *bufptr = buffer;
  String token;
  
  if(inputCompoundCats) {
    listForEachDelete(inputCompoundCats,listDelete);
    inputCompoundCats = NULL;
  }

  strcpy (buffer, *var);
  token = strtok_r(buffer, ",", &bufptr);
  while (token) {
    inputCompoundCats = listAppendElement(inputCompoundCats, strRegister(token));
    token = strtok_r(NULL, ",", &bufptr);
  }

}

/* ----------------------------------------------------------------------------
 * Parse inputAnnoCatString into inputAnnoCats.
 */
void inputParseAnnoCats(String name, String *var)
{
  char buffer[strlen(*var)+1];
  char *bufptr = buffer;
  String token;
  
  if(inputAnnoCats) {
    listForEachDelete(inputAnnoCats,listDelete);
    inputAnnoCats = NULL;
  }

  strcpy (buffer, *var);
  token = strtok_r(buffer, ",", &bufptr);
  while (token) {
    inputAnnoCats = listAppendElement(inputAnnoCats, scanPath(token));
    token = strtok_r(NULL, ",", &bufptr);
  }

#if 0
  {
    List l,ll;
    cdgPrintf(CDG_DEBUG, "DEBUG: inputAnnoCats=");
    for (l = inputAnnoCats; l; l = listNext(l)) {
      for (ll = listElement(l); ll; ll = listNext(ll)) {
	cdgPrintf(CDG_DEBUG, "%s", listElement(ll));
	if (listNext(ll)) {
	  cdgPrintf(CDG_DEBUG, ":");
	}
      }
      if (listNext(l)) {
	cdgPrintf(CDG_DEBUG, ",");
      }
    }
    cdgPrintf(CDG_DEBUG, "\n");
  }
#endif
}

/* ----------------------------------------------------------------------------
   Parse inputSpontaneousUpcaseString into inputSpontaneousUpcase.
*/
void inputParseSpontaneousUpcase(String name, String *var)
{
  char buffer[strlen(*var)+1];
  char *bufptr = buffer;
  String token;
  
  if(inputSpontaneousUpcase) {
    listForEachDelete(inputSpontaneousUpcase,listDelete);
    inputSpontaneousUpcase = NULL;
  }

  strcpy (buffer, *var);
  token = strtok_r(buffer, ",", &bufptr);
  while (token) {
    inputSpontaneousUpcase =
      listAppendElement(inputSpontaneousUpcase, strRegister(token));
    token = strtok_r(NULL, ",", &bufptr);
  }
}


/* ----------------------------------------------------------------------------
 * Initialize the input module
 */
void inputInitialize(void) 
{
  unsigned long int oldVerbosity = hkVerbosity;

  hkVerbosity = CDG_ERROR | CDG_WARNING;

  /* initialize the preprocessor settings */
  inputPreprocessor = strRegister(M4 " -s -P");
  inputPrepSuffix = strRegister(".m4");
  inputCallback(strRegister("preprocessor"), &inputPreprocessor);

  /* setting the locale */
  inputLocale = strRegister("");
  inputFeatureHierarchy = strRegister("");
  inputCallback(strRegister("locale"), &inputLocale);

  /* initialize section list */
  inputDefaultSection = (Section) memMalloc(sizeof(SectionStruct));
  inputDefaultSection->id = strRegister( "default" );
  inputDefaultSection->active = TRUE;
  inputDefaultSection->counter = 0;

  ErrorValue = (Value)memMalloc(sizeof(ValueStruct));
  ErrorValue->type = VTError;
  ErrorValue->index = 0;

  inputSections = listAppendElement(NULL, inputDefaultSection);

  /* structure for current grammar */
  inputCurrentGrammar = inputNewInput();

  /* get something better than NULL */
  inputAnnoCatString = strRegister("");
  inputCompoundString = strRegister("");
  inputSpontaneousUpcaseString = strRegister("");
  inputAnnodirs = strRegister(".");
  
  /* The ToDate flags are only for the tool XCDG. They are reset whenever
     the input changes, and should never be set from within C code.

     Strictly speaking, the value should be FALSE initially, since input
     changes from `undefined' to `empty', and inputNewInput() sets them
     accordingly. But visually, the list of no input looks exactly like the
     list of the empty input, so we set them here (and only here) to save
     XCDG a useless update. */
  inputCurrentGrammar->xcdgwordgraphsToDate = TRUE;
  inputCurrentGrammar->xcdglevelsToDate = TRUE;
  inputCurrentGrammar->xcdgconstraintsToDate = TRUE;
  inputCurrentGrammar->xcdglexiconToDate = TRUE;
  inputCurrentGrammar->xcdghierarchiesToDate = TRUE;

  setRegister("preprocessor", SET_STRING, &inputPreprocessor, NULL, NULL, inputCallback, NULL);
  setRegister("prepsuffix", SET_STRING, &inputPrepSuffix, NULL, NULL, inputCallback, NULL);
  setRegister("acoustics", SET_ENUM, &vmAcoustics,
	      NULL /* convert */, NULL /* validate */, NULL /* callback */,
	      "off", vmAcOff, "linear", vmAcLinear, "scaled", vmAcScaled, NULL);

  setRegister("anno-categories", SET_STRING, &inputAnnoCatString,
	      NULL, NULL, (VoidFunction *)inputParseAnnoCats,
	      NULL);
  setRegister("compound-categories", SET_STRING, &inputCompoundString,
	      NULL, NULL, (VoidFunction *)inputParseCompoundCats,
	      NULL);
  setRegister("capitalizable-categories", SET_STRING, &inputSpontaneousUpcaseString,
	      NULL, NULL, (VoidFunction *)inputParseSpontaneousUpcase,
	      NULL);
  setRegister("locale", SET_STRING, &inputLocale,
	      NULL, NULL, (VoidFunction *)inputCallback,
	      NULL);
  setRegister("featureHierarchy", SET_STRING, &inputFeatureHierarchy,
              NULL, NULL, NULL, NULL);
  setRegister("deduceCompounds", SET_BOOL, &inputDeduceCompounds,
              NULL, NULL, NULL, NULL);
  setRegister("annodirs", SET_STRING, &inputAnnodirs,
              NULL, NULL, NULL, NULL);

  setRegister("templates", SET_ENUM, &templateMode,
	      NULL /* convert */,
              NULL /* validate */,
              NULL /* callback */,
	      "never", 0,
              "bycategory", 1,
              "ifneeded", 2,
              "always", 3,
              NULL);


  hkVerbosity = oldVerbosity;

  /* this character set is used by commandEval(),
     so it has to be here, not in cdgp.c  */
  
  /* for selective .inputrc parsing */
  rl_readline_name = "CDG";
  /* quote characters */
  rl_completer_quote_characters = "'";
  rl_basic_word_break_characters = "\n\t ";

}

/* ----------------------------------------------------------------------------
 * called from registered string variables of the input module
 */
void inputCallback(String name, String *var) 
{
  /* preprocessor */
  if (strRegister("preprocessor") == name) {
    List l;
    int i;
    char buffer[strlen(inputPreprocessor)+1];
    char *bufptr = buffer;
    String token;
    List thisArgv;

    /* free the old preprocessor arguments */
    if (prepArgv != NULL) {
      for (i = 0; prepArgv[i]; i++) {
	cdgFreeString(prepArgv[i]);
      }
      memFree(prepArgv);
      prepArgv = NULL;
      prepArgNo = 0;
    }

    /* collect new preprocessor arguments */
    strcpy(buffer, inputPreprocessor);
    token = strtok_r(buffer, " ", &bufptr);
    thisArgv = NULL;

    while (token) {
      thisArgv = listAppendElement(thisArgv, (Pointer) strRegister(token));
      token = strtok_r(NULL, " ", &bufptr);
    }

    /* assert them; filename comes later */
    prepArgNo = listSize(thisArgv) + 1;
    prepArgv = memMalloc((prepArgNo +1)* sizeof(String));
    for (i = 0, l = thisArgv; l; l = listNext(l), i++) {
      token = listElement(l);
      prepArgv[i] = token;
    }
    prepArgv[prepArgNo-1] = NULL;
    prepArgv[prepArgNo] = NULL;

    listDelete(thisArgv);
  } 
  
  /* locale */
  else if (strRegister("locale") == name) {
    String newLocale = setlocale(LC_CTYPE, inputLocale);
    if (!newLocale) {
      newLocale = setlocale(LC_CTYPE, "");
      cdgPrintf(CDG_ERROR, 
                "ERROR: unknown locale `%s', falling back to defaults (`%s')\n",
                inputLocale, newLocale);
    }
    inputLocale = strRegister(newLocale);
  }

}

/* ----------------------------------------------------------------------------
 * Finalize the input module
 */
void inputFinalize(void)
{

  inputFreeInput(inputCurrentGrammar);

  memFree(inputDefaultSection);
  listDelete(inputSections);

  if (prepArgv) {
    int i;
    for (i = 0; prepArgv[i]; i++) {
      cdgFreeString(prepArgv[i]);
    }
    memFree(prepArgv);
  }

  if (ErrorValue)
    memFree(ErrorValue);

  listForEachDelete(inputAnnoCats, listDelete);
  listDelete(inputCompoundCats);
  
}


/* ----------------------------------------------------------------------------
   Returns the first Level that has its mainlevel attribute set.
*/
Level inputGetMainlevel(Input i) {

  Level level;
  List l;
  String prop;

  if (!i) { return NULL; }

  for(l = i->levels; l != NULL; l = listNext(l)) {
    level = listElement(l);
    prop = getLevelProperty(level, "mainlevel");
    if(prop && !strcmp(prop, "yes")) {
      return level;
    }
  }

  return NULL;
}


/* ----------------------------------------------------------------------------
 * Returns the weight of a constraint:
 usually, the penalty;
 with variable-penalty constraints, only the constant factor.
*/
Number inputGetWeight(Constraint c) {

  Term t = c->penaltyTerm;

  if(t) {
    if(t->type == TTMultiply &&
       t->data.operation.op1->type ==TTNumber) {
      return t->data.operation.op1->data.number;
    } else {
      return 1.0;
    }
  } else {
    return c->penalty;
  }
}


/* ----------------------------------------------------------------------------
 * Sets the weight of a constraint.
 Normally, sets c->penalty.
 With variable-penalty constraints, multiplies the Term by x.
 With "absolute" set, forces constant penalty even for those.
*/
void inputSetWeight(Constraint c, Number x, Boolean absolute) {

  Term t;
  
  if(absolute && c->penaltyTerm) {
    freeTerm(c->penaltyTerm);
    c->penaltyTerm = NULL;
  }

  if(c->penaltyTerm) {
    t = memMalloc(sizeof(TermStruct));
    t->type = TTMultiply;
    t->data.operation.op1 = memMalloc(sizeof(TermStruct));
    t->data.operation.op1->type = TTNumber;
    t->data.operation.op1->data.number = x;
    t->data.operation.op2 = c->penaltyTerm;
    c->penaltyTerm = t;
  } else {
    c->penalty = x;
  }

  /* Since changing the weight of a constraint can invalidate the scores
     of LVs, all constraint nets must be deleted */
  cdgDeleteComputed();
}


/* -------------------------------------------------------------- 
 * Constructor for a new pair of attribute and value 
 */
Value newValueAV(String att, Value v) {
  AVNode att_value;
  Value newNode = (Value)memMalloc(sizeof(ValueStruct));
  newNode->type = VTAVNode;
   
  att_value = (AVNode)memMalloc(sizeof(AVNodeStruct));
  att_value->attribute = strRegister(att);
  att_value->value = v;
   
  newNode->data.avnode = (struct AVNodeStruct *)att_value;
  return newNode;
}
   
/* ------------------------------------------------------------
   An arc constructor, which belongs to the ArcStruct
*/
Arc newArc(int from, int to, int no, 
           String word, Number penalty,
           Value info,
           int fstart, int fstop)
{
  Arc arc = (Arc)memMalloc(sizeof(ArcStruct));
  arc->from = from;
  arc->to = to;
  arc->no = no;
  arc->word = strRegister(word);
  arc->penalty = penalty;
  arc->info = info;
  arc->fstart = fstart;
  arc->fstop = fstop;
  return arc;
}

/* --------------------------------------------------------------
 * An Lexical Entry Constructor
 */
LexiconItem newLexiconItem(String word, Value v, String filename, int lineNo)
{
  LexiconItem le;

  le = (LexiconItem) memMalloc(sizeof(LexiconItemStruct));
  le->word = strRegister( word );
  le->filename = strRegister(filename);
  le->lineNo = lineNo;
  le->value = v;
  le->description = strRegister(le->word);
  le->values = NULL;
  le->distinctive_features = NULL;
  
  return le;
} 

/* ----------------------------------------------------------------------------
 * clone a LexiconItem 
 */
LexiconItem inputCloneLI(LexiconItem le) {
  LexiconItem result =
    newLexiconItem(le->word,
                   cloneValue(le->value),
                   le->filename,
                   le->lineNo);
  memcpy(&result->re, &le->re, sizeof(regex_t));
  inputCacheLI(inputCurrentGrammar, result);
  result->description = strRegister(le->description);
  result->distinctive_features = listClone(le->distinctive_features);
  return result;
}

/* ----------------------------------------------------------------------------
 * Count differences in feature structure between items
 */
int leCompare(Input input, LexiconItem a, LexiconItem b, Boolean verbose) {

  List l, m, n, path;
  int i = 0;
  Value va, vb;
  int result = 0;
  Boolean found;
  
  for(l = input->allAttributes; l != NULL; l = listNext(l), i++) {
    path = listElement(l);
    va = a->values[i];
    vb = b->values[i];

    if(va == ErrorValue && vb != ErrorValue) {
      result++;
      if(verbose) {
        cdgPrintf(CDG_INFO, "Only in item B::%s: ", a->description);
        printPath(CDG_INFO, listElement(l));
        cdgPrintf(CDG_INFO, "\n");
      }
      continue;
    } else if(va != ErrorValue && vb == ErrorValue) {
      result++;
      if(verbose) {
        cdgPrintf(CDG_INFO, "Only in item A::%s: ", b->description);
        printPath(CDG_INFO, listElement(l));
        cdgPrintf(CDG_INFO, "\n");
      }
      continue;
    } else if(va == ErrorValue && vb == ErrorValue) {
      continue;
    }
    if(va->type != vb->type) {
      result++;
      if(verbose) {
        cdgPrintf(CDG_INFO, "Type mismatch:\n  %s:", a->description);
        printPath(CDG_INFO, listElement(l));
        cdgPrintf(CDG_INFO, " = ");
        printValue(CDG_INFO, va, 0);
        cdgPrintf(CDG_INFO, "\n");
        cdgPrintf(CDG_INFO, "  %s:", b->description);
        printPath(CDG_INFO, listElement(l));
        cdgPrintf(CDG_INFO, " = ");
        printValue(CDG_INFO, vb, 0);
        cdgPrintf(CDG_INFO, "\n");
      }
      continue;
    }
    switch ( va->type ) {
    case VTString:
      if(va->data.string != vb->data.string) {
        result++;
        if(verbose) {
          cdgPrintf(CDG_INFO, "Feature mismatch:\n  %s:", a->description);
          printPath(CDG_INFO, listElement(l));
          cdgPrintf(CDG_INFO, " = %s\n", va->data.string);
          cdgPrintf(CDG_INFO, "  %s:", b->description);
          printPath(CDG_INFO, listElement(l));
          cdgPrintf(CDG_INFO, " = %s\n", vb->data.string);
        }
      }
      break;
    case VTNumber:
      if ( va->data.number != vb->data.number ) {
        result++;
        if(verbose) {
          cdgPrintf(CDG_INFO, "Feature mismatch:\n  %s:", a->description);
          printPath(CDG_INFO, listElement(l));
          cdgPrintf(CDG_INFO, " = %f\n", va->data.number);
          cdgPrintf(CDG_INFO, "  %s:", b->description);
          printPath(CDG_INFO, listElement(l));
          cdgPrintf(CDG_INFO, " = %f\n", vb->data.number);
        }
      }
      break;
    case VTList:
      found = FALSE;
      m = va->data.list;
      n = vb->data.list;
      while(m) {
        if(((Value)listElement(m))->data.string != ((Value)listElement(n))->data.string) {
          found = TRUE;
          break;
        }
        if(((Value)listElement(listNext(m)))->data.number != ((Value)listElement(listNext(n)))->data.number) {
          found = TRUE;
          break;
        }
        m = listNext(listNext(m));
        n = listNext(listNext(n));
      }
      if(found) {
        result++;
        if(verbose) {
          cdgPrintf(CDG_INFO, "Feature mismatch:\n  %s:", a->description);
          printPath(CDG_INFO, listElement(l));
          cdgPrintf(CDG_INFO, " = ");
          printValue(CDG_INFO, va, 0);
          cdgPrintf(CDG_INFO, "\n");
          cdgPrintf(CDG_INFO, "  %s:", b->description);
          printPath(CDG_INFO, listElement(l));
          cdgPrintf(CDG_INFO, " = ");
          printValue(CDG_INFO, vb, 0);
          cdgPrintf(CDG_INFO, "\n");
        }
      }
      break;
    default:
      break;
    }
  }

  return result;

}

/* ----------------------------------------------------------------------------
 * Compare LexiconItem structures by their description.
 */
Boolean leCompareByName(LexiconItem a, LexiconItem b) {

  if(a->description && b->description) {
    return strcmp(a->description, b->description) < 0;
  }

  /* force stable sort */
  return (Pointer)a > (Pointer)b;

}

/* ------------------------------------------------------------
   An ad hoc method to convert VM Lattice penalty field
   to DAWAI 0.0-1.0 penalty range.
   This will be worked on.
*/
Number vmArcPenalty(Number score, Number frame_start, Number frame_end)
{
#if 1
  return score;
#else
  double ret;
  double frame_diff = (frame_end - frame_start) * 100.0;

  if (frame_diff <= DBL_MIN) return (Number)0.0;
  ret = 1.0 - score / frame_diff;
  if (ret <= DBL_MIN) return (Number)0.0;

  return (Number)ret;
#endif
}

/*  ----------------------------------------------------------------------------
 * vmAdjustAcousticScoreNorm
 */
void vmAdjustAcousticScoreNorm(Lattice lat)
{
  Number best=DBL_MAX, worst=-DBL_MAX;
  List l;
  
  if (!lat->arcs) {
    cdgPrintf(CDG_WARNING, "vmAdjustAcousticScoreNorm: no arcs?\n");
    return;
  }

  for(l=lat->arcs; l; l=listNext(l)) {
    Arc a=listElement(l);
    /*
      int frames=a->fstop-a->fstart;
      Number p;
      if (frames<1) { vmAcoustics=vmAcOff; break; }
      p=a->penalty/frames;
    */
    Number p=a->penalty;
    best=min(best, p);
    worst=max(worst, p);
  }

  if (best>=worst) { vmAcoustics=vmAcOff; }

  /*   cdgPrintf(CDG_WARNING, "best %6.5e worst %6.5e\n", best, worst); */
  
  for(l=lat->arcs; l; l=listNext(l)) {
    Arc a=listElement(l);
    /*     int frames=a->fstop-a->fstart; */
    switch (vmAcoustics)
      {
      case vmAcOff:
	a->penalty=1.0;
	break;
      case vmAcLinear:
        /* 	a->penalty/=frames; */
	/*
	  it really should be 

	  a->penalty=(exp(-a->penalty) - exp(-worst))/(exp(-best) - exp(worst));

	  but then we'll have arithmetic underflow. exp(-worst) is
	  small anyway and
	  exp(-a->penalty)/exp(-best)=
	  exp(-a->penalty - -best)=
	  exp(best - a->penalty)
	*/
	a->penalty=exp(best - a->penalty);
	break;
      case vmAcScaled:
	a->penalty=(worst-a->penalty)/(worst-best);
	break;
      }
  }
}

/*  ----------------------------------------------------------------------------
 * vmAdjustAcousticScore
 */
void vmAdjustAcousticScore(Lattice lat)
{
  switch (vmAcoustics)
    {
    case vmAcOff: 
    case vmAcLinear: 
    case vmAcScaled: vmAdjustAcousticScoreNorm(lat); break;
    default:
      cdgPrintf(CDG_ERROR, "unknown acoustics method %d\n", vmAcoustics);
      exit(1);
    }
}

/* ------------------------------------------------------------
   Takes a VM string and converts i.e. ("a) to .
   Works in place.
*/
String vm2String(String str) {
  char* s = str;
  char* t = str;

  while(*s != '\0') {
    if (*s == '\"') {
      s++; /* skip quote */
      if (*s == '\0') break; /* end of string */
      switch (*s) {
      case 'a':
        *t++ = '';
        s++;
        break;
      case 'o':
        *t++ = '';
        s++;
        break;
      case 'u':
        *t++ = '';
        s++;
        break;
      case 'A':
        *t++ = '';
        s++;
        break;
      case 'O':
        *t++ = '';
        s++;
        break;
      case 'U':
        *t++ = '';
        s++;
        break;
      case 's':
        *t++ = '';
        s++;
        break;     
      } 
    } else {
      *t++ = *s++;
    }
  }
  *t = '\0'; /* End t string */
  return str;
}

/* --------------------------------------------------------------------------
   Is X composed of only uppercase letters?
*/
Boolean inputALLUPPER(String x)
{
    if (strspn(x, "ABCDEFGHIJKLMNOPRSTUVWXYZ") == strlen(x)) {
      return TRUE;
    }
  return FALSE;
}

/* ----------------------------------------------------------------------------
   Does this Arc carry a word that might indicate a sentence boundary?

   This is only true of some punctuation marks.
*/
Boolean inputSentenceBoundary(Arc arc) {
  if (arc &&
      (!strcmp(arc->word, "\"") ||
       !strcmp(arc->word, "``") ||
       !strcmp(arc->word, "`") ||
       !strcmp(arc->word, "'") ||
       !strcmp(arc->word, "(") ||
       !strcmp(arc->word, ")") ||
       !strcmp(arc->word, "[") ||
       !strcmp(arc->word, "]") ||
       !strcmp(arc->word, "-") ||
       arc->word[strlen(arc->word)-1] == '.' ||
       !strcmp(arc->word, ",") ||
       !strcmp(arc->word, "?") ||
       !strcmp(arc->word, "!") ||
       !strcmp(arc->word, ":"))) {
    return TRUE;
  }
  return FALSE;
  
}

/* ----------------------------------------------------------------------------
   A lexical template may capture substrings which are then
   accessible as backreferences in the newly created Value.
   This function turns the string '\1' into whatever was matched.
*/
void inputSubstBackrefs(Value v, String word, regmatch_t *matches) {

  List l;
  
  switch ( v->type ) {
  case VTString:
    if(strlen(v->data.string) == 2 &&
       v->data.string[0] == '\\' &&
       strspn(&v->data.string[1], "0123456789") == 1 ) {
      int i;
      int from, to;
      String s = strCopy(word);
      sscanf(&v->data.string[1], "%d", &i);
      from = matches[i].rm_so;
      to = matches[i].rm_eo;
      if(to < 0) {
        cdgPrintf(CDG_ERROR, "ERROR: backreference %d not matched!\n", i);
        return;
      }
      cdgFreeString(v->data.string);
      strncpy(s, &word[from], to - from);
      s[to-from] = '\0';

      /* maybe cast the result back to number */
      if(strspn(s, "0123456789-.") == strlen(s)) {
        int n;
        v->type = VTNumber;
        sscanf(s, "%d", &n);
        v->data.number = n;
      } else {
        v->data.string = strRegister(s);
      }
      memFree(s);
    }
    break;
  case VTAVNode:
    inputSubstBackrefs(((AVNode)v->data.avnode)->value, word, matches);
    break;
  case VTList:
  case VTDisjunction:
  case VTConjunction:
    for ( l = v->data.list; l != NULL; l = listNext(l) ) {
      inputSubstBackrefs(listElement(l), word, matches);
    }
    break;
  case VTNumber:
  default:
    break;
  }  
}

/* ----------------------------------------------------------------------
   Search the lexicon for an exact match with WORD.

   This function checks if the exact form WORD is present in the lexicon;
   if so, it returns all lexicon items for it. It does *not* consult
   templates (see inputTemplatesGet), conflate spurious upper/lower case
   (see lgNewIter), or do compound analysis (see inputLexiconGet).

   However, it does understand spontaneous upper case: if a category is
   listed in `capitalizable-cats', then the lower case form is allowed even
   if WORD is in upper case. This allows processing of `die Deutsche Bahn'
   without duplicating all entries for `deutsch'.

   Do not confuse spontaneous upper case (caused by being part of a proper
   name) with spurious upper case (caused by being the first word of a
   sentence). The latter must be dealt with in context, which is why it's
   done in lgNewIter(); but the former can happen to all words of the
   appropriate category, so we can do it here.

   The only thing it *does* abstract from is the difference between lexicon
   items in RAM (in the field inputCurrentGrammar->lexicon) and those in
   the external data base (hiding behind cdgdb::database).
*/
List inputLexiconGetExactly(String word) {
  
  List result = NULL;
  List l, m;
  String low = strCopy(word);
  low[0] = tolower(low[0]);

  /* First, query the database. If there are entries available, dbLoad()
     adds them to inputCurrentGrammar->lexicon. It takes care not to
     add the same items twice, so we can call it unconditionally. */
  if(dbAvailable()) {
    dbLoad(word);
    if(isupper(word[0])) {
      dbLoad(low);
    }
  }
  
  /* Now look up the form in the lexicon. */
  m = hashGet(inputCurrentGrammar->lexicon, word);
  for(l = m; l != NULL; l = listNext(l)) {
    result = listPrependElement(result, inputCloneLI(listElement(l)));
  }
  
  /* Maybe this word is in spontaneous upper case.
     Look up lower-case variants to check that. */
  if(taggerCategoryIndex >= 0 &&
     isupper(word[0])) {
    Boolean found = FALSE;
    List n = hashGet(inputCurrentGrammar->lexicon, low);
    for(l = n; l != NULL; l = listNext(l)) {
      LexiconItem li = listElement(l);
      Value v = li->values[taggerCategoryIndex];
      if (v->type == VTString &&
          listContains(inputSpontaneousUpcase, v->data.string)) {

        /* If so, generate the upcase version on the fly
           and store it for later reuse. */
        LexiconItem li2 = inputCloneLI(li);
        String new_word = strCopy(li2->word);
        String new_description = strCopy(li2->description);
        new_word[0] = toupper(li2->word[0]);
        new_description[0] = toupper(li2->description[0]);
        li2->word = new_word;
        li2->description = new_description;
        result = listPrependElement(result, li2);
        found = TRUE;
      }
    }
    
    /* The new items and the old items must be annotated with distinctive
       features, since by definition they have the same reading. */
    if(found) {
      for(l = result; l != NULL; l = listNext(l)) {
        LexiconItem li = listElement(l);
        cdgFreeString(li->description);
        li->description = strRegister(li->word);
      }
      resolvePartition(inputCurrentGrammar,
                       result,
                       inputCurrentGrammar->allAttributes);
    }
  }
  return result;
}

/* ----------------------------------------------------------------------
   Search existing templates for items matching WORD.

   Only templates are considered, not normal lexicon items. All templates
   that match WORD are used to produce new LexiconItems which are then
   owned by the caller.

   Templates that have a category (as defined by taggerCategoryPath)
   listed in CATEGORIES are not used.
   
*/
List inputTemplatesGet(String word, List categories) {
  List l, result = NULL;
  
  /* templates are stored in the lexicon under the key "" */
  for(l = hashGet(inputCurrentGrammar->lexicon, strRegister(""));
      l != NULL; l = listNext(l)) {
    LexiconItem template = listElement(l);
    regmatch_t matches[10];

    /* check regex match */
    if (!regexec(&template->re, word, 10, matches, 0)) {
      LexiconItem le = inputCloneLI(template);
      le->word = strRegister(word);
      le->description = strRegister(word);
      inputSubstBackrefs(le->value, word, matches);

      /* maybe suppress */
      if(NULL != categories && taggerCategoryIndex >= 0) {
        Value v = le->values[taggerCategoryIndex];
        if (v->type == VTString &&
            listContains(categories, v->data.string)) {
          freeLexiconItem(le);
          continue;
        }
      }
      result = listAppendElement(result, le);
    }
  }
  
  return result;

}

/* ----------------------------------------------------------------------
   Probe for templates matching WORD.

   TRUE is returned exactly if inputTemplatesGet() would not return NULL, 
   but no new items are actually generated or returned.
   
*/
Boolean inputTemplatesQuery(String word) {
  List l;
  
  /* templates are stored in the lexicon under the key "" */
  for(l = hashGet(inputCurrentGrammar->lexicon, strRegister(""));
      l != NULL; l = listNext(l)) {
    LexiconItem template = listElement(l);
    regmatch_t matches[10];

    /* check regex match */
    if (!regexec(&template->re, word, 10, matches, 0)) {
      return TRUE;
    }
  }
  
  return FALSE;

}

/* ----------------------------------------------------------------------
   Does the lexicon contain at least one item with the reading WORD?
*/
Boolean inputLexiconQuery(String word) {

  if(dbAvailable()) {
    dbLoad(word);
  }
  if(hashGet(inputCurrentGrammar->lexicon, word)) {
    return TRUE;
  }
  return inputTemplatesQuery(word);
}


/* ----------------------------------------------------------------------
   Search current lexicon for matching words.

   This function returns a List of LexiconItem structures that match WORD. 
   The simplest kind of match is if a known lexicon item has the exact
   reading WORD.

   Near-matches of various kinds are tried if no exact match is found:

   If inputDeduceCompounds is set, a lexicon item may match if it is a
   compound of WORD marked with a hyphen: a query for `Halb-Elf' will
   succeed by finding `Elf'.

   If inputUnmarkedCompounds is set, a lexicon item may match if it is any
   compound of word: `Halbelf' will match on `Elf'. The word that is
   actually in the lexicon must have the same casing as WORD, and it must
   belong to the categories allowed by inputUnmarkedCompounds.
   
   These three lookup methods are tried in order, so if there is an exact
   match in the normal lexicon, no compounding will be tried.

   If lexicon templates are loaded, any of them whose regex matches WORD
   may be returned. This behaviour is further controlled by the setting of
   templateMode.
   
*/
List inputLexiconGet(String word) {

  List result, l;
#ifdef DEBUGLGSEARCHLEXICON
  if (word) {
    printf("DEBUG: Querying word `%s'\n", word);
    printf("DEBUG: (Actual characters:");
    for (i = 0; i < strlen(word); i++) {
      printf(" %c", word[i]);
    }
    printf(")\n");
    printf("DEBUG: (Actual data:");
    for (i = 0; i < strlen(word); i++) {
      printf(" %x", word[i]);
    }
    printf(")\n");
  }
#endif

  /* detect inconsistent option setting:
     per-category template use is only possible
     if a category string is defined. */
  if(2 == templateMode &&
     taggerCategoryIndex < 0) {
    templateMode = 1;
    cdgPrintf(CDG_WARNING, "WARNING: no category string defined.\n");
    cdgPrintf(CDG_WARNING, "WARNING: `templates' was reset to `ifneeded'.\n");
  }

  /* look for exact match */
  result = inputLexiconGetExactly(word);

  if(!result) {
  
    /* try to find entry for a marked suffix
       if no exact match is present */
    if(inputDeduceCompounds) {
      String suffix, prefix;
      if((NULL != (suffix = strchr(word, '-'))
          ||
          (NULL != (suffix = strchr(word, '/'))))
         && strlen(suffix) > 1) {

        if(word < suffix) {
          int i = suffix-word;
        
          prefix = strCopy(word);
          prefix[i+1] = '\0';
          
          result = inputLexiconGet(suffix+1);
          for(l = result; l != NULL; l = listNext(l)) {
            LexiconItem le = listElement(l);
            le->word = strRegister(word);
            le->description = strAppend(prefix, le->description, NULL);
          }
          memFree(prefix);
        }
      }
    }
  }

  if(!result) {
    
    /* try to find entry for an unmarked suffix
       if not even a marked suffix is present */
    if(strlen(word) > 4) {
      int i;
      Boolean upper = FALSE;
      if(isupper(word[0])) {
        upper = TRUE;
      }
      for(i = 2; i < strlen(word)-3; i++) {
        String suffix = strCopy(word+i);
        String prefix = strCopy(word);
        List meanresult;
        prefix[i] = '\0';
        if(upper) {
          suffix[0] = toupper(suffix[0]);
        }
#ifdef DEBUGINPUTLEXICONGET
        cdgPrintf(CDG_DEBUG, "i == %d  prefix == `%s'  suffix == `%s'\n",
                  i, prefix, suffix);
#endif
        meanresult = inputLexiconGetExactly(suffix);
        if(meanresult) {
          for(l = meanresult; l != NULL; l = listNext(l)) {
            LexiconItem le = listElement(l);

            /* check allowed categories */
            if(taggerCategoryIndex >= 0) {
              Value v = le->values[taggerCategoryIndex];
              if (v->type == VTString &&
                  !listContains(inputCompoundCats, v->data.string)) {
                continue;
              }

              /* force lower case after compound break */
              {
                String newd = strCopy(le->description);
                newd[0] = tolower(newd[0]);
                le->description = strRegister(newd);
              }
            
              result = listPrependElement(result, le);
              le->word = strRegister(word);
              le->description =
                strRegister(strAppend(prefix, le->description, NULL));
              inputLexiconPut(inputCloneLI(le), inputCurrentGrammar->lexicon);
            }
          }
          if(result) {
            cdgPrintf(CDG_DEBUG,
                      "DEBUG: assuming `%s' is a compound of `%s'\n",
                      word, suffix);
            break;
          }
        } else {
          memFree(suffix);
          memFree(prefix);
        }
      }
    }
  }
  
  /* templateMode determines when to look for matching templates. */

  /* never */
  if(0 == templateMode) {

    /* no issues with newly created items, we can return immediately */
    return result;
  }
  
  /* by category */
  if(1 == templateMode) {
    List cats = NULL;
    List l;
    for(l = result; l != NULL; l = listNext(l)) {
      LexiconItem li = listElement(l);
      Value v = li->values[taggerCategoryIndex];
      if (v->type == VTString) { 
        cats = listAddUniqueElement(cats, v->data.string);
      }
    }
    result = listAppendList(result, inputTemplatesGet(word, cats));
    listDelete(cats);
  }

  /* if needed */
  else if(2 == templateMode) {
    if(!result) {
      result = inputTemplatesGet(word, NULL);
    }
  }
  
  /* always */
  else if(3 == templateMode) {
    result = listAppendList(result, inputTemplatesGet(word, NULL));
  }

  /* Items created from templates are brand new,
     and have not received distinctive feature lists yet,
     so we have to install them now. */
  for(l = result; l != NULL; l = listNext(l)) {
    LexiconItem li = listElement(l);
    cdgFreeString(li->description);
    li->description = strRegister(li->word);
  }
  resolvePartition(inputCurrentGrammar,
                   result,
                   inputCurrentGrammar->allAttributes);

  return result;
}


/* ----------------------------------------------------------------------------
 * Add a lexical entry to the lexicon.
 */
void inputLexiconPut(LexiconItem le, Hashtable lexicon) {
  List l = hashGet(lexicon, le->word);
  l = listAppendElement(l, le);
  hashSet(lexicon, le->word, l);
}

/* ----------------------------------------------------------------------
   Insert the lexical entry LE into the lexicon of INPUT.

   If LE is disjunctive, it is expanded first and the list of results is
   inserted instead. The original disjunctive item is deallocated.
*/
void inputInsertLI(Input input, LexiconItem le) {
  List x, l;
  
  /* accept normal item */
  if(!disjunction(le->value)) {
    inputLexiconPut(le, input->lexicon);
    return;
  }

  /* expand disjunctive item */
  x = expandValue(le->value);
  for (l = x; l; l = listNext(l)) {
    LexiconItem leNew =
      newLexiconItem(strRegister( le->word ),
                     (Value) listElement(l),
                     le->filename,
                     le->lineNo);
    memcpy(&leNew->re, &le->re, sizeof(regex_t));
    inputLexiconPut(leNew, input->lexicon);
  }
  listDelete(x);
  freeLexiconItem(le);
}


/* ----------------------------------------------------------------------------
   Detect transliterations for German umlauts
   and replace them with the real thing.
*/
String inputEncodeUmlauts(String word) {

  String result;
  String s;
  int i, j;

  if(!word) {
    return NULL;
  }

  if(!cdgEncodeUmlauts) {
    return strRegister(word);
  }
  
  result = memMalloc(strlen(word) + 1);
  strcpy(result, "");
  for(i = 0; i < strlen(word); i++) {

    /* These rules transform "a into , etc.

    This is for the case where your keyboard doesn't have . */
    if(strstr(word+i, "\"a") == word+i) {
      strcat(result, "");
      i++;
    }
    else if(strstr(word+i, "\"o") == word+i) {
      strcat(result, "");
      i++;
    }
    else if(strstr(word+i, "\"u") == word+i) {
      strcat(result, "");
      i++;
    }
    else if(strstr(word+i, "\"A") == word+i) {
      strcat(result, "");
      i++;
    }
    else if(strstr(word+i, "\"O") == word+i) {
      strcat(result, "");
      i++;
    }
    else if(strstr(word+i, "\"U") == word+i) {
      strcat(result, "");
      i++;
    }
    else if(strstr(word+i, "\"s") == word+i) {
      strcat(result, "");
      i++;
    }

    else {
      j = strlen(result);
      result[j] = word[i];
      result[j+1] = '\0';
    }
  }

  if(strcmp(word, result)) {
    cdgPrintf(CDG_DEBUG, "DEBUG: normalised `%s' to `%s'\n", word, result);
  }

  s = strRegister(result);
  memFree(result);
  return s;

}

/* ----------------------------------------------------------------------------
 * Reads CDG structures from filename into parseResult.
 *
 * side-effects: 
 *	- produce a newly created input structure in parseResult
 *	- store the loaded filename in parseCurrentFile
 */
Boolean inputLoad(String filename) 
{
  FILE *inFile = NULL;
  int pid=0;
  int fds[2];
  struct stat buffer;
  char filepath[PATH_MAX];
  char cwd[1024] = "";
  String dir;

  
  /* check that it isn't a directory, which would crash the lexer */
  stat(filename, &buffer);
  if (S_ISDIR(buffer.st_mode)) {
    cdgPrintf(CDG_ERROR, "ERROR: `%s' is a directory.\n", filename);
    return FALSE;
  }

  /* get the real location of the cruft */
  realpath(filename, filepath);
  filename = strRegister(filepath);
  dir = strCopy(filename);

  /* remember original directory */
  getcwd(cwd, 1024);

  /* enter specified directory,
     in case the file contains relative includes */
  if(strcmp(cwd, "")) {
    chdir(dirname(dir));
  }

  /* load a dll grammar */
  if (strlen(filename) >= 3 &&
      strcmp(&(filename[strlen(filename) - 3]), ".so") == 0) {
#ifndef DARWIN
    parseResult = comLoad(filename);
#endif
    parseCurrentFile = strRegister(filename);
    chdir(cwd);
    memFree(dir);
    if (!parseResult) 
      return FALSE;

    return TRUE; /* we are done */
  }

  /* preprocess a file */
  if (strlen(filename) >= strlen(inputPrepSuffix) &&
      strcmp(&(filename[strlen(filename) - strlen(inputPrepSuffix)]), 
             inputPrepSuffix) == 0) {
    cdgPrintf(CDG_INFO,
              "INFO: using preprocessor `%s' for `%s'\n", 
              inputPreprocessor, filename);
    if (pipe(fds) != 0) {
      cdgPrintf(CDG_ERROR, "ERROR: calling pipe\n");
      chdir(cwd);
      memFree(dir);
      return FALSE;
    }
    if ((pid = fork()) < 0) {
      cdgPrintf(CDG_ERROR, "ERROR: can't fork\n");
      chdir(cwd);
      memFree(dir);
      return FALSE;
    }
    if (pid == 0) {
      /* 990615, IS
         I don't know whether this fixes the occasional crashes when 
         hitting Ctrl-C while the m4 process was active. Shouldn't hurt.
      */
      /* 20020305, IS: re-setting signal handler to default */
      signal(SIGINT, SIG_DFL);
      signal(SIGTERM, SIG_DFL);
      signal(SIGXCPU, SIG_DFL);

      /* child process */

      close(1);
      dup(fds[1]);
      close(fds[1]);
      close(fds[0]);
      /* 990318, IS: 
       * `m4 -s' inserts line like `#line 3 "prg.c"' into the output stream
       * which is nice if you want correct information about the origin of 
       * your input (Hint: macro expansion & file inclusion). We hacked 
       * the lexer to account for these lines and update the `currentFile' 
       * and `lineNo' variables accordingly. Unfortunately, the handling
       * of keeping track of loaded files (e. g. in cmdUnload) depends on
       * unique filenames. Therefore I disabled the `-s' flag again.
	  
       * 990319, IS:
       * Ok, we (Kilian & I) decided to abandon the concept of automatic 
       * version control by filenames in cmdUnload and cmdLoad. Each time 
       * something is loaded all items are deleted.
       * old version:
       * execl(inputPreprocessor, basename(inputPreprocessor), filename, (char *)NULL);	
       *
       * 990413, IS
       * Next change. Now it's possible to load more than one file but you 
       * can only reset the whole system, i. e. discard all loaded and 
       * computed information. Loaded items are identified by their ID and
       * previously loaded items are overwritten by later ones. 
       * CHANGE: No warning is issued when an item is overwritten since this
       * will happen quite often.
       *
       */
      prepArgv[prepArgNo-1] = strRegister(filename);
      execv(prepArgv[0], prepArgv);
      cdgPrintf( CDG_ERROR, "ERROR: can't exec `%s'\n", inputPreprocessor );
      /* this exits the child process */
      _exit(1);
    } else {
      /* mother process */
      /* close write side */
      close(fds[1]);
      inFile = fdopen(fds[0], "r");
      if (!inFile) {
        cdgPrintf(CDG_ERROR, 
                  "ERROR: `load': can't read pipe for `%s' from m4\n",
                  filename);
	chdir(cwd);
	memFree(dir);
        return FALSE;
      }
    }
  } else {
    inFile = fopen(filename, "r");
    if (!inFile) {
      cdgPrintf(CDG_ERROR,
                "ERROR: `load': no such file `%s'\n", 
                filename);
      chdir(cwd);
      memFree(dir);
      return FALSE;
    }
  }

  /* 990319, IS: The problem here is that we need a registered filename
   * while parsing. What is even worse is that new files may appear 
   * while parsing. But if we detect an error we don't want these 
   * filenames in cdgFiles, instead we want the old filenames.
   */
  if (parseResult) {
    cdgPrintf(CDG_WARNING, "WARNING: structure `parseResult' is not NULL.");
  }
  parseResult = inputNewInput();
  parseCurrentFile = inputRegisterFilename( parseResult, filename );
  /* This is necessary for the old flex.
   * yyrestart(yyin);
   */
  yyin = inFile;
  parseLineNo = parseNoOfWarnings = parseNoOfErrors = 0;
  yyparse();
  fclose(inFile);
  if (pid) {
    (void)wait4(pid, (int *)NULL, 0, NULL);
  }

  
  /* restore cwd */
  chdir(cwd);
  memFree(dir);

  return TRUE;
}

/* ----------------------------------------------------------------------------
 * read a cdg init rc file.
 *
 * @param initFileName the name of a file or NULL
 * @returns TRUE if an init file has successfully been loaded, otherwise FALSE
 *
 * if no init file has been provided then try to find the default one
 * @ref DEFAULTINITFILE 
 * - in the current directory
 * - or in the home directory
 */
Boolean inputReadInitFile(String initFileName)
{
  FILE *initFile = NULL;
  int len;
  char buffer[MAXBUFFER];

  if (initFileName == NULL) {
    String homeDir = NULL;

    /* open the default one ... */
    initFileName = strRegister(DEFAULTINITFILE);

    /* ... in the current directory */
    if ((initFile = fopen(initFileName, "r")) == NULL) {
      initFileName = NULL;

      /* ... or in the HOME directory */
      if ((homeDir = getenv("HOME")) != NULL) {

	/* check the amount of stuff which will be sprinted into buffer */
	if (strlen(homeDir) + strlen(DEFAULTINITFILE) > MAXBUFFER - 2) {
	  cdgPrintf(CDG_WARNING, "WARNING: path to init file too long\n");
	} else {
	  initFileName = strPrintf("%s/%s", homeDir, DEFAULTINITFILE);
	  initFile = fopen(initFileName, "r");
	}
      }
    }
  } else {
    initFile = fopen(initFileName, "r");
  }

  /* no init file hanging around anywhere */
  if (!initFileName || !initFile) {
    return FALSE;
  }

  /* we've found an init file that we can use */
  cdgPrintf(CDG_INFO, "INFO: processing initialization file `%s'\n",
	    initFileName);

  /* 
   * From man exec(2): File descriptors open in the calling process image remain open in
   * the new process image, except for those whose close-on-exec flag FD_CLOEXEC is set;
   * (see fcntl(2)).  For those file descriptors that remain open, all attributes of the
   * open file description, including file locks, remain unchanged.
   * 
   * This file should be closed by all children since we are the only ones who have a
   * legitimate interest in it. */
  if (fcntl(fileno(initFile), F_SETFD, FD_CLOEXEC) == -1) {
    cdgPrintf(CDG_ERROR, "ERROR: fcntl failed: %s\n", strerror(errno));
    fclose(initFile);
    return FALSE;
  }

  /* read all init stuff */
  while (!feof(initFile)) {
    if (!fgets(buffer, MAXBUFFER, initFile)) {
      break;
    }
    len = strlen(buffer);
    if (len >= MAXBUFFER - 1) {
      cdgPrintf(CDG_WARNING,
		"WARNING: line too long, skipping rest of initialization file\n");
      fclose(initFile);
      return FALSE;
    }
    if (len > 0 && buffer[len - 1] == '\n') {
      buffer[len - 1] = '\0';	       /* strip newline */
    }
    cdgPrintf(CDG_DEFAULT, "%s\n", buffer);
    commandEval(buffer);
  }
  fclose(initFile);

  return TRUE;
}

/* ----------------------------------------------------------------------------
   Return string version of LE's value for attribute FEATURE.
*/
String inputFeatureOfLI(LexiconItem le, String feature) {
  Value v;
  int index;
  
  if(!le) {
    return NULL;
  }
  index = inputLookupPath(scanPath(feature), inputCurrentGrammar);
  if(index == -1) {
    return NULL;
  }

  v = le->values[index];
  if(v == ErrorValue) {
    return NULL;
  }

  return formatValue(v, 0);

}

/* ----------------------------------------------------------------------------
   Are these two values similar enough to count as a `hit' when
   verifying parse results?
*/
Boolean inputCompareFeatures(String a, String b, Hierarchy h) {
  String bottom = strRegister("bot");
  int x,y;
  
  /* `bot' matches everything, no need to even look at h */
  if(bottom == a || bottom == b) {
    return TRUE;
  }

  /* `nom' matches `nom' */
  if(a == b) {
    return TRUE;
  }

  /* `nom_dat_acc' should match `nom',
     but that only works if the user named a hierarchy to use for lookup */
  if(h) {
    if (0 > (x = (int)hashGet(h->types, a) - 1)) {
      return FALSE;
    }
    if (0 > (y = (int)hashGet(h->types, b) - 1)) {
      return FALSE;
    }
#ifdef DEBUGINPUTCOMAREFEATURES
    if (((int)arrayElement(h->distance, x, y) >= 0) ||
        ((int)arrayElement(h->distance, y, x) >= 0)) {
      cdgPrintf(CDG_DEBUG, "DEBUG: %s matches %s.\n", a, b);
    }
#endif
    return (((int)arrayElement(h->distance, x, y) >= 0) ||
            ((int)arrayElement(h->distance, y, x) >= 0)); 
  }

  /* nope, too different */
  return FALSE;
  
}

/* ----------------------------------------------------------------------
   Allocate an Annotation with the specified fields.
*/
Annotation newAnnotation(int from, int to, String word) {
  Annotation result = memMalloc( sizeof(AnnotationStruct) );
  result->from = from;
  result->to = to;
  result->word = word;
  result->specs = NULL;
  return result;
}

/* ----------------------------------------------------------------------
   Allocate a Specification with the specified fields.
*/
Specification newSpecification(SpecType type, String kind,
                               String name, int position) {
  Specification result = memMalloc( sizeof(SpecificationStruct) );
  result->type = type;
  result->kind = strRegister(kind);
  result->name = strRegister(name);
  result->position = position;
  return result;
}

/* ----------------------------------------------------------------------
   Add a Specification to an Annotation.
*/
void inputSpecificationAdd(Annotation a, Specification s) {
  a->specs = listInsertSorted(a->specs, s, inputSpecCompare);
}

/* ----------------------------------------------------------------------
   Change the field ->name of a Specification.
*/
void inputSpecSetName(Specification spec, String s) {
  spec->name = strRegister(s);
}


/* ----------------------------------------------------------------------
   Sort two strings intelligently like "ls -v" does, i.e.
   s10 is sorted after s2.
*/
int smartly(const void *a, const void *b) {
  String pa = *((String *)a);
  String pb = *((String *)b);
  int x, y;
  int i;
  while(*pa && *pb) {
    
    /* remove identical non-numeric parts */
    while(pa && pb && pa[0] == pb[0] && !isdigit(pa[0])) {
      pa++;
      pb++;
    }

    /* compare different non-numeric parts alphabetically */
    if(!isdigit(pa[0]) || !isdigit(pb[0])) {
      if(strcmp(pa,pb) < 0) {
        return -1;
      } else {
        return 1;
      }
    }

    /* compare numeric parts numerically */
    sscanf(pa, "%d", &x);
    sscanf(pb, "%d", &y);
    if(x > y) {
      return 1;
    }
    if(x < y) {
      return -1;
    }

    i = strspn(pa, "0123456789");
    pa += i;
    pb += i;
  }
  return 0;

}


/* ----------------------------------------------------------------------
   Return sorted list of all known lattices' names.
*/
List inputGetSortedLatticeNames(Input input) {
  int nr = listSize(input->lattices);
  List result = NULL;
  String *field = (String *)memMalloc(sizeof(String) * nr);
  int i = 0;
  List l;
  
  for(l = input->lattices; l != NULL; l = listNext(l)) {
    Lattice lat = listElement(l);
    field[i++] = lat->id;
  }
  qsort(field, nr, sizeof(String), smartly);
  for(i = nr; i;) {
    i--;
    result = listPrependElement(result, field[i]);
  }

  return result;
}

/* ----------------------------------------------------------------------
   Set the fields quotes, parens, and phraseQuotes in each Arc.

   The field parens indicates within how many levels of parentheses a word
   stands. Thus, in the sentence

     We have (we believe) ample evidence.

   the words `we' and `believe' would receive a value of 1, and all other
   words, 0. This information can be retrieved in a constraint with the
   predicate parens(). For instance, a constraint might say that it is okay
   to have multiple sentence roots if they are on different levels of
   parenthesis nesting, but not otherwise.

   The field quotes counts the number of quotation marks that a word is
   in. In the sentence

     "Thus", said Gandalf, "passes Denethor."

   the words `said' and `Gandalf' will receive values of 0, and all other
   words, 1. The information can be retrieved in a constraint with the
   predicate quotes(). For instance, it might say that a matrix clause must
   not have a higher value that its object clause.

   The field phrasequotes is a variant of quotes that only considers
   quotations that are not also preceded with commas. For instance, in the
   previous example it would remain 0 for all words. It would assign a
   value of 1 to the word `shall' in the sentence

     Mark you his absolute "shall"?
     
   This information can be retrieved in a constraint with the predicate
   phrasequotes(). For instance, it could say that words of other classes
   may function as nouns, i.e. take determiners, be subjects etc., if they
   are phrase-quoted.
     
     
   All of this information is limited to what can be detected within one
   lattice. For instance, if in a long paragraph, several sentences are
   contained within one pair of parentheses, then all but the first will
   have no trace of that and carry parens values of 0.
   
*/
void inputScanLattice(Lattice lat) {
  List l;
  int p = 0;
  int q = 0;
  int pq = 0;
  Arc last, next;
  int minparen = 0;
  int minquote = 0;
  
  if(latticeBranches(lat)) {
    return;
  }

  last = NULL;
  for(l = lat->arcs; l != NULL; l = listNext(l)) {
    Arc a = listElement(l);
    next = listNext(l) ? listElement(listNext(l)) : NULL;
    
#ifdef DEBUGINPUTSCANLATTICE
    cdgPrintf(CDG_DEBUG, "DEBUG: `%s'\n", a->word);
#endif
    a->parens = p;
    a->quotes = q;
    a->phraseQuotes = pq;

    /* parenthesis are unambiguous */
    if(!strcmp(a->word, "(") ||
       !strcmp(a->word, "[")) {
      p++;
    }
    if(!strcmp(a->word, ")") ||
       !strcmp(a->word, "]")) {
      p--;
      if(p < minparen) {
        minparen = p;
      }
    }
    
    /* But quotation marks are more difficult when direct speech is
       distributed across sentences, so that the entire lattice has only
       one quote.

       At the start or end of a sentence it is clear that they start or end
       a quotation. But the first or last part may also be accompanied by a
       speech attribution. In that case we investigate the context on
       either side.

       Isolated closing quotes tend to be preceded by sentence-final
       punctuation:

          Die Wrfel sind gefallen!" sagte er.
          Wo bi-" begann er.

       or followed by commas:

          "Komm her, Junge! Das ist doch keine Gesellschaft fr dich",
          sagte er.

       while isolated opening quotes tend to be preceded by commas or
       semicolons:

         Er hub an, "Ich will Euch eine lange und wunderbare Geschichte
         erzhlen. Als ich im letzten Sommer...

       (Note that this last criterion is German-specific and fails badly in
       English, which has the opposite convention.)
 
    */
    if(!strcmp("\"", a->word)) {

      /* sentence-initial quote */
      if(!last) {
        q++;
        pq++;
#ifdef DEBUGINPUTSCANLATTICE
        cdgPrintf(CDG_DEBUG, "DEBUG: q++ for sentence-initial quote\n");
#endif
      }

      /* sentence-final quote */
      else if(!next) {
#ifdef DEBUGINPUTSCANLATTICE
        cdgPrintf(CDG_DEBUG, "DEBUG: q-- for sentence-final quote\n");
#endif
        q--;
        pq = 0;
        if(q < minquote) {
          minquote = q;
        }
      }

      /* second quote */
      else if(1 == q) {
#ifdef DEBUGINPUTSCANLATTICE
        cdgPrintf(CDG_DEBUG, "DEBUG: q-- for closing quote\n");
#endif
        q--;
        pq = 0;
        if(q < minquote) {
          minquote = q;
        }
      }

      /* hard case, look at the context */
      else if(strspn(last->word, "!?.")) {
#ifdef DEBUGINPUTSCANLATTICE
        cdgPrintf(CDG_DEBUG, "DEBUG: q-- for quote after `?'\n");
#endif
        q--;
        if(q < minquote) {
          minquote = q;
        }
      }

      else if(strspn(next->word, ",")) {
#ifdef DEBUGINPUTSCANLATTICE
        cdgPrintf(CDG_DEBUG, "DEBUG: q-- for quote before `,'\n");
#endif
        q--;
        if(q < minquote) {
          minquote = q;
        }
      }
      else if(strspn(last->word, "(:,")) {
#ifdef DEBUGINPUTSCANLATTICE
        cdgPrintf(CDG_DEBUG, "DEBUG: q++ for quote after `:,'\n");
#endif
        q++;
      }
      else {
#ifdef DEBUGINPUTSCANLATTICE
        cdgPrintf(CDG_DEBUG, "DEBUG: q++ by default\n");
#endif
        q++;
        pq++;
      }
    }
#ifdef DEBUGINPUTSCANLATTICE
    cdgPrintf(CDG_DEBUG, "%30s q=%d  p=%d  pq=%d\n", a->word, q, p, pq);
#endif
    last = a;
  }

  /* If we found an isolated closing quote or parenthesis,
     then some values may be below 0. That is no problem as long as we only
     compare quotelevels against each other, but if a constraint tests the
     quote level of a word against a constant the results become
     inconsistent. Therefore we adjust the level so that no value is
     negative. */
  for(l = lat->arcs; l != NULL; l = listNext(l)) {
    Arc a = listElement(l);
    if(minquote < 0) {
      a->quotes -= minquote;
    }
    if(minparen < 0) {
      a->parens -= minparen;
    }
  }
}

/* ----------------------------------------------------------------------
   Return the Arc in LAT whose field no is NO, or NULL.
*/
Arc inputFindArc(Lattice lat, int no) {
  List l;
  for(l = lat->arcs; l != NULL; l = listNext(l)) {
    Arc a = listElement(l);
    if(a->no == no) {
      return a;
    }
  }
  return NULL;
}

/* ---------------------------------------------------------------------- */
/* ---------------------------------------------------------------------- */
/** @} */
