/*
 * Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without 
 * modifications, as long as this notice is preserved.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Birth: 10/2/97
 * Author: Ingo Schroeder (see also AUTHORS and THANKS for more)
 * $Id: input.h,v 1.141 2004/09/01 13:57:58 micha Exp $
 */

/* -------------------------------------------------------------------------
 * @addtogroup Input 
 * @{ */

#ifndef INPUT_H
#define INPUT_H

/* ---------------------------------------------------------------------- */
#include <cdg.h>
#include <regex.h>

/* -- MACROS ------------------------------------------------------------ */
#ifndef SWIG
#define BUFFERSIZE 255
#define MAXBUFFER 4096
#define MAXARGNO 16384
#define DEFAULTINITFILE ".cdgrc"
#endif

/* ---------------------------------------------------------------------- */
#ifndef SWIG
typedef double Number;
#endif

typedef struct {
   String id;			       /* section name */
   Boolean active;		       /* use or ignore section */
   int counter;			       /* no of constraints in this section */
} SectionStruct;
typedef SectionStruct *Section;

typedef struct {
   String id;			       /* level name */
   List properties;		       /* list of Flags for extra info */
   List labels;			       /* list of labels */
   int no;			       /* index from zero on */
   Boolean showflag;		       /* show it or not */
   Boolean useflag;		       /* use it or ignore it */

   String filename;		       /* source file */
   int lineNo;			       /* line number in source file */
} LevelStruct;
typedef LevelStruct *Level;

#ifndef SWIG
typedef struct {
  String name;                         /* name of the property */
  String value;                        /* value of the property */
} PropertyStruct;
typedef PropertyStruct *Property;

typedef enum {			       /* How are two edges connected? */
   Top,				       /* at the top */
   Bottom,			       /* at the bottom */
   Over,			       /* edge X is vertically over edge Y */
   Under,			       /* edge X is vertically under edge Y */
   None,			       /* not at all */
   Inverse,			       /* edge X is the inversion of edge Y */
   Identical,			       /* edges X and Y share Top and Bottom */
   AnyCon                              /* any of the above */
} Connexion;			       

#define minConnexion ((int)Top)
#define maxConnexion ((int)Identical)
#define nrConnexions (maxConnexion - minConnexion + 1)

typedef enum {			       /* where does the edge lead? */
   Left,			       /* to the left */
   Right,			       /* to the right */
   Nil,				       /* nowhere */
   NonNil,			       /* somewhere */
   AnyDir                              /* any of the above */
} Direction;			       

#define minDirection ((int)Left)
#define maxDirection ((int)Nil)
#define nrDirections (maxDirection - minDirection + 1)

typedef struct {
   int arity;			       /* number of edges */
   Connexion conn;		       /* configuration of edges */
   Level l1;		       /* level of edge 1 */
   Level l2;		       /* level of edge 2 */
   Direction d1;		       /* direction of edge 1 */
   Direction d2;		       /* direction of edge 2 */
} SignatureStruct;
typedef SignatureStruct *Signature;

/* EdgeType is an algebraic integer type to represent one of the possible
   types of dependency edges, distinguished by level, direction and label. 
   Say that there are defined the levels SYN # SUBJ,OBJ,VERB and SEM #
   AGENT,PATIENT, then there would be 18 (2x3x3) different types of edges.

   EdgeTypes should only be accessed with the functions etDecodeLabel(),
   etDecodeLevel(), etDecodeDirection() and etEncode().
*/
typedef int EdgeType;

#endif

typedef struct {
   String varname;        /* print name of the variable ("X" in {X:syn})*/
   String levelname;      /* only valid for this level ("syn" in {X:syn} ) */
   Level level;       /* should replace levelname in the future */
   Pointer levelvalue;    /* pointer to level value */
   Direction dir;         /* declared direction of the edge */
} VarInfoStruct;
typedef VarInfoStruct *VarInfo;

typedef enum {
   TTTopPeek,           	       /* lexical info from top node: X^cat */
   TTBottomPeek,        	       /* lexical info from bottom node: X@cat */
   TTLabel,             	       /* label of subordination: X.label */
   TTLevel,             	       /* level of subordination: X.level */
   TTAdd,               	       /* addition */
   TTSubtract,          	       /* subtraction */
   TTMultiply,          	       /* multiplication */
   TTDivide,            	       /* division */
   TTFunction,          	       /* function: e.g. distance(X@id, X^id) */
   TTString,            	       /* string constant: 'foo' */
   TTNumber,            	       /* number constant: 0.025 */
} TermType;

struct TermStruct;
typedef struct TermStruct TermStruct, *Term;
struct TermStruct {
   TermType type;		       /* kind of term */
   union {
      struct {
	 VarInfo varinfo;	       /* pointer to variable information */
	 unsigned int pathindex;       /* index for path in table */
	 List path;		       /* path to peek */
      } peek;			       /* TopPeek and BottomPeek */
      VarInfo label;		       /* pointer to variable information */
      VarInfo level;		       /* pointer to variable information */
      struct {
	 Term op1, op2; /* operands */
      } operation;
      struct { 
	String functor;	               /* functor */
	PointerFunction *function;     /* function to call */
	List args;		       /* arguments */
      } function;
      String string;		       /* string constant */
      Number number;		       /* number constant */
      Pointer pointer;		       /* pointer */
   } data;
};

#ifndef SWIG
typedef enum {
   FTEqual,			       /* equal: 2 = 3 */
   FTGreater,			       /* greater: 3 > 2 */
   FTLess,			       /* less: 3 < 2 */
   FTGreaterEqual,		       /* greater or equal: 3 >= 2 */
   FTLessEqual,			       /* less or equal: 2 <= 3 */
   FTNotEqual,			       /* not equal: 2 != 3 */

   FTConjunction,		       /* conjunction: false & true */
   FTDisjunction,		       /* disjunction: false | true */
   FTImplication,		       /* implication: true -> false */
   FTBiimplication,		       /* biimplication: true <-> false */

   FTPredicate,			       /* predicate */
   FTNegation,			       /* negation */
   FTConstant,			       /* constant: true or false */

   FTConnexion,			       /* Connexion between two edges */
   FTDirection			       /* Direction of an edge */
} FormulaType;

struct FormulaStruct;
typedef struct FormulaStruct FormulaStruct, *Formula;
struct FormulaStruct {
  int index;		     /* equal indices => equivalent formulas */
  FormulaType type;	     /* kind of formula */
  union {
    Boolean constant;	     /* boolean value for constant */
    Formula formula;	     /* negation */
    struct {		     /* predicate */
      String functor;	     /* predicate head */
      BooleanFunction *function;	/* function to call */
      List args;	     /* predicate arguments */
    } predicate;
    struct {		     /* junction */
      Formula arg1, arg2;    /* junction arguments */
    } junction;
    struct {		     /* relation */
      Term op1, op2;	     /* operands of junction */
    } relation;
    struct {
      Connexion c;
      VarInfo var1, var2;
    } connexion;
    struct {
      Direction d;
      VarInfo var;
    } direction;
  } data;
};
#endif

typedef enum {
   VTNumber,			       /* number */
   VTString,			       /* string */
   VTAVNode,			       /* attribute & value */
   VTList,                             /* list of values */
   VTConjunction,		       /* conjunction of values */
   VTDisjunction,		       /* disjunction of values */
   VTError,                            /* Value has invalid Data */ 
   VTLexemNode,                        /* LexemNode from TermValue */
   VTGraphemNode,                      /* GraphemNode from TermValue */
} ValueType;

#ifndef SWIG
struct AVNodeStruct;
typedef struct AVNodeStruct AVNodeStruct, *AVNode;
typedef struct {
   ValueType type;		       /* kind of value */
   int index;			       /* marker for distributed disjunctions */
   union {
      Number number;		       /* number */
      String string;		       /* string */
      List list;		       /* list */
      AVNode avnode;                   /* AVNode */
      LexemNode lexemnode;             /* LexemNode */
      GraphemNode graphemnode;         /* GraphemNode */
   } data;
} ValueStruct;
typedef ValueStruct *Value;
typedef Value VTFunction();

struct AVNodeStruct {
   String attribute;		       /* attribute */
   Value value;			       /* value */
};		    	               /* node within a dag */
#endif

typedef struct {
   List vars;			       /* info about variables */
   Signature sig;		       /* what configurations matter? */
   String id;			       /* constraint identification */
   Section section;		       /* section name */
   Term penaltyTerm;		       /* penalty term */
   Number penalty;		       /* penalty */
   Formula formula;		       /* formula */
   Boolean active;		       /* use or ignore constraint */
   int no;                             /* for unique identification */
   String filename;		       /* source file */
   int lineNo;			       /* line number in source file */
   BooleanFunction *cfunction;         /* compiled constraint */
   Boolean is_context_sensitive;       /* does it use c.s. functions/preds? */
} ConstraintStruct;
typedef ConstraintStruct *Constraint;

typedef struct {
   String description;                 /* identifier */
   String word;			       /* lexem */
   Value value;			       /* tree of lexical information */
   Value *values;                      /* Array of Values accessible per index */
   List distinctive_features;          /* features that distinguish this item
                                          from its homonyms */
   regex_t re;                         /* regular expression that must be matched */
  
   String filename;		       /* source file */
   int lineNo;			       /* line number in source file */
} LexiconItemStruct;
typedef LexiconItemStruct *LexiconItem;

typedef struct {
  int from;			       /* start node */
  int to;			       /* end node */
  int no;                              /* unique identifier */
  String word;			       /* lexem */
  Number penalty;		       /* penalty factor */
  Value  info;                         /* infostring */
  int fstart;                          /* start frame */
  int fstop;                           /* end frame */

  int quotes;                          /* nesting level of quotation */
  int phraseQuotes;                    /* nesting level of phrase quotation */
  int parens;                          /* nesting level of parentheses */

  /* See inputScanLattice() for an explanation of these three values. */
}
ArcStruct;
typedef ArcStruct *Arc;

typedef struct {
   String id;			       /* identification */
   List arcs;			       /* list of arcs */

   String filename;		       /* source file */
   int lineNo;			       /* line number in source file */
} LatticeStruct;
typedef LatticeStruct *Lattice;

typedef struct {
   String id;			       /* identification */
   String lattice;                     /* id of lattice */
   List annos;			       /* list of annotations */

   String filename;		       /* source file */
   int lineNo;			       /* line number in source file */
} AnnoEntryStruct;
typedef AnnoEntryStruct *AnnoEntry;

#ifdef SWIG
%mutable;
#endif
typedef struct {
   int from;			       /* start node */
   int to;			       /* end node */
   String word;			       /* lexem */
   List specs;			       /* list of annotations */
} AnnotationStruct;
typedef AnnotationStruct *Annotation;
#ifdef SWIG
%immutable;
#endif

typedef enum {
   STTag,			       /* word tag */
   STDep			       /* dependency */
} SpecType;

#ifdef SWIG
%mutable;
#endif
typedef struct {
   SpecType type;		       /* kind of specification */
   String kind;			       /* level name or category kind */
   String name;			       /* tag name or label name */
   int position;		       /* position of modifying word */
} SpecificationStruct;
typedef SpecificationStruct *Specification;
#ifdef SWIG
%immutable;
#endif


/* Store a cdg grammar.

The type Input is basically a tuple of eight containers holding each kind
of CDG grammar elements. These are the fields constraints, levels,
lexicon, lattices, annotations, hierarchies, maps, and parameters.

Additional fields hold various related information about the grammar. In
general, all these other fields must be recomputed via inputCacheInput()
whenever a grammar element is added or removed.

The list pragmas contains a list of commands that were declared with the
#pragma directive in the input. They are executed after the entire
load operation has succeeded. Usually this List is NULL.

The Vector levelPrio holds a user-defined order between the
levels if this is to be different from the declaration order.

The field files holds a list of names of all files that contributed data
to this grammar.

The field dllHandle points to a compiled grammar that corresponds to the
loaded constraints; see compile.c for more information.

The field noOfLevels holds the number of known grammar levels.

The field maxLabelLength holds the length of the longest label
defined in the grammar (this is used for formatted output of LVs).

The field maxNrLabels holds the maximal number of labels that one level in
the grammar defines. This is only ever used within inputCollectFeatures()
to determine how many abstract types of dependency edges are possible.

The various xcdg `to date' flags are reset whenever new grammar structures
have been introduced into the current grammar. This tells the graphical
parser xcdg to update its displays. They are only ever set by xcdg.

The `current' flags serve a similar function for the cdg library itself.
They are reset whenever a grammar element has been added to the grammar and
not cached. The field attsCurrent is set when there is, or may be, an
attribute in a constraint, lexicon item or annotation that has not been
indexed by inputRegisterPath() yet.

The field hashedBCs is a five-dimensional matrix holding all known binary
constraints, ordered by their signatures. Each cell is a List of
constraints applicable to this particular configuration. If a constraint is
declared with an underspecified signature, it may appear in more than one
of the cells. The recommended way of accessing this matrix is to use
inputBCIndex():


// calculate the mutual score of LVs A and B:

   LevelValue A,B;
   int levelA = A->level->no;
   int levelB = B->level->no;
   Direction dirA = getDir(A);
   Connexion conn = getCon(A, B);
   Direction dirB = getDir(B);

   int index = inputBCIndex(levelA, levelB, dirA, conn, dirB);

   List relevantConstraints = cdgConstraintMatrix[index];

   // now apply all of these constraints to (A,B)


The field hashedUCs is used in a similar way to store all known unary
constraints ordered by their signatures. As with hashedBCs, it should only
be accessed with the corresponding index function inputUCIndex.

The fields hashedUCsize and hashedBCsize hold the number of cells in
hashedUCs and hashedBCs, respectively.

The field levelMatrixCounter counts the number of binary
constraints between each pair of two levels.

The field allAttributes holds a List of all feature paths
occurring in the grammar. This List is used to translate explicit
feature paths into index numbers.

The field is_context_sensitive notes whether any of the constraints in the
grammar is context-sensitive. (If the entire grammar is free of such
constraints, some processing steps can be omitted; cf. evalInContext(),
lock_tree().)

The fields upFeatures and downFeatures are Vectors of
BitStrings that indicate which lexical features of words are relevant
for a dependency edges of a particular type. The n-th bit in the
m-th Bitstring of upFeatures is set iff the modifiees of edges
of type m (as encoded by etEncode) distinguish between
different values for the n-th feature in allAttributes. (This
detailed information helps in building fewer LVs for the same problem
and increase parsing efficiency considerably.) The corresponding relation
holds for modifiers and the field downFeatures.

The field no_has counts all instances of the predicate has() in the
constraint formulas; this is used to determine how large a cache must be
allocated during locked evaluation (cf. eval.c).

*/


typedef struct {
  Hashtable constraints;          /* loaded constraints */
  List levels;                    /* level declarations */
  Hashtable lexicon;              /* hashtable of lexical entries */
  List lattices;                  /* loaded lattices */
  Hashtable annotations;          /* loaded annotations */
  Hashtable hierarchies;          /* loaded hierarchies */
  Hashtable maps;                 /* loaded data maps */
  List parameters;                /* list of parameters */

  List pragmas;                   /* #pragma commands in the input */

  List files;                     /* list of loaded files (type String) */
  void *dllHandle;                /* a loaded dll grammar */
  int noOfLevels;                 /* number of loaded levels */
  int maxLabelLength;             /* length of longest CDG label */
  int maxNrLabels;                /* maximal number of labels in one level */

  Boolean xcdgwordgraphsToDate;       /* Flags indicating whether the */
  Boolean xcdglevelsToDate;       /* fields changed. Used by XCDG */
  Boolean xcdgconstraintsToDate;      
  Boolean xcdglexiconToDate;
  Boolean xcdghierarchiesToDate;
  Boolean xcdgannotationsToDate;

  Boolean annosCurrent;
  Boolean attsCurrent;
  Boolean constraintsCurrent;
  Boolean levelsCurrent;
  Boolean lexiconCurrent;
  
  List *hashedUCs;                /* unary constraints by signature */
  List *hashedBCs;                /* binary constraints by signature */
  int hashedUCsize;               /* number of cells in hashedUCs */
  int hashedBCsize;               /* number of cells in hashedBCs */
  Vector levelMatrixCounter;      /* counts # of constraints between levels */
  List allAttributes;             /* Used to flatten LexiconItems and 
				     Constraints which are trees */
  Boolean is_context_sensitive;   /* does it contain c.s. constraints? */

  Vector upFeatures;              /* contains bitstrings that indicate which */
  Vector downFeatures;            /* edge types look at which features  */   
  int no_has;                     /* counts instances of has() */
  
} InputStruct;
typedef InputStruct *Input;

#ifndef SWIG
typedef struct {
  String id;			       /* sort identifikation */
  List subsorts;		       /* list of subsorts */
  int depth;                           /* # of levels below root */
} SortStruct;
typedef SortStruct *Sort;

typedef struct {
  String type;			       /* type identifikation */
  Boolean subsumesFlag;                /* TRUE if type subsumes types */
                                       /* FALSE if types subsume type */
  List types;                          /* list of types */
} SubsumptionStruct;
typedef SubsumptionStruct *Subsumption;
#endif

typedef struct {
  String id;			       /* identifikation */
  Hashtable types;                     /* hash table of all defined types */
  /* array of int
   * distance[2, 4] == 3 means there is a path of length 3 
   * from node 2 to node 4
   * distance[...] == 0 means there is NO such path
   */
  Array distance;  

  String filename;		       /* source file */
  int lineNo;			       /* line number in source file */
} HierarchyStruct;
typedef HierarchyStruct *Hierarchy;

#ifndef SWIG
typedef struct {
   Value v;
   int decisions[10];
} ExpansionStruct;
typedef ExpansionStruct *Expansion;

/*
 * How to integrate acoustic scores with CDG scores
 */
typedef enum {
  vmAcOff,
  vmAcLinear,
  vmAcScaled,
} VmAcousticsType;
#endif

/* ---------------------------------------------------------------------- */
extern String inputFeatureOfLI(LexiconItem le, String feature);
extern List inputGetSortedLatticeNames(Input input);
extern String getLevelSort(void);
extern void freeAnnoEntry(AnnoEntry a);
extern String formatConstraint(Constraint c);
extern Constraint findConstraint(String id);
extern void inputSpecSetName(Specification spec, String s);

#ifndef SWIG

/*
 * These functions are called on program start and program end
 * respectively.
 */
extern void inputInitialize(void);
extern void inputFinalize(void);
extern void inputCallback(String name, String *var);
extern String inputRegisterFilename( Input input, String s );

extern Boolean trivialise(Lattice lat);
extern Boolean graphContainsWord(Lattice g, String w);
extern void showFeatures(Level l);

extern Hierarchy inputSortTreeToHierarchy( char *id, Sort s );
extern Hierarchy inputSubsumptionListToHierarchy( char *id, List l );
extern Hashtable inputHashFromList(List items);

extern Boolean inputLoad(String filename);
extern Input inputNewInput(void);
extern void inputCacheLexicon(Input input);
extern void inputCacheParses(Input input, Hashtable ht);
extern void inputFreeInput(Input i);
extern void inputCacheInput(Input i);

extern void inputFreePath(List path);
extern int inputLookupPath(List path, Input input);
extern int inputRegisterPath(List path, Input input);

extern void mergeInput(Input goal, Input source);

extern void printAnnoEntry(unsigned long mode, AnnoEntry a);
extern void printConstraint(unsigned long mode, Constraint c);
extern void printFormula(unsigned long mode, Formula f, FormulaType enclosing, int indent);
extern String formatFormula(Formula f, FormulaType enclosing, int indent);
extern void printHierarchy(unsigned long mode, Hierarchy h);
extern void printLattice(unsigned long mode, Lattice l);
extern void printLevel(unsigned long mode, Level l);
extern void printLexiconItem(unsigned long mode, LexiconItem l);
extern void printPath(unsigned long mode, List p);
extern String formatPath(List p);
extern String formatTerm(Term t, Boolean termContext);
extern void printTerm(unsigned long mode, Term t, Boolean termContext);
extern void printValue(unsigned long mode, Value v, int indent);
extern String formatLexiconItem(LexiconItem l);
extern String formatValue(Value v, int indent);
extern String formatIdentifier(String s);
extern String formatTime(unsigned long time);

extern String sprintPath(List p);
extern String sprintValue(Value v);

extern List scanPath(String s);

extern void freeAnnoEntry(AnnoEntry a);
extern void freeConstraint(Constraint c);
extern void freeFormula(Formula f);
extern void freeHierarchy(Hierarchy h);
extern void freeLattice(Lattice l);
extern void freeLevel(Level l);
extern void freeLevelProp(Property l);
extern void freeLexiconItem(LexiconItem l);
extern void freeParameter(Constraint c);
extern void freeSection(Section s);
extern void freeSort(Sort s);
extern void freeTerm(Term t);
extern void freeValue(Value v);

extern int inputUCIndex(int level, int dir);
extern int inputBCIndex(int levela, int levelb, int dira, int conn, int dirb);

extern Connexion inverseConnexion(Connexion x);
extern Boolean subsumesConnexion(Connexion a, Connexion b);
extern void printConnexion(unsigned long mode, Connexion x);
extern String formatConnexion(Connexion x);
extern Boolean subsumesDirection(Direction a, Direction b);
extern void printDirection(unsigned long mode, Direction x);
extern String formatDirection(Direction x);

extern Boolean matchLevel(Level a, Level b);
extern void invertSignature(Signature x);
extern String formatSignature(Signature x);

extern Boolean setLevelSort(String levelIds[]);

extern String getLevelProperty(Level level, String name);
extern Boolean latticeBranches(Lattice lat);
extern Boolean latticeHasInfo(Lattice lat);
extern Lattice inputCloneLattice(Lattice lat);
extern Number inputGetWeight(Constraint c);
extern Arc inputLatticePrevArc(Lattice lat, Arc a);

extern void inputSetWeight(Constraint c, Number x, Boolean absolute);

extern Value newValueAV(String attribute, Value value);
extern Value ErrorValue;

extern int leCompare(Input input, LexiconItem a, LexiconItem b, Boolean verbose);
extern Boolean leCompareByName(LexiconItem a, LexiconItem b);
extern LexiconItem inputCloneLI(LexiconItem le);
extern LexiconItem newLexiconItem(String word, Value v, String filename, int lineNo);

extern Boolean inputLexiconQuery(String word);
extern List inputLexiconGet(String word);
extern void inputLexiconPut(LexiconItem le, Hashtable lexicon);
extern List inputTemplatesGet(String word, List categories);

extern Number vmArcPenalty(Number score, Number frame_start, Number frame_end);  
#if 0
extern String staticBasename(String fq_filename);
#endif
extern String vm2String(String parsestring);
extern Arc newArc(int from, int to, int no, String word, Number penalty, Value info, int fstart, int fstop);

/*
 * adjust the acoustic score so that it can be used as a CDG score
 * (see parser.y & functions.c)
 */
extern void vmAdjustAcousticScore(Lattice lat);

extern void traverseItem(int pass, LexiconItem lex, Value v, List p, Input i);
#endif

extern Annotation newAnnotation(int from, int to, String word);
extern Specification newSpecification(SpecType type, String kind,
                                      String name, int position);
extern void inputSpecificationAdd(Annotation a, Specification s);

extern Level inputGetMainlevel(Input i);
extern Level findLevel(String levelId);
extern List findAnnotations(Lattice lat, Boolean autoload);
extern AnnoEntry findAnnoByName(String id, Boolean autoload);
extern AnnoEntry findAnnoForLattice(Lattice lat, Boolean autoload);
extern Lattice findLattice(String id);
extern Lattice findLatticeInInput(Input ip, String id);
extern Section findSection(String id);
extern Hierarchy findHierarchy(String id);

/* Are A and B distinguished by one of FEATURES? */
extern Boolean inputCompareLeByAtts(LexiconItem a, LexiconItem b, BitString features);
extern Boolean inputCompareLeByAtt(LexiconItem a, LexiconItem b, List feature, Boolean unify);


/* ---------------------------------------------------------------------- */

/* NULL if no lattice has been created, or a pointer to the last one */
extern Lattice inputMostRecentlyCreatedLattice;

/* default section, used if no section is  */
extern Section inputDefaultSection;

/*
 * This variable holds the collected info for the current grammar.
 */
extern Input inputCurrentGrammar;

/*
 * List of all sections (type Section) of the system. This list cannot
 * be a member of the Input structure because that would make the
 * mergeInput() function too complicated.
 */
extern List inputSections;

/*
 * UNIX including path to preprocess m4 files
 */
extern String inputPreprocessor;

/*
 * suffix of files to be preprosessed
 */
extern String inputPrepSuffix;

/*
 * List of categories to record in annotations.
 */
extern List inputAnnoCats;

/*
 * currently used locale as returned by setlocale
 */
extern String inputLocale;

/*
 * name of hierarchy used for comparing features
 */
extern String inputFeatureHierarchy;

/* how to integrate acoustic scores with CDG scores */
extern VmAcousticsType vmAcoustics;

/* in which order should two Specifications written into an annotation? */
extern Boolean inputSpecCompare(Specification a, Specification b);

/* detect umlaut translitarations and replace them with the real thing */
extern String inputEncodeUmlauts(String word);

/* do two paths read the same? */
extern Boolean comparePath(List a, List b);

extern Boolean inputReadInitFile(String initFileName);

extern int etEncode(Level level, String label, Direction dir);
extern Level etDecodeLevel(EdgeType et); 
extern String etDecodeLabel(EdgeType et);
extern Direction etDecodeDirection(EdgeType et);
extern int etMax(void);

/* are these two features values similar enough to count as a `hit'
   when verifying a parse result? */
extern Boolean inputCompareFeatures(String a, String b, Hierarchy h);


/* Does this Arc carry a word that might indicate a sentence boundary? */
extern Boolean inputSentenceBoundary(Arc arc);

extern Boolean inputALLUPPER(String x);

extern Boolean inputCompareValue(Value v1, Value v2);

extern Boolean inputSubsumes(Hierarchy h, String up, String down);
extern void inputInsertLI(Input input, LexiconItem le);
extern void inputScanLattice(Lattice lat);
extern void printTypeError(unsigned long mode, Value v, ValueType vt);

extern void inputCacheConstraints(Input input);
extern Arc inputFindArc(Lattice lat, int no);

/* ---------------------------------------------------------------------- */
/** @} */
#endif /* don't insert anything after this #endif */
