/* 
 * Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without 
 * modifications, as long as this notice is preserved.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: timer.c,v 1.28 2004/09/27 17:07:06 micha Exp $ 
 */

/* -------------------------------------------------------------------------
 * @addtogroup Timer Timer - timekeeping functions
 * @author Michael Schulz
 * @date somewhen in 1998
 *
 *
 * This module exports functions for measuring the time spent in various
 * parts of the application. The time measured is always user time + kernel
 * time.
 * @{
 */

/* ---------------------------------------------------------------------- */
#include <config.h>
#include <sys/time.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/resource.h>
#include <signal.h>
#include <blah.h>
#include <limits.h>
#include "hook.h"
#include "timer.h"

/* -- VARIABLES --------------------------------------------------------- */

/* -------------------------------------------------------------------------
 * This variable \b timerExpired= is a flag to be set after a
 * user-defined timespan has elapsed.
 */
Boolean timerExpired;

static double timerClockTicks = -1; /**< initialized correctly in 
				      * timerInitialize */

/* ----------------------------------------------------------------------------
 *  The fast timer -- faster, but expires over 2470 sec
 */

/* ----------------------------------------------------------------------------
 * return clocks elapse since program start (user + system)
 * 
 * this function returns the time used so far by the process, as computed
 * by the system function \b clock()=.
 */
TimerFast timerStartFast(void)
{
  return clock();
}

/* ----------------------------------------------------------------------------
 *	return milliseconds elapsed since last timerStartFast
 * 
 * this function returns the difference between the time used so far and
 * a time determined earlier by \b timerstartfast() in milliseconds.
 */
unsigned long timerElapsedFast(TimerFast timer) 
{
  return (unsigned long)((double)(clock() - timer) / timerClockTicks);
}

/* ----------------------------------------------------------------------------
   The standard timer -- slower, but only expires over 2,470,000 sec
*/

/* ----------------------------------------------------------------------------
 * Store number of cycles elapsed since program start (user + system)
   in timer. 

   * This function performs a call to \b getrusage() and fills the
   * corresponding fields in the \b Timer structure.
 */
void timerStart(Timer timer)
{
  struct rusage t;

  if (timer == NULL) {
    return;
  }
  getrusage(RUSAGE_SELF, &t);
  timer->tv_sec = t.ru_utime.tv_sec + t.ru_stime.tv_sec;
  timer->tv_usec = t.ru_utime.tv_usec + t.ru_stime.tv_usec;
}

/* ----------------------------------------------------------------------------
 *	allocates memory for the timer and calls timerStart
 */
Timer timerNew(void)
{
  Timer timer = (Timer) memMalloc(sizeof (TimerStruct));

  timerStart(timer);
  return timer;
}

/* ----------------------------------------------------------------------------
 *	frees the timer memory
 * 
 * this function deallocates a \b timer.
 */
void timerFree(Timer timer)
{
  if (timer == NULL) {
    return;
  }
  memFree(timer);
}

/* ----------------------------------------------------------------------------
 * Return milliseconds elapsed since last timerStart() on this timer.
 * 
 * this function returns the difference between the time used so far and
 * a time determined earlier by \b timerstart() in milliseconds.
 */
unsigned long timerElapsed(Timer timer)
{
  struct rusage t;
  long s, u;

  getrusage(RUSAGE_SELF, &t);
  s = t.ru_utime.tv_sec + t.ru_stime.tv_sec - timer->tv_sec;
  u = t.ru_utime.tv_usec + t.ru_stime.tv_usec - timer->tv_usec;

  if (u < 0) {
    u += 1000000;
    s--;
  }
  if ((s < LONG_MAX / 1000) || ((s == LONG_MAX / 1000) && (LONG_MAX - s * 1000 >= u / 1000))) {
    return (unsigned long)(s) * 1000 + u / 1000;
  } else {
    cdgPrintf(CDG_ERROR, "ERROR: the timer has overflow\n");
    return 0;
  }
}

/* ----------------------------------------------------------------------------
 *  Signal handler for when timer expires.
 * 
 * this function sets \b timerexpired.
 */
static void expire(int n)
{
   timerExpired = TRUE;
}

/* ----------------------------------------------------------------------------
 *  Set an alarm-clock that will set the flag timerExpired after ``limit'' 
 *  milliseconds.
 * 
 * this function sets a time limit for the completion of a task. after
 * \b limit milliseconds the variable \b timerexpired is
 * asynchronously set to~1 by \b expire(). this can be used as follows:
 * 
 * <i>
 * \/\* build lvs for one second \*\/
 * 
 * void fastbuildvalues(
 *   constraintnet net,
 *   leveldecl level,
 *   ConstraintNode cn,
 *   List modifiers) {
 * 
 *  LexemNode modifiee;
 * *   LevelValue lv;
 * 
 *  int i;
 *  List l;
 *  String label;
 * 
 *  timerSetAlarm(1000);
 * 
 *  for (i = 0; i < vectorSize(net->lexemgraph->nodes); i++) {
 *    modifiee = (LexemNode) vectorElement(net->lexemgraph->nodes, i)
 * 
 *      for (l = level->labels; l != NULL; l = l->next) {
 *        label = (String) l->item;
 * 
 *        lv = lvNew(modifiers, level, label, modifiee);
 *        vectorAddElement(node->values, lv);
 *      }
 * 
 *    if(timerExpired) { break; }
 * 
 *  }
 * 
 * }
 * </i>
 */
void timerSetAlarm(long unsigned int limit)
{
  struct itimerval value;
  ldiv_t l;
#ifdef NO_SIG_PROF
  return;
#endif
  timerExpired = FALSE;

  value.it_interval.tv_sec = 0;
  value.it_interval.tv_usec = 0;

  l = ldiv(limit, 1000);
  value.it_value.tv_sec = l.quot;
  value.it_value.tv_usec = l.rem * 1000;

  signal(SIGPROF, expire);
  setitimer(ITIMER_PROF, &value, (struct itimerval *)NULL);
}

/* ----------------------------------------------------------------------------
 * stop the alarm clock
 * 
 * This function cancels the alarm set by \b timerSetAlarm().
 */
void timerStopAlarm(void)
{
#ifdef NO_SIG_PROF
  return;
#endif
   signal(SIGPROF, SIG_IGN);
  timerExpired = FALSE;
}

/* ----------------------------------------------------------------------------
 * timerInitialize
 * 
 * this function stops any alarms pending and initializes the variable
 * \b timerclockticks.
 */
void timerInitialize(void) 
{
  timerStopAlarm();

  /* get system clock ticks */
  timerClockTicks = CLOCKS_PER_SEC / 1000.0;
}
