/*
 * File: vars.c
 * Author: Horia F. Pop
 * Purpose: variables encapsulation
 * Birth: February 9, 2000
 *
 * $Id: variables.c,v 1.18 2004/09/27 17:07:06 micha Exp $
 */

/* -- INCLUDES ---------------------------------------------------------- */
#include <config.h>

#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "variables.h"
#include "cdg.h"
#include "hook.h"

/* -- MACROS ------------------------------------------------------------ */

/* -- TYPE DEFINITIONS -------------------------------------------------- */

struct VariableStruct {
  String name;
  int type;
  Pointer ptr;
  BooleanFunction *funcValidate;
  VoidFunction *funcCallback;
  VoidFunction *funcUserSet;
};

/* -- VARIABLES --------------------------------------------------------- */

static List varList = NULL;

/* -- FUNCTIONS --------------------------------------------------------- */

/* ----------------------------------------------------------------------
 * varIteratorNew
 */
void varIteratorNew(VarIterator *vip)
{
  *vip=(VarIterator)varList;
}

/* ----------------------------------------------------------------------
 * varIteratorNextName
 */
String varIteratorNextName(VarIterator *vip)
{
  String s=NULL;
  List l=(List)*vip;
  
  if (l) {
    s=((Variable)listElement(l))->name;
    *vip=(VarIterator)listNext(l);
  }

  return s;
}

/* ----------------------------------------------------------------------
 * varIteratorFree
 */
void varIteratorFree(VarIterator *vip)
{
  *vip=NULL;
}

/* ----------------------------------------------------------------------
 * Allocate a Variable and sets its fields
 */
Variable varAllocate(String name, int type, Pointer ptr, 
		     BooleanFunction *funcValidate,
		     VoidFunction *funcCallback,
		     VoidFunction *funcUserSet)
{
  Variable v = (Variable) memMalloc(sizeof(VariableStruct));

  v->name = strRegister(name);
  v->type = type;
  v->funcValidate = funcValidate;
  v->funcCallback = funcCallback;
  v->funcUserSet = funcUserSet;
  v->ptr = ptr;

  return v;
}

/* ----------------------------------------------------------------------
 * return the variable name
 */
String varName(Variable v)
{
  return v->name;
}

/* ----------------------------------------------------------------------
 * return the variable type
 */
int varType(Variable v)
{
  return v->type;
}

/* ----------------------------------------------------------------------
 * return the variable pointer
 */
Pointer varPtr(Variable v)
{
  return v->ptr;
}

/* ----------------------------------------------------------------------
 * return the variable validation function
 */
BooleanFunction * varValidate(Variable v)
{
  return v->funcValidate;
}

/* ----------------------------------------------------------------------
 * return the variable callback function
 */
VoidFunction * varCallback(Variable v)
{
  return v->funcCallback;
}

/* ----------------------------------------------------------------------
 * return the variable userset function
 */
VoidFunction * varUserSet(Variable v)
{
  return v->funcUserSet;
}

/* ----------------------------------------------------------------------
 * Registers a variable. The arguments are the name, type, address, and 
 * address of validation function, callback function and user set function
 * - funcValidate(type) if is a simple type, or (Pointer) otherwise
 * - funcCallback()
 * - funcUserSet(Pointer, Pointer)
 */
Boolean varRegister(String name, int type, Pointer ptr, 
		    BooleanFunction *funcValidate,
		    VoidFunction *funcCallback,
		    VoidFunction *funcUserSet)
{
  List l;
  Variable v;

  for (l=varList; l; l=listNext(l)) {
    Variable i = (Variable) listElement(l);
    if (strcmp(i->name, name) == 0) {
      return FALSE;
    }
  }

  v = varAllocate(name, type, ptr, funcValidate, funcCallback, funcUserSet);

  varList = listAppendElement(varList, v);
  return TRUE;
}

/* ----------------------------------------------------------------------
 * Sets the value of a previously registered variable.
 * The second argument must be of the type associated to the given name.
 * Returns TRUE if the named variable does exist and if the value is valid.
 * It is SetVar's responsibility to set the value; the validating function
 * only validates it.
 */
Boolean varSet(String name, ...)
{
  List l;
  Variable v = NULL;
  Boolean cont = TRUE;
  va_list ap;

  for (l=varList; l && cont; l=listNext(l)) {
    v = (Variable) listElement(l);
    if (strcmp(v->name, name) == 0) {
      cont = FALSE;
    }
  }
  if (cont) {
    return FALSE;
  }

  va_start(ap, name);
  switch (v->type) {
    case VAR_INT: {
      int i = va_arg(ap, int);
      if (v->funcValidate == NULL || v->funcValidate(i)) {
	*((int *)v->ptr) = i;
      } else {
	return FALSE;
      }
      break;
    }
    case VAR_FLOAT: {
      double f = va_arg(ap, double);
      if (v->funcValidate == NULL || v->funcValidate(f)) {
	*((double *)v->ptr) = f;
      } else {
	return FALSE;
      }
      break;
    }
    case VAR_BOOL: {
      Boolean b = (Boolean)va_arg(ap, int);
      if (v->funcValidate == NULL || v->funcValidate(b)) {
	*((Boolean *)v->ptr) = b;
      } else {
	return FALSE;
      }
      break;
    }
    case VAR_STRING: {
      String s = va_arg(ap, String);
      if (v->funcValidate == NULL || v->funcValidate(s)) {
	*((String *)v->ptr) = strRegister(s);
      } else {
	return FALSE;
      }
      break;
    }
    case VAR_USER: {
      Pointer p = va_arg(ap, Pointer);
      if ((v->funcValidate == NULL || v->funcValidate(p)) &&
          (v->funcUserSet)) {
	v->funcUserSet(v->ptr, p);
      } else {
        return FALSE;
      }
      break;
    }
    default:
      return FALSE;
  }

  va_end(ap);
  if (v->funcCallback) {
    v->funcCallback();
  }
  return TRUE;
}

/* ----------------------------------------------------------------------
 * Gets the value of a previously registered variable.
 * The second argument must be an address to the type associated with the
 * given name. Returns TRUE if the named variable does exist. On the exit 
 * of this function, the second argument will hold the value of that 
 * variable.
 */
Boolean varGet(String name, Pointer ptr)
{
  List l;
  Boolean cont = TRUE;
  Variable v = NULL;

  for (l=varList; l && cont; l=listNext(l)) {
    v = (Variable) listElement(l);
    if (strcmp(v->name, name) == 0) {
      cont = FALSE;
    }
  }

  if (cont) {
    return FALSE;
  }

  switch (v->type) {
    case VAR_INT: 
      *((int *)ptr) = *((int *)v->ptr);
      break;
    case VAR_FLOAT:
      *((float *)ptr) = *((float *)v->ptr);
      break;
    case VAR_BOOL:
      *((Boolean *)ptr) = *((Boolean *)v->ptr);
      break;
    case VAR_STRING:
      *((String *)ptr) = *((String *)v->ptr);
      break;
    case VAR_USER:
      if (v->funcUserSet) {
        v->funcUserSet(ptr, v->ptr);
      } else {
	return FALSE;
      }
      break;
    default:
      return FALSE;
  }
  return TRUE;
}

/* ----------------------------------------------------------------------
 * Returns a list of Strings holding the names of all variables.
 */
List varReturnListOfNames()
{
  List l, rez = NULL;

  for (l=varList; l; l=listNext(l)) {
    String s = ((Variable) listElement(l))->name;
    rez = listAppendElement(rez, s);
  }
  return rez;
}

/* ----------------------------------------------------------------------
 * Returns the type of the given variable, or VAR_NONE otherwise.
 */
int varReturnType(String name)
{
  List l;

  for (l=varList; l; l=listNext(l)) {
    Variable i = (Variable) listElement(l);
    if (strcmp(i->name, name) == 0) {
      return i->type;
    }
  }
  return VAR_NONE;
}

/* ----------------------------------------------------------------------
 * Deletes the list of registered variables
 */
void varFinalize()
{
  List l;

  for (l = varList; l; l=listNext(l)) {
    Variable v = (Variable) listElement(l);
    cdgFreeString(v->name);
    if (v->type == VAR_STRING) {
      cdgFreeString(*((String *)v->ptr));
    }
    memFree(v);
  }
  listDelete(l);
  varList = NULL;
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
