#!/usr/bin/perl
# cdg/utils/deutsch-tagger.pl
#
# Wrapper for a POS tagger for use by cdgp. Different third-party
# taggers can be employed, and special heuristics are superimposed on
# their results to correct some common serious assignment errors.
use warnings;
use strict;
use IO::Handle;

sub complain {
  my $m = shift;
  if ($m !~ /\n$/) {
    $m .= "\n";
  }
  print STDERR "WARNING: $m";
}

# raise score of @pos and lower previous best choice
sub prefer {
  my($word,@pos) = @_;
  complain "`$$word{word}' is probably @pos, not $$word{best}.\n";

  $$word{scores}{$$word{best}} /= 2;
  for my $p (@pos) {
    $$word{scores}{$p} = 1;
    $$word{best} = $p;
  }

}

# raise score of @pos to the of the best choice
sub alternate {
  my($word,@pos) = @_;
  complain "`$$word{word}' could also be @pos, not $$word{best}.\n";

  for my $p (@pos) {
    $$word{scores}{$p} = 1;
    $$word{best} = $p;
  }

}

# neutralize tagger scores for this word
sub equalize {
  my $word = shift;
  complain "`$$word{word}' is probably detected wrong.\n";
  for my $x (keys %{$$word{scores}}) {
    $$word{scores}{$x} = 1;
  }
  $$word{best} = '';

}

sub disprefer_finite_verb {
  my $x = shift;

  if ($$x{best} eq 'VVFIN' &&
      $$x{word} =~ /ten?$/) {
    prefer $x, 'ADJA', 'VVINF';
  } elsif ($$x{word} =~ /e[lr]?n$/) {
    (my $p = $$x{best}) =~ s/FIN/PP/;
    (my $i = $$x{best}) =~ s/FIN/INF/;
    prefer $x, $p, $i;
  } elsif ($$x{word} =~ /[nt]$/) {
    (my $p = $$x{best}) =~ s/FIN/PP/;
    prefer $x, $p;
  }
}

my %pronouns =
  (
   ich => '1sg',
   mir => '1sg',
   mich => '1sg',
   du => '2sg',
   dir => '2sg',
   dich => '2sg',
   er => '3sg',
   sie => '3sg',
   es => '3sg',
   wir => '1pl',
   uns => '1pl',
   ihr => '2pl',
   euch => '2pl'
  );
sub pernum {
  my $w = lc shift;
  if (exists $pronouns{$w}) {
    return $pronouns{$w};
  }
  return undef;
}


my %APZR;
for( qw (
       an:entlang am:entlang an:vorbei am:vorbei auf:herab auf:hin
       auf:zu aus:heraus aus:hinaus hinter:her hinter:hervor neben:her
       in:hinein um:herum um:willen unter:durch von:an von:aus von:her
       von:herab von:wegen vom:an vom:aus vom:her vom:herab vom:wegen
       vor:her zu:hin zur:hin ber:hin ber:hinaus ber:hinweg
             bers:hin bers:hinaus bers:hinweg ) ) {
  $APZR{$_}++;
}



# Use various patterns to remove obvious errors in the POS
# assignments
sub postprocess {
  my $x = shift;
  my @words;
  my @phrases;

  # parse lines
  my $i = 0;
  for (@$x) {
    unless(s/^(\S+)\s+//) {
      push @words, { word => '' };
      next;
    }

    my $w = $1;
    my %s;
    my $record = 0;
    my $best = '';
    while (s/(\S+)\s+([0-9\.e+-]+)\s*//) {
      my ($pos,$x) = ($1,$2);
      $s{$pos} = $x;
      if ($x>$record) {
        $record = $x;
        $best = $pos;
      }
    }
    push @words, { nr => $i++, word => $w, scores => \%s, best => $best };
  }

  # cut into phrases
  my @p;
  my $j;
  for ($j = 0; $j <= $#words; $j++) {
    my $prev = $j ? $words[$j-1] : undef;
    my $this = $words[$j];
    my $next = $j < $#words ? $words[$j+1] : undef;

    # ignore quotation marks
    next if $$this{word} =~ /^[\'\"]$/;

    if ($$this{word} =~ /^([,.:;\-!?]|--+|\z)$/) {
      if (defined $prev && $$prev{best} eq 'ADJA' &&
          defined $next && $$next{best} eq 'ADJA') {
        push @p, $this;
      } else {
        push @phrases, [@p] if @p;
        @p=();
      }
    } else {
      push @p, $this;
    }
  }
  push @phrases, [@p] if(@p);

  # Apply error patterns

  # `noch' is ambiguous between KON and ADV, but it is certainly KON
  # when `weder' precedes, no matter how far back.
  my $weder = 0;
  for my $w (@words) {
    if ($$w{word} =~ /^weder$/i) {
      $weder = 1;
    }
    if ($weder &&
        $$w{word} =~ /^noch$/i) {
      $weder = 0;
      if ($$w{best} ne 'KON') {
        complain "Conjunction `weder' followed by `noch'/ADV\n";
        prefer $w, 'KON';
      }
    }
  }

  # Imperatives do not occur in direct questions.
  my $question = 0;
  for my $w (reverse @words) {
    if ($$w{word} eq '?') {
      $question = 1;
    }
    if ($$w{word} =~ /^[!\.:;]$/) {
      $question = 0;
    }
    if ($question && $$w{best} =~ /IMP/) {
      complain "Imperative `$$w{best}' followed by question mark!\n";
      (my $fin = $$w{best}) =~ s/IMP/FIN/;
      prefer $w, $fin;
    }
  }

  for my $p (@phrases) {

    my @w = @$p;

    # Words like `mich' are ambiguous between a reflexive and a
    # non-reflexive reading. There is no syntactic way of distinguishing
    # them; but if the corresponding personal pronoun is in the
    # vicinity, the reflexive form is much more likely.
    for (0..$#w) {
      if ($w[$_]{word} =~ /^(mich|mir|dich|dir|uns|euch)$/i) {

        # after imperatives, `dich'/`dir' are invariably right.
        if($_ > 0 &&
           $w[$_]{word} =~ /^[dD]/ &&
           $w[$_-1]{best} =~ /IMP/) {
          if($w[$_]{best} eq 'PPER') {
            complain "Personal `$w[$_]{word}' follows imperative!";
            prefer $w[$_], 'PRF';
          }
          next;
        }

        # find previous personal pronoun
        my $this = pernum($w[$_]{word});
        my $i = $_-1;
        while ($i >= 0) {
          my $that = pernum($w[$i]{word});
          if ($that && $that ne $this &&
              $w[$_]{best} eq 'PRF') {
            complain "Reflexive `$w[$_]{word}' near `$w[$i]{word}!'";
            prefer $w[$_], 'PPER';
            last;
          }
          if ($that && $that eq $this &&
              $w[$_]{best} eq 'PPER') {
            complain "Personal `$w[$_]{word}' near `$w[$i]{word}!'";
            prefer $w[$_], 'PRF';
            last;
          }
          $i--;
        }
      }
    }

    # Imperatives do not follow most other words.
    for (1..$#w) {
      if ($w[$_]{best} =~ /IMP/ &&
          $w[$_-1]{best} !~ /^(KON|ITJ|PTKANT|ADV|\$\()$/) {
        complain "Imperative following `$w[$_-1]{word}'!";
        (my $fin = $w[$_]{best}) =~ s/IMP/FIN/;
        prefer $w[$_], $fin;
      }
    }

    # PTKZU cannot precede a finite form.
    for (0..$#w-1) {
      if ($w[$_]{word} =~ /^zu$/i &&
          $w[$_+1]{best} =~ /FIN$/i) {
        complain "Particle `zu' followed by finite verb!";
        (my $inf = $w[$_+1]{best}) =~ s/FIN/INF/;
        prefer $w[$_+1], $inf;
      }
    }

    # `zu' in front of an active participle is almost certainly
    # PTKZU, not PTKA: "der zu zahlende Betrag"
    for (0..$#w-1) {
      if ($w[$_]{word} =~ /^zu$/i &&
          $w[$_]{best} ne 'PTKZU' &&
          $w[$_+1]{best} =~ /ADJA/ &&
          $w[$_+1]{word} =~ /e[rl]?nde[nmsr]?$/) {
        complain "Adjectival `zu' in front of active participle!";
        prefer $w[$_], 'PTKZU';
      }
    }

    # `am' in front of `-sten' is akmost certainly PTKA.
    for (0..$#w-1) {
      if ($w[$_]{word} =~ /^am$/i &&
          $w[$_]{best} ne 'PTKA' &&
          $w[$_+1]{word} =~ /[s]ten$/) {
        complain "Preposition `am' in front of superlative!";
        prefer $w[$_], 'PTKA';
      }
    }

    # `wie' within a clause is probably KOKOM, not PWAV.
    for (1..$#w-1) {
      if ($w[$_]{word} =~ /^wie$/i &&
          $w[$_]{best} ne 'KOKOM' &&
          $w[$_+1]{best} !~ /ADJ|ADV|^VV/ &&
          $w[$_-1]{best} !~ /KON|\$\(/) {
        complain "wie/PWAV within a clause!";
        prefer $w[$_], 'KOKOM';
      }
    }

    # Some verb forms look like adjectives. But if they are introduced
    # by a subordinating conjunction, they are almost certainly not
    # adjectives.
    if ($w[0]{best} eq 'KOUS' &&
        $w[-1]{best} eq 'ADJD' &&
        exists $w[-1]{scores}{VVFIN}) {
      complain "Subclause with `$w[0]{word}' ends with deverbal adjective!\n";
      prefer $w[-1], 'VVFIN';
    }

    # `bis', `whrend' etc are ambiguous between KOUS and APPR.
    # If an infinitive ends the subclause, APPR is more likely;
    # if a finite verb ends it, KOUS is more likely.
    if ($w[0]{word} =~ /^(bis|seit|statt|whrend|so(weit|lange|viel|wenig))$/i) {
      if ($w[0]{best} eq 'KOUS' &&
          $w[-1]{best} =~ /INF/) {
        complain "Subclause with `$w[0]{word}' ends with infinitive!\n";
        prefer $w[0], 'APPR';
      } elsif ($w[0]{best} =~ /APPR|ADV/ &&
               $w[-1]{best} =~ /FIN/) {
        complain "PP with `$w[0]{word}' ends with finite verb!\n";
        prefer $w[0], 'KOUS';
      }
    }

    # `anstatt' etc. are ambiguous between KOUI and APPR. But KOUI is
    # unlikely if there is no `zu' to be found.
    if ($w[0]{word} =~ /^(um|ohne|statt|anstatt|ausser|auer)$/i) {
      if ($w[0]{best} eq 'KOUI' && $w[1]{word} !~ /^da(|ss)$/) {
        my $zu = undef;
        for (0..$#w) {
          if ($w[$_]{word} =~ /zu/) {
            $zu = $w[$_];
            last;
          }
        }
        if(!defined $zu) {
          complain "`$w[0]{word}' has no `zu'!\n";
          prefer $w[0], 'APPR';
        }
      }
    }

    # `mit', `ab' and others are ambiguous as APPR/PTKVZ. The category
    # PTKVZ is particularly hard to get right. But if a verb follows,
    # it is almost certainly not a preposition.
    for (0..$#w-1) {
      if ($w[$_]{best} =~ /^APPR/ &&
          $w[$_+1]{best} =~ /^V/) {
        complain "Suspicious preposition followed by verb!";
        equalize $w[$_];
      }
    }

    # initial `doch', `denn', `aber' are virtually always KON, not ADV.
    if ($w[0]{word} =~ /^(denn|aber|doch|jedoch)$/i &&
        $w[0]{best} eq 'ADV') {
      complain "Adverb `$w[0]{word}' starts phrase\n";
      prefer $w[0], 'KON';
    }

    # The sequence `ber'+CARD is always tagged as APPR, but it is
    # much more often an ADV. The same goes for some other
    # prepositions/adverbs.
    for (0..$#w-1) {
      if ($w[$_]{word} =~ /^(unter|ber|zwischen)$/i &&
          $w[$_+1]{best} =~ /CARD/) {
        complain "Suspicious preposition followed by number!";
        prefer $w[$_], 'ADV';
      }
    }

    # Successive articles are nearly always wrong.
  WORD:
    for (0..$#w-2) {
      my($l,$r) = ($w[$_], $w[$_+1]);
      if ($$l{best} eq 'ART' &&
          $$r{best} eq 'ART') {

        # There are really only two cases where this is legit:
        #
        # 1. `der ein oder andere'.
        #
        if ($$l{word} =~ /^der$/i &&
            $$r{word} =~ /^ein$/i &&
            $w[$_+2]{word} =~ /^oder$/i) {
          next WORD;
        }

        # 2. Truly nested NPs. These are always formed with an adjective
        # that licences the inner NP:
        #
        #  `die den Kunstwerken entsprechende Analyse ist...'
        #
        my $ADJA = 0;
        for my $x ($_+1..$#w) {
          if ($w[$x]{best} =~ /^N[NE]$/) {
            for my $y ($x+1..$#w) {
              if ($w[$y]{best} eq 'ADJA') {
                next WORD;
              }
            }
          }
        }

        # Everything else is almost certainly a classification error.
        complain 'Duplicate articles!';
        if ($$l{word} =~ /^d/i && $$r{word} =~ /^eine[nm]?$/i) {
          prefer $r, 'ADJA';
        } elsif ($$l{word} =~ /^d/i) {
          prefer $l, 'PDS';
        } else {
          prefer $l, 'PIS';
        }
      }
    }

    # The sequence P..AT ART is possible, but only for some
    # restricted constructions.
    for (0..$#w-1) {
      my($l,$r) = ($w[$_], $w[$_+1]);
      if ($$l{best} =~ /^P.+AT$/ &&
          $$r{best} =~ /^(ART|P.+AT|V|PTKZU)/) {

        print STDERR "[A]\n";

        # `manch/PIDAT ein/ART Unternehmen'
        next if $$l{word} =~ /^(manch|solch|welch)$/i && $$r{word} =~ /^ein/;

        # `die/ART beiden/PIDAT Schwestern'
        next if $$r{word} =~ /^(all|beid)/;

        # `dessen/PRELAT eines/ART Ende'
        next if $$l{best} eq 'PRELAT' && $$r{word} =~ /^e/;

        # `diese/PDAT eine/ART Szene'
        next if $$l{best} eq 'PDAT' && $$r{word} =~ /^e/;

        # `weil wir alle/PIDAT die/ART Verfassung achten'
        next if $$l{word} eq 'alle' && $w[$_-1]{best} eq 'PPER';

        # `mit seinen/PPOSAT die/ART Blechinstrumente/NN anfeuernden/NN Fusten'
        my($NN,$participle);
        for my $j ($_+2..$#w) {
          if ($NN &&
              $w[$j]{best} eq 'ADJA' &&
              $w[$j]{word} =~ /nde[nrms]?$/) {
            $participle = $w[$j];
            last;
          }
          if ($w[$j]{best} =~ /^[NP]/) {
            $NN = $w[$j];
          }
        }
        next if $participle;

        complain "Attributive pronoun precedes non-NP material!";
        if ($$l{word} =~ /^ihr$/i) {
          prefer $l, 'PPER'
        } elsif ($$l{word} =~ /^(dies|jen|dess)/i) {
          prefer $l, 'PDS'
        } elsif ($$l{word} =~ /^wenig/i) {
          prefer $l, 'PIS'
        } elsif ($$l{word} =~ /^w/i) {
          prefer $l, 'PWS'
        } else {
          prefer $l, 'PIS'
        }
      }
    }

    # The sequence ART V.FIN is certainly wrong. Either the article is
    # a substituting pronoun, or the finite verb is something else.
    for (0..$#w-1) {
      if ($w[$_]{best} eq 'ART' &&
          $w[$_+1]{best} =~ /FIN/) {
        complain "Article precedes finite verb\n";

        if ($w[$_]{word} =~ /^e/) {
          alternate $w[$_], 'PIS';
        }

        if ($w[$_+1]{word} =~ /^(.+-)?[A-Z]/) {
          alternate $w[$_+1], 'NN';
        } elsif ($w[$_+1]{word} =~ /en?$/) {
          alternate $w[$_+1], 'ADJA';
        } elsif ($w[$_+1]{word} =~ /[nt]$/) {
          (my $p = $w[$_+1]{best}) =~ s/FIN/PP/;
          alternate $w[$_+1], $p;
        }
      }
    }

    # Now for the confusion between articles and relative pronouns. If
    # an article opens and an infinite verb and a finite verb end a
    # phrase, then the article is usually a relative pronoun really.
    if ($w[0]{best} eq 'ART' &&
        $w[0]{word} =~ /^(der|die|das|den|dem)$/ &&
        $#w > 0 &&
        $w[-2]{best} =~ /^V/ &&
        $w[-1]{best} =~ /FIN/) {
      complain "Article starts subclause\n";
      prefer $w[0], 'PRELS';
    }

    # Articles rarely end phrases.
    if ($w[-1]{best} eq 'ART') {
      complain "Article ends subclause!\n";
      if( $w[-1]{word} =~ /^(der|die|das|den|dem)$/) {
        prefer $w[-1], 'PDS';
      } else {
        prefer $w[-1], 'PIS';
      }
    }

    # Another sign of trouble is an infinitive that occurs on its own,
    # with no other verb near. This is almost certainly a finite form instead.
    my $inf = '';
    my $next = '';
    my $verbs = 0;
    for (@w) {
      $verbs++ if $$_{best} =~ /^V|PTKZU/;
      $inf = $_ if $$_{best} =~ /INF/;
    }
    if ($inf &&
        1 == $verbs &&
        $$inf{nr} < $w[-1]{nr} &&
        $words[$$inf{nr}+1]{best} ne 'KOKOM' ) {
      complain "Lonely infinitive `$$inf{word}'\n";

      (my $fin = $$inf{best}) =~ s/INF/FIN/;
      prefer $inf, $fin;
    }

    # Much the same goes for participles.
    my $pp = '';
    $verbs = 0;
    for (@w) {
      $verbs++ if $$_{best} =~ /^V/;
      $pp = $_ if $$_{best} =~ /^V.PP$/;
    }
    if ($pp &&
        1 == $verbs &&
        $$pp{nr} < $w[-1]{nr}) {
      complain "Lonely participle `$$pp{word}'\n";

      (my $fin = $$pp{best}) =~ s/PP/FIN/;
      prefer $pp, $fin;
    }

    # Another characteristic error occurs when the finite form
    # signalled by an initial subordinating conjunction is misdetected
    # as an infinitive or participle.
    if ($w[0]{best} =~ /PREL|KOUS|PWAV/ ||
        defined $w[1] && $w[1]{best} =~ /PREL|KOUS/) {

      # There are, of course, true WH questions, in which
      # PWAV starts and a participle ends the phrase:
      #
      # `Warum hast du das nicht gleich gesagt?'
      #
      # These are characterized by a finite verb in the second position.
      if(!(defined $w[1] && $w[1]{best} =~ /FIN/)) {

        if ($w[-1]{best} =~ /INF$/ && $w[-2]{word} ne 'zu') {
          (my $fin = $w[-1]{best}) =~ s/INF/FIN/;
          complain "Infinitive `$w[-1]{word}' ends subclause\n";
          prefer $w[-1], $fin;
        } elsif ($w[-1]{best} =~ /PP$/ && $w[-1]{word} !~ /^ge/) {
          (my $fin = $w[-1]{best}) =~ s/PP/FIN/;
          complain "Participle `$w[-1]{word}' ends subclause\n";
          prefer $w[-1], $fin;
        }
      }
    }

    # A sure sign of trouble is two finite verbs not separated by any
    # phrase boundary or conjunction. This is almost certainly wrong,
    # but hard to correct because there are three different common
    # reasons:
    #
    # - passive verbal adjectives: `_verkaufte_ Gerte'
    # - infinitives: `hat viele Abonnements verkaufen _knnen_'
    # - participles with unstressed prefixes: `hat den Prsidenten beraten'
    my $left = '';
    my $right = '';

    for (@w) {
      if ($$_{best} =~ /KOUS|KON|\$\(/) {
        $left = '';
      }

      if ($$_{best} =~ /FIN/) {
        # second finite verb detected
        if ($left) {
          $right = $_;
          complain "Two finite words: `$$left{word}' and `$$right{word}'\n";

          # if one of the two is an auxiliary verb, try to change the
          # other tag.
          if ($$left{best} =~ /^V[MA]/) {
            disprefer_finite_verb $right;
          } elsif ($$right{best} =~ /^V[MA]/) {
            disprefer_finite_verb $left;
          } else {
            disprefer_finite_verb $right;
          }
        }

        # first finite verb detected
        else {
          $left = $_;
        }
      }
    }

    # `als' in front of a finite Verb is certainly KOKOM, not KOUS.
    for (0..$#w-1) {
      if ($w[$_]{word} =~ /^als$/i &&
          $w[$_]{best} ne 'KOKOM' &&
          $w[$_+1]{best} =~ /FIN/) {
        complain "KOUS followed by finite verb!";
        prefer $w[$_], 'KOKOM';
      }
    }

    # Anyway, `als' is never APPR; it is either KOKOM or KOUS.
    for (0..$#w-1) {
      if ($w[$_]{word} =~ /^als$/i &&
          $w[$_]{best} eq 'APPR') {
        complain "`als' tagged as APPR!";

        # non-initial `als' is mostly KOKOM
        if($_ > 0) {
          prefer $w[$_], 'KOKOM';
        }

        # otherwise, `als' that ends with a finite verb is KOUS
        elsif ($w[-1]{best} =~ /FIN/) {
          prefer $w[$_], 'KOUS';
        }

        # otherwise, assume KOKOM
        else {
          prefer $w[$_], 'KOKOM';
        }
      }
    }

    # `etwas' and some related words are ambiguous between ADV and
    # PIS. IN front of other adverbials they are more often PIS.
    for(0..$#w) {
      if ($w[$_]{word} =~ /^(mehr|weniger|etwas)$/i
          && $w[$_]{best} eq 'ADV'
          && ($_ == $#w || $w[$_+1]{best} =~ /APPR|VVPP|PROAV/)) {
        complain "Suspicious ADV before adverbial!";
        prefer $w[$_], 'PIS';
      }
    }

    # Particles which could be circumpositions are probably
    # circumpositions if a fitting preposition precedes.
    my $PTKVZ = '';
    for (reverse @w) {
      $PTKVZ = $_ if $$_{best} eq 'PTKVZ';
      $PTKVZ = '' if $$_{best} =~ /FIN/;
      if($PTKVZ &&
         $$_{best} =~ /^APPR/ &&
         exists($APZR{"$$_{word}:$$PTKVZ{word}"})) {
        complain "`$$_{word}' ... `$$PTKVZ{word}' probably belongs together.";
        prefer $PTKVZ, 'APZR';
      }
    }
  }

  # regenerate lines
  $i = 0;
  for my $w (@words) {
    my $line = "$$w{word}";
    for (sort {$$w{scores}{$b} <=> $$w{scores}{$a}} keys %{$$w{scores}}) {
      $line .= "\t$_\t$$w{scores}{$_}";
    }
    print "$line\n";
  }
}


# decide which tagger to use
use Getopt::Std;
our($opt_n,$opt_t,$opt_b);
getopts 'bnt:';
my $tagger = $opt_t ? $opt_t : 'tnt';
my $nocorrect = $opt_n ? 1 : 0;

# look for the tagger binary
my $base;
my $cmdline;
for (qw(/data/linux/opt /usr/local /usr /opt/pkg)) {
  if ($tagger eq 'tnt') {
    if (-x "$_/tnt/tnt") {
      $base = "$_/tnt";
      # use an extra lexicon tuned to our grammar, if present
      my $lex = "$base/models/papa.lex";
      my $EXTRALEXICON = -f $lex ? "-b$lex" : '';
      if($opt_b) {
        $EXTRALEXICON = '';
      }
      $cmdline="$base/tnt $EXTRALEXICON -v0 -z200 $base/models/negra.tnt";
      last;
    }
  } elsif ($tagger eq 'TreeTagger') {
    if (-x "$_/TreeTagger/bin/tree-tagger") {
      $base="$_/TreeTagger";
      $cmdline="$base/bin/tree-tagger -quiet -token -threshold 0.01 -prob $base/lib/german.par "
    }
  }
}

die "Can't find tagger `$tagger'\n" unless $cmdline;

my $in = "/tmp/tagger.in.$$";
my $out = "/tmp/tagger.out.$$";
open(IN, ">$in") or die "Can't write `$in': $!\n";
print IN while(<>);
close(IN);

# open a tagger with that file
system "$cmdline $in > $out";

# read output
my @output = ();
open(OUT, "$out") or die "Can't read `$out': $!\n";
while (<OUT>) {
  push @output, $_;
  if (/^$/) {
    # remove obvious errors
    postprocess \@output;
    @output = ();
  }
}
if ($nocorrect) {
  print @output;
} else {
  postprocess \@output;
}

close(OUT);
unlink $in, $out;

