/* 
 * Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without 
 * modifications, as long as this notice is preserved.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * $Id: indexer.c,v 1.11 2004/10/13 14:20:49 micha Exp $ 
 *
 * Description:
 * make a berkeley database which holds an index of the position
 * of each lexical entry in a cdg lexicon file.
 *
 */


#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include "db.h"
#include "blah.h"


Boolean writeNextKey(FILE *stream);
Boolean validChar(Boolean qouted, int c);
Boolean dbOpen(String filename);
Boolean dbAdd(String word, int pos[]);

static DB *database;	     // the database handle


/* ------------------------------------------------------------------------- 
 * write the next valid key from a stream to the database
 * searches through the given stream till the next valid lexical entry is
 * found and adds it to the database 
 * @param *stream the stream (usually a file) in cdg format
 * @returns TRUE if success, otherwise FALSE
 */
Boolean writeNextKey(FILE *stream)
{
  int maxKeySize = 200;
  char key[maxKeySize];
  int pos[2] = { -1, -1 };   // the position of lex entry in file

  // pos[0] = beginning, pos[1] = end
  int c;		     // the character, actually read from file
  int keysize = 0;
  int stage = 1;
  Boolean slashFound = FALSE;
  int commentType = 0;	     // type 1: '//', type 2: '/*'
  Boolean starFound = FALSE; // expecting end of '/*' comment? ('*' found) 
  Boolean isQouted = FALSE; // are we processing a quoted string? 
  int quoteSign = 0; // integer value of qoute sign
  Boolean isEscaped = FALSE; // is this an escaped char?
  
  /* stages are: 
   * 1 = beginning, you will stay here until a possible key char
   *   occurs. Here you will  
   * 2 = there was a valid char, you will stay here and collect the key 
   * 3 = there was a space, stay here until all spaces are done 
   * 4 = there was a ':', this could be a key 
   * 5 = there was a '=', this is a key, stay here till final ';' was 
   *   found 
   * 6 = end of lex entry found, work done... return 
   */

  while ((c = getc(stream)) && !feof(stream)) {
    
    // comment handling
    if (slashFound) {		       // there was an '/' indicating a
				       // comment
      if (commentType == 1) {
	if (c == (int)'\n') {	       // comment ends with end of line 
	  slashFound = FALSE;
	  commentType = 0;
	}
      }
      if (commentType == 2) {
	if ((c == (int)'/') && starFound) {	// comment ends with '*/'
	  slashFound = FALSE;
	  commentType = 0;
	  starFound = FALSE;
	}
	if (c == (int)'*') {	       // this MAY be the end of the comment
	  starFound = TRUE;
	} else {
	  starFound = FALSE;
	}
      }
      if (commentType == 0) {
	switch (c) {
	  case ((int)'/'):
	    commentType = 1;	       // this is a line comment
	    break; // so it has to be deleted there
	  case ((int)'*'):
	    commentType = 2;	       // possibly more than one line
	    break;
	  default:
	    slashFound = FALSE;      // this is no comment
	}
      }
    }				       // end if(slashFound)

    if (!slashFound && (c == (int)'/')) {	// this MAY be a comment
      slashFound = TRUE;
      commentType = 0;
    }

    // indexing
    if (commentType == 0) {
      switch (stage) {
	case 1:
	  if ((c == (int) '"')||(c == (int)'\'')) { // this is the beginning of a quoted string
	    isQouted = TRUE;
	    quoteSign = c;
	    pos[0] = ftell(stream); 
	    break;
	  }
	  if (validChar(isQouted, c)) {
	    stage = 2;
	    if(!isQouted) {
	      pos[0] = ftell(stream);    // set offset only if not a qouted string
	    }
	    if (c == (int)'\\') { // don't add the escape character '\' to the key
	      isEscaped = TRUE;
	      break;
	    }
	    key[0] = c;		       // collect first char
	    keysize = 1;
	    break;
	  } else {
	    break;
	  }
	case 2:
	  if (validChar(isQouted, c)) {
	    if (c == (int)'\\') { // don't add the escape character '\' to the key
	      isEscaped = TRUE;
	      break;
	    }
	    if (!isEscaped && c == quoteSign) {
	      isQouted = FALSE;
	      quoteSign = 0;
	      break;
	    } else {
	      key[keysize] = c;
	      keysize++;
	      isEscaped = FALSE;
	      break;
	    }
	  }
	  if (isspace(c)) {
	    isEscaped = FALSE;
	    stage = 3;
	    break;
	  }
	  if (c == (int)':') {
	    isEscaped = FALSE;
	    stage = 4;
	    break;
	  }
	  // if nothing appropriate was found
	  isEscaped = FALSE;
	  stage = 1;
	  keysize = 0;
	  break;
	case 3:
	  if (isspace(c)) {
	    break;
	  }
	  if (c == (int)':') {
	    stage = 4;
	    break;
	  }
	  if ((c == (int)'"')||((int)'\'')) { // this was no key, perhaps now there is a new one
	    isQouted = TRUE;
	    quoteSign = c;
	    pos[0] = ftell(stream);
	    stage = 2;
	    break;
	  }
	  if (validChar(isQouted, c)) { // this was no key, perhaps now there is a new one
	    stage = 2;
	    if(!isQouted) {
	      pos[0] = ftell(stream);
	    }
	    key[0] = c;
	    keysize = 1;
	    break;
	  }
	  stage = 1;
	  break;
	case 4:
	  if (c == (int)'=') {
	    stage = 5;
	    break;
	  } else {
	    stage = 1;
	    keysize = 0;
	    break;
	  }
	case 5:
	  if (c == (int)';') {
	    stage = 6;
	    pos[1] = ftell(stream);
	    break;
	  } else {
	    break;
	  }
	case 6:
	  key[keysize] = '\0';
	  dbAdd(key, pos);
	  return TRUE;
      }
    }
  }

  return FALSE;

}

      
/* ------------------------------------------------------------------------- 
 * checks if this is a valid char for key
 * tests if this char is allowed to be part of a keyname, distinquish
 * between qouted and not qouted strings
 * @param quoted Boolean flag if this is a qouted key
 * @param char the char which is to be checked
 * @returns TRUE if this is a valid char, FALSE otherwise
 */
Boolean validChar(Boolean quoted, int c) 
{
  if (quoted) {
    return TRUE;
  }
  else {
    return
      (isalnum(c) ||
       (c > 0x80));
    }
}


/*------------------------------------------------------------------------- 
 * opens a database.
 * this function opens a database, specified by the parameter 'filename'
 * and stores the handle in the static variable 'database'. If there is no
 * database file with this filename, a new database is created.
 * @param filename the filename of the database file
 * @returns TRUE if success, FALSE otherwise
 */
Boolean dbOpen(String filename)
{
  int ret = 0;		     // return code for database operations

  // create database handle
  if ((ret = db_create(&database, NULL, 0)) != 0) {
    fprintf(stderr, "ERROR: (db_create) - %s\n", db_strerror(ret));
    return FALSE;
  }
  // setting database to accept more than one data item for one key
  database->set_flags(database, DB_DUP);

  // open database 'filename', if the files already exists, the error EEXIST
  // is thrown by database->open
#if (DB_VERSION_MAJOR < 4 || \
     DB_VERSION_MAJOR == 4 && DB_VERSION_MINOR == 0)
  if((ret = database->open(database, filename, NULL, DB_BTREE, 
      DB_CREATE | DB_EXCL | DB_TRUNCATE, 0664)) != 0) {
#else /* db 4.1.x or later */
  if((ret = database->open(database, NULL, filename, NULL, DB_BTREE, 
      DB_CREATE | DB_EXCL | DB_TRUNCATE, 0664)) != 0) {
#endif
    fprintf(stderr, "ERROR: (database->open) - %s\n", db_strerror(ret));
    return FALSE;
  } 

  return TRUE;
}

/* ------------------------------------------------------------------------- 
 * close the database
 * write database to disk and close it.
 * @returns TRUE if success, FALSE otherwise
 */

Boolean dbClose(void)
{
  int ret = 0;		     // return code for database operations

  if (database == NULL) {
    fprintf(stderr, "ERROR: No open database found for closing\n");
    return FALSE;
  }

  if ((ret = database->close(database, 0)) != 0) {
    fprintf(stderr, "ERROR: (database->close) - %s\n", db_strerror(ret));
    return FALSE;
  }
  return TRUE;
}

/* ------------------------------------------------------------------------- 
 * adds an entry to the database 
 * adds a key and its position in the lexicon file to the database
 * @param word the word which is the key
 * @param pos element [0]: the beginning, element [1] the end of this entry
 * @returns TRUE if success, FALSE otherwise
 */
Boolean dbAdd(String word, int pos[])
{
  int ret = 0;		     // return code for database operations
  DBT key, value;	     // DBT structs for key and value


  
  if (database == NULL) {
    fprintf(stderr, "ERROR: No database opened");
    return FALSE;
  }


  memset(&key, 0, sizeof (key));       // cleaning memory
  memset(&value, 0, sizeof (value));
  key.data = word;
  key.size = strlen(word);
  value.data = pos;
  value.size = sizeof(int) * 2;

  if ((ret = database->put(database, NULL, &key, &value, 0)) == 0) {
    //printf("added : %s->[%d][%d]\n", key.data, *((int *)value.data), *((int *)(value.data)+1));
    return TRUE;
  } else {
    fprintf(stderr, "ERROR: (database->put) - %s\n", db_strerror(ret));
    return FALSE;
  }
}

/* ---------------------------------------------------------------------- */ 

int main(int argc, char **argv)
{
  String lexFilename; 
  String dbFilename;
  FILE *stream;
  int keysFound = 0;
  
  if((argc > 3) || (argc == 1)) {
    fprintf(stderr, "syntax: %s <cdg_file> <db_file>\n", argv[0]);
    exit(1);
  }

  lexFilename = argv[1];
  dbFilename = argv[2];

  if (!dbOpen(dbFilename)) {
    fprintf(stderr, "ERROR: cannot open db file '%s'\n", dbFilename);
    exit(1);
  }
  // open lexicon file (readonly)
  if ((stream = fopen(lexFilename, "r")) == NULL) {
    fprintf(stderr, "ERRPR: cannot open lex file '%s'\n", lexFilename);
    exit(1);
  }
  while(writeNextKey(stream)) {
    keysFound++;
#if 0
    if (keysFound % 100 == 0) {
      printf("%-9d\r", keysFound);
      fflush(stdout);
    }
#endif
  }
  printf("%d valid keys found\n",keysFound);
  fclose(stream);
  dbClose();
  return 0;
  
}
