eval '(exit $?0)' && eval 'exec perl -S $0 ${1+"$@"}'
 & eval 'exec perl -S $0 $argv:q'
 if 0;

# cdg/utils/memdebug.pl
# 2000/3/1 Ingo Schroeder

# This script aids in debugging cdg memory leaks. It takes a
# self-contained call to cdg and a fully qualified binary that was
# used in that call. A typical invocation would be
#
#  memdebug.pl 'cdgp > /dev/null' `type cdgp` > memdebug.log
#
# where your .cdgrc contains code to drive the cdgp binary, including
# a quit statement.
#
# The binary must have been compiled with DMALLOC defined. It will
# then produce a detailed log of all allocated blocks that leaked in
# this run. This script scans the dmalloc logfile for faulty
# allocations. Then it re-creates the exact call and uses dmalloc to
# obtain a core file created at the exact point where the leaking
# block was allocated. Not content with that, it then uses gdb to
# translate the call stack into human-readable form. Thus, you'll get
# something similar to
#
# 952616404: 49176: not freed: '0x12a008|s1' (16386 bytes) from 'ra=0x86dc8'
#   #0                       0x6ff08244 _libc_kill
#   #1                       0x6feba608 abort
#   #2         error.c:377   0x2e7b4 _dmalloc_die
#   #3         error.c:413   0x2e844 dmalloc_error
#   #4        malloc.c:316   0x27ce0 check_pnt
#   #5        malloc.c:612   0x28070 _loc_malloc
#   #6        malloc.c:758   0x28318 malloc
#   #7      program.c:1878   0x86dd0 fooConstructor
#   #8      program.c:1562   0x86854 initFoos
#   #9       program.c:669   0x846f8 interact
#   #10     program.c:1234   0x846f8 main
#
# So a block that was allocated by fooConstructor() is never freed.
# This doesn't really mean that the constructor is at fault -- the
# constructor might allocate redundant blocks, or else the destructor
# might fail to free them. In fact, you would get this exact same
# output even if there were no destructor at all! But at least you can
# pick up the pieces from there and chase the guilty memory block
# through the program until you realize why it wasn't freed.
#
# ``Pretend that you're Hercule Poirot: Examine all clues, and deduce
#   the truth by order and method.''
#                                                    - Donald E. Knuth

require "getopts.pl";

if ($0 =~ m/.*\/([^\/]*)/) {
  $cmd = $1;
} else {
  $cmd = $0;
}

&Getopts('h');

# educate user
die "Usage: $cmd [-h] testcommand binary\n"
  if defined($opt_h) || $#ARGV!=1;

$testcommand = $ARGV[0] . " > /dev/null";
$binary = $ARGV[1];
$options="debug=0x14e40583,inter=100,log=dmalloc.log";
$gdb="/opt/bin/gdb";
$responsefile="commands.$$";

open(W,">$responsefile") or die "can't write gdb batch file: $!";
print W "where\n";
close(W);

# verify that $binary was built with dmalloc
die
  "You must link the dmalloc library into your program for this tool to work.\n"
  if `strings $binary` !~ /dmalloc/;

# A list of known leaks:
while(<DATA>) {
  chomp;
  $known_leaks .= "|$_"
}
$known_leaks =~ s/^\|//;

# Set up the environment variable that controls dmalloc
$ENV{DMALLOC_OPTIONS}=$options;

# Call the program. Note that this call must terminate without user
# input, so be sure to get a `quit' into the input stream somehow.
system("$testcommand");

# Find all instances of unfreed blocks
open(LOG, "< dmalloc.log") || die "$cmd: Can't open log file: $!\n";
while ($l=<LOG>) {
  printf "%s", $l;

  # match lines like this:
  #  not freed: '0x124008|s1' (5128 bytes) from 'ra=0x6fee3ca8'
  if ( $l=~m/not freed: '(0x.*)\|s(\d+)' .* 'ra=(0x.*)'/ ) {
    $notfreed{"$1:$2"}=$l;
  }
}
close(LOG);

$total = keys %notfreed;
printf STDERR "Found %d unfreed block%s.\n", $total, $total == 1 ? "" : "s";
printf "Found %d unfreed block%s.\n", $total, $total == 1 ? "" : "s";
printf "\nStack frames for allocations without de-allocations:\n";

# Repeat the call, watching this memory location like a hawk.
foreach $l (keys %notfreed) {

  printf STDERR "%d%% complete\n", 100 * $current++ / $total;

  $stacktrace = sprintf "\n%s", $notfreed{$l};
  $ENV{DMALLOC_OPTIONS}="$options,addr=$l";
  unlink("core");
  system("$testcommand");

  # Pipe the result through gdb to get symbolic backtraces
  open(GDB, "$gdb -batch -x $responsefile $binary core 2>&1|")
    || die "$cmd: Can't open gdb: $!\n";
  @raw = <GDB>;

  # split output into frames delimited by #<number>,
  # keeping the delimiters
  @frames = split (/(?=\#\d+\b)/, join(" ",@raw));

  # Pretty-print each frame
  foreach (@frames) {
    tr/\n/ /;

    # skip first frame (printed twice by gdb)
    next if /\(gdb\)/;

    # Match lines like this:
    #  #8  0xa1530 in rl_add_funmap_entry (name=0xbd8a0 "clear-screen", function=0x9dd9c <rl_clear_screen>) at funmap.c:154
    if (/(\#\d+)\s+(0x[^ ]+)\s+in\s+([^\(]+)\s+\(.*\)\s+(at (.*):(.*))?/) {
      $stacktrace .= sprintf "%4s  %15.15s%s%-6s %s %s\n",
        $1, $5, defined($5) ? ":" : " ", $6, $2, $3;
    }
    close(GDB);
  }

  if(known_leak($stacktrace)) {
    $stacktrace = "\nThis leak is known already: " . $stacktrace;
  } else {
    $stacktrace = "\nThis leak seems to be new: " . $stacktrace;
  }

  push @traces, $stacktrace;

}

unlink($responsefile);

print sort @traces;

sub known_leak {
  shift =~ /$known_leaks/
}


# There follows a list of leaks that are already known to the cdg maintainers.
__DATA__
vprintf
getpwnam
add_history
fgets
ctime
