#!/usr/bin/env perl
# cdg/utils/shift-reduce.pl

use warnings;
use strict;

BEGIN {
  use File::Basename;
  unshift @INC, dirname(readlink($0) || $0); 
  require 'cdgsetup.pl';
}
use CDG;

our $invocation = "$0 @ARGV";

use Getopt::Std;
our $opt_a = 0;
our $opt_b = 0;
our $opt_c = 'cat';
our $opt_d = 0;
our $opt_f = '';
our $opt_h = 1;
our $opt_i = 0;
our $opt_l = 1;
our $opt_m = 'SYN';
our $opt_n = 0;
our $opt_p = 0;
our $opt_r = '';
our $opt_s = 0;
our $opt_t = 'trans.tbl';
our $opt_v = 0;
our $opt_w = 0;
our $opt_x = 0;
getopts('abc:df:h:il:m:n:pst:r:v:wx');

$opt_p = 1 if $opt_d;
$opt_w = 1 if $opt_x;

die "Usage: $0 <filename>

    -a    use top word attachment as a feature (FALSE)
    -b    use backtracking (FALSE)
    -c S  category feature (cat)
    -d    debug parsing decisions (implies -p and -v2)
    -f S  use file for category tying (off)
    -h N  threshold for backoff (1)
    -i    encode distance between visible words as a feature (FALSE)
    -l N  lookahead window (1)
    -m S  syntax level name (SYN)
    -n N  use only N utterances (all)
    -p    parse (default: train)
    -r S  use real tagger program (default: simulate)
    -s    allow special (nonprojective) transitions where necessary (FALSE)
    -t S  transition table file name (trans.tbl)
    -v N  verbosity level (0)
    -w    write parse result as new annotation (FALSE)
    -x    immediately launch the tree viewer (implies -w) (FALSE)

" if $#ARGV;

our $input = shift;

# possible parse actions
our @actions = qw(SHIFT REDUCE LEFT RIGHT);
push @actions, qw(FARRIGHT FARLEFT) if($opt_s);

# table of category ties
our %ties = ();
if($opt_f) {
  open(TIES, "$opt_f") or die "Can't read `$opt_f': $!\n";
  while(<TIES>) {
    if(/^(.+)=(.+)$/) {
      $ties{$1} = $2;
    }
  }
}

close(TIES);

# table of predicted parse actions per state
our %table;

# list of sentence structures pulled from CDG
our @annos;

# number of features used in this run
our $dim = 0;

# number of events before we trust a parsing decision
our $threshold = $opt_h;

# number of permissible agenda items
our $agenda_size = 10;

our ($good, $bad);

# conditionally print output
sub debug {
  print @_ if $opt_v >= shift;
}

# quote and escape string suitably for CDG input
sub print_identifier {
  my $x = shift;
  (my $result = $x) =~ s/([\'\"])/\\$1/g;
  if ($result =~ /[^a-zA-Z\x80=\xff]/) {
    $result = "'$result'";
  }
  return $result;
}

# sort like ls -v does
sub byVName {

  my ($a, $b) = @_;

  # don't change the names themselves
  my ($x,$y) = ($a,$b);

  while ($x ne "" && $y ne "") {

    # skip identical alphabetical parts
    while (substr($x,0,1) eq substr($y,0,1) &&
           $x !~ /^[0-9]/) {
      substr($x,0,1) = "";
      substr($y,0,1) = "";
    }

    # compare alphabetical parts normally
    if ($x !~ /^\d/ || $y !~ /^\d/) {
      return $x cmp $y;
    }

    # compare numeric parts numerically
    $x=~s/^(\d+)//;
    my $nx = $1;
    $y=~s/^(\d+)//;
    my $ny = $1;
    my $result = $nx <=> $ny;
    next unless $result;
    return $result;
  }
}



# create signature of current parse state,
# e.g. "VVFIN/NN,KON"
sub encode_state {
  my $s = shift;
  my $x = $$s{anno};
  die unless exists $s->{stack};

  my $dimensionality = 0;

  # encode top stack category
  my $sig = @{$s->{stack}} ? $$x{tags}[$s->{stack}[0]] : 'NIL';

  if ($opt_a) {

    # encode environment of top word
    if (@{$s->{stack}} &&
        $s->{struc}[$s->{stack}[0]]) {
      $sig .= ",1";
    } else {
      $sig .= ",0";
    }

    my $has_left_child = 0;
    my $has_right_child = 0;
    if (@{$s->{stack}}) {
      for (1..$s->{dot}-1) {
        if ($s->{struc}[$_] == $s->{stack}[0]) {
          if($_ < $s->{stack}[0]) {
            $has_left_child = 1;
          } else {
            $has_right_child = 1;
          }
        }
      }
    }
    if ($has_left_child) {
      $sig .= "1";
    } else {
      $sig .= "0";
    }
    if ($has_right_child) {
      $sig .= "1";
    } else {
      $sig .= "0";
    }

    $dimensionality++;
  }

  $sig .= '/';

  # encode distance between top of stack and input pointer
  if ($opt_i) {
    my $d = @{$s->{stack}} ? $s->{dot}-$s->{stack}[0] : 0;
    $sig .= "$d/";
    $dimensionality++;
  }

  # encode lookahead categories
  for ($s->{dot}..$s->{dot}+$opt_l-1) {
    $dimensionality++;
    if (defined($$x{tags}[$_])) {
      $sig .= $$x{tags}[$_];
    } else {
      $sig .= "\$";
    }
    if ($_ < $s->{dot}+$opt_l-1) {
      $sig .= '|';
    }
  }

  $dim = $dimensionality unless $dim;

  return $sig;
}

# Given a particular state specification, return one that has
# less contextual information.
#
# The possible features are stripped in the following order:
#
# - if there is distance information, that is stripped:
#   ART,01/3/NN|APPR ==> ART,01/NN|APPR
#
# - if there are lookahead tokens, on is removed:
#   ART,01/NN ==> ART,01/NN
#
# - if there is stack attachment information, it is removed:
#   ART,01/ ==> ART/
#
#
# reduce_context() does not care whether the new state has any
# information attached to it or not, it just backs off by exactly
# one feature. (A better solution might be to back of to the
# neighbouring state that has the most weight, or to pool the
# information of neighbouring states.) To find the nextmost
# *existing* state, use backoff().
sub reduce_context {
  my $sig = shift;

  # remove distance information
  return if $$sig =~ s!/\d+/!/!;

  # remove lookahead tokens:
  # (a) one of several
  return if $$sig =~ s/(\|[^\|]+)$//;

  # (b) the last one
  return if $$sig =~ s!/.+$!/!;

  # remove stack attachment info
  $$sig =~ s/,[01]+//;
}

# if no action is programmed for this state,
# return one that is programmed for a similar state with less context
sub backoff {
  my $sig = shift;
  my $n = shift;

  # reduce lookahead
  while ($n > 0 &&
         (!exists($table{$n}{$$sig})
          ||
          $table{$n}{$$sig}{total} < $threshold)) {

    # remove one of several lookahead tokens
    reduce_context $sig;
    $n--;
  }
  if (exists $table{$n}{$$sig}) {
    return ($table{$n}{$$sig}{action}, $n);
  } else {
    warn "Can't backoff to `$$sig'!\n";
    return ('shift', 0);
  }
}

# read annotation into perl hash
sub read_anno {
  my $ae = shift;

  my $x =
    {
     id => AnnoEntryStruct_id_get($ae),
     annotated => [ undef ],
     normalized => [ undef ],
     tags => [''],
     words => ['NIL'],
     labels => [''],
     projective => [''],
     history => []
    };

  for (my $m = AnnoEntryStruct_annos_get $ae;$m;$m = listNext $m) {
    my $a = listElement $m;
    my $cat = undef;
    my $struc = undef;

    push @{$x->{words}}, AnnotationStruct_word_get $a;

    for (my $n = AnnotationStruct_specs_get $a;$n;$n = listNext $n) {
      my $s = listElement $n;
      my $t = SpecificationStruct_type_get $s;

      # note tag
      if ($t eq $STTag) {
        my $att = SpecificationStruct_kind_get $s;
        if ($att eq $opt_c) {
          $cat = SpecificationStruct_name_get $s;
        }
      }

      # note edge
      else {
        my $level = SpecificationStruct_kind_get $s;
        if ($level eq $opt_m) {
          $struc = SpecificationStruct_position_get $s;
          push @{$x->{labels}}, SpecificationStruct_name_get $s;
        }
      }
    }

    # note missing cat
    if (!defined $cat) {
      print STDERR "WARNING: no category for word `$$x{words}[-1]' in $$x{id}\n";
      $cat = '';
    }

    # maybe tie cat
    if(%ties && exists $ties{$cat}) {
      $cat = $ties{$cat};
    }

    # note missing structure
    if (!defined $struc) {
      print STDERR
        "WARNING: no syntactic structure for word `$$x{words}[-1]' in $$x{id}\n";
      $struc = 0;
    }

    push @{$x->{projective}}, 1;
    push @{$x->{annotated}}, $struc;
    push @{$x->{tags}}, $cat;
  }
  debug 1, "$$x{id}: @{$x->{words}}\n";

  return $x;
}



# return earlier end of (annotated) edge #$i
sub min {
  my $x = shift;
  my $i = shift;
  return $i < $$x{annotated}[$i] ? $i : $$x{annotated}[$i];
}
# return later end of (annotated) edge #$i
sub max {
  my $x = shift;
  my $i = shift;
  return $i > $$x{annotated}[$i] ? $i : $$x{annotated}[$i];
}
# is (annotated) edge #$i attached at the left?
sub left {
  my $x = shift;
  my $i = shift;
  return $i < $$x{annotated}[$i];
}
# is (annotated) edge #$i attached at the right?
sub right {
  my $x = shift;
  my $i = shift;
  return $$x{annotated}[$i] && $i > $$x{annotated}[$i];
}

# detect nonprojective edges
sub check_anno {

  my $x = shift;

  my @normalized = @{$x->{annotated}};
  my @projective = (1) x ($#{@$x{words}}+1);
  my @labels = @{$$x{labels}};
  my @tags = @{$$x{tags}};
  my @words = @{$$x{words}};

  # check all edge pairs
  for my $X (1..$#{$x->{words}}) {
    next unless $normalized[$X];
    for my $Y ($X+1..$#{$x->{words}}) {
      next unless $normalized[$Y];

      if (min($x,$X) < min($x,$Y) &&
          max($x, $X) < max($x,$Y) &&
          max($x,$X) > min($x,$Y)
          ||
          $X == $normalized[$Y] &&
          right($x,$Y) &&
          left($x,$X) &&
          $normalized[$X] < $Y
          ||
          $Y == $normalized[$X] &&
          right($x,$Y) &&
          left($x,$X) &&
          $normalized[$Y] > $X
         ) {

        debug 3, "WARNING: nonprojective edges " .
          "$X:$$x{words}[$X]($$x{labels}[$X]) $Y:$$x{words}[$Y]($$x{labels}[$Y])\n";

        # do not punish sn edge twice
        if ($projective[$X] && $projective[$Y]) {

          # try very hard to find the `responsible' edge and mark it
          # !This is heavily grammar-dependent!
          if ($labels[$X] eq 'RE') {
            $projective[$X] = 0;
          } elsif ($labels[$Y] eq 'RE') {
            $projective[$Y] = 0;
          } elsif ($labels[$X] =~ '^(KOM|CJ|KON)$') {
            $projective[$X] = 0;
          } elsif ($labels[$Y] =~ '^(KOM|CJ|KON)$') {
            $projective[$Y] = 0;
          } elsif ($labels[$X] eq 'REL') {
            $projective[$X] = 0;
          } elsif ($labels[$Y] eq 'REL') {
            $projective[$Y] = 0;
          } elsif ($labels[$X] eq 'OBJC' && $tags[$normalized[$X]] eq 'NN') {
            $projective[$X] = 0;
          } elsif ($labels[$Y] eq 'OBJC' && $tags[$normalized[$Y]] eq 'NN') {
            $projective[$Y] = 0;
          } elsif ($labels[$X] =~ /^OBJ|SUBJ|PRED/) {
            $projective[$X] = 0;
          } elsif ($labels[$Y] =~ /^OBJ|SUBJ|PRED/) {
            $projective[$Y] = 0;
          } else {
            # give up
            debug 3, "WARNING: weird nonprojective edges " .
              " $words[$X]($labels[$X]) $words[$Y]($labels[$Y])\n";
          }
        }
      }
    }
  }

  $$x{normalized} = [ @normalized ];
  $$x{projective} = [ @projective ];

}

# print current state to screen,
# e.g. "[ soll Produkte ] . und"
sub dump_state {
  my $s = shift;
  my $x = $$s{anno};
  return '' unless exists $s->{dot};
  my $string = '';
  $string .= '[';
  for (reverse @{$s->{stack}}) {
    die "no word $_!" unless defined $$x{words}[$_];
    $string .= " $$x{words}[$_]";
  }
  $string .= ' ] . ';
  if ($s->{dot} == $#{@$x{words}}+1) {
    $string .= '$';
  } else {
    $string .= "$$x{words}[$s->{dot}]";
  }
  $string .= "\n";
  return $string;
}

# given the current state of affairs, decide on an action. Also
# returns the state that was actually used to make the decision,
# since the actual current state may be unknown in the table.
sub decide {
  my $sig = shift;
  my $n = shift;
  my $action;

  if (exists $table{$n}{$sig}) {
    $action = $table{$n}{$sig}{action};
  } else {
    ($action, $n) = backoff \$sig, $n;
  }

  if (!defined $action) {
    $action = 'SHIFT';
    $n = 0;
  }
  return ($action, $sig, $n);

}

# determine accuracy and maybe write back
sub postprocess {

  my $s = shift;
  my $x = $$s{anno};
  return if exists $s->{no} && $s->{no} == 0;
  my @parsed = @{$s->{struc}};

  # determine accuracy
  my $correct =
    grep {
      $$x{annotated}[$_] eq $parsed[$_] ||
        (debug(1,
               sprintf
               "ERROR: governor of `%s' is `%s', should be `%s'\n",
               $$x{words}[$_],
               $$x{words}[$parsed[$_]],
               $$x{words}[$$x{annotated}[$_]]) && 0)
      } (1..$#{@$x{words}});
  my $incorrect = $#{@$x{words}} - $correct;
  $good += $correct;
  $bad += $incorrect;

  if (defined $correct) {
    printf "$$x{id}: %d/%d (%4.2f%%)\n",
      $correct, $correct+$incorrect, 100*$correct/($correct+$incorrect);
  }


  # create new anno
  if ($opt_w) {
    my $annoname = "$$x{id}.sr";
    open(ANNO, ">$annoname")
      or die "Can't write `>$annoname': $!\n";
    print ANNO "'$$x{id}' : '$$x{id}' <->\n";
    for (1..$#{@$x{words}}) {
      printf ANNO "%d %d %s\n", $_-1, $_, print_identifier $$x{words}[$_];
      print ANNO "$opt_c/'$$x{tags}[$_]'\n";
      print ANNO "'$opt_m' -> '' -> $parsed[$_]\n";
      print ANNO ($_ == $#{@$x{words}} ? ';' : ',');
      print ANNO "\n";
    }
    close(ANNO);
    if($opt_x) {
      system "cdg -x -s showparse '$annoname' '$$x{id}'"
    }
  }

}

# insert search state sorted by P into array ref
sub insert_sorted {
  my $a = shift;
  my $s = shift;

  my $i = 0;
  $i++ while $i <= $#$a && $$a[$i]{p} > $s->{p};
  if ($i < $agenda_size) {
    splice @$a, $i, 0, $s;
  }

}

# parse sentence, finding most probable tree
sub parse_anno {

  my $x = shift;

  my @agenda = ();
  my $no = 0;
  my $solution;
  my @history = @{$$x{history}};
  my $state_error = 0;

  if ($opt_d) {
    $opt_v = 2;
  }

  # initial state
  push @agenda,
    {
     struc => [(0) x ($#{@$x{words}} + 1)],
     stack => [],
     p => 1.0,
     dot => 1,
     no => 0,
     anno => $x
    };

  while (@agenda) {
    my $s = shift @agenda;

    debug 3, "read: ";
    debug 1, dump_state $s;

    my ($action, $n);
    my $sig = encode_state($s);

    next if $solution && $s->{p} < $solution->{p};
    ($action, $sig, $n) = decide($sig, $dim);

    my $branch = 0;
    for (@actions) {

      next if !$opt_b && $_ ne $action;

      # ignore zero probability transitions
      my $c = $table{$n}{$sig}{$_};
      next unless $c;

      my $top = $s->{stack}[0];
      my $dot = $s->{dot};

      # backing off may have introduced a nonzero p for SHIFT, LEFT
      # or REDUCE although the input has been used up.
      next if $dot > $#{@$x{words}} && $_ ne 'REDUCE';

      my $p = $c / $table{$n}{$sig}{total};
      debug 2, "$sig: ";
      debug 1, uc $_;
      debug 2, sprintf(" (%.3f)", $p);
      my $suc = { struc => [@{$s->{struc}}],
                  stack =>  [@{$s->{stack}}],
                  dot => $s->{dot},
                  p => $s->{p} * $p,
                  no => ++$no,
                  anno => $x
                };

      # transit
      if (/^RIGHT$/) {

        debug 1, " ($$x{words}[$top]<----$$x{words}[$dot])";
        if ($opt_d && $$x{annotated}[$dot] != $top) {
          print STDERR "WARNING: correct governor would be ",
            "`$$x{words}[$$x{annotated}[$dot]]'\n";
        }

        $suc->{struc}[$dot] = $top;
        unshift @{$suc->{stack}}, $suc->{dot}++;
      } elsif (/^LEFT$/) {

        debug 1, " ($$x{words}[$top]---->$$x{words}[$dot])";
        if ($opt_d && $$x{annotated}[$top] != $dot) {
          print STDERR "WARNING: correct governor would be ",
            "`$$x{words}[$$x{annotated}[$top]]'\n";
        }

        $suc->{struc}[$top] = $dot;
        shift @{$suc->{stack}};
      }

      elsif (/^REDUCE$/) {
        shift @{$suc->{stack}};
      }

      elsif (/^SHIFT$/) {
        unshift @{$suc->{stack}}, $suc->{dot}++;
      }

      debug 1, "\n";

      # annotate wrong transitions
      if ($opt_d && !$state_error) {
        my $h = shift @history;
        if ($_ ne $h) {
          print STDERR "WARNING: correct action would have been $h, not $_.\n";

          # after a wrong transition, the transition history gets
          # out of sync, and we can no longer reliably detect wrong
          # transitions just by comparing to @history.
          # Therefore we disable transition debugging altogether.
          $state_error = 1;
        }
      }

      # detect end state
      if ($suc->{dot} > $#{@$x{words}} &&
          (!$solution || $solution->{p} < $suc->{p})) {
        $solution = $suc;
        debug 1, sprintf("Found solution with p = %.3e\n", $suc->{p});
      }

      # sort new state into agenda
      else {
        $branch++;
        insert_sorted \@agenda, $suc;
        debug 3, "wrote: " . dump_state $suc;
      }
    }
    unless($branch || $solution) {
      warn "ARRGH! No transition possible for $sig.\n";
      return;
    }
  }
  postprocess $solution;
}

# reconstruct shift/reduce history
sub deparse_anno {

  # make statistics about transition frequency
  sub count {
    my $action = shift;
    my $sig = shift;
    my $x = shift;
    if ($opt_d) {
      # don't write to $table,
      # just record the individual parse history
      push @{$x->{history}}, $action;
    } else {
      # count the transition for later
      $table{$dim}{$sig}{$action}++;
      debug 2, "[$sig]: " . uc $action. "\n";
    }
  }

  my $x = shift;
  my $s =
    {
     dot => 1,
     stack => [],
     struc => [ (0) x ($#{@$x{words}}+1) ],
     anno => $x
    };

  check_anno $x;

  while ($$s{dot} <= $#{@$x{words}}) {

    debug 2, dump_state $s;
    my $sig = encode_state $s;

    my $top = $$s{stack}[0];
    my $dot = $$s{dot};

    # detect left action
    if ($top && $$x{normalized}[$top] == $dot) {
      debug 1, "$$x{words}[$top]---->$$x{words}[$dot]\n";
      count('LEFT', $sig, $x);
      $$s{struc}[$top] = $dot;
      shift @{$$s{stack}};
      next;
    }

    # if the top stack word is already subordinated,
    # and has no normal (projective) dependents on the right,
    # reduce it
    if ($top && $$x{normalized}[$top] < $top) {
      my $deps = 0;
      for ($dot..$#{@$x{words}}) {
        if ($$x{projective}[$_] && $$x{normalized}[$_] == $top) {
          $deps = 1;
          last;
        }
      }
      if (!$deps) {
        count('REDUCE', $sig, $x);
        shift @{$$s{stack}};
        next;
      }
    }

    # if the next word has a dependent on the left,
    # but that dependent is buried in the stack,
    # then it must be nonprojective.
    # Postulate a special transition.
    my $deps = 0;
    for (@{$$s{stack}}) {
      if ($$x{normalized}[$_] == $dot &&
          !$$s{struc}[$_]) {
        if ($opt_s) {
          count('FARLEFT', $sig, $x);
          debug 1, "far left attachment needed for `$$x{words}[$_]'\n";
          $$s{struc}[$_] = $dot;
        } else {
          debug 1, "WARNING: forcing `$$x{words}[$_]' to NIL\n";
          $$x{normalized}[$_] = 0;
        }
        next;
      }
    }

    # detect right action
    if ($top && $$x{normalized}[$dot] == $top) {
      debug 1, "$$x{words}[$top]<----$$x{words}[$dot]\n";
      count('RIGHT', $sig, $x);
      $$s{struc}[$dot] = $top;
      unshift @{$$s{stack}}, $$s{dot}++;
      next;
    }

    # if the next word is right attached,
    # but its attachment point is not available
    # (either already gone from the stack or buried in it),
    # then it must be nonprojective.
    # Postulate a special transition.
    if (right($x,$dot)) {

      if ($opt_s) {
        debug 1, "far right attachment needed for `$$x{words}[$dot]' ($dot)\n";
        count('FARRIGHT', $sig, $x);
        $$s{struc}[$dot] = $$x{normalized}[$dot];
      } else {
        debug 1, "WARNING: forcing `$$x{words}[$dot]' ($dot) to NIL\n";
        $$x{normalized}[$dot] = 0;
        count('SHIFT', $sig, $x);
      }
      unshift @{$$s{stack}}, $$s{dot}++;
      next;
    }

    # if stack is empty, shift
    if (!$top) {
      count('SHIFT', $sig, $x);
      unshift @{$$s{stack}}, $$s{dot}++;
      next;
    }


    # if the attachment is farther to the right, shift
    if ($$x{normalized}[$dot] > $dot || $$x{normalized}[$dot] == 0) {
      count('SHIFT', $sig, $x);
      unshift @{$$s{stack}}, $$s{dot}++;
      next;
    }

    # if we can't recreate this word's edge, pretend it is a root edge
    print STDERR "WARNING: Can't deal with word $$x{words}[$dot] in $$x{id}!\n";
    $$x{normalized}[$dot] = 0;
    unshift @{$$s{stack}}, $$s{dot}++;
  }

  # check to see whether the parsed structure is imperfect
  for (1..$#{@$x{words}}) {
    if ($$x{normalized}[$_] != $$s{struc}[$_]) {
      die "ERROR: $$x{words}[$_] has governor $$s{struc}[$_]," .
        " should be $$x{normalized}[$_]\n";
    }
  }
}


# read annotations from CDG input
sub read_annos {

  # read annos
  my $all = InputStruct_annotations_get($inputCurrentGrammar);
  for (my $l = hashListOfKeys $all; $l; $l = listNext $l) {
    my $ae = hashGet($all, listElement($l));
    my $x = read_anno $ae;
    push @annos, $x;
  }
  @annos = sort { byVName($$a{id}, $$b{id}) } @annos;

  # maybe re-tag annos
  if ($opt_r && $opt_p) {
    my @allwords;
    my @alltags;

    # collect words
    for my $x (@annos) {
      push @allwords, '';
      push @allwords, @{$x->{words}};
    }

    print "INFO: tagging $#allwords words...\n";

    # tag
    my $out = "/tmp/$$.out";
    my $in = "/tmp/$$.in";
    open(OUT, ">$out") or die "Can't write `$out': $!\n";
    for (@allwords) {
      next if /^NIL$/;
      print OUT "$_\n";
    }
    close(OUT);
    system("$opt_r < $out > $in")
      and die "Can't call tagger `$opt_r': $!\n";
    open(IN, "$in")
      or die "Can't read `$in': $!\n";
    while (<IN>) {
      if(/^[^\t]+\t+([^\t]+)/) {
        unshift @alltags, $1;
      } elsif (/^$/) {
        # do nothing
      } else {
        die "unexpected tagger output `$_'\n";
      }
    }
    close(IN);

    # distribute tags
    for my $x (@annos) {
      for(1..$#{$$x{words}}) {
        $$x{tags}[$_] = pop @alltags;
        print "$$x{words}[$_]/$$x{tags}[$_]\n";
      }
    }
    print "@alltags\n";
  }

}


# read transition table from disk
sub read_transitions {

  print "INFO: reading transition table `$opt_t'\n";

  open(IN, "$opt_t")
    or die "Can't read `$opt_t': $!\n";

  my $count = 0;
  while (<IN>) {

    # recreate settings in effect when this table was created
    if(/shift-reduce.pl (.+)/) {
      my $options = $1;
      while ($options =~ s/-([^ ]+)//) {
        my $o = $1;
        if($o =~ /^(.)(.+)$/) {
          eval "\$opt_$1 = $2";
        } else {
          eval "\$opt_$1 = 1";
        }
      }
    }

    # read state statistics
    next unless /^transition (\d+) ([^:]+): (.+)$/;
    $count++;

    my $i = $1;
    my $sig = $2;
    my $dist = $3;

    # record highest possible dimensionality
    $dim = $i if $i > $dim;

    # default to 0
    for (@actions) {
      $table{$i}{$sig}{$_} = 0;
    }

    while ($dist =~ s/ *([A-Z]+) ([0-9]+)//) {
      my ($a,$c) = ($1,$2);
      $table{$i}{$sig}{$a} = $c;

      # The first noted action is taken to be the preferred action.
      # This only works because we write the actions ordered by frequency!
      $table{$i}{$sig}{action} = $a
        unless exists $table{$i}{$sig}{action};

      # tally total
      $table{$i}{$sig}{total} += $c;

    }
  }
  close(IN);
  print "INFO: read $count transitions.\n";

}

# write transition table to disk
sub write_transitions {
  # create backoff states and write transition table
  open(OUT, ">$opt_t")
    or die "Can't write `>$opt_t': $!\n";
  print OUT "Automatically generated by $invocation\n";

  my $i = $dim;
  while ($i >= 0) {

    # write n-ary transitions
    for (sort keys %{$table{$i}}) {
      print OUT "transition $i $_:";
      for my $t ( sort
                  {
                    $table{$i}{$_}{$b} <=> $table{$i}{$_}{$a}}
                  keys %{$table{$i}{$_}}) {
        print OUT " $t $table{$i}{$_}{$t}";
      }
      print OUT "\n";
    }

    # create (n-1)-ary backoff transitions from n-ary transitions
    if ($i > 0) {
      while (my ($sig, $dist) = each %{$table{$i}}) {
        my $newsig = $sig;
        reduce_context \$newsig;

        for (keys %$dist) {
          $table{$i-1}{$newsig}{$_} += $$dist{$_};
        }
      }
    }
    $i--;
  }

  close(OUT);
}

# MAIN #

# detect attempt to overwrite existing table
if (!$opt_p && -f $opt_t) {
  die "ERROR: won't overwrite transition table `$opt_t'.\n";
}

if(-d $input) {
  chdir $input;
  system "find|grep .cda\$|xargs cat > TREES.$$";
  commandEval "load TREES.$$";
  unlink "TREES.$$";
} else {
  commandEval "load $input";
}

read_annos;

my $i = 0;
if ($opt_p) {

  read_transitions;

  # parse all sentences
  for my $x (@annos) {
    if($opt_n) {
      last if ++$i > $opt_n;
    }
    if ($opt_d) {
      deparse_anno $x;
    }
    parse_anno $x;
  }

  if (defined $good) {
    printf "Overall: %d/%d (%4.2f%%)\n",
      $good, $good+$bad, 100*$good/($good+$bad);
  }

} else {

  # deparse all annotations
  for my $x (@annos) {
    if($opt_n) {
      last if ++$i > $opt_n;
    }
    deparse_anno $x;
  }

  write_transitions;

}
