#!/usr/bin/env perl
=head1 NAME

tagger - call a tagger for each sentence to be tagged

=head1 SYNOPSIS

tagger 
B<-cmd> I<command> 
[B<-debug>] 
[B<-syntax|help|?>] 
[B<-man>]
[I<inputFile1> I<inputFile2> ...]

=head1 DESCRIPTION

This programm is part of the CDG parsing suite.
It is a wrapper to fork a real tagger in the background
for each sentence that is provided to be tagged. A sentence
is read from one or more input files or from standard input if
none is given on the command line. Each sentence is given
one word per line; the end of a sentence is marked by a 
blank line or by the end of the input. 

=head1 OPTIONS

The option I<command> specifies the command of the real tagger.
The option B<-debug> switches on the debug mode where a logfile
is produced in /tmp.

=head1 EXAMPLE

In order to use this wrapper from within the cdgp system
you have to set the appropriate scripting variable I<taggerCommand>
to something like this:

   set taggerCommand 'tagger.pl -c "tnt -v0 -z200 negra.tnt"'

=head1 AUTHORS

Michael Daum E<lt>micha@nats.informatik.uni-hamburg.deE<gt>

=head1 COPYRIGHT

Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 
This file is free software; as a special exception the author gives
unlimited permission to copy and/or distribute it, with or without 
modifications, as long as this notice is preserved.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.


=cut

use Getopt::Long;
use Pod::Usage;
use File::Temp "tempfile";
use warnings;
use strict;
use IO::Handle;

STDOUT->autoflush(1);

my $debug;
my ($debugFileHandle, $debugFileName);

#------------------------------------------------------------------------------
sub printDebug(@) {
  if ($debug) {
    print $debugFileHandle @_;
    print $debugFileHandle "\n";
  }
}

#------------------------------------------------------------------------------
sub main {
  my $man = 0;
  my $help = 0;
  my $taggerCmd;
  my ($tempFileHandle, $tempFileName);

  $help = 1 unless 
    GetOptions(
      "cmd:s" => \$taggerCmd,
      "debug+" => \$debug,
      "syntax|help|?" => \$help,
      "man" => \$man,
    ) || pod2usage(2);

  pod2usage(-exitval =>1, -message => "\nThe tagger wrapper \n") if $help;
  pod2usage(-verbose => 2) if $man;

  die "ERROR: no tagger command specified\n"
    if ! defined $taggerCmd;

  # open a debug file
  if ($debug) {
    ($debugFileHandle, $debugFileName) = 
      tempfile("taggerDebugXXXXXX", SUFFIX => '.log', DIR => "/tmp");
    $debugFileHandle->autoflush(1);
  }

  printDebug "tagger command is $taggerCmd";

  # main loop

  while(my $inLine = <>) {
    chomp($inLine);

    # get a tempfile
    if (! defined $tempFileHandle) {
      ($tempFileHandle, $tempFileName) = 
	tempfile("taggerInputXXXXXX", DIR => "/tmp");
      printDebug "using tempfile $tempFileName";
    }

    # write all into a tempfile til ...
    printDebug "read: $inLine";
    print $tempFileHandle "$inLine\n";

    # ... an empty line signals end of input
    if ($inLine =~ /^$/ or eof) {
      close($tempFileHandle);

      # then open a tagger with that file ...
      open(TAGGER, "$taggerCmd $tempFileName |") or die "couldn't open $taggerCmd\n";
      TAGGER->autoflush(1);

      # ... and read all output the tagger produces and print it to stdout
      while (my $outLine = <TAGGER>) {
	chomp($outLine);
	printDebug "write: $outLine";
	print "$outLine\n";
      }

      # close the tagger, remove the used tempfile and create a new one
      close(TAGGER);
      unlink($tempFileName);
      undef $tempFileHandle;
      undef $tempFileName;
      printDebug "waiting for new sentence";
    }
  }
  printDebug "terminating";
  if (defined $tempFileHandle) {
    unlink($tempFileName);
  }
  if ($debug) {
    close($debugFileHandle);
  }
}

#------------------------------------------------------------------------------

&main();
