#!/usr/bin/env perl
# cdg/grammar/heise/ttc-wrapper.pl

# Call the TreeTagger as a chunker, insert head markers, and transform
# the output to CDG format

use warnings;
use strict;
use File::Temp "tempfile";

our @raw_input;     # what we read from cdgp
our @input;         # raw_input minus some punctuation
our @map;           # from input to raw_input positions

our @output;        # what we write to cdgp

our $prefix = '/data/linux/opt/TreeTagger';

use Getopt::Std;
our ($opt_d,        # give debug output
     $opt_p);       # path to tree tagger

getopts('dp:');

if ($opt_p) {
  $prefix = $opt_p;
}

our $debug = $opt_d;


our $chunker = "$prefix/bin/tree-tagger";
our $format_file = "$prefix/lib/german-chunker.par";
our $chunker_options = "-token -sgml -eps 0.00000001 -quiet";
our ($tmpFile, $tmpFileName) = tempfile("ttc-inputXXXX", DIR => "/tmp");
our $command = "$chunker $format_file $chunker_options < $tmpFileName 2>/dev/null";

our @stack;
our @chunks;

sub noun_like {
  my $w = shift;
  return
    ($$w{tag} =~ /^(NE|NN|FM)$/ &&
     $$w{word} !~ /^(Dutzend|(Mio|Mrd)\.?|Million(en)?|Milliarden?|Prozent|\%)$/);
}
sub min {
  my($x,$y) = @_;
  $x < $y ? $x : $y;
}
sub max {
  my($x,$y) = @_;
  $x > $y ? $x : $y;
}


# read input from cdgp
while (<>) {
  last if /^$/;
  /^([^ ]+)\t([^ ]+)$/ or die "ERROR: Invalid input: `$_'\n";
  push @raw_input, { word=>$1, tag=>$2 };
}

# filter punctuation that confuses the TreeTagger:
# sentence-internal . and "
{
  my($i, $j);
  for (@raw_input) {
    $i++;
    if ($_->{word} =~ /^[\.\"]$/) {
      next;
    }
    $j++;
    $input[$j] = $_;
    $map[$j] = $i;
  }
}

# write to disk
print $tmpFile "$_->{word}-$_->{tag}" for(@input[1..$#input]);
close($tmpFile);

# call chunker
if (!open(CHUNKER, "$command |")) {
  unlink $tmpFileName;
  die "ERROR: Can't call $command: $!\n";
}

sub current_chunk {
  my $i = shift;
  if (@stack) {
    return $stack[-1]
  }
  return { from => $i, to => $i, type => 'XC', head => $i }
}

# parse chunker output
my $i = 1;
my $current_type = '';


sub clear_stack () {
  while (@stack) {
    my $c = pop @stack;
    $c->{to} = $i-1;
    push @chunks, $c;
  }
}

while (my $line = <CHUNKER>) {
  chomp $line;
  unless ($line =~ /^([^ ]+)-[^ ]+\t[^ ]+\/([^ ]+)$/) {
    print STDERR "ERROR: unexpected chunker output `$line', ignoring...\n";
    next;
  }
  my $word = $1;
  my $x = $2;

  # handle illegal I-?C
  if ($x =~ /^I-(.+)$/ &&
      $current_type ne $1) {
    print STDERR "ERROR: illegal chunker output $x, fixing...\n";
    $x = "B-$1";
  }
  $current_type = ($x =~ /-([A-Z]C)$/ ? $1 : '0');

  # handle '0'
  if ($x eq '0') {
    $current_type = '0';
    clear_stack();
    push @chunks, { from => $i, to => $i, type => 'XC' };
  }

  # handle legal B-?C
  elsif ($x =~ /B-([A-Z]C)/) {

    # close previous chunks unless it was a PC enclosing an NC
    unless(@stack &&
           $stack[-1]->{type} eq 'PC' &&
           $current_type eq 'NC') {
      clear_stack();
    }

    # start new chunk
    my $c = { from => $i, type => $current_type };
    push @stack, $c;
  }

  $i++;
}

close(CHUNKER);
#unlink $tmpFileName;
clear_stack();

# correct some systematic errors of TreeTagger

# 1) It often includes words in an NC when this is clearly wrong. E.g.
# it will cheerfully produce a chunk such as this:
#
#  [NC seitdem/ADV *alle/PIDAT]
#
# although an ADV can never ever be in a noun chunk. We correct this
# by breaking NCs at such `stop' words.
my @new;
for my $c (@chunks) {

  if ($$c{from} == $$c{to}) {
    push @new, $c;
    next;
  }

  my @stop = ();
  for ($$c{from}..$$c{to}) {
    # break NC at ADV, PWS, etc.
    if ($$c{type} eq 'NC' &&
        $input[$_]->{tag} =~
        /PWS|PPER|PIS|ADV|PTKZU|ADJD|PRELS|KOKOM|PROAV|KON|TRUNC|PRF/) {
      print STDERR "WARNING: breaking NC at `$input[$_]->{word}'\n";
      push @stop, $_;
    }
    # break any chunk at ( ) - ,
    elsif ($input[$_]->{word} =~ /^[(),-]$/) {
      print STDERR "WARNING: breaking chunk at `$input[$_]->{word}'\n";
      push @stop, $_;
    }
  }

  # break a PC that starts with an APPO
  if ($$c{type} eq 'PC' &&
      $input[$$c{from}]->{tag} =~ /APPO/) {
    print STDERR "WARNING: breaking PC at `$input[$_]->{word}'\n";
    push @stop, $$c{from};
  }


  if (!@stop) {
    push @new, $c;
    next;
  }

  # create partial NC from words preceding the first stop word
  if ($stop[0] < $$c{from}) {
    push @new, { from=>$$c{from}, to=>$stop[0]-1, type => $$c{type} };
  }
  while (@stop) {
    $_ = pop @stop;
    # create XC for stop word
    push @new, { from=>$_, to=>$_, type =>'XC'};
    if ($_ < $$c{to}) {
      # create new partial NC
      my $from = $_ + 1;
      my $to = @stop ? $stop[-1] - 1 : $$c{to};
      push @new, { from=>$from, to=>$to, type=>$$c{type} };
    }
  }
}
@chunks = @new;

# calculate head positions
for my $c (@chunks) {

  # assume right-headedness
  $$c{head} = $$c{to};
  next if $$c{type} eq 'XC';

  # anticipate many TreeTagger errors: in particular, it often builds
  # long NC chunks that cannot possibly be correct. It includes stuff
  # like ADV and PROAV which does not belong there at all. Also, it
  # often includes things like PPER or PWS into a longer chunk
  # although these are always in chunks of their own. We correct all
  # these errors here.

  # NP chunks:
  if ($$c{type} eq 'NC') {
    # NP chunks are usually right-branching.
    # special case: `Microsofts Server' is left-branching.
    # We cannot detect this perfectly, but well enough.
    if ($input[$$c{from}]->{tag} =~ 'NE' &&
        $input[$$c{from}]->{word} =~ /s$/) {
      $$c{head} = $$c{to};
      next;
    }

    # if the chunk is all CARD, assume right-branching
    if ($input[$$c{from}]->{tag} =~ /CARD/ &&
        $input[$$c{to}]->{tag} =~ /CARD/) {
      $$c{head} = $$c{from};
      next;
    }

    # normal case: the first noun-like word from the left is the head.
    for ($$c{from} ..$$c{to}) {
      if (noun_like($input[$_])) {
        $$c{head} = $_;
        last;
      }
    }
  }

  # PP chunks: if there is a preposition, it is the head,
  # otherwise there had better be a postposition at the end,
  # and it will remain the head.
  elsif ($$c{type} eq 'PC') {
    if ($input[$$c{from}]{tag} =~ 'APPR(ART)?') {
      $$c{head} = $$c{from};
    }
  }

  # VC chunks: right-headed unless they begin with a finite verb.
  elsif ($$c{type} eq 'VC') {
    if ($input[$$c{from}]{tag} =~ /FIN$/) {
      $$c{head} = $$c{from};
    }
  } else {
    die "ERROR: cannot compute head for $$c{type} chunk!\n";
  }

  if ($debug) {
    print STDERR "Assigned head to $$c{type} chunk: ";
    for ($$c{from}..$$c{to}) {
      if ($_ eq $$c{head}) {
        print STDERR "*";
      }
      print STDERR "$input[$_]->{word} ";
    }
    print STDERR "\n";
  }

}


sub dump_chunks() {
  for my $c (@chunks) {
    print STDERR "[$$c{type}";
    for ($$c{from}..$$c{to}) {
      print STDERR " $input[$_]->{word}/";
      chomp (my $x = $input[$_]->{tag});
      print STDERR $x;
    }
    print STDERR "]\n";
  }
}

# prepare output

# reverse so that [PC [NC ]] is handled correctly
for my $c (reverse @chunks) {
  for ($$c{from}..$$c{to}) {
    $output[$map[$_]] =
      {
       word => $input[$_]->{word},
       type => $$c{type},
       from => $map[$$c{from}],
       to => $map[$$c{to}],
       head => $map[$$c{head}]
      }
    }
}

# give output
for (1..$#raw_input+1) {
  my $line = $output[$_];
  if(defined $line) {
    print "$line->{word}\t";
    print "$line->{type}\t";
    print "$line->{from}\t";
    print "$line->{to}\t";
    print "$line->{head}\n";
  }
  else {
    print "$raw_input[$_-1]->{word}\tXC\t$_\t$_\t$_\n";
  }

}

