# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## VisDirectory - a directory browser
##
## \author Michael Daum
##
## $Id: VisDirectory.tcl,v 1.10 2004/02/25 14:42:09 micha Exp $
## ----------------------------------------------------------------------------
class VisDirectory {
  inherit iwidgets::Scrolledcanvas

  # -- variables -------------------------------------------------------------
  public variable chdircommand ""
  public variable selectioncommand ""
  public variable finishcommand ""

  private variable _fileObjects {}
  private variable _selectedFiles {}
  private variable _maxWidth 70
  private variable _maxHeight 0
  private variable _filter "*"
  private variable _showHidden 1
  private variable _selectionAnchor ""
  private variable _browseMode "multi"
  private variable _fileMode "both"
  private variable _doneFlag 0
  
  # -- methods ---------------------------------------------------------------
  public method show {dirName}; ## \type TclString
  public method hide {}
  public method getMaxFileWidth {}
  public method getMaxFileHeight {}
  public method getSelection {}
  public method getFilter {}
  public method getShowHidden {}
  public method getMode {}
  public method setFilter {filter}; ## \type TclString
  public method setShowHidden {showHidden}; ## \type TclNumber
  public method setMode {mode}; ## \type TclString
  public method clearSelection {}
  public method select {fileNames}; ## \type TclList
  public method new {{dirName ""}}; ## \type TclString

  constructor {args} {}; ## \type TclList
  destructor {}

  private method _cmpFiles {fileObj1 fileObj2}; ## \type VisFile, VisFile
  private method _draw {}
  private method _b1Click {}
  private method _ctrlB1Click {}
  private method _shiftB1Click {}
  private method _doubleB1Click {}
  private method _select {fileObjs}; ## \type TclList
  private method _unselect {fileObjs}; ## \type TclList
  private method _entryOkCommand {}
  private method _entryAbortCommand {}
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body VisDirectory::constructor {args} {

  itk_component add renameEntry {
    entry $itk_interior.renameEntry \
      -borderwidth 2
  } {}

  # binding
  ::bind $itk_component(canvas) <Configure> [code $this _draw]
  ::bind $itk_component(renameEntry) <Return> [code $this _entryOkCommand]
  ::bind $itk_component(renameEntry) <Escape> [code $this _entryAbortCommand]
  
  # canvas bindings
  bind all <ButtonPress-1> [code $this _b1Click]
  bind all <ButtonPress-2> [code $this _ctrlB1Click]
  bind all <ButtonPress-3> [code $this _shiftB1Click]
  bind all <Control-ButtonPress-1> [code $this _ctrlB1Click]
  bind all <Shift-ButtonPress-1> [code $this _shiftB1Click]
  bind all <Double-Button-1> [code $this _doubleB1Click]
  

  eval itk_initialize $args
}

## ----------------------------------------------------------------------------
## destructor
## ----------------------------------------------------------------------------
body VisDirectory::destructor {} {
  hide
}

## ----------------------------------------------------------------------------
## hide
## ----------------------------------------------------------------------------
body VisDirectory::hide {} {
  foreach obj $_fileObjects {
    itcl::delete object $obj
  }
  set _fileObjects ""
  set _selectedFiles ""
  set _selectionAnchor ""
  set _maxWidth 70
  set _maxHeight 0
}

## ----------------------------------------------------------------------------
## show
## ----------------------------------------------------------------------------
body VisDirectory::show {dirName} {

  # clear the previous view
  hide

  # change the current directory to be shown
  cd $dirName
  set dirName [pwd]

  # get all files in the actual directory
  set _fileObjects ""
  foreach file [eval glob -nocomplain .* *] {
    if {[info exists allFiles($file)]} {
      continue
    }

    if {$dirName == "/" && ( $file == "." || $file == ".." )} {
      continue
    }

    if {!$_showHidden && \
	$file != "." && $file != ".." && \
	[string index $file 0] == "."} {
      continue
    } 

    if {[file isdirectory $file]} {
      set icon [YadaImages::get defaultFolder]
      set type "directory"
    } else {
      if {$_fileMode == "dir"} {
	continue
      }

      set doesMatch 0
      foreach pattern [split $_filter] {
	if {[string match $pattern $file]} {
	  set doesMatch 1
	  break
	}
      }
      if {!$doesMatch} {
	continue
      }

      set icon [YadaImages::get textFile]
      set type "file"
    }

    set fileObj [VisFile #auto $this \
      -name $file \
      -icon $icon \
      -type $type]

    lappend _fileObjects [namespace which $fileObj]
    set allFiles($file) 1
    
    # keep track of the maximum width and height of VisFiles
    set width [$fileObj getWidth]
    if {$width > $_maxWidth} {
      set _maxWidth $width
    }
    set height [$fileObj getHeight]
    if {$height > $_maxHeight} {
      set _maxHeight $height
    }
  }
  
  set _fileObjects [lsort -command [code $this _cmpFiles] $_fileObjects]
  set _selectionAnchor [lindex $_fileObjects 0]
  _draw

  # call the chdir handle
  $chdircommand
}

## ----------------------------------------------------------------------------
## _cmpFiles
## ----------------------------------------------------------------------------
body VisDirectory::_cmpFiles {fileObj1 fileObj2} {
  set file1 [$fileObj1 cget -name]
  set file2 [$fileObj2 cget -name]

  if {$file1 == "."} {
    return -1
  }
  if {$file2 == "."} {
    return 1
  }
  if {$file1 == ".."} {
    return -1
  }
  if {$file2 == ".."} {
    return 1
  }

  if {[file isdirectory $file1]} {
    if {[file isdirectory $file2]} {
      return [string compare $file1 $file2]
    } else {
      return -1
    }
  } 

  if {[file isdirectory $file2]} {
    return 1
  } else {
    return [string compare $file1 $file2]
  }
}

## ----------------------------------------------------------------------------
## _draw
## ----------------------------------------------------------------------------
body VisDirectory::_draw {} {

  set canvas $itk_component(canvas)
  set canvasHeight [expr [winfo height $canvas] - $_maxHeight]
  if {$canvasHeight < 0} {
    return
  }

  set padding [expr [$canvas cget -highlightthickness] + [$canvas cget -bd]]
  if {$padding < 2} {
      set padding 2
  }
  set x $padding
  set y $padding
  set dx [expr $_maxWidth + $padding]
  set dy [expr $_maxHeight + $padding]

  set canvasHeight [expr $canvasHeight - 2 * $padding]

  foreach fileObj $_fileObjects {

    $fileObj setCoords $x $y
    $fileObj show

    set y [expr $y + $dy]
    if {$y > $canvasHeight} {
      set y $padding
      set x [expr $x + $dx]
    }
  }

  # adjust the view 
  configure -scrollregion [bbox all]
  xview moveto 0
}

## ----------------------------------------------------------------------------
## _shiftB1Click
## ----------------------------------------------------------------------------
body VisDirectory::_shiftB1Click {} {
  if {$_browseMode == "single"} {
    return [_b1Click]
  }

  if {$_selectionAnchor == ""} {
    _ctrlB1Click
    return
  }

  set oldAnchor $_selectionAnchor
  clearSelection
  set _selectionAnchor $oldAnchor
  set id [find withtag current]
  set tags [itemcget $id -tags]
  set currentFileObj [lindex $tags 0]
  set fromIndex [lsearch $_fileObjects $_selectionAnchor]
  set toIndex [lsearch $_fileObjects $currentFileObj]

  if {$fromIndex > $toIndex} {
    set tmp $toIndex
    set toIndex $fromIndex
    set fromIndex $tmp
  }

  set selectedRange [lrange $_fileObjects $fromIndex $toIndex]

  _select $selectedRange
}

## ----------------------------------------------------------------------------
## _ctrlB1Click
## ----------------------------------------------------------------------------
body VisDirectory::_ctrlB1Click {} {
  if {$_browseMode == "single"} {
    return [_b1Click]
  }
  
  set id [find withtag current]
  set tags [itemcget $id -tags]
  set fileObj [lindex $tags 0]
  
  if {[$fileObj getSelection] == 0} {  
    _select $fileObj
    set _selectionAnchor $fileObj
  } else {
    _unselect $fileObj
  }

}

## ----------------------------------------------------------------------------
## _b1Click
## ----------------------------------------------------------------------------
body VisDirectory::_b1Click {} {
  set id [find withtag current]
  set tags [itemcget $id -tags]
  set fileObj [lindex $tags 0]

  set wasSelected [$fileObj getSelection]
  clearSelection 
  if {$wasSelected} {
    _unselect $fileObj
  } else {
    _select $fileObj
    set _selectionAnchor $fileObj
  }
}

## ----------------------------------------------------------------------------
## _doubleB1CLick
## ----------------------------------------------------------------------------
body VisDirectory::_doubleB1Click {} {
  set id [find withtag current]
  set tags [itemcget $id -tags]
  set fileObj [lindex $tags 0]
  set fileName [$fileObj cget -name]
  clearSelection 

  # change directory
  if {[$fileObj cget -type] == "directory"} {
    show $fileName
  } else {
    # select file
    _select $fileObj
    set _selectionAnchor $fileObj
    $finishcommand
  }
}

## ----------------------------------------------------------------------------
## getMaxFileWidth
## ----------------------------------------------------------------------------
body VisDirectory::getMaxFileWidth {} {
  return $_maxWidth
}

## ----------------------------------------------------------------------------
## getMaxFileHeight
## ----------------------------------------------------------------------------
body VisDirectory::getMaxFileHeight {} {
  return $_maxHeight
}

## ----------------------------------------------------------------------------
## clearSelection
## ----------------------------------------------------------------------------
body VisDirectory::clearSelection {} {
  _unselect $_fileObjects
}

## ----------------------------------------------------------------------------
## getSelection
## ----------------------------------------------------------------------------
body VisDirectory::getSelection {} {
  set selection ""
  foreach fileObj $_selectedFiles {
    lappend selection [$fileObj cget -name]
  }

  return $selection
}

## ----------------------------------------------------------------------------
## setFilter
## ----------------------------------------------------------------------------
body VisDirectory::setFilter {filter} {
  set _filter $filter
  clearSelection
}

## ----------------------------------------------------------------------------
## getFilter
## ----------------------------------------------------------------------------
body VisDirectory::getFilter {} {
  return $_filter
}

## ----------------------------------------------------------------------------
## setShowHidden
## ----------------------------------------------------------------------------
body VisDirectory::setShowHidden {showHidden} {
  set _showHidden $showHidden
  clearSelection
}

## ----------------------------------------------------------------------------
## getShowHidden
## ----------------------------------------------------------------------------
body VisDirectory::getShowHidden {} {
  return $_showHidden
}

## ----------------------------------------------------------------------------
## _select
## ----------------------------------------------------------------------------
body VisDirectory::_select {fileObjs} {
  foreach fileObj $fileObjs {
    $fileObj setSelection 1

    if {($_fileMode == "file" && [$fileObj cget -type] == "directory") ||
	($_fileMode == "dir" && [$fileObj cget -type] == "file")} {
      continue
    }

    if {[lsearch $_selectedFiles $fileObj] < 0} {
      lappend _selectedFiles $fileObj
    }
  }

  $selectioncommand
}

## ----------------------------------------------------------------------------
## _unselect
## ----------------------------------------------------------------------------
body VisDirectory::_unselect {fileObjs} {
  foreach fileObj $fileObjs {
    $fileObj setSelection 0
    set index [lsearch $_selectedFiles $fileObj]

    if {$index >= 0} {
      set _selectedFiles [lreplace $_selectedFiles $index $index]
    }
  }

  if {$_selectedFiles != ""} {
    set _selectionAnchor [lindex $_selectedFiles end]
  } else {
    set _selectionAnchor ""
  }
  $selectioncommand
}

## ----------------------------------------------------------------------------
## getMode
## ----------------------------------------------------------------------------
body VisDirectory::getMode {} {
  return $_mode
}

## ----------------------------------------------------------------------------
## set the selection mode.
## In effect the variables _browseMode and _fileMode are set.
## \param mode possible values are
## - multi: lets you select multiple files and directories
## - single: lets you select one file or directory
## - multidir: lets you select multiple directories
## - multifile: lets you select multiple files
## - singledir: lets you select one directory
## - singlefile: lets you select one file
## ----------------------------------------------------------------------------
body VisDirectory::setMode {mode} {

  switch $mode {
    "multi" {
      set _browseMode "multi"
      set _fileMode "both"
    }
    "single" {
      set _browseMode "single"
      set _fileMode "both"
    }
    "multidir" {
      set _browseMode "multi"
      set _fileMode "dir"
    }
    "multifile" {
      set _browseMode "multi"
      set _fileMode "file"
    }
    "singledir" {
      set _browseMode "single"
      set _fileMode "dir"
    }
    "singlefile" {
      set _browseMode "single"
      set _fileMode "file"
    }
    default {
      error "ERROR: unknown mode `$mode'"
    }
  }
}

## ----------------------------------------------------------------------------
## select
## ----------------------------------------------------------------------------
body VisDirectory::select {fileNames} {
  clearSelection

  set toBeSelectedObjects ""
  foreach fileObj $_fileObjects {
    set fileName [file join [pwd] [$fileObj cget -name]]
    if {[lsearch $fileNames $fileName] >= 0} {
      lappend toBeSelectedObjects $fileObj
    }
  }

  _select $toBeSelectedObjects
  set _selectionAnchor [lindex $_selectedFiles end]
}

## ----------------------------------------------------------------------------
## new
## ----------------------------------------------------------------------------
body VisDirectory::new {{dirName ""}} {
  if {$dirName == ""} {

    # get the next file position
    set lastFileObj [lindex $_fileObjects end]
    ::scan [$lastFileObj getCoords] "%f %d" x y
    set canvas $itk_component(canvas)
    set padding [expr [$canvas cget -highlightthickness] + [$canvas cget -bd]]
    if {$padding < 2} {
	set padding 2
    }
    set dx [expr $_maxWidth + $padding]
    set dy [expr $_maxHeight + $padding]
    set canvasHeight [expr [winfo height $canvas] - $_maxHeight - 2 * $padding]
    set y [expr $y + $dy]
    if {$y > $canvasHeight} {
      set y $padding
      set x [expr $x + $dx]
    }

    set _doneFlag 0
    $itk_component(renameEntry) delete 0 end
    $itk_component(renameEntry) insert  0 "NewFolder"
    set entryTag [\
      create window $x $y \
	-anchor nw \
	-width $_maxWidth \
	-window $itk_component(renameEntry)]
    ::focus $itk_component(renameEntry)

    vwait [scope _doneFlag]
    delete $entryTag
    set dirName [$itk_component(renameEntry) get]

    if {$_doneFlag == 2 || $dirName =="" } {
      return
    } else {
      set _doneFlag 0
    }
  } else {
    set baseDir [file dirname $dirName]
    show $baseDir
  }

  file mkdir $dirName
  show [pwd]
}

## ----------------------------------------------------------------------------
## _entryOkCommand
## ----------------------------------------------------------------------------
body VisDirectory::_entryOkCommand {} {
  set _doneFlag 1
}

## ----------------------------------------------------------------------------
## _entryAbortCommand
## ----------------------------------------------------------------------------
body VisDirectory::_entryAbortCommand {} {
  set _doneFlag 2
}
