# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaJob - bundled information about a yada job.
## A YadaJob is being created with the help of a YadaExperiment and organizes the
## computation of a wordgraph on a YadaMachine.
## \ingroup YadaScheduler
##
## \author Michael Daum
##
## $Id: YadaJob.tcl,v 1.7 2004/02/25 14:42:08 micha Exp $
## ----------------------------------------------------------------------------
class YadaJob {

  ## the name of the YadaExperiment for which we created the job.
  public variable experiment ""

  ## the wordgraph which we are going to start YadaJob::experiment on
  public variable wordgraphName ""

  ## the YadaRunnableDocument that is responsible for this job.
  ## especialy all messages that want to be printed are displayed on this
  ## document.
  public variable runner ""

  ## consumber callback.
  ## After the job has finished we call this tcl command 
  ## \code $consumerCommand $thisJon \endcode
  ## to process the results. \see Process::consumerCommand
  public variable consumerCommand ""

  ## the CdgProcess that is connected to a running job.
  public variable process ""

  ## the YadaMachine on which we run this job.
  public variable machine ""

  ## the actual node of the machine that this job runs on.
  public variable node ""

  ## state of the job.
  ## A YadaJob passes a series of states from creation time of this
  ## object til its destruction. Possible values are:
  ## - created
  ## - dequeuing
  ## - running
  ## - consuming
  ## - stopping
  private variable _state "created"

  ## list of passed states. \see _state
  private variable _passedStates ""

  # methods ------------------------------------------------------------------
  public method setState {state}; ## \type TclString
  public method getState {}
  public method getPassedStates {}
  public method hasPassedState {state}; ## \type TclString

  constructor {args} {}; ## \type TclList
  destructor {}
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaJob::constructor {args} {

  eval configure $args
}

## ----------------------------------------------------------------------------
## destructor
## ----------------------------------------------------------------------------
body YadaJob::destructor {} {
  if {$process != ""} {
    catch {itcl::delete object $process}
  }
}

## ----------------------------------------------------------------------------
## setStatus
## ----------------------------------------------------------------------------
body YadaJob::setState {state} {
  lappend _passedStates $_state
  set _state $state
  if {0} {
    if {$process != ""} {
      set printCommand [$process cget -printCommand]
      if {$printCommand != ""} {
	$printCommand "INFO: $wordgraphName - $_state\n"
      }
    }
  }
}

## ----------------------------------------------------------------------------
## getState
## ----------------------------------------------------------------------------
body YadaJob::getState {} {
  return $_state
}

## ----------------------------------------------------------------------------
## getPassedStates
## ----------------------------------------------------------------------------
body YadaJob::getPassedStates {} {
  return $_passedStates
}

## ----------------------------------------------------------------------------
## hasPassedState
## ----------------------------------------------------------------------------
body YadaJob::hasPassedState {state} {
  return [expr [lsearch $_passedStates $state] >= 0]
}
