# Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
#  
# This file is free software; as a special exception the author gives
# unlimited permission to copy and/or distribute it, with or without 
# modifications, as long as this notice is preserved.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

## ----------------------------------------------------------------------------
## YadaLog - a text output widget.
## This class defines a scrollable text window used to display the informations,
## warnings and errors that occure during YADA.
##
## \author Michael Daum
##
## $Id: YadaLog.tcl,v 1.9 2004/02/25 14:42:09 micha Exp $
## ----------------------------------------------------------------------------
class YadaLog {
  inherit iwidgets::Scrolledtext

  # variables ----------------------------------------------------------------

  ## list of highlightened patterns in the text.
  private variable _tagging {}

  ## flag used in print().
  private variable _needsLineFeed 0

  ## flag indicating whether new text should be focused or not.
  public variable followData 1

  # methods ------------------------------------------------------------------
  public method autotag {pattern args}; ## \type TclString, TclList
  public method print {message}; ## \type TclString
  public method clear {}

  constructor {args} {}; ## \type TclList
};

## ----------------------------------------------------------------------------
## constructor
## ----------------------------------------------------------------------------
body YadaLog::constructor {args} {
  eval itk_initialize $args
}

## ----------------------------------------------------------------------------
## print a message on the widget.
## The text is added to the end of the already printed text and highlightened
## according to the configured autotags. \see autotag(). After inserting the
## text tcltk is given an update loop to refresh its widges.
## \param message the text to be displayed.
## ----------------------------------------------------------------------------
body YadaLog::print {message} {

   # remember first position
   ::scan [index end] %d logFirst
   incr logFirst -1

  # split string by formfeeds if available

  set splitMessages [split "$message" "\r"]  
  set lastMessage [lindex $splitMessages end]
  set splitMessages [lreplace $splitMessages end end]

  foreach splitMessage $splitMessages {
    if {$_needsLineFeed} {
      mark set insert end
      delete "insert linestart" "insert lineend"
    }
    insert end "$splitMessage"
  }

  if {$lastMessage == ""} {
    set _needsLineFeed 1
  } else {
    if {$_needsLineFeed} {
      mark set insert end
      delete "insert linestart" "insert lineend"
    }
    insert end "$lastMessage"
    set _needsLineFeed 0
  }

  # remember last position
  ::scan [index end] %d logLast

  # apply autotagging
  foreach pattern $_tagging {
    for {set i $logFirst} {$i < $logLast} {incr i} {
       mark set last $i.0
       while {[regexp -indices $pattern [get last "last lineend"] indices]} {
	  
	  mark set first "last + [lindex $indices 0] chars"
	  mark set last  "last + 1 chars + [lindex $indices 1] chars"
	  tag add $pattern first last
       }
    }
  }

  if {$followData} {
    see end
  }
  update
}

## ----------------------------------------------------------------------------
## clear the text.
## This simply deletes all previousply printed text.
## ----------------------------------------------------------------------------
body YadaLog::clear {} {
  delete 1.0 end 
}

## ----------------------------------------------------------------------------
## configure syntax highlightning.
## This method lets you define a pattern as a regular expression and associate
## tag configurations (see tk::text).
## \param pattern a regular expression matching a part of printed text.
## \param args list of tag configurations in the text widget
## ----------------------------------------------------------------------------
body YadaLog::autotag {pattern args} {
   lappend _tagging $pattern
   eval tag configure $pattern $args
}

