/* 
 * Copyright (C) 1997-2004 The CDG Team <cdg@nats.informatik.uni-hamburg.de>
 *  
 * This file is free software; as a special exception the author gives
 * unlimited permission to copy and/or distribute it, with or without 
 * modifications, as long as this notice is preserved.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY, to the extent permitted by law; without even the
 * implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

/* ---------------------------------------------------------------------------
 * @addtogroup MGLS MGLS - Multiple Guided Local Search
 * @author Michael Daum
 * $Id: gls.c,v 1.85 2004/09/27 17:07:04 micha Exp $
 *
 * @{
 */

/* -- INCLUDES ---------------------------------------------------------- */
#include <config.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <blah.h>
#include "gls.h"
#include "eval.h"
#include "write.h"
#include "cdg.h"
#include "hook.h"
#include "timer.h"
#include "constraintnet.h"
#include "parse.h"

/* -- MACROS ------------------------------------------------------------ */

/* command line errors */
#define CMDGLS_NO_ERROR 0
#define CMDGLS_NOT_ENUF_PARAMETERS -1
#define CMDGLS_EXPECTED_FLOAT -2
#define CMDGLS_EXPECTED_INT -3
#define CMDGLS_CANNOT_PARSE_FLOAT -4
#define CMDGLS_CANNOT_PARSE_INT -5
#define CMDGLS_EXPECTED_BOOLEAN -6
#define CMDGLS_CANNOT_PARSE_BOOLEAN -7
#define CMDGLS_HELP -8
#define CMDGLS_VERSION -9
#define CMDGLS_UNKNOWN_NET -10
#define CMDGLS_UNKNOWN_OPTION -11
#define CMDGLS_NO_NET -12
#define CMDGLS_CANNOT_PARSE_PARAMETER -13
#define CMDGLS_EXPECTED_PARAMETER -14
#define CMDGLS_EXPECTED_FILENAME -15
#define CMDGLS_FILEERROR -16
#define CMDGLS_UNKNOWN_CMP -17
#define CMDGLS_EXPECTED_CMP -18
#define CMDGLS_EXPECTED_INITMODE -19
#define CMDGLS_UNKNOWN_INITMODE -20
#define CMDGLS_UNKNOWN_GLSMODE -22
#define CMDGLS_DEFAULT_ERROR -99


/* compare functions */
#define GLS_CMP_BADNESS 0
#define GLS_CMP_COSTS 1

/* init modes */
#define GLS_INIT_BEST 0
#define GLS_INIT_NORMAL 1
#define GLS_INIT_ROOT 2
#define GLS_INIT_LAST 3

/* constants */
#define GLS_BUFFERSIZE 255
#define GLS_DEBUG1 1 /* deletions */
#define GLS_DEBUG2 2 /* penalzing */
#define GLS_DEBUG3 3 /* job queuing */
#define GLS_DEBUG4 4 /* repairing, freezing bindings */
#define GLS_DEBUG5 5 /* unused */
#define GLS_DEBUG6 6 /* detour */

/* code flags */
#define INLINE inline

#define SECURE_COPYGLSSEARCHSTATE 0
#define SECURE_GLSPRINTBINDING 0
#define SECURE_NEWCOMPNODE 0
#define DEBUG_NODE 0

#define MY_MAXDOUBLE 999999.99


/* -- TYPE DEFINITIONS -------------------------------------------------- */

/* forward declaration */
struct GlsNetStruct;
struct GlsDomainStruct;
struct GlsBindingStruct;
struct GlsTimepointStruct;

/* ---------------------------------------------------------------------- 
 * GlsTag:
 *	tag the state of values, value pairs and domains
 */
typedef enum {
  GlsDefault = 0,	     /* no specific activation */
  GlsActive = (1L << 0),     /* to be processed any further */
  GlsChanged = (1L << 1),    /* changed in a job */
  GlsFrozen = (1L << 2),     /* cannot be processed any more */
  GlsSorted = (1L << 3),     /* tagged item is sorted correctly within its container */
  GlsPenalized = (1L << 4),   /* tagged item has been penalized */
  GlsImproved = (1L << 5),    /* tagged item improved */
  GlsDeleted = (1L << 6)      /* tagged item is useless */
} GlsTag;

/* ---------------------------------------------------------------------- 
 * GlsPhase
 *	tag the state of the search efford on a gls net
 */
typedef enum {
  GlsInit,		     /* we are in init state */
  GlsProlong,		     /* we prolongating */
  GlsTerminate		     /* we try to find the end of the show */
} GlsPhase;

/* ---------------------------------------------------------------------- 
 * A GlsJobType is used to mark the type of a repairment of a GlsBinding
 */
typedef enum {
  GlsNoJob,		     /**< no job needed */
  GlsUnaryJob,		     /**< reassign a binding */
  GlsModifierJob,	     /**< reassign a binding by only exchanging the
			          modifier */
  GlsModifieeJob	     /**< reassign a binding by only exchanging the
			          modifiee */ ,
} GlsJobType;

/* ---------------------------------------------------------------------- 
 * GlsValuePair: 
 *	information about a pair of levelvalues 
 */
typedef struct GlsValuePairStruct {

  int index;		     /* identification of a value pair */

  /* workload */
  int value1Index;	     /* pair of values */
  int value2Index;
  int node1Index;
  int node2Index;

  /* bookkeeping */
  Number score;		     /* score of value-pair */
  Number costs;		     /* costs of value-pair */

  /* information */
  int noSoftViolations;	     /* number of soft constraints violated by this value pair */
  int noHardViolations;	     /* number of hard constraints violated by this value pair */
  List violations;	     /* list of constraints violations */

  /* extra structuring */
  struct GlsValuePairStruct *next;	/* pointer to next value pair */

} GlsValuePairStruct;
typedef GlsValuePairStruct *GlsValuePair;


/* ---------------------------------------------------------------------- 
 * GlsValue:
 *	information about one levelvalue gathered during search 
 */
typedef struct GlsValueStruct {
  LevelValueStruct lv;	     /* levelvalue */

  GlsTag tag;
  /* 
   * state of value: 
   * 
   *   GlsDefault   = no information specified
   *   GlsActive    = used in current analysis
   *   GlsChanged   = <not used>
   *   GlsFrozen    = <not used>
   *   GlsSorted    = the value is placed in its correct position in the domain
   *   GlsPenalized = <not used>
   *   GlsImproved  = <not used>
   *   GlsDeleted   = the value is not part of the domain any more
   */

  int index;		     /* identification of a value */

  /* extra structuring */
  struct GlsValueStruct *prev;	/* values double-linked */
  struct GlsValueStruct *next;

  /* upwards links */
  struct GlsDomainStruct *domain;	/* every glsValue belongs to a domain */
  ConstraintNode constraintNode;	/* every glsValue belongs to a constraint node */

  /* workload */
  Number penalty;	     /* weight to punish usage */
  Number reinforcement;	     /* weight to encourage usage */
  Number costs;		     /* costs of value */
  Number augCosts;	     /* augmented costs of value */
  int noSoftViolations;	     /* number of soft constraints violated by this value */

} GlsValueStruct;
typedef GlsValueStruct *GlsValue;

/* ---------------------------------------------------------------------- 
 * GlsDomain:
 *	domain of values that can be bound
 */
typedef struct GlsDomainStruct {

  int index;		     /** identification of a domain, the index in GlsNet->domains */

  /* upwards links */
  struct GlsNetStruct *net;  /** every gls domain belongs to a gls net */

  /* value domain */
  int noValues;		     /** number of values */
  GlsValue firstValue;	     /** first (=best) value in domain */
  GlsValue lastValue;	     /** last value in domain */
  GlsValue annoValue;        /** corresponds to annotation */

  /* information */
  ConstraintNode constraintNode; /** example constraint node in the nodes list */
  List constraintNodes;	     /** list of unified constraint nodes */
  List lexemes;		     /** list of unified corresponding lexemes */

} GlsDomainStruct;
typedef GlsDomainStruct *GlsDomain;

/* ---------------------------------------------------------------------- 
 * GlsState
 * 
 */
typedef struct {

  /* upwards links */
  struct GlsNetStruct *net;  /* every state belongs to a gls net */

  /* main components */

  struct GlsBindingStruct *firstBinding; /** first one in the set of bindings */
  Vector bindings;	     /* collection of GlsBindings */
  Vector context;	     /* same but using pure level values */

  List conflictPairs;	     /* conflicting value pairs */
  List violations;	     /* list of non-local constraint violations */
  RingBuffer allJobs;	     /* queued jobs */
  RingBuffer interimJobs;    /* jobs which must be done to fullfill a
			      * complete job */
  ByteVector isDeletedLexem;  /* local lexem information */

  /* bookkeeping */
  Number costs;		     /* costs of the solution candidate */
  Number augCosts;	     /* augmented costs */
  Number maxCostsSoFar;      /* maximum costs used so far */
  Number score;		     /* score of candidate */
  int noSoftViolations;	     /* number of soft violations */
  int noHardViolations;	     /* number of hard violations */

  /* information */
  int noFrozenBindings;	     /* number of frozen bindings */
  Number maxUtility;	     /* maximum utility in all bindings */
  Number utility;	     /* utility of the search state */

} GlsStateStruct;
typedef GlsStateStruct *GlsState;

/* ---------------------------------------------------------------------- 
 * GlsBinding: 
 *	information on one assignment 
 */
typedef struct GlsBindingStruct {

  GlsTag tag;
  /* 
   * state of binding:
   * 
   *   GlsDefault   = no information specified
   *   GlsActive    = <not used>
   *   GlsChanged   = another value has been bound
   *   GlsFrozen    = no alternatives are tried any more
   *   GlsSorted    = <not used>
   *   GlsPenalized = the used value has been penalized
   *   GlsImproved  = <not used>
   *   GlsDeleted   = <not used>
   */

  int index;		     /* identification in its container */

  /* extra structuring */
  struct GlsBindingStruct *next;	/* pointer to next binding */

  /* upwards links */
  GlsState state;	/* every binding belongs to a state */

  /* workload */
  GlsDomain domain;	     /* domain of values to a bind */
  GlsValue value;            /* current binding */
  GlsJobType jobType;	     /* the kind of a scheduled repairment */

  /* bookkeeping */
  Number costs;		     /* cost contribution by the whole solution candidate */
  Number augCosts;	     /* augmented costs of this binding */
  Number utility;	     /* utility to repair this binding */

} GlsBindingStruct;
typedef GlsBindingStruct *GlsBinding;

/* ---------------------------------------------------------------------- 
 * GlsStatRecord
 *	facts we want to collect in every cycle of a gls search
 */
typedef struct GlsStatRecordStruct {
  unsigned long cycle;	     /* cycle which this record describes */
  struct GlsNetStruct *net;  /* pointer to the net which generated this record */
  Boolean isSolution;	     /* indicate a cycle with a solution */
  Number score;		     /* actual score */
  Number utility;	     /* actual utility */
  Number minUtility;         /* minimal utility ever needed */
  Number maxUtility;         /* maximum utility ever reached */
  Number costs;		     /* costs */
  Number augCosts;	     /* augmented costs */
  Number maxCosts;           /* maximum augmented costs needed */
  Number errorRate;          /* ratio of erroneous edges */
  Number minErrorRate;       /* minimal ratio of erroneous edges */
  unsigned long maxDetour;  /* longest detour needed */
  unsigned long noBlindAlleys;	/* number of blind alleys we trapped in during this cycle */
  unsigned long noHardViolations;	/* number of hard violations */
  unsigned long noSoftViolations;	/* number of soft violations */
  unsigned long noRepairs;   /* numbers of repairs til now */
  unsigned long time;	     /* time spent til now */
  long eta;                  /* estimated time of arrival */
  unsigned long noPenalties; /* number of penalties applied to the net til now */
  Parse parse;		     /* partial result we have now */
  ByteVector isDeletedLexem;  /* copy of the according state property */

  struct GlsStatRecordStruct *next;	/* extra structuring */
  struct GlsStatRecordStruct *prev;	

  struct GlsStatRecordStruct *prevSolution; /* link each record to the previous solution */

} GlsStatRecordStruct;
typedef GlsStatRecordStruct *GlsStatRecord;

/* ---------------------------------------------------------------------- 
 * GlsNet: 
 *	extension of a constraint net 
 */
typedef struct GlsNetStruct {

  GlsTag tag;
  /* 
   * state of network:
   * 
   *   GlsDefault   = no information specified
   *   GlsActive    = the constraint net is valid
   *   GlsChanged   = the net must be adapted to the best net in the pool
   *   GlsFrozen    = the constraint has expired
   *   GlsSorted    = <not used>
   *   GlsPenalized = devaluation happened
   *   GlsImproved  = a better searchstate was found in the last cycle
   *   GlsDeleted   = <not used>
   */

  int index;			/**< identification in the pool */

  /* extra structuring */
  struct GlsNetStruct *next;	/**< pointer to the next network in the pool */

  /* workload */
  Vector domains;		/**< vector of gls domains */
  Vector values;		/**< vector of all values */
  Vector timePoints;		/**< vector of all timepoints */
  LexemGraph lexemgraph;	/**< local lexem information */
  Vector noModifiersOfLexem;	/**< reference counters of lexemes being used as modifiers */
  GlsState state;		/**< actual search state */
  GlsState bestState;		/**< local best search state so far */

  /* parameter */
  Number beta;			/**< normalization factor */
  Number cutoff;		/**< cutoff of cost function */
  Number lambda1;		/**< strength of penalties */
  Number lambda2;		/**< strength of reinforcement */

  unsigned long fromCycle;	/**< start computing at this cycle */
  unsigned long toCycle;	/**< end of this net */
  unsigned long maxDetour;	/**< maximum number of cycles without improovment */
  unsigned long maxTime;	/**< number of milliseconds to try with this net */
  Number initialUnaryPruning;	/**< max costs of domains to be pruned in init phase */
  Number prolongUnaryPruning;	/**< max costs of domains to be pruned in prolongation phase */
  Number terminationUnaryPruning;	/**< max costs of domains to be pruned in termination phase */
  Number maxCosts;		/**< maximum allowed costs */
  Number maxUtility;		/**< maximum utility of a termination */
  Number toleranceFraction;	/**< tolerated utility of a binding to be repaired */
  unsigned long priority;	/**< number of cycles this net is scheduled */

  /* information */
  unsigned long currentDetour;	/**< current number of cycles without improvments */
  int noValidValues;		/**< number of values contained in a domain somewhere */
  int noPenalties;		/**< number of all applied penalties */
  GlsStatRecord statRecord;	/**< actual statistics on this net */
  GlsPhase phase;		/**< mark the state of the search progress */

} GlsNetStruct;
typedef GlsNetStruct *GlsNet; /**< type definition of a GlsNet */

/* ---------------------------------------------------------------------- 
 * GlsLexem:
 *	This is a derivation of a LexemNode used in the MGLS module,
 *	enriching it with weights the same way as GlsValues are derived
 *	from LevelValues. So you can cast a GlsLexem to a LexemNode safely.
 *	Each GlsLexem is part of a GlsTimepoint. A GlsTimepoint owns
 *	all GlsLexem objects during the same timepoint in the utterance.
 *	All GlsLexem objects are double linked to each other and sorted
 *	according to their penalization to speed up the search of the
 *	best path thru all timepoints.
 */
typedef struct GlsLexemStruct {

  /* derivation of a LexemNode */
  LexemNodeStruct lexemNode;	/**< copy of the LexemNode */

  /* workload */
  Number penalty;		/**< weight to punish usage */
  Number reinforcement;		/**< weight to encourage usage */

  /* maintenance information */
  int index;			/**< identification in a GlsNet */
  int refCounter;		/**< no values using this lexem */

  /* upwards link */
  struct GlsTimepointStruct *timePoint; /**< every lexem belongs to a timepoint */

  /* extra structuring */
  struct GlsLexemStruct *next;	/**< next lexem during the same timepoint */
  struct GlsLexemStruct *prev;	/**< previous lexem during the same timepoint */
} GlsLexemStruct;
typedef GlsLexemStruct *GlsLexem; /**< type definition of a GlsLexem */

/* ---------------------------------------------------------------------- 
 * GlsTimepoint
 */
typedef struct GlsTimepointStruct {
  /* maintenance information */
  int index;			/**< identification in a GlsNet */

  /* upwards link */
  GlsNet net;			/**< every timepoint belongs to a GlsNet */

  /* workload */
  GlsValue firstLexem;		/**< first (=best) lexem in timepoint */
  GlsValue lastLexem;		/**< last value in timepoint */
  GlsValue annoLexem;		/**< corresponds to annotation */
  Boolean isActive;		/**< is this timepoint compatible with others */

} GlsTimepointStruct;
typedef GlsTimepointStruct *GlsTimepoint; /**< type definition of a GlsTimepoint */

/* ---------------------------------------------------------------------- 
 * GlsMode:
 *	gls might subdivide the given problem in several ways, solve the
 *	sub problems individually and reassemble a global solution from
 *	the collected partial solution (inspired by frobbing dynamic)
 */
typedef enum {
  GlsNormal,		     /** do not subdivide, work as normal */
  GlsDynamic,		     /** comparable mode as frobbing dynamic: 
			       subdivide the problem */
} GlsMode;

/* ---------------------------------------------------------------------- 
 * GlsPool:
 *	burry a farm of GlsNets which are scheduled one after the other
 *	in order to solve the same problem using different heuristics.
 */
typedef struct GlsPoolStruct {

  /* workload */
  GlsNet firstNet;	     /** list of gls networks */
  GlsNet lastNet;	     /** last net in the list of gls networks */
  ConstraintNet constraintNet;	/** reference to constraintnet */

  GlsValuePair *cache;	     /** cache of value pairs 
				 (only used if using full caching) */

  GlsValuePair valuePairs;   /** list of all value pairs 
				 (only used if using full caching) */

  /* information */
  unsigned long sliceCycles; /** length of active slice in cycles */
  GlsNet currentNet;	     /** active network */
  GlsNet bestNet;	     /** network containing the best we have */
  int noFrozenNets;	     /** number of expired contracts */
  unsigned long cycle;	     /** total of all cycles in all networks */
  Timer profileTimer;	     /** ovreral profiling */
  Parse annoParse;	     /** decorated parse derived from the annotation */

  /* statistics */
  GlsStatRecord firstRecord; /** head of all statistics records */
  GlsStatRecord lastRecord;  /** tail of all statistics records */

  /* parameters */
  unsigned long maxTime;     /** overall maximum time allowed to spend on this problem */
  unsigned long maxCycles;   /** overall maximum cycles */
  int debugLevel;	     /** amount of nifty messages */
  int interrupt;	     /** 0: interactive cycle, 
			         >0: skip a number of cycles, 
				 <0: no interruption */
  String statisticsFileName; /** file into which statistics are printed out */
  Boolean cooperativeFlag;   /** should networks learn from each other */
  int compareFunction;	     /** 0: badness, 1: costs */

  Boolean initMode;	     /** 0: best parse, 
				 1: no initial state, 
				 2: best root values */
  Boolean useCache;	     /** switch on/off value pair caching */
  GlsMode mode;		     /** operate in specified modes @see GlsMode */
  List span;

} GlsPoolStruct;
typedef GlsPoolStruct *GlsPool;

/* ---------------------------------------------------------------------- 
 * GlsInteractionInfo:
 *	helper structure in glsInteraction
 */
typedef struct GlsInteractionInfoStruct {
  int bufferSize;	     /* how much space is there in my buffer */
  int bufferPos;	     /* current parsing position */
  String buffer;	     /* place where input is gathered */
  String token;		     /* currently parsed chunk from buffer */
  int intValue;		     /* scanned token value */
  Number floatValue;	     /* scanned token value */
} GlsInteractionInfoStruct;
typedef GlsInteractionInfoStruct *GlsInteractionInfo;

/* -- DECLARATIONS -------------------------------------------------------- */

static void newGlsPool(void);
static GlsNet newGlsNet(void);
static GlsState newGlsState(GlsNet);
static GlsDomain newGlsDomain(GlsNet, List);
static GlsValuePair newGlsValuePair(GlsValue, GlsValue);
static GlsBinding newGlsBinding(GlsDomain);
static GlsValue newGlsValue(LevelValue);
static GlsStatRecord newGlsStatRecord(GlsNet);
static GlsInteractionInfo newGlsInteractionInfo(void);
static Boolean glsInitPool(void);
static Boolean glsInitNet(GlsNet);
static Boolean glsInitState(GlsState);
static Boolean glsInitStateRoot(GlsState);
static Boolean glsInitStateBest(GlsState);
static Boolean glsInitStateLast(GlsState);
static Boolean glsInitStateParse(GlsState, Parse);
static Vector glsBuildDomains(GlsNet);
static void freeGlsPool(void);
static void freeGlsState(GlsState);
static void freeGlsDomain(GlsDomain);
static void freeGlsValuePair(GlsValuePair);
static void freeGlsBinding(GlsBinding);
static void freeGlsValue(GlsValue);
static void freeGlsNet(GlsNet);
static void freeGlsStatRecord(GlsStatRecord);
static void freeGlsInteractionInfo(GlsInteractionInfo);
static void copyGlsState(GlsState, GlsState);
static void glsAdaptToBestNet(GlsNet);
static GlsValuePair cloneGlsValuePair(GlsValuePair);
static INLINE void glsInsertValue(GlsValue);
static void glsInsertValueAt(GlsValue, GlsValue);
static INLINE void glsRemoveValue(GlsValue);
static void glsResortValue(GlsValue);
static INLINE int glsHashIndex(int,int);
static INLINE Number glsPenaltyToCosts(GlsNet, Number);
static INLINE Boolean glsAreDeletedLexemes(ByteVector, List);
static INLINE Boolean glsIsValidValue(GlsState, GlsValue);
static INLINE Boolean glsIsValidState(GlsState);
static char *glsTagToString(GlsTag);
static char glsStateOfValue(GlsValue value);
static char glsGetPropeller(void);
static Number glsGetPruning(GlsNet);
static int glsDeleteIncompatibles(ByteVector, GlsValue);
static INLINE Boolean glsJobNeedsToBeDone(GlsBinding);
static INLINE void glsQueueJob(RingBuffer, GlsJobType, GlsBinding);
static INLINE GlsBinding glsDequeueJob(RingBuffer);
static INLINE GlsBinding glsRequeueJob(RingBuffer, RingBuffer);
static INLINE void glsClearQueue(RingBuffer);
static void glsActivateInvalids(GlsState);
static void glsPrintBinding(unsigned long, ByteVector, GlsBinding);
static void glsPrintValue(unsigned long, ByteVector, GlsValue);
static void glsPrintLexemes(unsigned long, ByteVector, List);
static void glsPrintDomain(unsigned long, ByteVector, GlsDomain);
static void glsPrintState(unsigned long, GlsState);
static void glsPrintStatistics(unsigned long);
static void glsPrintSettings(unsigned long);
static void glsWriteStatistics(void);
static Boolean gls(void);
static void glsNormal(void);
static void glsDynamic(void);
static GlsNet glsScheduleNet(void);
static GlsValuePair glsEvalBinary(GlsValue, GlsValue);
static void glsEvalState(GlsState);
static void glsLearn(GlsState);
static void glsPenalizeBinding(GlsBinding, Number penalty);
static Boolean glsImprove(GlsState);
static Boolean glsTransform(GlsState);
static Boolean glsRepairUnary(GlsBinding);
static Boolean glsRepairModifier(GlsBinding);
static Boolean glsRepairModifiee(GlsBinding);
static void glsRegisterBestNetSoFar(GlsNet);
static void glsRegisterBestStateSoFar(GlsState);
static void glsRestoreBestState(GlsNet);
static int glsAssignValue(GlsBinding, GlsValue);
static int glsCompareStates(GlsState, GlsState);
static void glsComputeUtility(GlsNet);
static INLINE Number glsAugCostsOfValue(GlsState, GlsValue);
static void glsPrunePool(void);
static int glsPruneNet(GlsNet);
static Boolean glsPruneValue(GlsValue);
static Boolean glsOptimizeNet(GlsNet);
static void glsOptimizePool(void);
static void glsRegisterSolution(void);
static Parse glsToParse(GlsState);
static void glsCompleteStatistics(GlsNet);
static int glsFreezeBinding(GlsBinding);
static int glsFreezeDomain(GlsDomain);
static INLINE void glsFreezeNet(GlsNet); 
static INLINE void glsUnfreezeNet(GlsNet); 
static void glsRegisterModifiers(GlsValue);
static void glsUnRegisterModifiers(GlsValue);
static GlsStatRecord glsGetSoftRecord(void);
static INLINE GlsStatRecord glsGetSolutionRecord(void);
static String glsTimeToString(unsigned long);
static void glsCheckPhase(GlsNet);
static Boolean glsCvCompare(ConstraintViolation, ConstraintViolation);
static Boolean glsMatchLVs(LevelValue a, LevelValue b);

/* -- GLOBAL VARIABLES ---------------------------------------------------- */
static const String glsVersion = "$Revision: 1.85 $";
static GlsPool glsCurrentPool = NULL; /* allocated and freed by glsCommand */

/* -- IMPLEMENTATION ------------------------------------------------------ */

/* ---------------------------------------------------------------------------
 * check validity of a value
 *
 * A value is invalid with respect to the search state it is thought to be 
 * used in if it is globally not active (@see GlsTag) or all used lexemes
 * are deleted, that is incompatible with the otherwise used lexemes.
 *
 * @param state the search state in which we want to use the \a value
 * @param value the value to be checked for validity
 * @returns TRUE if the value can be used together with the others in the
 * search state
 */
INLINE Boolean glsIsValidValue(GlsState state, GlsValue value)
{
  LevelValue lv = (LevelValue)value;

  if (!value) /* pathological case */
    return FALSE;
    
  return (value->tag & GlsActive &&
	  !glsAreDeletedLexemes(state->isDeletedLexem, lv->modifiers) &&
	  !glsAreDeletedLexemes(state->isDeletedLexem, lv->modifiees));
}

/* ---------------------------------------------------------------------------
 * check all bound values.
 *
 * A search state is invalid if not all bindings are fullfilled or the set
 * of used values is incompatible. @see glsIsValidValue
 *
 * @param state the search state to be tested
 * @returns TRUE on success, FALSE otherwise
 */
INLINE Boolean glsIsValidState(GlsState state)
{
  GlsBinding binding;

  for (binding = state->firstBinding; binding; binding = binding->next) {
    if (!glsIsValidValue(state, binding->value)) {
#if 0
      cdgPrintf(CDG_DEBUG, "DEBUG: binding not valid:\n");
      glsPrintBinding(CDG_DEBUG, state->isDeletedLexem, binding);
#endif
      return FALSE;
    }
  }

  return TRUE;
}

/* ---------------------------------------------------------------------------
 * glsAreDeletedLexemes: (see lgAreDeletedNodes)
 */
INLINE Boolean glsAreDeletedLexemes(ByteVector isDeletedLexem, List lexemes)
{
  List l;
  LexemNode ln;

  if (!spec(lexemes))
    return FALSE;

  for (l = lexemes; l != NULL; l = listNext(l)) {
    ln = (LexemNode) (listElement(l));
    if (!bvElement(isDeletedLexem, ln->no))
      return FALSE;
  }

  return TRUE;
}

/* ---------------------------------------------------------------------------
 * glsHashIndex
 *	computes an index based on two integers in order to address a 
 *	symetric matrix excluding the main diagonal 
 *	stored into a linear vector in the illustrated way
 *
 *	      | y = 0   1   2   3   4   5   ...
 *	----------------------------------
 *	x = 0 |         0   1   3   6   10
 *	      |                           
 *	    1 |             2   4   7   11
 *            |                           
 *	    2 |                 5   8   12
 *	      |                           
 *	    3 |                     9   13
 *	      |                           
 *	    4 |                         14
 *	      |                           
 *	  ... |                           
 *
 *	                    y * (y - 1)
 *	formula: f(x,y) :=  -----------  + x	with 0 <= x < y
 *	                         2
 *
 * CAUTION: the fraction cannot be split (in order to multiply smaler numbers) 
 *	    because of rounding errors while dividing by two. 
 *	    y * (y-1) is alway dividable by two. so no rounding erros and
 *	    we stay within integers.
 *
 */
INLINE int glsHashIndex(int x, int y) 
{
  return (x < y)?  (((y * (y-1)) >> 1) + x) : (((x * (x-1)) >> 1) + y);
}

/* ---------------------------------------------------------------------------
 * glsPenaltyToCosts
 *	calulates score to costs 
 */
INLINE Number glsPenaltyToCosts(GlsNet net, Number score)
{
  if (score == 0)
    return net->cutoff;
  else {
#if 0
    /* 
     * mathematicaly speaking this is the same as the
     * function below but 
     * - gives too much numerical problems
     *   which leads to a completely different behavior of gls :(++
     * - is slower
     */
    Number q = pow(score, 2 * net->beta);
    return (1-q)/(1+q);
#else
    return tanh(-log(score)* net->beta) * net->cutoff; 
#endif
  }

}
  
/* ---------------------------------------------------------------------------
 * glsResortValue
 */
void glsResortValue(GlsValue value)
{
  GlsValue prev, next;
  if (value->tag & GlsDeleted) {
#if 0
    cdgPrintf(CDG_WARNING, "WARNING: why do you wanna resort a deleted value %d?\n",
	value->index);
    cdgExecHook(HOOK_GLSINTERACTION, value->domain->net, "");
#endif
    return;
  }

  for (prev = value->prev; prev && !(prev->tag & GlsSorted); prev = prev->prev);
  for (next = value->next; next && !(next->tag & GlsSorted); next = next->next);

  if ((prev && prev->augCosts > value->augCosts) ||
      (next && next->augCosts < value->augCosts)) {
    glsRemoveValue(value);
    glsInsertValue(value);
  } else {
    value->tag |= GlsSorted;
  }
}

/* ---------------------------------------------------------------------------
 * glsRemoveValue
 *	remove value from sorting
 */
INLINE void glsRemoveValue(GlsValue oldValue)
{
  GlsDomain domain = oldValue->domain;
  GlsNet net = domain->net;

#if DEBUG_NODE
  {
    int n = 0;
    GlsValue value;
    Boolean foundOldValue = FALSE;
    Boolean foundError = FALSE;

    if (oldValue->tag & GlsDeleted) {
      cdgPrintf(CDG_ERROR, "ERROR: old value %03d already deleted in domain %03d of net %d\n",
	  oldValue->index, domain->index, net->index);
      foundError = TRUE;
    }

    for (value = domain->firstValue; value != NULL && !foundError; value = value->next) {
      n++;

      if (value == oldValue)
	foundOldValue = TRUE;

      if (value->next == value) {
	cdgPrintf(CDG_ERROR, "ERROR: remove detected a next-cycle at value %03d in domain %03d (before)\n",
	    value->index, domain->index);
	foundError = TRUE;
      }

      if (value->prev == value) {
	cdgPrintf(CDG_ERROR, "ERROR: remove detected a prev-cycle at value %03d in domain %03d (before)\n",
	    value->index, domain->index);
	foundError = TRUE;
      }
    }

    if (!foundError && n != domain->noValues) {
      cdgPrintf(CDG_ERROR, "ERROR: remove counted %d values but it should be %d in domain %03d (before)\n",
	  n, domain->noValues, domain->index);
      foundError = TRUE;
    }

    if (!foundError && !foundOldValue) {
      cdgPrintf(CDG_ERROR, "ERROR: old value %03d not found in domain %03d (before)\n",
	  oldValue->index, domain->index);
      foundError = TRUE;
    }

    if (foundError) {
      cdgExecHook(HOOK_GLSINTERACTION, net, "");
      abort();
    }
  }
#endif

  if (domain->firstValue == oldValue) 
    domain->firstValue = oldValue->next;
  if (domain->lastValue == oldValue) 
    domain->lastValue = oldValue->prev;
  if (oldValue->prev != NULL)
    oldValue->prev->next = oldValue->next;
  if(oldValue->next != NULL)
    oldValue->next->prev = oldValue->prev;

  /* tag the value unsorted and deleted */
  oldValue->tag |= GlsDeleted;
  oldValue->tag &= ~GlsSorted;
  oldValue->prev = NULL;
  oldValue->next = NULL;

  domain->noValues--;
  net->noValidValues--;

#if DEBUG_NODE
  {
    int n = 0;
    GlsValue value;

    for (value = domain->firstValue; value != NULL; value = value->next) {
      n++;

      if (value == oldValue) {
	cdgPrintf(CDG_ERROR, "ERROR: removed value %03d still in domain %03d (after)\n",
	    value->index, domain->index);
	abort();
      }

      if (value->next == value) {
	cdgPrintf(CDG_ERROR, "ERROR: remove detected a next-cycle at value %03d in domain %03d (after)\n",
	    value->index, domain->index);
	abort();
      }

      if (value->prev == value) {
	cdgPrintf(CDG_ERROR, "ERROR: remove detected a prev-cycle at value %03d in domain %03d (after)\n",
	    value->index, domain->index);
	abort();
      }
    }

    if (n != domain->noValues) {
      cdgPrintf(CDG_ERROR, "ERROR: remove counted %d values but it should be %d in domain %03d (after)\n",
	  n, domain->noValues, domain->index);
      abort();
    }

  }
#endif
}

/* ---------------------------------------------------------------------------
 * glsBuildDomains
 *	this method builds gls domains by collecting a set of constraint nodes
 *	which point to the same graheme and which are part of the same level
 */
Vector glsBuildDomains(GlsNet net)
{
  Vector domains;
  List allNodes, someNodes;
  List l, m, k, kk;
  ConstraintNode cn1, cn2;
  ConstraintNet constraintNet = glsCurrentPool->constraintNet;
  GlsDomain domain;
  int noLevels;
  int noDomains;
  int i;
  Level level;

  /* count used levels of the current grammar */
  noLevels = 0;
  for (l = inputCurrentGrammar->levels; l; l = listNext(l)) {
    level = listElement(l);
    noLevels++;
  }

  /* be serious about used levels */
  if (noLevels == 0) {
    cdgPrintf(CDG_WARNING, "WARNING: no levels in use.\n");
    return NULL;
  }
  noDomains = vectorSize(constraintNet->lexemgraph->graphemnodes) * noLevels;
  

  /* allocate space */
  domains = vectorNew(noDomains);
  vectorSetAllElements(domains, NULL);

  /* allNodes is a list of all constraint nodes.
   * building gls domains deletes arbitraty constraint nodes from
   * this list while we loop thru it
   */
  allNodes = vectorToList(constraintNet->nodes);
  for (m = allNodes; m != NULL; m = listNext(m)) {
    cn1 = (ConstraintNode) (listElement(m));

    /* skip superfluous constraint nodes */
    if (cn1->noValidValues <= 0)
      continue;

    /* cn1 is the pivot for this gls domain */
    someNodes = listPrependElement(NULL, cn1);
    
    /* collect constraint nodes refering to the same grapheme */
    for (kk = m, k = listNext(m); k != NULL;) {
      cn2 = (ConstraintNode) (listElement(k));

      if (cn2->noValidValues > 0 &&
	  cn2->level == cn1->level &&
	  cn1->gn == cn2->gn) {

	someNodes = listPrependElement(someNodes, cn2);

	/* remove domain and and step thru allNodes */
	listSetNext(kk, listNext(k));
	memFree(k);
	k = listNext(kk);
      } else {
	/* step thru allNodes */
	kk = k;
	k = listNext(k);
      }
    }

    /* now, we have all constraint nodes which make up a gls domain,
     * so that we can build it an store it in the gls net 
     */
    domain = newGlsDomain(net, someNodes);

    /* check if this domain has been built */
    if (domain == NULL) {
      for (i = vectorSize(domains)-1; i >= 0; i--) {
	domain = (GlsDomain) vectorElement(domains, i);
	freeGlsDomain(domain);
      }
      vectorDelete(domains);
      return NULL;
    }

    /* store the domain at a defined cell in the GlsNet->domains vector */
    domain->index = domain->constraintNode->level->no
      + domain->constraintNode->gn->arc->from * noLevels;
#if 0
    {
      GlsDomain otherNode = vectorElement(domains, domain->index);
      if (otherNode != NULL) {
	cdgPrintf(CDG_ERROR, "PANIC: other domain found in same slot %d\n",
	    domain->index);
	abort();
      }
    }
#endif
    vectorSetElement(domains, domain, domain->index);
  }

  /* if there remain some superfluous constraint nodes they make
   * up the rest of this list
   */
  listDelete(allNodes);

  return domains;
}

/* ---------------------------------------------------------------------------
 * newGlsPool:
 *	allocate a new pool of gls networks
 */
void newGlsPool(void)
{
  glsCurrentPool = (GlsPool) memMalloc(sizeof(GlsPoolStruct));
  glsCurrentPool->constraintNet = NULL;
  glsCurrentPool->interrupt = -1; /* automode by default */
  glsCurrentPool->cooperativeFlag = TRUE;
  glsCurrentPool->debugLevel = (hkVerbosity & CDG_DEBUG)? 1:0;
  glsCurrentPool->sliceCycles = 1;
  glsCurrentPool->currentNet = NULL;
  glsCurrentPool->firstNet = NULL;
  glsCurrentPool->lastNet = NULL;
  glsCurrentPool->bestNet = NULL;
  glsCurrentPool->maxTime = 0;
  glsCurrentPool->maxCycles = 0;
  glsCurrentPool->cycle = 1;
  glsCurrentPool->noFrozenNets = 0;
  glsCurrentPool->cache = NULL; 
  glsCurrentPool->valuePairs = NULL;
  glsCurrentPool->statisticsFileName = NULL;
  glsCurrentPool->compareFunction = GLS_CMP_BADNESS;
  glsCurrentPool->firstRecord = NULL;
  glsCurrentPool->lastRecord = NULL;
  glsCurrentPool->initMode = GLS_INIT_NORMAL;
  glsCurrentPool->profileTimer = timerNew();
  glsCurrentPool->useCache = TRUE;
  glsCurrentPool->mode = GlsNormal;
  glsCurrentPool->annoParse = NULL;
}

/* ---------------------------------------------------------------------------
 * glsInitPool:
 *	construct components within a pool. networks are already inserted but
 *	themself not initialized. we make that for all of them from here.
 *
 * @returns FALSE if no initial pool can be constructed; this is the case when
 * all configured nets prune too strong in advance
 */
Boolean glsInitPool(void) 
{
  GlsNet net;
  int maxPairs;
  int maxNoValues;
  Boolean oldUseCache = glsCurrentPool->useCache;
  AnnoEntry anno;

  /* get a decorated parse from an annotation */
  anno = findAnnoForLattice(glsCurrentPool->constraintNet->lexemgraph->lattice, TRUE);
  if (anno) {
    glsCurrentPool->annoParse = parseFromAnno(anno);
    if (glsCurrentPool->annoParse) {
      if (!parseDecorate(glsCurrentPool->annoParse, glsCurrentPool->constraintNet->lexemgraph, 
	  anno)) {
	parseDelete(glsCurrentPool->annoParse);
	glsCurrentPool->annoParse = NULL;
      }
    }
  }

  /* init all gls networks */
  glsCurrentPool->maxTime = 0;
  glsCurrentPool->useCache = FALSE; /* trick to do initial evaluations */
  for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {

    if (!glsInitNet(net)) {
      return FALSE;
    }

    /* get the maximum time and cycles in the pool */
    if (glsCurrentPool->maxTime < net->maxTime) {
      glsCurrentPool->maxTime = net->maxTime;
    }
    if (glsCurrentPool->maxCycles < net->toCycle) {
      glsCurrentPool->maxCycles = net->toCycle;
    }
  }

  /* start the alarm timer */
  if (glsCurrentPool->maxTime > 0) {
    timerSetAlarm(glsCurrentPool->maxTime);
  } else {
    timerExpired = FALSE;
  }
  

  /* get the maximum number of values in a net */
  maxNoValues = 0; 
  for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
    if (maxNoValues < vectorSize(net->values)) {
      maxNoValues = vectorSize(net->values);
    }
  }

  /* we need at least two values ;) */
  if (maxNoValues < 2) {
    cdgPrintf(CDG_WARNING, "WARNING: not enough values in the pool\n");
    return FALSE;
  }

  /* build the value pair cache */
  glsCurrentPool->useCache = oldUseCache;
  if (glsCurrentPool->useCache) {
    maxPairs = glsHashIndex(maxNoValues-1, maxNoValues);
    glsCurrentPool->cache = memMalloc(sizeof(GlsValuePair) * maxPairs);
    memset(glsCurrentPool->cache, 0, sizeof(GlsValuePair) * maxPairs);
  }

  /* optimize the initial pool here */
  glsOptimizePool();

  return TRUE;
}

/* ---------------------------------------------------------------------------
 * freeGlsPool
 */
void freeGlsPool(void)
{
  GlsNet net, nextNet;
  GlsValuePair valuePair, oldValuePair;
  GlsStatRecord record, oldRecord;
    
  /* free all gls networks */
  for (net = glsCurrentPool->firstNet; net != NULL; net = nextNet) {
    nextNet = net->next;
    freeGlsNet(net);
  }

  /* free the value pairs */
  valuePair = glsCurrentPool->valuePairs;
  while(valuePair != NULL) {
    oldValuePair = valuePair;
    valuePair = valuePair->next;
    freeGlsValuePair(oldValuePair);
  }
  if (glsCurrentPool->useCache)
    memFree(glsCurrentPool->cache);
  cdgFreeString(glsCurrentPool->statisticsFileName);
  
  /* free all statistic records */
  for (record = glsCurrentPool->firstRecord; record != NULL; ) {
    oldRecord = record;
    record = record->next;
    freeGlsStatRecord(oldRecord);
  }

  /* stop the timer which was started in newGlsPool */
  timerFree(glsCurrentPool->profileTimer);
  timerStopAlarm();

  /* free the annotation parse */
  if (glsCurrentPool->annoParse) {
    parseDelete(glsCurrentPool->annoParse);
  }

  /* free myselfe */
  memFree(glsCurrentPool);
  glsCurrentPool = NULL;
  
}

/* ---------------------------------------------------------------------------
 * newGlsNet: 
 *	allocate a new GlsNet structure
 */
GlsNet newGlsNet()
{
  GlsNet net;

  if (glsCurrentPool == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: no current pool available\n");
    abort();
  }

  net = (GlsNet) memMalloc(sizeof (GlsNetStruct));
  memset(net, 0, sizeof(GlsNetStruct));
  net->tag = GlsDefault;
  net->phase = GlsInit;
  net->next = NULL;
  net->statRecord = NULL;

  /* append the new net to the pool */
  if (glsCurrentPool->firstNet == NULL) {
    net->index = 0;
    glsCurrentPool->firstNet = net;
    glsCurrentPool->lastNet = net;

    /* default parameters */
    net->beta = 1.0;
    net->cutoff = 10;
    net->lambda1 = 1.0;
    net->lambda2 = 0.0;
    net->maxCosts = 0;
    net->fromCycle = 0;
    net->toCycle = UINT_MAX;
    net->maxDetour = 0;
#if 0
    net->maxUtility = MY_MAXDOUBLE;
#else
    net->maxUtility = 100;
#endif
    net->toleranceFraction = 0.0;
    net->maxTime = cdgTimeLimit; 
    net->initialUnaryPruning = 0.0;
    net->prolongUnaryPruning = 0.0;
    net->terminationUnaryPruning = 0.0;
    net->priority = 1;
  } else {
    net->index = glsCurrentPool->lastNet->index + 1;

    /* default parameters taken from the last net */
    net->beta = glsCurrentPool->lastNet->beta;
    net->cutoff = glsCurrentPool->lastNet->cutoff;
    net->lambda1 = glsCurrentPool->lastNet->lambda1;
    net->lambda2 = glsCurrentPool->lastNet->lambda2;
    net->maxCosts = glsCurrentPool->lastNet->maxCosts;
    net->fromCycle = glsCurrentPool->lastNet->fromCycle;
    net->toCycle = glsCurrentPool->lastNet->toCycle;
    net->maxDetour = glsCurrentPool->lastNet->maxDetour;
    net->maxUtility = glsCurrentPool->lastNet->maxUtility;
    net->toleranceFraction = glsCurrentPool->lastNet->toleranceFraction;
    net->maxTime = glsCurrentPool->lastNet->maxTime;
    net->initialUnaryPruning = glsCurrentPool->lastNet->initialUnaryPruning;
    net->prolongUnaryPruning = glsCurrentPool->lastNet->prolongUnaryPruning;
    net->terminationUnaryPruning = glsCurrentPool->lastNet->terminationUnaryPruning;
    net->priority = glsCurrentPool->lastNet->priority;

    glsCurrentPool->lastNet->next = net;
    glsCurrentPool->lastNet = net;
  }

  /* reset information */
  net->currentDetour = 1;
  net->noValidValues = 0;
  net->noPenalties = 0;

  /* following structures are build in glsInitNet */
  net->lexemgraph = NULL;
  net->values = NULL;
  net->state = NULL;
  net->timePoints = NULL;
  net->noModifiersOfLexem = NULL;

  return net;
}

/* ---------------------------------------------------------------------------
 * glsInitNet: 
 *	construct all components within a gls net
 *
 * @returns TRUE on success, FALSE on failure
 */
Boolean glsInitNet(GlsNet net)
{
  int i, j;
  GlsDomain domain;
  GlsValue value;
  Boolean isValid = FALSE;

  /* generate local lexem information */
  net->lexemgraph =
    lgNew(glsCurrentPool->constraintNet->lexemgraph->lattice);
  net->noModifiersOfLexem = vectorNew(vectorCapacity(net->lexemgraph->nodes));
  vectorSetAllElements(net->noModifiersOfLexem, (void *)0);

  /* build domains */
  net->domains = glsBuildDomains(net);

  if (net->domains == NULL) {
    cdgPrintf(CDG_WARNING, "WARNING: cannot initialize net %d\n", net->index);
    glsFreezeNet(net);
    return FALSE;
  }

  /* store all values into the net */
  net->values = vectorNew(net->noValidValues);
  for (i = vectorSize(net->domains) - 1; i >= 0; i--) {
    domain = (GlsDomain) vectorElement(net->domains, i);
    if(!domain)
      continue;

    for (value = domain->firstValue; value != NULL; value = value->next) {
      value->index = vectorAddElement(net->values, value);
      /* the value index is importand to find value pairs */
    }
  }

  /* mark annoValues */
  if (glsCurrentPool->annoParse) {
    Parse parse = glsCurrentPool->annoParse;

    /* loop thru all level values */
    for (i = 0; i < vectorSize(parse->LVs); i++) {
      LevelValue lv = vectorElement(parse->LVs, i);
      if (!lv)
	continue;

      /* loop thru all domains */
      for (j = vectorSize(net->domains) - 1; j >= 0; j--) {
	domain = (GlsDomain) vectorElement(net->domains, j);
	if(!domain)
	  continue;

	/* loop thru the domain */
	for (value = domain->firstValue; value != NULL;
	    value = value->next) {
	  if (glsMatchLVs(lv, (LevelValue)value)) {
	    domain->annoValue = value;
	    break; /* looping thru domain */
	  }
	}
      }
    }
  }
  

  /* build space for assignments */
  net->state = newGlsState(net);
  net->bestState = newGlsState(net);

  /* build the first statistics records */
  glsCurrentPool->currentNet = net; 
  newGlsStatRecord(net);

  if (glsCurrentPool->mode & GlsNormal) {
    /* chose the init mode */
    switch (glsCurrentPool->initMode) {
      case GLS_INIT_NORMAL:
	isValid = glsInitState(net->state);
	break;
      case GLS_INIT_ROOT: 
	isValid = glsInitStateRoot(net->state); 
	break;
      case GLS_INIT_BEST:
	isValid = glsInitStateBest(net->state);
	break;
      case GLS_INIT_LAST:
	isValid = glsInitStateLast(net->state);
	break;
      default: /* never reach */
	cdgPrintf(CDG_ERROR, "illegal code reached\n");
	abort();
	break;
    }
    glsCompleteStatistics(net);
    glsCurrentPool->cycle++;

    /* check again if everything is bound */
    if (isValid) {
      isValid = glsIsValidState(net->state);
    }

    /* check the search net phase */
    if (isValid)
      glsCheckPhase(net);
    else
      cdgPrintf(CDG_WARNING, "WARNING: didn't get a valid initial search state\n");
  }

  return TRUE;
}

/* ---------------------------------------------------------------------------
 * initialize a search state by scheduling repair jobs.
 *
 * Here we actually don't assign bindings yet but simply prepair the 
 * first repairment loop by putting one repair job of an unfrozen binding into 
 * the queue. All other unassigned and bindings get repaired casually. 
 *
 * @param state the current search state that we pepair
 * @returns TRUE in any case
 */
Boolean glsInitState(GlsState state)
{
  GlsBinding binding;

  glsClearQueue(state->allJobs);
  for (binding = state->firstBinding; binding; binding = binding->next) {
    if (!(binding->tag & GlsFrozen)) {
      glsQueueJob(state->allJobs, GlsUnaryJob, binding);
      break;
    }
  }

  /* repair all unfrozen bindings */
  if (!rbIsEmpty(state->allJobs)) {
    glsImprove(state);
  } else {
    cdgPrintf(CDG_DEBUG, "DEBUG: all bindings frozen while initializing\n");
  }

  return TRUE;
}

/* ---------------------------------------------------------------------------
 * initialize the search state in GLS_INIT_ROOT mode
 *
 * This function searches the best root bindings.
 *
 * @param state the current search state that we try na find an initial state for
 * @returns TRUE if the initial search state is valid, otherwise FALSE
 */
Boolean glsInitStateRoot(GlsState state)
{
  GlsBinding binding;
  GlsValue bestValue, value;

  /* assign best valid root bindings */
  for (binding = state->firstBinding; binding; binding = binding->next) {
    if (binding->tag & GlsFrozen)
      continue;

    bestValue = NULL;
    for (value = binding->domain->firstValue; value; value = value->next) {
      if (!glsIsValidValue(state, value))
	continue;

      if (((LevelValue)value)->modifiee == NULL &&
	  (bestValue == NULL || value->costs < bestValue->costs)) {
	bestValue = value;
      }
    }

    if (bestValue != NULL) {
      glsAssignValue(binding, bestValue);
    }
  }

  /* assign best valid bindings for those which we didn't find root bindings for */
  glsClearQueue(state->allJobs);
  for (binding = state->firstBinding; binding; binding = binding->next) {
    if (binding->tag & GlsFrozen || binding->tag & GlsChanged)
      continue;

    bestValue = NULL;
    for (value = binding->domain->firstValue; value; value = value->next) {
      if (!glsIsValidValue(state, value))
	continue;

      if (bestValue == NULL || value->costs < bestValue->costs) {
	bestValue = value;
      }
    }

    if (bestValue != NULL) {
      glsAssignValue(binding, bestValue);
    } else {
      glsQueueJob(state->allJobs, GlsUnaryJob, binding);
    }
  }

  /* repair if some bindings stayed unassigned */
  if (!rbIsEmpty(state->allJobs)) {
    glsImprove(state);
  }

  return TRUE; 
}

/* ----------------------------------------------------------------------
 * returns true if two levelvalues have the same properties
 * otherwise false
 *
 * WARNING: This function assumes that a is a levelvalue that comes from
 * a decorated annotation parse.
 *
 * @param a a levelvalue from a decorated annotation parse
 * @param b a levelvalue from a connstraint net
 * @returns true iff a matches b
 */
Boolean glsMatchLVs(LevelValue a, LevelValue b)
{
  if (a->level != b->level)
    return FALSE;

  if (a->label != b->label) 
    return FALSE;

  if (a->modifier->arc->from != b->modifier->arc->from ||
      a->modifier->arc->to != b->modifier->arc->to) 
    return FALSE;

  if (a->modifiee == NONSPEC && b->modifiee != NONSPEC) 
    return FALSE;
  
  if (a->modifiee != NONSPEC && b->modifiee == NONSPEC) 
    return FALSE;

  if (a->modifiee == NULL && b->modifiee != NULL) 
    return FALSE;

  if (a->modifiee != NULL && b->modifiee == NULL) 
    return FALSE;

  if (spec(a->modifiee) &&
      (a->modifiee->arc->from != b->modifiee->arc->from ||
       a->modifiee->arc->to   != b->modifiee->arc->to))
    return FALSE;

  if (a->modifier->arc->from != b->modifier->arc->from ||
      a->modifier->arc->to   != b->modifier->arc->to)
    return FALSE;

  if (spec(a->modifiee)) {
    List l;
    Boolean found = FALSE;
    for (l = b->modifiees; l; l = listNext(l)) {
      LexemNode ln = listElement(l);
      if (ln->lexem->description == a->modifiee->lexem->description) {
	found = TRUE;
	break;
      }
    }

    if (!found) {
#if 0
      cdgPrintf(CDG_DEBUG, "DEBUG: modifiees incompatible\n");
      lvPrint(CDG_DEBUG, a, NULL); cdgPrintf(CDG_DEBUG, "\n");
      lvPrint(CDG_DEBUG, b, NULL); cdgPrintf(CDG_DEBUG, "\n");
#endif
      return FALSE;
    }
  }

  {
    List l;
    Boolean found = FALSE;
    for (l = b->modifiers; l; l = listNext(l)) {
      LexemNode ln = listElement(l);
      if (ln->lexem->description == a->modifier->lexem->description) {
	found = TRUE;
	break;
      }
    }

    if (!found) {
#if 0
      cdgPrintf(CDG_DEBUG, "DEBUG: modifiers incompatible\n");
      lvPrint(CDG_DEBUG, a, NULL); cdgPrintf(CDG_DEBUG, "\n");
      lvPrint(CDG_DEBUG, b, NULL); cdgPrintf(CDG_DEBUG, "\n");
#endif
      return FALSE;
    }
  }

  return TRUE;
}

/* ---------------------------------------------------------------------------
 * initialize the search state in GLS_INIT_LAST mode
 *
 * This function use the last parse found in the constraint net to
 * install initial bindings. 
 * If this fails, we fall back to GLS_INIT_NORMAL mode, the default init mode.
 *
 * @param state the current search state that we try na find an initial state for
 * @returns TRUE if the initial search state is valid, otherwise FALSE
 */
Boolean glsInitStateLast(GlsState state)
{
  Parse parse = NULL;
  List l;

  /* get the first parse in the list if parses */
  for (l = glsCurrentPool->constraintNet->parses; l; l = listNext(l)) {
    parse = listElement(l);
    if (parse->LVs) {
      break;
    } else {
      parse = NULL;
    }
  }

  return glsInitStateParse(state, parse);
}

/* ---------------------------------------------------------------------------
 * initialize the search state in GLS_INIT_BEST mode
 *
 * This function generates a decorated
 * parse from the annotation and uses the contained values in there.
 * If all this fails for some reason we fall back to GLS_INIT_NORMAL mode, 
 * the default init mode.
 *
 * @param state the current search state that we try na find an initial state for
 * @returns TRUE if the initial search state is valid, otherwise FALSE
 */
Boolean glsInitStateBest(GlsState state)
{
  return glsInitStateParse(state, glsCurrentPool->annoParse);
}

/* ---------------------------------------------------------------------------
 * initialize the search state with the given parse.
 *
 * @param state the current search state that we try na find an initial state for
 * @param parse a decorated parse 
 * @returns TRUE if the initial search state is valid, otherwise FALSE
 */
Boolean glsInitStateParse(GlsState state, Parse parse)
{
  Boolean isValid = TRUE; 
  GlsBinding binding;
  GlsValue value;
  int i;
  Boolean found = FALSE;

  glsClearQueue(state->allJobs);

  /* parse must be decorated */
  if (!parse) {
    cdgPrintf(CDG_WARNING, "WARNING: no decorated parse found, "
                           "falling back to init mode 'normal'\n");
    return glsInitState(state);
  }

  /* loop thru all level values */
  for (i = 0; i < vectorSize(parse->LVs); i++) {
    LevelValue lv = vectorElement(parse->LVs, i);
    if (!lv)
      continue;

    /* loop thru all bindings of the current gls net */
    for (binding = state->firstBinding; binding != NULL;
	binding = binding->next) {

      /* already bound */
      if (binding->value != NULL)
	continue;

      /* loop thru the domain */
      found = FALSE;
      for (value = binding->domain->firstValue; value != NULL;
	  value = value->next) {
	if (glsMatchLVs(lv, (LevelValue)value)) {
	  found = TRUE;
	  break; /* looping thru domain */
	}
      }

      /* found right place for that level value */
      if (found) {
	glsAssignValue(binding, value);
	break; /* looping binding values */
      }
    }
  }

  /* repair if some bindings stayed unassigned */
  if (!rbIsEmpty(state->allJobs)) {
    glsImprove(state);
  } else {
    glsEvalState(state);
    glsComputeUtility(state->net);
    glsRegisterBestStateSoFar(state);
    glsRegisterBestNetSoFar(state->net);
    glsRegisterSolution();
    glsPrunePool();
  }
  
  return isValid;
}

/* ---------------------------------------------------------------------------
 * freeGlsNet:
 *	free things allocated by newGlsNet exept the constraintnet
 */
void freeGlsNet(GlsNet net)
{
  int i;
  GlsDomain domain;
  GlsValue value;

  /* free domains */
  if (net->domains != NULL) {
    for (i = vectorSize(net->domains)-1; i >= 0; i--) {
      domain = (GlsDomain) vectorElement(net->domains, i);
      if(!domain)
	continue;
      freeGlsDomain(domain);
    }
    vectorDelete(net->domains);
    net->domains = NULL;
  }

  /* free values */
  if (net->values != NULL) {
    for (i = vectorSize(net->values)-1; i >= 0; i--) {
      value = (GlsValue) vectorElement(net->values, i);
      freeGlsValue(value);
    }
    vectorDelete(net->values);
    net->values = NULL;
  }

  /* free all search states */ 
  freeGlsState(net->state);
  freeGlsState(net->bestState);

  /* free local lexeme information, this is only the dynamic part */
#if 0
  if (net->lexemgraph) {
    if(net->lexemgraph->distance != NULL) {
      arrayDelete(net->lexemgraph->distance);
    }
    memFree(net->lexemgraph->noOfPathsFromStart);
    memFree(net->lexemgraph->noOfPathsToEnd);
    bvDelete(net->lexemgraph->isDeletedNode);
    net->lexemgraph = NULL;
  }
#else
  lgDelete(net->lexemgraph);
#endif
  if (net->noModifiersOfLexem) {
    vectorDelete(net->noModifiersOfLexem);
    net->noModifiersOfLexem = NULL;
  }

  /* free myselfe */
  memFree(net);
}

/* ---------------------------------------------------------------------------
 *  glsInsertValue
 */
INLINE void glsInsertValue(GlsValue newValue)
{
  glsInsertValueAt(newValue->domain->firstValue, newValue);
}

/* ---------------------------------------------------------------------------
 * glsInsertValueAt
 *	insert value into sorted domain at the given position
 */
void glsInsertValueAt(GlsValue atValue, GlsValue newValue)
{
  GlsValue value;
  GlsDomain domain = newValue->domain;
  GlsNet net = domain->net;

  if (domain->noValues == 0) {
    /* first value in domain */
    domain->firstValue = newValue;
    domain->lastValue = newValue;
    newValue->prev = NULL;
    newValue->next = NULL;
  } else {

    /* search position */
    for (value = atValue; value != NULL; value=value->next) {

      if((value->tag & GlsSorted) && 
	  value->augCosts >= newValue->augCosts) 
	break;
    }

    if (value != NULL) {
      /* insert before value */
      newValue->next = value;
      newValue->prev = value->prev;
      if (value->prev != NULL) 
	value->prev->next = newValue;
      value->prev = newValue;
      if (value == domain->firstValue)
	domain->firstValue = newValue;
    } else {
      /* insert at the end of the domain */
      newValue->prev = domain->lastValue;
      newValue->next = NULL;
      if (domain->lastValue != NULL) {
	domain->lastValue->next = newValue;
      }
      domain->lastValue = newValue;
    }
  }

  /* tag the new value as sorted */
  newValue->tag |= GlsSorted;
  newValue->tag &= ~GlsDeleted;
  
  domain->noValues++;
  net->noValidValues++;
  
#if DEBUG_NODE
  {
    int n = 0;
    Boolean foundAtValue = (atValue == NULL) ? TRUE:FALSE; 
    Boolean foundNewValue = FALSE;

    for (value = domain->firstValue; value != NULL; value = value->next) {
      n++;
      if (value == atValue)
	foundAtValue = TRUE;
      if (value == newValue)
	foundNewValue = TRUE;

      if (value->next == value) {
	cdgPrintf(CDG_ERROR, "ERROR: insert detected a next-cycle at value %03d in domain %03d\n",
	    value->index, domain->index);
	abort();
      }

      if (value->prev == value) {
	cdgPrintf(CDG_ERROR, "ERROR: insert detected a prev-cycle at value %03d in domain %03d\n",
	    value->index, domain->index);
	abort();
      }
    }

    if (n != domain->noValues) {
      cdgPrintf(CDG_ERROR, "ERROR: insert counted %d values but it should be %d in domain %03d\n",
	  n, domain->noValues, domain->index);
      abort();
    }

    if (!foundAtValue) {
      cdgPrintf(CDG_ERROR, 
	  "ERROR: inserted new value %03d at value %03d but didn't find %03d any more in domain %03d\n",
	  newValue->index, atValue->index, atValue->index, domain->index);
      abort();
    }

    if (!foundNewValue) {
      cdgPrintf(CDG_ERROR, 
	  "ERROR: inserted new value %03d at value %03d but didn't find %03d any more in domain %03d\n",
	  newValue->index, atValue->index, newValue->index, domain->index);
      abort();
    }
  }
#endif

}

/* ---------------------------------------------------------------------------
 * newGlsDomain
 */
GlsDomain newGlsDomain(GlsNet net, List nodes)
{
  GlsDomain domain;
  ConstraintNode cn;
  List l;
  int i;
  LevelValue lv;
  GlsValue value;

#if SECURE_NEWCOMPNODE
  if (nodes == NULL || listElement(nodes) == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: newGlsDomains called with invalid arguments!\n");
    abort();
  }
#endif

#if 0
  cdgPrintf(CDG_DEBUG, "DEBUG: new gls domain with %d constraint nodes(s)\n",
      listSize(nodes));
#endif

  domain = (GlsDomain) memMalloc(sizeof(GlsDomainStruct));
  domain->constraintNode = (ConstraintNode) (listElement(nodes));
  domain->constraintNodes = nodes;
  domain->net = net;
  domain->noValues = 0;
  domain->firstValue = NULL;
  domain->lastValue = NULL;
  domain->annoValue = NULL;
  domain->index = -1;
  domain->lexemes = NULL;

  /* collect all corresponding lexemes */
  for (l = nodes; l != NULL; l = listNext(l)) {
    cn = (ConstraintNode)(listElement(l));
    domain->lexemes = listAppendList(domain->lexemes, listClone(cn->gn->lexemes));
  }

  /* 
   * unify domains:
   * this reverts some work done by constructing partitions from lexemlists 
   * via lgPartitions
   *
   * why does this work:
   *	normaly each (gaphem x level) is related to a set of constraint nodes.
   *	this happens by taking into account the different lexical readings.
   *	on the other hand each gaphem x level is allowed to be bound once.
   *	so some constraint nodes are left without binding. formaly this means
   *	that out of the unification of the set of these constraint nodes only
   *	one value is chosen. 
   *	as a consequence these constraint nodes can be colapsed to one domain
   *	unifiing their domains and only one value of this compound domain
   *	must be chosen. no extra checks have to be done to prevent a
   *	graphem x level pair to be bound twice.
   *
   *	this aproach could be used in the whole system. this function gets
   *	obsolete then. 
   */

  /* collect all levelvalues and build GlsValues */
  for (l = domain->constraintNodes; l != NULL; l = listNext(l)) {
    cn = (ConstraintNode) (listElement(l));

    for (i = vectorSize(cn->values)-1; i >= 0; i--) {
      lv = (LevelValue) vectorElement(cn->values, i);
      
      if (lv->isDeleted)
	continue;

      value = newGlsValue(lv);
      value->domain = domain; 
      value->constraintNode = cn;
      value->noSoftViolations = listSize(lv->constraints);
      value->costs = glsPenaltyToCosts(net, lv->score);
      value->augCosts = value->costs; /* initialy the same */
      glsInsertValue(value); /* insert sorted */
      glsRegisterModifiers(value);
#if 0
      lvPrint(CDG_DEBUG, lv, NULL);
      cdgPrintf(CDG_DEBUG, "\n");
#endif
    }
  }

  if (domain->noValues == 0) {
    cdgPrintf(CDG_WARNING, "WARNING: cannot build domain, no values\n");
    freeGlsDomain(domain);
    domain = NULL;
  }

  return domain;
}

/* ---------------------------------------------------------------------------
 * freeGlsValue
 */
void freeGlsValue(GlsValue value)
{
  memFree(value);
}


/* ---------------------------------------------------------------------------
 * freeGlsDomain:
 *	free stuff allocated by newGlsDomain. 
 */
void freeGlsDomain(GlsDomain domain)
{
  if (domain != NULL) {
    listDelete(domain->constraintNodes);
    listDelete(domain->lexemes);
    memFree(domain);
  }
}

/* ---------------------------------------------------------------------------
 * newGlsValue
 */
GlsValue newGlsValue(LevelValue lv)
{
  GlsValue value;

  value = (GlsValue)memMalloc(sizeof(GlsValueStruct));
  memcpy(value, lv, sizeof(LevelValueStruct));
  value->tag = GlsDefault | GlsActive;
  value->index = -1;
  value->prev = NULL;
  value->next = NULL;
  value->domain = NULL; 
  value->constraintNode = NULL;
  value->penalty = 0.0;
  value->reinforcement = 0.0;
  value->costs = 0.0;
  value->augCosts = 0.0;
  value->noSoftViolations = 0;

  return value;
}

/* ---------------------------------------------------------------------------
 * newGlsValuePair
 */
GlsValuePair newGlsValuePair(GlsValue v1, GlsValue v2)
{
  GlsValuePair valuePair;

  /* allocate a new value pair */
  valuePair = (GlsValuePair) memMalloc(sizeof (GlsValuePairStruct));
  valuePair->index = glsHashIndex(v1->index, v2->index);
  if (v1->index < v2->index) {
    valuePair->value1Index = v1->index;
    valuePair->value2Index = v2->index;
    valuePair->node1Index = v1->domain->index;
    valuePair->node2Index = v2->domain->index;
  } else {
    valuePair->value1Index = v2->index;
    valuePair->value2Index = v1->index;
    valuePair->node1Index = v2->domain->index;
    valuePair->node2Index = v1->domain->index;
  }
  valuePair->costs = 0.0;
  valuePair->score = 1.0;
  valuePair->noSoftViolations = 0;
  valuePair->noHardViolations = 0;
  valuePair->violations = NULL;

  /* store the new value pair into the cache */
  if (glsCurrentPool->useCache) {
    glsCurrentPool->cache[valuePair->index] = valuePair;
    valuePair->next = glsCurrentPool->valuePairs;
    glsCurrentPool->valuePairs = valuePair;
  }

  return valuePair;
}

/* ---------------------------------------------------------------------------
 * allocate a twin.
 *
 * @param origPair the original value pair
 * @returns a copy
 *
 * This copy is not asserted into the value pair cache.
 */
GlsValuePair cloneGlsValuePair(GlsValuePair origPair)
{
  GlsValuePair valuePair;
  List l;

  /* allocate a new value pair */
  valuePair = (GlsValuePair) memMalloc(sizeof (GlsValuePairStruct));

  /* set all values to those in origPair */
  valuePair->index = origPair->index;
  valuePair->value1Index = origPair->value1Index;
  valuePair->value2Index = origPair->value2Index;
  valuePair->node1Index = origPair->node1Index;
  valuePair->node2Index = origPair->node2Index;
  valuePair->costs = origPair->costs;
  valuePair->score = origPair->score;
  valuePair->noSoftViolations = origPair->noSoftViolations;
  valuePair->noHardViolations = origPair->noHardViolations;
  valuePair->violations = NULL;
  for (l = origPair->violations; l; l = listNext(l)) {
    ConstraintViolation cv = listElement(l);
    valuePair->violations = listPrependElement(valuePair->violations, cvClone(cv));
  }

  return valuePair;
}

/* ---------------------------------------------------------------------------
 * freeGlsValuePair
 */
void freeGlsValuePair(GlsValuePair valuePair)
{
  if (valuePair != NULL) {
    listForEachDelete(valuePair->violations, cvDelete);
    memFree(valuePair);
  }
}

/* ---------------------------------------------------------------------------
 * newGlsState: construct a new search state structure
 */
GlsState newGlsState(GlsNet net)
{
  GlsState state;
  GlsBinding binding = NULL, prevBinding;
  GlsDomain domain;
  int noDomains = vectorSize(net->domains);
  int i;

  state = (GlsState) memMalloc(sizeof (GlsStateStruct));
  state->net = net;
  state->conflictPairs = NULL;
  state->violations = NULL;
  state->noSoftViolations = INT_MAX;
  state->noHardViolations = INT_MAX;
  state->costs = MY_MAXDOUBLE;
  state->augCosts = 0.0;
  state->maxCostsSoFar = 0.0;
  state->score = 0.0;
  state->maxUtility = 0.0;
  state->utility = 0.0;
  state->noFrozenBindings = 0;
  state->isDeletedLexem = bvClone(net->lexemgraph->isDeletedNode);
  state->interimJobs = rbNew(noDomains);
  state->allJobs = rbNew(noDomains);
  state->bindings = vectorNew(noDomains);
  state->context = vectorNew(noDomains);
  
  for (prevBinding = NULL, i = 0; i < noDomains; i++, prevBinding = binding) {

    /* get loop item */
    domain = (GlsDomain) vectorElement(net->domains, i);
    if (!domain)
      continue;

    /* build a binding */
    binding = newGlsBinding(domain);
    binding->state = state;

    /* add binding to bindings list */
    if (prevBinding == NULL) {
      state->firstBinding = binding;
    } else {
      prevBinding->next = binding;
    }
    
    /* store binding at the same index as its domain */
    vectorSetElement(state->bindings, binding, i);

    /* set all context slots to NULL */
    vectorSetElement(state->context, NULL, i);
  }

  /* propagate domain states to binding states */
  for (binding = state->firstBinding; binding != NULL;
       binding = binding->next) {

    domain = binding->domain;
    if (domain->noValues == 1)
      glsFreezeDomain(domain);
  }


  return state;
}

/* ---------------------------------------------------------------------------
 * freeGlsState:
 *	free things allocated by newGlsState.
 */
void freeGlsState(GlsState state)
{
  GlsBinding binding;
  int i;

  if (state == NULL)
    return;

  bvDelete(state->isDeletedLexem);
  rbDelete(state->interimJobs);
  rbDelete(state->allJobs);

  for (i = vectorSize(state->bindings)-1; i >=0; i--) {
    binding = (GlsBinding)vectorElement(state->bindings, i);
    if (!binding)
      continue;
    freeGlsBinding(binding);
  }

  if (glsCurrentPool->useCache) {
    listDelete(state->conflictPairs);
  } else {
    listForEachDelete(state->conflictPairs, freeGlsValuePair);
  }
  listForEachDelete(state->violations, cvDelete);

  vectorDelete(state->bindings);
  vectorDelete(state->context);
  memFree(state);
}

/* --------------------------------------------------------------------------------
 * copyGlsState
 */
void copyGlsState(GlsState dst, GlsState src)
{
  GlsBinding dstBinding, srcBinding;
  int i;
  List l;

#if SECURE_COPYGLSSEARCHSTATE
  if (vectorSize(dst->bindings) != vectorSize(src->bindings)) {
    cdgPrintf(CDG_ERROR, "ERROR: cannot copy bindings of different size\n");
    abort();
  }

  if (dst->net != src->net) {
    cdgPrintf(CDG_ERROR, "ERROR: cannot copy states of different networks\n");
    abort();
  }
#endif

  /* copy the bindings */
  for (srcBinding = src->firstBinding,
       dstBinding = dst->firstBinding;
       srcBinding != NULL;
       srcBinding = srcBinding->next,
       dstBinding = dstBinding->next) {
       
    /* dstBinding->domain = srcBinding->domain; */
    dstBinding->tag = srcBinding->tag;
    dstBinding->value = srcBinding->value;
    dstBinding->costs = srcBinding->costs;
    dstBinding->utility = srcBinding->utility;
  }

  /* copy pair information */
  if (glsCurrentPool->useCache) {
    dst->conflictPairs = listCopy(dst->conflictPairs, src->conflictPairs);
  } else {
    listForEachDelete(dst->conflictPairs, freeGlsValuePair);
    dst->conflictPairs = NULL;
    for (l = src->conflictPairs; l; l = listNext(l)) {
      GlsValuePair valuePair = cloneGlsValuePair(listElement(l));
      dst->conflictPairs = listAppendElement(dst->conflictPairs, valuePair);
    }
  }

  /* copy conflicts information */
  listForEachDelete(dst->violations, cvDelete);
  dst->violations = NULL;
  for (l = src->violations; l; l = listNext(l)) {
    ConstraintViolation viola = listElement(l);
    dst->violations = listPrependElement(dst->violations, cvClone(viola));
  }

  /* copy job information */
  bvCopy(dst->isDeletedLexem, src->isDeletedLexem);
  
  /* copy information */
  dst->costs = src->costs;
  dst->augCosts = src->augCosts;
  dst->maxCostsSoFar = src->maxCostsSoFar;
  dst->score = src->score;
  dst->noFrozenBindings = src->noFrozenBindings;
  dst->maxUtility = src->maxUtility;
  dst->utility = src->utility;

  /* copy violation information */
  dst->noSoftViolations = src->noSoftViolations;
  dst->noHardViolations = src->noHardViolations;

  /* copy context */
  for (i = 0; i < vectorSize(src->context); i++) {
    vectorSetElement(dst->context, vectorElement(src->context, i), i);
  }
}

/* --------------------------------------------------------------------------------
 * newGlsBinding
 */
GlsBinding newGlsBinding(GlsDomain domain)
{
  GlsBinding binding;

  binding = (GlsBinding)memMalloc(sizeof(GlsBindingStruct));
  binding->tag = GlsDefault;
  binding->index = domain->index; /* same position */
  binding->next = NULL;
  binding->state = NULL; /* set by the calling procedure */
  binding->domain = domain;
  binding->value = NULL;
  binding->costs = 0.0;
  binding->augCosts = 0.0;
  binding->utility = MY_MAXDOUBLE;
  binding->jobType = GlsNoJob;

  return binding;
}

/* ---------------------------------------------------------------------------
 * freeGlsBinding
 */
void freeGlsBinding(GlsBinding binding)
{
  memFree(binding);
}


/* ---------------------------------------------------------------------------
 * newGlsStatRecord
 *	append a new statistics record to the pool.
 *	first all values are defaulting to zero here and provided within the
 *	cycle this record wants to describe
 */
GlsStatRecord newGlsStatRecord(GlsNet net)
{
  GlsStatRecord record = (GlsStatRecord) memMalloc(sizeof(GlsStatRecordStruct));
  GlsStatRecord prevRecord = net->statRecord;

  net->statRecord = record;
  record->net = net;
  record->cycle = glsCurrentPool->cycle;
  record->isSolution = FALSE;
  record->score = 0.0;
  record->utility = 0.0;
  record->costs =  0.0;
  record->augCosts =  0.0;
  record->errorRate = 0.0;
  record->minErrorRate = MY_MAXDOUBLE;
  record->noHardViolations = 0;
  record->noSoftViolations = 0;
  record->noPenalties = 0;
  record->parse = NULL;
  record->isDeletedLexem = bvNew(1);
  record->prevSolution = glsGetSolutionRecord();

  /* list handling */
  record->next = NULL;
  record->prev = glsCurrentPool->lastRecord;
  if (glsCurrentPool->lastRecord != NULL) {
    glsCurrentPool->lastRecord->next = record;
  }
  glsCurrentPool->lastRecord = record;
  if (glsCurrentPool->firstRecord == NULL) {
    glsCurrentPool->firstRecord = record;
  }

  if (prevRecord == NULL) {
    record->minUtility = MY_MAXDOUBLE;
    record->maxUtility = 0;
    record->maxCosts = 0;
    record->maxDetour = 0;
    record->noBlindAlleys = 0;
    record->noRepairs = 0;
    record->time = 0;
    record->eta = 0;
  } else {
    record->minUtility = prevRecord->minUtility;
    record->maxUtility = prevRecord->maxUtility;
    record->maxCosts = prevRecord->maxCosts;
    record->maxDetour = prevRecord->maxDetour;
    record->noBlindAlleys = prevRecord->noBlindAlleys;
    record->noRepairs = prevRecord->noRepairs;
    record->time = prevRecord->time;
    record->eta = prevRecord->eta;
    record->minErrorRate = prevRecord->minErrorRate;
  }

  return record;
}

/* ---------------------------------------------------------------------------
 * freeGlsStatRecord
 */
void freeGlsStatRecord(GlsStatRecord record)
{
  bvDelete(record->isDeletedLexem);
  memFree(record);
}

/* ---------------------------------------------------------------------------
 * newGlsInteractionInfo
 */
GlsInteractionInfo newGlsInteractionInfo(void)
{
  GlsInteractionInfo gii = memMalloc(sizeof(GlsInteractionInfoStruct));

  gii->bufferSize = GLS_BUFFERSIZE;
  gii->bufferPos = 0;
  gii->buffer = memMalloc(sizeof(char) * GLS_BUFFERSIZE);
  gii->token  = memMalloc(sizeof(char) * GLS_BUFFERSIZE);
  gii->intValue = 0;
  gii->floatValue = 0.0;

  return gii;
}

/* ---------------------------------------------------------------------------
 * freeGlsInteractionInfo
 */
void freeGlsInteractionInfo(GlsInteractionInfo gii)
{
  memFree(gii->buffer);
  memFree(gii->token);
  memFree(gii);
}

/* ---------------------------------------------------------------------------
 * glsFreezeBinding
 *
 *	returns the number of deleted lexemes which is a possible 
 *	consequence of freezing this binding.
 */
int glsFreezeBinding(GlsBinding binding)
{
  GlsValue value;
  int noDeletedLexemes = 0;
  GlsState state = binding->state;
  GlsNet net = state->net;
  GlsDomain domain = binding->domain;
  Boolean found;

  /* security check */
  if (binding->tag & GlsFrozen) {
    return 0;
  }

  binding->utility = 0.0;
  binding->tag = GlsFrozen;
  state->noFrozenBindings++;

  /* check if we freeze the last binding */
  if (state->noFrozenBindings == vectorSize(state->bindings) && !(net->tag & GlsFrozen)) {
    cdgPrintf(CDG_INFO, "INFO: all bindings frozen, freezing net\n");
    glsFreezeNet(net);
  }

  /* assign value and delete all incompatible lexemes */
  if (binding->value) {
    value = binding->value;
    found = TRUE;
  } else {
    found = FALSE;
    for (value = domain->firstValue; value; value = value->next) {
      if (glsIsValidValue(state, value)) {
	found = TRUE;
	break;
      }
    }
  }

  if (found) {
    noDeletedLexemes = glsAssignValue(binding, value);
    glsDeleteIncompatibles(net->lexemgraph->isDeletedNode, value);
#if 0
    cdgPrintf(CDG_DEBUG, "DEBUG: freezing binding %03d\n", binding->index);
#endif
  } else {
    cdgPrintf(CDG_WARNING, 
	"WARNING: cannot find a valid value while freezing binding %03d\n",
	binding->index);
  }

  return noDeletedLexemes;
}

/* ---------------------------------------------------------------------------
 * glsEvalBinary: 
 *	violated constraints are combined using a costs-scheme which is
 *	additive instead of a score scheme which is multipicative
 *
 */
GlsValuePair glsEvalBinary(GlsValue va, GlsValue vb)
{
  GlsValuePair valuePair = 
    glsCurrentPool->useCache?
      glsCurrentPool->cache[glsHashIndex(va->index, vb->index)]: NULL;

  if (valuePair == NULL) {
    GlsValuePair otherPair;
    List l;
    ConstraintViolation cv;
    GlsNet net = va->domain->net;
    ConstraintNet constraintNet = glsCurrentPool->constraintNet;
    GlsNet otherNet;
    static BadnessStruct badnessStruct;
    Badness badness = &badnessStruct;
    GlsValue otherValue1, otherValue2;
    
#if 0
    cdgPrintf(CDG_DEBUG, "DEBUG: eval binary btw values in domains %03d and %03d",
	va->domain->index, vb->domain->index);
#endif

    /* initialize badness */
    badness->no = 0;
    badness->hard = 0;
    badness->soft = 1.0;

    /* build a value pair from badness and conflicts */
    valuePair = newGlsValuePair(va, vb);
    evalBinary((LevelValue)va, (LevelValue)vb, constraintNet, NULL, FALSE, badness, &(valuePair->violations));
    valuePair->score = badness->soft;
    valuePair->noHardViolations = badness->hard;
    valuePair->noSoftViolations = badness->no - badness->hard;
    valuePair->costs = 0.0;
    
    /* compute costs */
    for (l = valuePair->violations; l; l = listNext(l)) {
      cv = listElement(l);
      valuePair->costs += glsPenaltyToCosts(net, cv->penalty);
      cv->nodeBindingIndex1 = valuePair->node1Index;
      cv->nodeBindingIndex2 = valuePair->node2Index;
    }
    
    /* 
     * build value pairs for all other nets. 
     * this should be faster than reevaluating the grammar.
     */
    for (otherNet = glsCurrentPool->firstNet; 
	otherNet != NULL; otherNet = otherNet->next) {

      if (otherNet == net || (otherNet->tag & GlsFrozen))
	continue;

      if (!otherNet->values)
	continue;

      otherValue1 = (GlsValue) vectorElement(otherNet->values, va->index);
      otherValue2 = (GlsValue) vectorElement(otherNet->values, vb->index);

      otherPair = newGlsValuePair(otherValue1, otherValue2);
      otherPair->score = valuePair->score;
      otherPair->costs = valuePair->costs;
      otherPair->noSoftViolations = valuePair->noSoftViolations;
      otherPair->noHardViolations = valuePair->noHardViolations;
      for (l = valuePair->violations; l; l = listNext(l)) {
	otherPair->violations = listPrependElement(otherPair->violations,
	    cvClone(listElement(l)));
      }
    }
  }

  return valuePair;
}

/* ---------------------------------------------------------------------------
 * glsPrintStatistics
 */
void glsPrintStatistics(unsigned long mode)
{
  if (hkVerbosity & mode ) {
    Number solutionPercentage, softSolutionPercentage;
    GlsStatRecord currentRecord = glsCurrentPool->lastRecord;
    GlsStatRecord solutionRecord = glsGetSolutionRecord();
    GlsStatRecord softSolutionRecord = glsGetSoftRecord();

    if (!solutionRecord) {
      cdgPrintf(CDG_INFO, "INFO: no solution found so far\n");
    }

    if (currentRecord->time == 0 || !solutionRecord) {
      solutionPercentage = 0;
      softSolutionPercentage = 0;
    } else {
      solutionPercentage = solutionRecord->time / (Number) currentRecord->time * 100.0;
      if (softSolutionRecord) {
	softSolutionPercentage = softSolutionRecord->time / (Number) currentRecord->time * 100.0;
      } else {
	softSolutionPercentage = 0.0;
      }
    }

    cdgPrintf(mode,
	      "\n  statistics:\n"
	      "  ----------------------------\n"
	      "  total         :   %7lu ms     %6lu cycles     %7lu repairs%7lu penalties\n"
	      "  solution      :   %7lu ms     %6lu cycles     %7.2f %%        %4.3e score\n"
	      "  soft solution :   %7lu ms     %6lu cycles     %7.2f %%        %4.3e score\n"
	      "  benchmark     :   %7.0f c/sec  %6.0f r/sec     %7.0f r/c\n"
	      "  blind alleys  :   %7lu\n"
	      "  detour        :   %7lu max    %7lu needed\n"
	      "  costs         :   %7.2f max    %7.2f needed\n"
	      "  utility       : %4.3e max  %4.3e needed   %4.3e min\n"
	      "\n",
	      currentRecord->time,
	      currentRecord->cycle,
	      currentRecord->noRepairs,
	      currentRecord->noPenalties,
	      solutionRecord? solutionRecord->time: 0,
	      solutionRecord? solutionRecord->cycle: 0, 
	      solutionPercentage,
	      solutionRecord? ((solutionRecord->noHardViolations > 0)? 0:solutionRecord->score): 0,
	      softSolutionRecord? softSolutionRecord->time: 0,
	      softSolutionRecord? softSolutionRecord->cycle: 0, 
	      softSolutionPercentage,
	      softSolutionRecord? softSolutionRecord->score:0,
	      1000.0 * (Number) currentRecord->cycle / (Number) currentRecord->time ,
	      1000.0 * (Number) currentRecord->noRepairs / (Number) currentRecord->time,
	      (Number) currentRecord->noRepairs / (Number) currentRecord->cycle,
	      currentRecord->noBlindAlleys,
	      currentRecord->maxDetour,
	      solutionRecord? solutionRecord->maxDetour: currentRecord->maxDetour, 
	      currentRecord->maxCosts,
	      solutionRecord? solutionRecord->maxCosts: currentRecord->maxCosts, 
	      currentRecord->maxUtility,
	      solutionRecord? solutionRecord->maxUtility: currentRecord->maxUtility,
	      currentRecord->minUtility);

    /* XML statistics */
    writeXmlStatistics(1, "gls", "general statistics on gls",
	"score", solutionRecord?((solutionRecord->noHardViolations > 0)? 0.0 : solutionRecord->score): 0, 
	"totalTime", (Number) currentRecord->time,
	"softTime", (Number) (softSolutionRecord? softSolutionRecord->time:0),
	"solutionTime", (Number) (solutionRecord? solutionRecord->time: 0), 
	"steps", (Number) (solutionRecord? solutionRecord->cycle: 0), 
	"cyclesPerSecond", 
	(Number) currentRecord->cycle / (Number) currentRecord->time * 1000.0,
	"repairsPerSecond", 
	(Number) currentRecord->noRepairs / (Number) currentRecord->time * 1000.0,
	"repairsPerCycle",
	(Number) currentRecord->noRepairs / (Number) currentRecord->cycle,
	"blindAlleys", (Number)currentRecord->noBlindAlleys,
	"maxCosts", (Number) currentRecord->maxCosts,
	"minUtility", (Number) currentRecord->minUtility,
	"maxDetour", (Number) currentRecord->maxDetour,
	"maxDetourNeeded", 
	(Number) (solutionRecord? solutionRecord->maxDetour: currentRecord->maxDetour), 
	"minUtilityNeeded", 
	(Number) (solutionRecord? solutionRecord->minUtility: currentRecord->minUtility),
	"maxUtilityReached", 
	(Number) currentRecord->maxUtility,
	"maxCostsNeeded", 
	(Number) (solutionRecord? solutionRecord->maxCosts: currentRecord->maxCosts), 
	NULL);
  }
}

/* ---------------------------------------------------------------------------
 * glsWriteStatistics
 *	write a data file containing the statistis we collected
 */
void glsWriteStatistics(void)
{
  GlsNet net;
  FILE *statFile;
  GlsStatRecord record;
#if 0
  unsigned long last;
#endif
  int plotIndex = 0;
  Number score;
  String wgId = glsCurrentPool->constraintNet->lexemgraph->lattice->id;
  int noNets;
  Number costs;

  /* return if we aren't in the moode */
  if (glsCurrentPool->statisticsFileName == NULL || 
      strcmp(glsCurrentPool->statisticsFileName, "/dev/null") == 0)
    return;
  
  /* open the statistics file */
  statFile = fopen(glsCurrentPool->statisticsFileName, "w");
  if (!statFile) {
    cdgPrintf(CDG_ERROR, "ERROR: can't write to file '%s'\n",
	glsCurrentPool->statisticsFileName);
    return;
  }

  /* count nets */
  noNets = 0;
  for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
    noNets++;
  }

  /* comment header */
  fprintf(statFile,
      "# Version             : %s\n"
      "# \n"
      "# wordgraph           : %s\n"
      "# parallelity         : %s\n\n",
      glsVersion,
      wgId,
      noNets == 1 ? "single" :
      noNets == 2 ? glsCurrentPool->cooperativeFlag ?  "cooperative duo" : "competitive duo" :
      glsCurrentPool->cooperativeFlag ? "cooperative multi" : "competitive multi");


  /* show net specific settings */
  for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
    fprintf(statFile, 
	"# net                 : %d\n"
	"# priority            : %lu\n"
	"# from - to           : %lu - %lu\n"
	"# penalty guide       : %f\n"
	"# reinforcement guide : %f\n"
	"# normalization       : %f\n"
	"# cutoff              : %f\n"
	"# tolerance           : %g %%\n"
	"# max detour          : %lu\n"
	"# max time            : %s\n"
	"# max costs           : %f\n"
	"# max utility         : %f\n" 
	"# pruning             : %g/%g/%g\n\n",
	net->index,
	net->priority,
	net->fromCycle,
	net->toCycle,
	net->lambda1,
	net->lambda2,
	net->beta,
	net->cutoff,
	net->toleranceFraction,
	net->maxDetour,
	glsTimeToString(net->maxTime),
	net->maxCosts,
	net->maxUtility,
	net->initialUnaryPruning,
	net->prolongUnaryPruning,
	net->terminationUnaryPruning);
    }


  /* score trace */
  score = -1.0;
  fprintf(statFile, "# %d.: score trace\n", plotIndex++);
  for (record = glsCurrentPool->firstRecord; record; record = record->next) {
    if (record->score > 0.0 && record->score != score) { /*  no hard, no double */
      score = record->score;
      fprintf(statFile, "\t%lu\t%g\n", 
	  record->time, record->score);
    }
  }
  fprintf(statFile, "\n\n");

  /* utility trace */
  fprintf(statFile, "# %d.: utility trace\n", plotIndex++);
  for (record = glsCurrentPool->firstRecord; record; record = record->next) {
    fprintf(statFile, "\t%lu\t%g\n", 
	record->time, record->utility);
  }
  fprintf(statFile, "\n\n");
    
  /* best costs trace */
  fprintf(statFile, "# %d.: best costs trace\n", plotIndex++);
  for (record = glsCurrentPool->firstRecord; record; record = record->next) {
    if (record->isSolution) { /*  only generate data for the best */
      fprintf(statFile, "\t%lu\t%g\n", 
	  record->time, record->costs);
    }
  }
  fprintf(statFile, "\n\n");
    
  /* costs trace */
  fprintf(statFile, "# %d.: costs trace\n", plotIndex++);
  for (record = glsCurrentPool->firstRecord; record; record = record->next) {
    fprintf(statFile, "\t%lu\t%g\n", 
	record->time, record->costs);
  }
  fprintf(statFile, "\n\n");

  /* augmented costs trace */
  fprintf(statFile, "# %d.: aug. costs trace\n", plotIndex++);
  for (record = glsCurrentPool->firstRecord; record; record = record->next) {
    fprintf(statFile, "\t%lu\t%g\n", 
	record->time, record->augCosts);
  }
  fprintf(statFile, "\n\n");

  /* max augmented costs trace */
  costs = -1.0;
  fprintf(statFile, "# %d.: max. aug. costs trace\n", plotIndex++);
  for (record = glsCurrentPool->firstRecord; record; record = record->next) {
    if (costs != record->maxCosts || record == glsCurrentPool->firstRecord) { 
      /* unique */
      fprintf(statFile, "\t%lu\t%g\n", 
	  record->time, record->maxCosts);
      costs = record->maxCosts;
    }
  }
  fprintf(statFile, "\n\n");

  /* cost difference to aug costs */
  fprintf(statFile, "# %d.: difference costs trace\n", plotIndex++);
  for (record = glsCurrentPool->firstRecord; record; record = record->next) {
    fprintf(statFile, "\t%lu\t%g\n", record->time, 
	record->augCosts - record->costs);
  }
  fprintf(statFile, "\n\n");

  cdgPrintf(CDG_INFO, "INFO: written statistics to '%s'\n", 
      glsCurrentPool->statisticsFileName);
  fclose(statFile);

}

/* ---------------------------------------------------------------------------
 * glsPrintSettings
 */
void glsPrintSettings(unsigned long mode)
{
  if (hkVerbosity & mode) {
    int i;
    GlsNet net;
    int noNets;

    /* count unfrozen nets */
    noNets = 0;
    for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
      if (net->tag & GlsFrozen)
	continue;
      noNets++;
    }
    

    cdgPrintf(mode, 
	      "\n"
	      "  --------------------------------------------------------------------------------\n"
	      "  settings:\n"
	      "  --------------------------------------------------------------------------------\n"
	      "  net                        : %s\n"
	      "  wordgraph                  : %s\n"
	      "  debug level                : %d\n"
	      "  method                     : %s\n",
	      glsCurrentPool->constraintNet->id,
	      glsCurrentPool->constraintNet->lexemgraph->lattice->id,
	      glsCurrentPool->debugLevel,
	      glsCurrentPool->mode == GlsNormal?"normal":"dynamic");

    cdgPrintf(mode, "  init mode                  : ");
    switch(glsCurrentPool->initMode) {
      case GLS_INIT_BEST:
	cdgPrintf(mode, "best\n");
	break;
      case GLS_INIT_NORMAL:
	cdgPrintf(mode, "normal\n");
	break;
      case GLS_INIT_ROOT:
	cdgPrintf(mode, "root\n");
	break;
      default:
	cdgPrintf(mode, "???\n");
	break;
    }
    
    cdgPrintf(mode, "  compare function           : ");
    switch(glsCurrentPool->compareFunction) {
      case GLS_CMP_BADNESS:
	cdgPrintf(mode, "badness\n");
	break;
      case GLS_CMP_COSTS:
	cdgPrintf(mode, "costs\n");
	break;
      default:
	cdgPrintf(mode, "???\n");
	break;
    }
    
    cdgPrintf(mode, "  interruption               : ");
    if (glsCurrentPool->interrupt == 0) {
      cdgPrintf(mode, "yes\n");
    } else if (glsCurrentPool->interrupt < 0) {
      cdgPrintf(mode, "no\n");
    } else {
      cdgPrintf(mode, "breaking after %d cycles\n", glsCurrentPool->interrupt);
    }

    cdgPrintf(mode,
	      "  parallelity                : %s\n"
	      "  --------------------------------------------------------------------------------\n",
	      noNets == 1 ? "single" :
	      noNets == 2 ? glsCurrentPool->cooperativeFlag ?  "cooperative duo" : "competitive duo" :
	      glsCurrentPool->cooperativeFlag ? "cooperative multi" : "competitive multi");

    
    /* print table of all gls network settings */

    /* heading */
    cdgPrintf(mode, "                           ");
    for (i = 0, net = glsCurrentPool->firstNet; net != NULL; net = net->next, i++) {
      cdgPrintf(mode, "%10d ", i);
    }
    cdgPrintf(mode, 
	      "\n"
	      "  --------------------------------------------------------------------------------\n");

    /* cutoff */
    cdgPrintf(mode, "  cutoff                 : ");
    for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
      cdgPrintf(mode, " %4.3e ", net->cutoff);
    }
    cdgPrintf(mode, "\n");

    /* penalty guide */
    cdgPrintf(mode, "  penalty guide          : ");
    for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
      cdgPrintf(mode, " %4.3e ", net->lambda1);
    }
    cdgPrintf(mode, "\n");
	      
    /* reinforcement guide */
    cdgPrintf(mode, "  reinforcement guide    : ");
    for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
      cdgPrintf(mode, " %4.3e ", net->lambda2);
    }
    cdgPrintf(mode, "\n");
	      
    /* beta */
    cdgPrintf(mode, "  normalization(beta)    : ");
    for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
      cdgPrintf(mode, " %4.3e ", net->beta);
    }
    cdgPrintf(mode, "\n");
	      
    /* unary pruning */
    cdgPrintf(mode, "  pruning                : ");
    for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
      cdgPrintf(mode, " %g/%g/%g", 
	  net->initialUnaryPruning,
	  net->prolongUnaryPruning,
	  net->terminationUnaryPruning);
    }
    cdgPrintf(mode, "\n");
	      
    /* priority */
    cdgPrintf(mode, "  priority               : ");
    for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
      cdgPrintf(mode, "%10lu ", net->priority);
    }
    cdgPrintf(mode, "\n");

    /* range of cycles */
    cdgPrintf(mode, "  from cycle             : ");
    for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
      cdgPrintf(mode, "%10lu ", net->fromCycle);
    }
    cdgPrintf(mode, "\n");
    cdgPrintf(mode, "  to cycle               : ");
    for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
      if (net->toCycle == UINT_MAX) 
	cdgPrintf(mode, "%10s ", "(unbound)");
      else 
	cdgPrintf(mode, "%10lu ", net->toCycle);
    }
    cdgPrintf(mode, "\n");
	      
    /* maximum time */
    cdgPrintf(mode, "  max time               : ");
    for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
      if (net->maxTime == 0)
	cdgPrintf(mode, "%10s ", "(unbound)");
      else
	cdgPrintf(mode, "%10s ", glsTimeToString(net->maxTime));
    }
    cdgPrintf(mode, "\n");

    /* maximum detour */
    cdgPrintf(mode, "  max detour             : ");
    for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
      if (net->maxDetour == 0)
	cdgPrintf(mode, "%10s ", "(unbound)");
      else
	cdgPrintf(mode, "%10lu ", net->maxDetour);
    }
    cdgPrintf(mode, "\n");
	      
    /* maximum costs */
    cdgPrintf(mode, "  max costs              : ");
    for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
      if (net->maxCosts == 0.0)
	cdgPrintf(mode, "%10s ", "(unbound)");
      else
	cdgPrintf(mode, " %9g ", net->maxCosts);
    }
    cdgPrintf(mode, "\n");

    /* minimum utility */
    cdgPrintf(mode, "  max utility            : ");
    for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
      if (net->maxUtility == 0.0)
	cdgPrintf(mode, "%10s ", "(unbound)");
      else
	cdgPrintf(mode, " %4.3e ", net->maxUtility);
    }
    cdgPrintf(mode, "\n");

    /* tolerance */
    cdgPrintf(mode, "  tolerance              : ");
    for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
      cdgPrintf(mode, " %g %%", net->toleranceFraction);
    }
    cdgPrintf(mode, "\n");

    cdgPrintf(mode, 
	      "\n"
	      "  --------------------------------------------------------------------------------\n"
	      "\n");

    cdgFlush();
  }
}


/* -------------------------------------------------------------------------
 * gls interaction token-handling
 */

/* getNextChunk ------------------------------------------------------------
 * read next token from input
 * returns TRUE if there's a token to read
 */
Boolean getNextChunk(const String delimiter, GlsInteractionInfo gii)
{
  int n;

  if (gii->bufferPos > gii->bufferSize)
    return FALSE;

  gii->bufferPos += strspn(gii->buffer + gii->bufferPos, delimiter);

  if (gii->bufferPos > GLS_BUFFERSIZE)
    return FALSE;

  n = strcspn(gii->buffer + gii->bufferPos, delimiter);

  if (gii->bufferPos + n > GLS_BUFFERSIZE)
    return FALSE;

  memcpy(gii->token, gii->buffer + gii->bufferPos, n);
  gii->token[n] = '\0';
  gii->bufferPos += n;

  return (n != 0);
}

/* getNextToken -------------------------------------------------------
 * read next token from input delimited by white spaces
 * returns TRUE if there's a token to read
 */
INLINE Boolean getNextToken(GlsInteractionInfo gii)
{
  return getNextChunk(" \t", gii);
}

/* getNextCommand------------------------------------------------------
 * read next token from input delimited by white spaces or numerics
 * returns TRUE if there's a token to read
 */
INLINE Boolean getNextCommand(GlsInteractionInfo gii)
{
  return getNextChunk(" \t0123456789.+-", gii);
}

/* getNextInteger -----------------------------------------------------
 * calls getNextToken and tries to read an integer
 * returns TRUE on success, parsed value in gii->intValue
 */
INLINE Boolean getNextInteger(GlsInteractionInfo gii)
{
  return (getNextToken(gii) && sscanf(gii->token, "%d", &gii->intValue));
}

/* getNextFloat -----------------------------------------------------
 * calls getNextToken and tries to read a float
 * returns TRUE on success, parsed value in gii->floatValue
 */
INLINE Boolean getNextFloat(GlsInteractionInfo gii)
{
  return (getNextToken(gii) && sscanf(gii->token, "%lf", &gii->floatValue));
}
 
/* ---------------------------------------------------------------------------
 * glsInteraction
 */
void glsInteraction(Hook hook, va_list ap)
{
  GlsNet net; 
  String message;
  Boolean loopFlag;
  GlsInteractionInfo gii = newGlsInteractionInfo();

  /* 
   * sometimes glsInteraction is used for debugging purpose on the
   * commandline; if stdin is no tty we'd propably like to get a
   * nice coredump; if one likes to prevent this, then switch off
   * the gls interaction hook on the commandline
   */
  if (!isatty(fileno(stdin))) {
    cdgPrintf(CDG_ERROR, "WARNING: no tty available\n");
    return;
  }
  

  /* parse va_list */
  net = va_arg(ap, GlsNet);
  message = va_arg(ap, String);

  /* start the interactive loop */
  cdgCtrlCTrapped = FALSE;
  loopFlag = TRUE;
  glsCurrentPool->interrupt = 0;

  cdgPrintf(CDG_DEFAULT, "\n%s\n", message);
  while (loopFlag && !cdgCtrlCTrapped) {

    gii->buffer[0] = '\0';
    gii->bufferPos = 0;

    cdgPrintf(CDG_DEFAULT, "gls %d >> ", glsCurrentPool->cycle);
    cdgFlush();


    cdgGetString(gii->buffer, GLS_BUFFERSIZE);

    /* strip white spaces */
    {
      String bp;

      for (bp = gii->buffer + strlen(gii->buffer) - 1; bp >= gii->buffer; bp--) {
	if (!isspace((int)*bp))
	  break;
      }
      *(bp+1) = '\0';
    }

    /* if line is empty, we continue with the analysis */
    if (strlen(gii->buffer) == 0)
      break; 

    while (loopFlag 
	   && !cdgCtrlCTrapped 
	   && getNextCommand(gii)) {
      if (!gii->token[0])
	continue;

      /* help */
      if (strcmp(gii->token, "help") == 0 ||
	  strcmp(gii->token, "h") == 0 ||
	  strcmp(gii->token, "?") == 0) {
	cdgPrintf(CDG_DEFAULT, 
		  "available commands:\n"
		  "   acc, a               : get the accuracy of the current bindings\n"
		  "   accbest, ab          : get the accuracy of the best bindings so far\n"
		  "   break, b <n>         : break at given cycle\n"
		  "   cont, c              : turn interaction off\n"
		  "   domain, d <n>        : show domain\n"
		  "   debug, db <n>        : change the debug level\n"
		  "   exit, x              : exit cdg\n"
		  "   guide, g <x>         : change strength of penalties (lambda1)\n"
		  "   jump, jb		   : jump to best state\n"
		  "   help, ?, h           : this text\n"
		  "   info, i              : show info on the net\n"
		  "   normal <x>           : change normalization factor (beta)\n"
		  "   print, p [<n>]       : print all current bindings or specified binding\n"
		  "   printbest, pb [<n>]  : print best bindings so far or specified binding\n"
		  "   prune <x>            : prune the gls pool\n"
		  "   quit, q, ctrl-c      : interrupt the analysis\n"
		  "   reinf, r <x>         : change strength reinforcement (lambda2)\n"
		  "   skip, s <n>          : skip given number of cycles\n"
		  "   value, v <n>         : show value\n"
		  "   <return>             : step thru analysis\n"
		  );
      } 
      /* accuracy */
      else if (strcmp(gii->token, "acc") == 0 ||
		 strcmp(gii->token, "a") == 0) {
	AnnoEntry anno = findAnnoForLattice(glsCurrentPool->constraintNet->lexemgraph->lattice, TRUE);

	if (!anno) {
	  cdgPrintf(CDG_WARNING, "WARNING: no annotation for '%s'\n",
	      glsCurrentPool->constraintNet->lexemgraph->lattice->id);
	} else {
	  Parse parse = glsToParse(net->state);

	  if (parse) {
	    Vector lvs = parse->LVs;
	    ParseVerification pv;

	    parse->LVs = NULL;
	    parseComputeStructure(parse, lvs, net->state->isDeletedLexem);
	    parseDecorate(parse, glsCurrentPool->constraintNet->lexemgraph, NULL);
	    vectorDelete(lvs);
	    parseEval(parse);
	    pv = parseVerify(parse, anno);
	    parseVerificationDelete(pv);
	    parseDelete(parse);
	  } else {
	    cdgPrintf(CDG_WARNING, "WARNING: no parse found\n");
	  }
	}
      }
      /* accuracy of best state*/
      else if (strcmp(gii->token, "accbest") == 0 ||
		 strcmp(gii->token, "ab") == 0) {
	AnnoEntry anno = findAnnoForLattice(glsCurrentPool->constraintNet->lexemgraph->lattice, TRUE);

	if (!anno) {
	  cdgPrintf(CDG_WARNING, "WARNING: no annotation for '%s'\n",
	      glsCurrentPool->constraintNet->lexemgraph->lattice->id);
	} else {
	  Parse parse = glsToParse(net->bestState);

	  if (parse) {
	    Vector lvs = parse->LVs;
	    ParseVerification pv;

	    parse->LVs = NULL;
	    parseComputeStructure(parse, lvs, net->bestState->isDeletedLexem);
	    parseDecorate(parse, glsCurrentPool->constraintNet->lexemgraph, NULL);
	    vectorDelete(lvs);
	    parseEval(parse);
	    pv = parseVerify(parse, anno);
	    parseVerificationDelete(pv);
	    parseDelete(parse);
	  } else {
	    cdgPrintf(CDG_WARNING, "WARNING: no parse found\n");
	  }
	}
      }
      /* debug */
      else if (strcmp(gii->token, "debug") == 0 ||
		 strcmp(gii->token, "db") == 0) {
	if(getNextInteger(gii)) {
	   glsCurrentPool->debugLevel = gii->intValue;
	   cdgPrintf(CDG_INFO, "INFO: debuglevel set to %lu\n",
		     gii->intValue);
	   if (gii->intValue > 0) {
	     hkVerbosity |= CDG_DEBUG;
	   } else {
	     hkVerbosity &= ~CDG_DEBUG;
	   }
	} else {
	  cdgPrintf(CDG_ERROR, "ERROR: integer expected\n");
	  break;
	}
      }
      /* break */
      else if (strcmp(gii->token, "break") == 0 ||
		 strcmp(gii->token, "b") == 0) {
	if(getNextInteger(gii)) {
	  if (gii->intValue <= glsCurrentPool->cycle) {
	    cdgPrintf(CDG_ERROR, "ERROR: must be greater than actual number of cycles\n");
	    break;
	  }
	  
	  glsCurrentPool->interrupt = gii->intValue - glsCurrentPool->cycle;
	  cdgPrintf(CDG_INFO, "INFO: breaking at cycle %d\n", 
	      glsCurrentPool->interrupt + glsCurrentPool->cycle);
	  loopFlag = FALSE;

	} else {
	  cdgPrintf(CDG_ERROR, "ERROR: integer expected\n");
	  break;
	}
      }
      /* skip */
      else if (strcmp(gii->token, "skip") == 0 ||
		 strcmp(gii->token, "s") == 0) {
	if(getNextInteger(gii)) {
	  glsCurrentPool->interrupt = gii->intValue;
	  cdgPrintf(CDG_INFO, "INFO: skipping %d cycles\n", glsCurrentPool->interrupt);
	  loopFlag = FALSE;
	} else {
	  cdgPrintf(CDG_ERROR, "ERROR: integer expected\n");
	  break;
	}
      }
      /* print current binding */
      else if (strcmp(gii->token, "print") == 0 ||
	       strcmp(gii->token, "p") == 0) {
	
	if(getNextInteger(gii)) {
	  if (gii->intValue >=0 
	      && gii->intValue < vectorSize(net->state->bindings)) {
	    glsPrintBinding(CDG_DEFAULT, NULL, (GlsBinding)
		vectorElement(net->state->bindings, gii->intValue));
	  } else {
	    cdgPrintf(CDG_ERROR, "ERROR: no binding %03d\n", gii->intValue);
	    break;
	  }
	} else {
	  glsPrintState(CDG_DEFAULT, net->state);
	}
      } 
      /* print best  bindings */
      else if (strcmp(gii->token, "printbest") == 0 ||
	       strcmp(gii->token, "pb") == 0) {

	if (!net->bestState) {
	  cdgPrintf(CDG_WARNING, "WARNING: no best state so far\n");
	} else {
	  if(getNextInteger(gii)) {
	    if (gii->intValue >= 0 
		&& gii->intValue < vectorSize(net->bestState->bindings)) {
	      glsPrintBinding(CDG_DEFAULT, NULL, (GlsBinding)
		  vectorElement(net->bestState->bindings, gii->intValue));
	    } else {
	      cdgPrintf(CDG_ERROR, "ERROR: no binding %03d\n", gii->intValue);
	      break;
	    }
	  } else {
	    glsPrintState(CDG_DEFAULT, net->bestState);
	  }
	}
      } 
      /* prune */
      else if (strcmp(gii->token, "prune") == 0) {
	if(getNextFloat(gii)) {
	  net->initialUnaryPruning = gii->floatValue;
	  net->prolongUnaryPruning = gii->floatValue;
	  net->terminationUnaryPruning = gii->floatValue;
	  cdgPrintf(CDG_INFO, "INFO: pruning set to %f\n", gii->floatValue);
	  glsPrunePool();
	} else {
	  cdgPrintf(CDG_ERROR, "ERROR: float expected\n");
	  break;
	}
      }
      /* quit */
      else if (strcmp(gii->token, "quit") == 0 ||
		 strcmp(gii->token, "q") == 0) {
	cdgCtrlCTrapped = TRUE;
	loopFlag = FALSE;
      } 
      /* continue */
      else if (strcmp(gii->token, "cont") == 0 ||
	       strcmp(gii->token, "c") == 0) {
	loopFlag = FALSE;
	glsCurrentPool->interrupt = -1;
	cdgCtrlCTrapped = FALSE;
      }
      /* domain */
      else if (strcmp(gii->token, "domain") == 0 ||
	       strcmp(gii->token, "d") == 0) {
	if (getNextInteger(gii)) {
	  if (gii->intValue < 0 || gii->intValue >= vectorSize(net->domains)) {
	    cdgPrintf(CDG_ERROR, "ERROR: invalid domain index %d\n", gii->intValue);
	    break;
	  } else {
	    GlsDomain domain = vectorElement(net->domains, gii->intValue);
	    if (!domain) {
	      cdgPrintf(CDG_INFO, "WARNING: no domain at %d\n", gii->intValue);
	    } else {
	      glsPrintDomain(CDG_DEFAULT, NULL, domain);
	    }
	  }
	}
      } 
      /* value */
      else if (strcmp(gii->token, "value") == 0 ||
	       strcmp(gii->token, "v") == 0) {
	if (getNextInteger(gii)) {
	  if (gii->intValue < 0 || gii->intValue >= vectorSize(net->values)) {
	    cdgPrintf(CDG_ERROR, "ERROR: invalid value index %d\n", gii->intValue);
	    break;
	  } else {
	    GlsValue value = (GlsValue) vectorElement(net->values, gii->intValue);
	    cdgPrintf(CDG_DEFAULT, "  %03d   %c  %7.2f  %7.2f  ",
		      value->index,
		      glsStateOfValue(value),
		      ((LevelValue)value)->score,
		      value->augCosts);
	    glsPrintValue(CDG_DEFAULT, NULL, value);
	    cdgPrintf(CDG_DEFAULT, "\n");
	  }
	}
      } 
      /* penalty guide */
      else if (strcmp(gii->token, "g") == 0 ||
	       strcmp(gii->token, "guide") == 0) {
	if(getNextFloat(gii)) {
	  net->lambda1 = gii->floatValue;
	  cdgPrintf(CDG_INFO, "INFO: guide set to %f\n", gii->floatValue);
	} else {
	  cdgPrintf(CDG_ERROR, "ERROR: float expected\n");
	  break;
	}
      }
      /* reinforcement guide */
      else if (strcmp(gii->token, "r") == 0 ||
	       strcmp(gii->token, "reinf") == 0) {
	if(getNextFloat(gii)) {
	  net->lambda2 = gii->floatValue;
	  cdgPrintf(CDG_INFO, "INFO: reinforcement set to %f\n", gii->floatValue);
	} else {
	  cdgPrintf(CDG_ERROR, "ERROR: float expected\n");
	  break;
	}
      }
      /* normalization */
      else if (strcmp(gii->token, "normal") == 0) {
	if(getNextFloat(gii)) {
	  net->beta = gii->floatValue;
	  cdgPrintf(CDG_INFO, "INFO: normalization factor set to %f\n", gii->floatValue);
	} else {
	  cdgPrintf(CDG_ERROR, "ERROR: float expected\n");
	  break;
	}
      }
      /* jump best */
      else if (strcmp(gii->token, "jump") == 0 ||
	       strcmp(gii->token, "jb") == 0) {
	glsRestoreBestState(net);
	glsEvalState(net->state);
	glsComputeUtility(net);
      }
      /* exit cdg */
      else if (strcmp(gii->token, "exit") == 0 ||
	       strcmp(gii->token, "x") == 0) {

#if 0
	cdgPrintf(CDG_DEFAULT, "exit cdg, are you sure ? (y/n) ");
	if (getchar() == 'y') {
	  exit(0);
	} else {
	  cdgPrintf(CDG_DEFAULT, "... ignored\n");
	}
#else
	exit(0);
#endif
      }
      /* info */
      else if (strcmp(gii->token, "info") == 0 ||
	       strcmp(gii->token, "i") == 0) {
	glsPrintSettings(CDG_INFO);
	glsPrintStatistics(CDG_INFO);
      }
	
      /* unknown command */ 
      else {
	cdgPrintf(CDG_ERROR, 
		  "ERROR: unknown command '%s', type help for more information\n",
		  gii->token);
	break;
      }
    }
  }

  freeGlsInteractionInfo(gii);
}

/* ---------------------------------------------------------------------------
 * glsDeleteIncompatibles:
 *	returns the number of deleted lexemes
 */
int glsDeleteIncompatibles(ByteVector isDeletedLexem, GlsValue value)
{
  if (value == NULL)
    return 0;
  else 
  {
    int i, noDeletedLexemes = 0;
    LexemNode ln;
    LexemGraph lg = glsCurrentPool->constraintNet->lexemgraph;
    LevelValue lv = (LevelValue)value;

    for (i = vectorSize(lg->nodes)-1; i >=0; i--) {
      ln = (LexemNode) vectorElement(lg->nodes, i);

      if (!bvElement(isDeletedLexem, ln->no) &&
	  (lgForbiddenBy(lg, ln, lv->modifiers) ||
	   lgForbiddenBy(lg, ln, lv->modifiees))) {

	bvSetElement(isDeletedLexem, TRUE, ln->no);
	noDeletedLexemes++;
      } 
    }

    return noDeletedLexemes;
  }
}

/* ---------------------------------------------------------------------------
 * glsActivateInvalids
 *	build new repair jobs for bindings which became invalid
 *
 */
void glsActivateInvalids(GlsState state)
{
  GlsBinding binding;
  GlsJobType jobType = GlsNoJob;

  /* loop thru all bindings */
  for (binding = state->firstBinding; binding; binding = binding->next) {

    if (binding->tag & GlsFrozen)
      continue;
    
    /* prolong job */

    /* which kind of a repairment fits best */
    if (!binding->value) {
      jobType = GlsUnaryJob;
    } else {
      if (glsAreDeletedLexemes(state->isDeletedLexem, 
	  ((LevelValue)binding->value)->modifiers)) {
	if (glsAreDeletedLexemes(state->isDeletedLexem, 
	    ((LevelValue)binding->value)->modifiees)) {
	  jobType = GlsUnaryJob;
	} else {
	  jobType = GlsModifierJob;
	}
      } else {
	if (glsAreDeletedLexemes(state->isDeletedLexem, 
	    ((LevelValue)binding->value)->modifiees)) {
	  jobType = GlsModifieeJob;
	}
      }
    }

    /* build a new one */
    if (jobType != GlsNoJob) {
      glsQueueJob(state->interimJobs, jobType, binding);
      jobType = GlsNoJob;
    }
  }
}

/* ---------------------------------------------------------------------------
 * glsPrintDomain
 */
void glsPrintDomain(unsigned long mode, ByteVector isDeletedLexem, GlsDomain domain)
{
  if (hkVerbosity & mode) {

    GlsValue value;
    List l;
    Constraint constraint;

    cdgPrintf(mode,
	      " val  tag   score    costs      penalty  reinforce\n"
	      "-------------------------------------------------------------------------------------------------\n");

    for (value = domain->firstValue; value != NULL; value = value->next) {
      if (!(value->tag & GlsActive))
	continue;

      cdgPrintf(mode, "%04d %c %c %7.2f  %7.2f  ",
		value->index,
		(domain->annoValue == value)?'*':' ',
		glsStateOfValue(value),
		((LevelValue)value)->score,
		value->augCosts);
      glsPrintValue(mode, isDeletedLexem, value);

      cdgPrintf(mode, "  ");
      for (l = ((LevelValue)value)->constraints; l != NULL; l = listNext(l)) {
	constraint = (Constraint) (listElement(l));
      	if (listNext(l) != NULL) 
	  cdgPrintf(mode, "%s, ", constraint->id);
	else 
	  cdgPrintf(mode, "%s", constraint->id);
      }
      cdgPrintf(mode, "\n");
    }
  }
}

/* ---------------------------------------------------------------------------
 * glsPrintLexemes
 */
void glsPrintLexemes(unsigned long mode, ByteVector isDeletedLexem,
		     List lexemes)
{
  Boolean isFirstLexeme = TRUE;
  List l;
  LexemNode ln;

  if (lexemes == NULL) {
    cdgPrintf(mode, "NIL");
    return;
  }
  if (lexemes == NONSPEC) {
    cdgPrintf(mode, "NONSPEC");
    return;
  }

  for (l = lexemes; l != NULL; l = listNext(l)) {
    ln = (LexemNode) (listElement(l));

    if (isDeletedLexem && bvElement(isDeletedLexem, ln->no))
      continue;

    if (isFirstLexeme) {
      isFirstLexeme = FALSE;
      cdgPrintf(mode, "%s", ln->lexem->description);
    } else {
      cdgPrintf(mode, "/%s", ln->lexem->description);
    }
  }
}

/* ---------------------------------------------------------------------------
 * glsPrintValue
 */
void glsPrintValue(unsigned long mode, ByteVector isDeletedLexem, GlsValue value)
{
  if (hkVerbosity & mode) {

    int i, j, k;
    LevelValue lv = (LevelValue)value;

    cdgPrintf(mode, "%7.0f %7.0f %11s: ", 
	value->penalty, value->reinforcement, lv->level->id);

    /* print modifiers */
    glsPrintLexemes(mode, isDeletedLexem, lv->modifiers);

    /* take care to make all --LABEL--> arrows the same length */
    cdgPrintf(mode, "--");

    /* compute length of arrow */
    i = inputCurrentGrammar->maxLabelLength - strlen(lv->label);
    j = i / 2;

    /* print arrow */
    for (k = 0; k < j; k++)
      cdgPrintf(mode, "-");
    cdgPrintf(mode, "%s", lv->label);
    if (i % 2 == 1)
      j++;
    for (k = 0; k < j; k++)
      cdgPrintf(mode, "-");
    cdgPrintf(mode, "-->");

    /* print modifiees */
    glsPrintLexemes(mode, isDeletedLexem, lv->modifiees);
  }
}

/* ---------------------------------------------------------------------------
 * glsTimeToString
 *	convert a time in milliseconds to a human readable string
 */
String glsTimeToString(unsigned long time)
{
  static char result[80], intString[10];
  int days, hours, minutes, seconds, milliSeconds;
  Boolean isFirst;

  if (time == 0) {
    return "0 ms";
  }
  
  /* calculate the units */
  days = time / 86400000;
  time = time - days * 86400000;
  hours = time / 3600000;
  time = time - hours * 3600000;
  minutes = time / 60000;
  time = time - minutes * 60000;
  seconds = time / 1000;
  milliSeconds = time - seconds * 1000;

  isFirst = TRUE;
  result[0] = 0;
  if (days) {
    sprintf(intString, "%s%d d", isFirst?"": " ", days);
    strcat(result, intString);
    isFirst = FALSE;
  }
  if (hours) {
    sprintf(intString, "%s%d h", isFirst?"":" ", hours);
    strcat(result, intString);
    isFirst = FALSE;
  }
  if (minutes) {
    sprintf(intString, "%s%d m", isFirst?"": " ", minutes);
    strcat(result, intString);
    isFirst = FALSE;
  }
  if (seconds) {
    sprintf(intString, "%s%d s", isFirst?"": " ", seconds);
    strcat(result, intString);
    isFirst = FALSE;
  }
  if (milliSeconds) {
    sprintf(intString, "%s%d ms", isFirst?"": " ", milliSeconds);
    strcat(result, intString);
    isFirst = FALSE;
  }

  return result;
}

/* ---------------------------------------------------------------------------
 * glsTagToString
 *
 * TODO: this is not thread save
 */
char *glsTagToString(GlsTag tag)
{
#define MAX_TAG 8
  static char str[MAX_TAG];
  int i = 0;

  memset(str, 0, MAX_TAG);
    
  if (tag & GlsActive)
    str[i++] = 'a';
  if (tag & GlsChanged)
    str[i++] = 'c';
  if (tag & GlsFrozen)
    str[i++] = 'f';
  if (tag & GlsSorted)
    str[i++] = 's';
  if (tag & GlsPenalized)
    str[i++] = 'p';
  if (tag & GlsDeleted)
    str[i++] = 'd';

  return str;
}


/* ---------------------------------------------------------------------------
 * glsStateOfValue
 */
char glsStateOfValue(GlsValue value)
{
  if (value->tag & GlsDeleted)
    return 'd';

  if (value->tag & GlsFrozen)
    return 'f';

  if (!(value->tag & GlsActive))
    return 'u';

  if (!glsIsValidValue(value->domain->net->state, value))
    return 'i';

  return ' ';
}

/* ---------------------------------------------------------------------------
 * glsPrintBinding
 */
void glsPrintBinding(unsigned long mode, ByteVector isDeletedLexem, 
    GlsBinding binding)
{
  if (hkVerbosity & mode) {

#if SECURE_GLSPRINTBINDING 
    if (binding == NULL) {
      cdgPrintf(CDG_ERROR,
		"ERROR: oops, glsPrintBinding called with null binding\n");
      abort();
    }
#endif


    if (binding->value == NULL) {
      cdgPrintf(mode,
		"%03d  ???  %-3s                     %10s: %-10s <not assigned>\n",
		binding->index, 
		glsTagToString(binding->tag),
		binding->domain->constraintNode->level->id,
		binding->domain->constraintNode->gn->arc->word);
    } else {
      cdgPrintf(mode, "%03d %04d %c %-3s %7.2f %7.2f", 
		binding->index,
		binding->value->index,
		(binding->value == binding->domain->annoValue)?'*':' ',
		glsTagToString(binding->tag),
		binding->costs,
		binding->utility);
      glsPrintValue(mode, isDeletedLexem, binding->value);

      cdgPrintf(mode, "\n");
    }
  }
}

/* --------------------------------------------------------------------
 * compares two constraint violations
 *
 * The violations \a cv1 and \a cv2 are sorted according to 
 * - penalty
 * - arrity
 * - ocurrence 
 * - name
 * with this priority.
 * 
 * @param cv1 the first violation
 * @param cv2 the second violation
 * @returns TRUE if cv1 < cv2, FALSE otherwise
 * @see cvCompare, cvCompareNatural
 */
Boolean glsCvCompare(ConstraintViolation cv1, ConstraintViolation cv2)
{
  int n;

  /* compare by penalty */
  if (cv1->penalty < cv2->penalty)
    return TRUE;
  if (cv1->penalty > cv2->penalty)
    return FALSE;

  /* compare by arity */
  if (!cv1->lv2 && cv1->lv2)
    return TRUE;
  if (cv1->lv2 && !cv2->lv2)
    return FALSE;

  /* compare by ocurrence */
  if (cv1->nodeBindingIndex1 < cv2->nodeBindingIndex1)
    return TRUE;
  if (cv1->nodeBindingIndex1 > cv2->nodeBindingIndex1)
    return FALSE;
if (cv1->lv2) {
    if (cv1->nodeBindingIndex2 < cv2->nodeBindingIndex2)
      return TRUE;
    if (cv1->nodeBindingIndex2 > cv2->nodeBindingIndex2)
      return FALSE;
  }

  /* compare names */
  n = strcmp(cv1->constraint->id, cv2->constraint->id);
  if (n < 0)
    return FALSE;
  if (n > 0)
    return TRUE;


  return FALSE;
}

/* ---------------------------------------------------------------------------
 * glsPrintState
 */
void glsPrintState(unsigned long mode,
			GlsState state)
{
  if (hkVerbosity & mode) {

    List l, m;
    GlsValuePair valuePair;
    Number totalUtility;
    Number totalPenalty;
    Number totalReinforcement;
    const String sep = 
      "  --------------------------------------------------------------------------------\n";
    GlsBinding binding;
    GlsNet net = state->net;

    totalUtility = 0.0;
    totalPenalty = 0.0;
    totalReinforcement = 0.0;
    

    /* recompute utility */
    glsComputeUtility(net);

     
    /* show state */
    cdgPrintf(mode, "\n  %s state of net %d:\n\n",
	(state == net->bestState)? "best": "current", net->index);
    cdgPrintf(mode, 
	"%s  nod val   tag   costs    util   penalty   reinforce       bindings\n%s", 
	sep, sep);
    
    /* loop thru all bindings */
    for (binding = state->firstBinding; binding != NULL;
	 binding = binding->next) {

      if (
#if 0
	  TRUE
#else
	  (binding->tag & GlsChanged) ||
	  binding->costs > 0.0 ||
	  !glsIsValidValue(state, binding->value) ||
	  binding->value->penalty > 0.0 ||
	  ((LevelValue)binding->value)->modifiee != NULL || 
	  (((LevelValue)binding->value)->modifiee == NULL && 
	   strlen(((LevelValue)binding->value)->label) > 0)
#endif
	  ) 
      {
	cdgPrintf(mode, "  ");
	glsPrintBinding(mode, state->isDeletedLexem, binding);
      }
      if (binding->value != NULL) {
	totalUtility += binding->utility;
	totalPenalty += binding->value->penalty;
	totalReinforcement += binding->value->reinforcement;
      }
    }
    cdgPrintf(mode, "%s  sum           %7.2f %7.2f %7.2f %7.2f\n%s",
	      sep, state->augCosts, 
	      totalUtility, totalPenalty, totalReinforcement, sep);


    /* 
     * show violations 
     */
    if (state->noSoftViolations + state->noHardViolations > 0) {
      List unaryViolations;
      List allViolations;
      ConstraintViolation cv;

      /* collect unary violations */
      unaryViolations = NULL;
      allViolations = NULL;
      for (binding = state->firstBinding; binding != NULL;
	   binding = binding->next) {

	if (binding->value == NULL || 
	    ((LevelValue)binding->value)->constraints == NULL)
	  continue;

	for (l = ((LevelValue)binding->value)->constraints; l; l = listNext(l)) {
	  cv = cvNew(listElement(l), (LevelValue)binding->value, NULL);
	  unaryViolations = listPrependElement(unaryViolations, cv);
	  cv->nodeBindingIndex1 = binding->index;
	  allViolations = listInsertSorted(allViolations, cv, glsCvCompare);
	}
      }

      /* collect binary violations */
      for (l = state->conflictPairs; l; l = listNext(l)) {
	valuePair = listElement(l);
	for (m = valuePair->violations; m != NULL; m = listNext(m)) {
	  cv = listElement(m);
	  allViolations = listInsertSorted(allViolations, cv, glsCvCompare);
	}
      }

      /* collect non local violations */
      for (l = state->violations; l; l = listNext(l)) {
	cv = listElement(l);
	allViolations = listInsertSorted(allViolations, cv, glsCvCompare);
      }

      /* print out list of sorted violations */
      for (l = allViolations; l; l = listNext(l)) {
	cv = listElement(l);
	if (cv->lv2 == NULL) {
	  cdgPrintf(mode, "  %03d       %4.3e : %s\n",
	      cv->nodeBindingIndex1, cv->penalty, cv->constraint->id);
	} else {
	  cdgPrintf(mode, "  %03d %03d   %4.3e : %s\n",
	      cv->nodeBindingIndex1, cv->nodeBindingIndex2, cv->penalty, cv->constraint->id);
	}
      }
      cdgPrintf(mode, "%s", sep);

      /* free tempory memory */
      listForEachDelete(unaryViolations, cvDelete);
      listDelete(allViolations);
    }

    cdgPrintf(mode, 
	"  score : %d/%d/%4.3e\n\n",
	state->noSoftViolations, state->noHardViolations, state->score);
  }
}

/* ---------------------------------------------------------------------------
 * schedule a repairment.
 *
 * This function prepairs a given binding to be repaired in a special way
 * defined by the GlsJobType parameter. The binding is then added to the 
 * job queue.
 * @param queue the interimJobs- or allJobs queue
 * @param type marks the kind of repairment that is thought to be promissing
 * @param binding the binding where the repairment shall happen
 */
INLINE void glsQueueJob(RingBuffer queue, GlsJobType type, GlsBinding binding)
{
  switch (binding->jobType) {
    case GlsNoJob:
      binding->jobType = type;
      break;
    case GlsUnaryJob:
      return;
    case GlsModifierJob:
      switch (type) {
	case GlsUnaryJob:
	case GlsModifieeJob:
	  binding->jobType = GlsUnaryJob;
	  break;
	case GlsModifierJob:
	  return;
	default: /* never reach */
	  cdgPrintf(CDG_ERROR, "ERROR: programming error in glsQueueJob\n");
	  abort();
      }
      break;
    case GlsModifieeJob:
      switch (type) {
	case GlsUnaryJob:
	case GlsModifierJob:
	  binding->jobType = GlsUnaryJob;
	  break;
	case GlsModifieeJob:
	  return;
	default: /* never reach */
	  cdgPrintf(CDG_ERROR, "ERROR: programming error in glsQueueJob\n");
	  abort();
      }
      break;
    default: /* never reach */
      cdgPrintf(CDG_ERROR, "ERROR: programming error in glsQueueJob\n");
      abort();
  }

#ifdef DEBUG_GLSQUEUEJOB
  if (glsCurrentPool->debugLevel >= GLS_DEBUG3) {
    switch (binding->jobType) {
      case GlsUnaryJob:
	cdgPrintf(CDG_DEBUG, "DEBUG: queued a unary repair job for %03d\n",
	    binding->index);
	break;
      case GlsModifierJob:
	cdgPrintf(CDG_DEBUG, "DEBUG: queued a modifier repair job for %03d\n",
	    binding->index);
	//cdgExecHook(HOOK_GLSINTERACTION, binding->state->net, "");
	break;
      case GlsModifieeJob:
	cdgPrintf(CDG_DEBUG, "DEBUG: queued a modifiee repair job for %03d\n",
	    binding->index);
	//cdgExecHook(HOOK_GLSINTERACTION, binding->state->net, "");
	break;
      default: /* never reach */
	cdgPrintf(CDG_ERROR, "ERROR: programming error in glsQueueJob\n");
	abort();
    }
  }
#endif

  rbAddTopElement(queue, binding);
}

/* ---------------------------------------------------------------------------
 * get the next job from a queue of jobs.
 * @param queue the set of jobs
 * @returns the binding which shall be repaired or NULL if there is none
 */
INLINE GlsBinding glsDequeueJob(RingBuffer queue)
{
  GlsBinding binding = NULL;
  
  if (!rbIsEmpty(queue)) {
    binding = rbRemoveBottomElement(queue);
  }

  return binding;
}

/* ---------------------------------------------------------------------------
 * schedule a job from one queue into another.
 *
 * Doing it this way is faster than calling glsDequeueJob and glsQueueJob
 * @param dstQueue the queue where a job is placed into
 * @paran srcQueue the origin where a job is taken from
 * @returns the binding which shall be repaired or NULL if there is none
 */
INLINE GlsBinding glsRequeueJob(RingBuffer dstQueue, RingBuffer srcQueue)
{
  GlsBinding binding = NULL;
  
  if (!rbIsEmpty(srcQueue)) {
    binding = rbRemoveBottomElement(srcQueue);
    rbAddTopElement(dstQueue, binding);
  }

  return binding;
}

/* ---------------------------------------------------------------------------
 * remove all queued jobs.
 * @param queue the set of jobs which should be emptied
 */
INLINE void glsClearQueue(RingBuffer queue)
{
  GlsBinding binding;
  
  while((binding = glsDequeueJob(queue))) {
    binding->jobType = GlsNoJob;
  }
}

/* ---------------------------------------------------------------------------
 * test wether a job has already been done
 * @param a possible repairment job
 * @returns TRUE if it is worth doing it, otherwise FALSE and we will skip
 * the job
 */
INLINE Boolean glsJobNeedsToBeDone(GlsBinding binding)
{
  /* a unary repair job doesnt need to be done if the corresponding
   * domain is assigned and frozen or already changed
   */
  if (binding->value != NULL) {

    if ((binding->tag & GlsFrozen) || (binding->tag & GlsChanged))
      return FALSE;
  }

  return TRUE;
}

/* ---------------------------------------------------------------------------
 * glsImprove
 *	do all jobs scheduled to improve the actual state
 *	
 *	returns TRUE if this could be approved
 */
Boolean glsImprove(GlsState state)
{
  Boolean hasImproved; /* TRUE if a better state was found */
  Boolean hasImpaired; /* FALSE if the last state was an improvement */
  GlsNet net = state->net;
  GlsBinding binding;

  /* enable lexemchanges */
  bvCopy(state->isDeletedLexem, net->lexemgraph->isDeletedNode);

  /* reset change info of bindings */
  for (binding = state->firstBinding; binding != NULL;
       binding = binding->next) {
    binding->tag &= ~GlsChanged;
  }

  /* 
   * reset improvement tag of the network.
   * if we improve, no penalties will be learned
   */
  net->tag &= ~GlsImproved;
  

  /* do jobs one by one */
  hasImproved = hasImpaired = FALSE;
  while ((binding = glsRequeueJob(state->interimJobs, state->allJobs))) {

    /* break if the net got frozen */
    if (net->tag & GlsFrozen) {
      glsClearQueue(state->allJobs);
      glsClearQueue(state->interimJobs);
      break;
    }

    if (!glsJobNeedsToBeDone(binding)) {
#if 0
      cdgPrintf(CDG_DEBUG, "DEBUG: no need to repair %03d binding\n",
	  binding->index);
#endif
      glsDequeueJob(state->interimJobs);
      continue;
    }

    /* do all jobs in the interim job queue. */
    if(glsTransform(state)) {

      if (!glsIsValidState(state)) 
	continue;

      /* job found a valid path thru the lexemgraph */


      /* evaluate the quality of the new state */
      glsEvalState(state);

      /* check for a global improvment */
      if (!glsCurrentPool->bestNet || 
	  glsCompareStates(state, glsCurrentPool->bestNet->bestState) > 0) {

	/* set the improve flag */
	net->tag |= GlsImproved;
    
	/* register net with the best state so far */
	glsRegisterBestStateSoFar(state);
	glsRegisterBestNetSoFar(net);
	glsRegisterSolution();
	glsPrunePool();

	hasImproved = TRUE;
      } else {
	/* check for a local improvment */
	if (glsCompareStates(state, net->bestState) > 0) {
	  
	  /* set the improve flag */
	  net->tag |= GlsImproved;
	  
	  /* register local best state */
	  glsRegisterBestStateSoFar(state);
	  hasImpaired = FALSE;
	} else  {
	  hasImpaired = TRUE;
	}
      }


    } else {
    /*
     * we were trapped into a blind alley.
     * so we restore the previous best state and go on with other
     * repair jobs 
     */
      copyGlsState(net->state, net->bestState); /* don't use glsRestoreBestState */
      glsEvalState(net->state);
      glsClearQueue(state->allJobs);
      break;
    }
  }

  /* copy back best state if the quality decreased after a new
   * best state has been found
   */
  if (hasImproved) {
    if (hasImpaired) {
      copyGlsState(net->state, net->bestState); 
#if 0
      if (glsCurrentPool->debugLevel >= GLS_DEBUG2) {
	cdgPrintf(CDG_DEBUG, "DEBUG: using best state\n");
      }
#endif
    }
    net->currentDetour = 1;
  } else { 

    /* copy back last best state if a detour expired 
     *
     *       (1) maximum detour length (implemented)
     *       (2) maximum detour costs
     *       (3) minimum detour:
     *           do a mimimum number of transformations before looking at its costs
     *           in order to terminate the detour
     *       (4) detour utility: 
     *           is the current detour worse it (detour utility := 1/(max costs) * 1/length)
     */
    if (net->maxDetour && net->currentDetour >= net->maxDetour) {

#if 0
      if (glsCurrentPool->debugLevel >= GLS_DEBUG6) {
	cdgPrintf(CDG_DEBUG, "DEBUG: detour expired, resetting to best state so far\n");
      }
#endif
      net->currentDetour = 1;

      /* if we own the cup then we just need a simple copy */
      if (glsCurrentPool->bestNet == net) {
	copyGlsState(net->state, net->bestState); 
      } 
      
      /* 
       * but if another net produced the current best state then we must adapt
       * that one to our own needs. this is flagged here and carried out in
       * glsScheduleNet
       */
      else {
	net->tag |= GlsChanged;
      }

    } else {
      net->currentDetour++;
    }
  } 

  /* use lexemes in values that did not change */
  for (binding = state->firstBinding; binding; binding = binding->next) {
    if (binding->tag & GlsChanged)
      continue;
    glsDeleteIncompatibles(state->isDeletedLexem, binding->value);
  }

#if 0
  cdgPrintf(CDG_DEBUG, "DEBUG: hasImproved = %s, hasImpaired = %s\n",
      hasImproved?"true":"false", hasImpaired?"true":"false");
  cdgExecHook(HOOK_GLSINTERACTION, net, "after improve");
#endif

  return hasImproved;
}

/* ---------------------------------------------------------------------------
 * glsRegisterBestNetSoFar:
 *	remember the net with the best state
 */
void glsRegisterBestNetSoFar(GlsNet net)
{

  /* regiser this net to be the best one so far */
  glsCurrentPool->bestNet = net;

  /* 
   * mark all bestStates in other networks to be changed in order to
   * propagate this newly registered best state
   */
  for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
    if (net == glsCurrentPool->bestNet) {
      net->tag &= ~GlsChanged;
    } else {
      net->tag |= GlsChanged;
    }
  }
}

/* ---------------------------------------------------------------------------
 * glsRestoreBestState
 *	copy back the best state so far and reintroduced possibly
 *	deleted values into the domains
 */
void glsRestoreBestState(GlsNet net)
{
  GlsBinding binding;
  GlsValue value;
  LevelValue lv;
  List l;
  GlsState bestState = net->bestState;

  for (binding = bestState->firstBinding; binding;
       binding = binding->next) {
    if (!binding)
      continue;
	  
    value = binding->value;

#if 0
    if (!glsIsValidValue(bestState, value)) 
      continue;
#endif

    if (!(value->tag & GlsDeleted))
      continue;
    
#if 1
    cdgPrintf(CDG_DEBUG, "DEBUG: reinserting deleted value %03d into domain %03d\n",
	value->index, value->domain->index);
#endif
    value->tag &= ~GlsDeleted; /* undeleting */
    value->penalty = 0.0;
    value->augCosts = glsAugCostsOfValue(bestState, value);
    glsInsertValue(value);
    glsRegisterModifiers(value);

    /* undelete modifiers and modifiees */
    lv = (LevelValue)value;
    for (l = lv->modifiers; l; l = listNext(l)) {
      LexemNode ln = listElement(l);
      bvSetElement(bestState->isDeletedLexem, FALSE, ln->no);
    }
    if (spec(lv->modifiee)) {
      for (l = lv->modifiees; l; l = listNext(l)) {
	LexemNode ln = listElement(l);
	bvSetElement(bestState->isDeletedLexem, FALSE, ln->no);
      }
    }
  }

  copyGlsState(net->state, bestState);
}

/* ---------------------------------------------------------------------------
 * glsRegisterBestStateSoFar:
 *	remember the best state localy in a net
 */
void glsRegisterBestStateSoFar(GlsState state)
{
  GlsNet net = state->net;
  GlsState bestState = net->bestState;
  GlsBinding binding;
  
  copyGlsState(bestState, state);

  for (binding = bestState->firstBinding; 
      binding != NULL; binding = binding->next) {

    if (binding->tag & GlsFrozen)
      continue;


    /* learn penalties for adapted better bindings */
    if (net->phase != GlsInit) {
      binding->value->penalty = 0.0; /* TODO: make this a parameter */
      binding->value->reinforcement += 1.0;
      binding->value->augCosts = glsAugCostsOfValue(state, binding->value);
      glsResortValue(binding->value);
    }
    binding->tag &= ~(GlsChanged | GlsPenalized);
  }
}

/* ---------------------------------------------------------------------------
 * glsTransform
 *	do all jobs in the interim jobs queue in order to transform one
 *	dependency tree into another.
 *
 *	return TRUE if this could be achieved
 *	       FALSE if we were trapped into a blind alley
 */
Boolean glsTransform(GlsState state)
{
  Boolean wasFulfilled = TRUE;
  GlsStatRecord currentRecord = state->net->statRecord;
  GlsBinding binding;

  /* 
   * loop until no more jobs exist 
   * or one job could not be wasFulfilled
   */
  while (wasFulfilled && !rbIsEmpty(state->interimJobs)) {

    /* break if the net got frozen */
    if (state->net->tag & GlsFrozen) {
      wasFulfilled = FALSE;
      break;
    }

    /* dequeue job */
    binding = glsDequeueJob(state->interimJobs);

    if (!glsJobNeedsToBeDone(binding))
      continue;

    switch (binding->jobType) {
      case GlsUnaryJob:
	wasFulfilled = glsRepairUnary(binding);
	break;
      case GlsModifierJob:
	wasFulfilled = glsRepairModifier(binding);
	if (!wasFulfilled)
	  wasFulfilled = glsRepairUnary(binding);
	break;
      case GlsModifieeJob:
	wasFulfilled = glsRepairModifiee(binding);
	if (!wasFulfilled)
	  wasFulfilled = glsRepairUnary(binding);
	break;
      default: /* never reach */
	cdgPrintf(CDG_ERROR, "ERROR: programming error in glsTransform\n");
	abort();
    }
    currentRecord->noRepairs++;
    binding->jobType = GlsNoJob;
  }

  /* remove jobs which werent fulfilled */
  if (!wasFulfilled) {
    glsClearQueue(state->interimJobs);
  }

  return wasFulfilled;
}

/* ---------------------------------------------------------------------------
 * glsAssignValue
 *	assign new value and recompute costs and augmented costs of binding
 *
 *	returns number of deleted incompatible lexemes 
 */
int glsAssignValue(GlsBinding binding, GlsValue value)
{
#define DEBUG_GLSASSIGN_VALUE
  GlsState state = binding->state;
  int noDeletedLexemes = 0;

  /* check for a real change */
  if (binding->value == value)
    return -1;
  
  /* settle assignment */
  binding->value = value;
  binding->tag |= GlsChanged;
  noDeletedLexemes = glsDeleteIncompatibles(state->isDeletedLexem, value);
  vectorSetElement(state->context, value, value->domain->index);

#ifdef DEBUG_GLSASSIGN_VALUE
  if (glsCurrentPool->debugLevel >= GLS_DEBUG4) {
    cdgPrintf(CDG_DEBUG, "%s: %03d %03d", 
	(binding->tag & GlsFrozen)?"FREEZE":"REPAIR",
	binding->index, value->index);
    glsPrintValue(CDG_DEBUG, state->isDeletedLexem, value);
    cdgPrintf(CDG_DEBUG, "\n");
  }
#endif

  return noDeletedLexemes;
}

/* ---------------------------------------------------------------------------
 * exchange a modifier
 *
 * This function tries to repair a binding by searching another one which
 * allows a different reading in modifier position keeping the same modifiee.
 * A label change is taken into account.
 * 
 * @param binding point of repairment
 * @returns TRUE if a value has be bound, FALSE if we might need a more invasive
 * repairment elsewhere.
 */
Boolean glsRepairModifier(GlsBinding binding)
{
  Boolean found = FALSE;
  GlsValue value;
  LexemNode oldModifiee, newModifiee; /* must stay */
  LexemNode oldModifier, newModifier; /* must change */
  GlsState state = binding->state;

  if (binding == NULL || binding->value == NULL) {
    return FALSE;
  }

  oldModifiee = ((LevelValue)binding->value)->modifiee;
  oldModifier = ((LevelValue)binding->value)->modifier;

  /* search next valid value */
  for (value = binding->domain->firstValue; value != NULL; value = value->next) {
    newModifiee = ((LevelValue)value)->modifiee;
    newModifier = ((LevelValue)value)->modifier;

    /* check root and nonspec modifiees */
    if ((int)oldModifiee <= 0) {
      if ((int)newModifiee <= 0) {
	if (oldModifiee == newModifiee &&
	    oldModifier != newModifier &&
	    glsIsValidValue(state, value)) {
	  found = TRUE;
	  break;
	}
      }
      continue;
    } 
    if ((int)newModifiee <= 0)
      continue;

    /* check specific values */
    if (oldModifiee == newModifiee && 
	oldModifier != newModifier &&
	glsIsValidValue(state, value)) {
      found = TRUE;
      break;
    }
  }

  /* bind new value if found */
  if (found) {
#ifdef DEBUG_GLSREPAIRMODIFIER
    cdgPrintf(CDG_DEBUG, "DEBUG: repairing modifier by replacing\n");
    glsPrintValue(CDG_DEBUG, state->isDeletedLexem, binding->value);
    cdgPrintf(CDG_DEBUG, "\nwith\n");
    glsPrintValue(CDG_DEBUG, state->isDeletedLexem, value);
    cdgPrintf(CDG_DEBUG, "\n");
#endif
    if(glsAssignValue(binding, value) >= 0) {
      glsActivateInvalids(state);
    }
  } 
#ifdef DEBUG_GLSREPAIRMODIFIER
  else {
    cdgPrintf(CDG_DEBUG, "DEBUG: no modifier repairment available for binding %03d\n",
	binding->index);
    //cdgExecHook(HOOK_GLSINTERACTION, state->net, "");
  }
#endif

  return found;
}

/* ---------------------------------------------------------------------------
 * exchange a modifiee
 *
 * This function tries to repair a binding by searching another one which
 * allows a different reading in modifiee position keeping the same modifier.
 * A label change is taken into account.
 * 
 * @param binding point of repairment
 * @returns TRUE if a value has be bound, FALSE if we might need a more invasive
 * repairment elsewhere.
 */
Boolean glsRepairModifiee(GlsBinding binding)
{
  Boolean found = FALSE;
  GlsValue value;
  LexemNode oldModifiee, newModifiee; /* must change */
  LexemNode oldModifier, newModifier; /* must stay */
  GlsState state = binding->state;

  if (binding == NULL || binding->value == NULL) {
    return FALSE;
  }

  oldModifiee = ((LevelValue)binding->value)->modifiee;
  oldModifier = ((LevelValue)binding->value)->modifier;

  /* search next valid value */
  for (value = binding->domain->firstValue; value != NULL; value = value->next) {
    newModifiee = ((LevelValue)value)->modifiee;
    newModifier = ((LevelValue)value)->modifier;

    if (oldModifier != newModifier)
      continue;

    /* check root and nonspec modifiees */
    if ((int)oldModifiee <= 0) {
      if ((int)newModifiee <= 0) {
	if (oldModifiee == newModifiee)
	  continue;
      }
      found = TRUE;
      break;
    } 

    /* check specific values */
    if (oldModifiee != newModifiee && 
	glsIsValidValue(state, value)) {
      found = TRUE;
      break;
    }
  }

  /* bind new value if found */
  if (found) {
#ifdef DEBUG_GLSREPAIRMODIFIER
    cdgPrintf(CDG_DEBUG, "DEBUG: repairing modifiee by replacing\n");
    glsPrintValue(CDG_DEBUG, state->isDeletedLexem, binding->value);
    cdgPrintf(CDG_DEBUG, "\twith\n");
    glsPrintValue(CDG_DEBUG, state->isDeletedLexem, value);
    cdgPrintf(CDG_DEBUG, "\n");
#endif
    if(glsAssignValue(binding, value) >= 0) {
      glsActivateInvalids(state);
    }
  } 
#ifdef DEBUG_GLSREPAIRMODIFIER
  else {
    cdgPrintf(CDG_DEBUG, "DEBUG: no modifiee repairment available for\n");
    glsPrintValue(CDG_DEBUG, state->isDeletedLexem, binding->value);
    cdgPrintf(CDG_DEBUG, "\n");
    //cdgExecHook(HOOK_GLSINTERACTION, state->net, "");
  }
#endif

  return found;
}

/* ---------------------------------------------------------------------------
 * glsRepairUnary
 *	try to repair one or all items of a single binding. the distinction
 *	is coded in the activation of the binding.
 *
 *	returns TRUE if everything could be bound
 *		FALSE if we found a blind alley in the lexemgraph
 */
Boolean glsRepairUnary(GlsBinding binding)
{
  GlsState state = binding->state;
  GlsNet net = state->net;
  GlsValue value;
  Boolean found = FALSE;
  GlsStatRecord currentRecord = net->statRecord;
  GlsBinding otherBinding;
  LexemGraph lg = net->lexemgraph;

  /* search next valid value */
  found = FALSE;
  for (value = binding->domain->firstValue; value != NULL; value = value->next) {
    if (glsIsValidValue(state, value)) {
      found = TRUE;
      break;
    }
  }

  /* check wether we ran into a blind alley */
  if (found) {
    /* bind new value */
    if(glsAssignValue(binding, value) >= 0) {
      glsActivateInvalids(state);
    }
  } else {
    /* blind alley treatment */
    int noPenalties = 0;
    currentRecord->noBlindAlleys++;

    /* choose the best we have here */
    value = binding->domain->firstValue;
    if(glsAssignValue(binding, value) >= 0) {
      glsActivateInvalids(state);
    }

#ifdef DEBUG_GLSREPAIRUNARY
    if (glsCurrentPool->debugLevel >= GLS_DEBUG3) {
      cdgPrintf(CDG_DEBUG, "DEBUG: blind alley, choosing %03d %03d ",
	  binding->index, value->index);
      glsPrintValue(CDG_DEBUG, NULL, value);
      cdgPrintf(CDG_DEBUG, "\n");
    }
#endif

    /* penalize all other valid bindings which conflict with firsValue */
    for (otherBinding = state->firstBinding; otherBinding != NULL;
	otherBinding = otherBinding->next) {

      /* skip the actual binding */
      if (otherBinding == binding)
	continue;

      /* this other binding is invalid too */
      if (otherBinding->value == NULL)
	continue;

      /* skip compatible values */
      if (lvCompatible(lg, (LevelValue)value, (LevelValue)otherBinding->value))
	continue;
      
      /* 
       * penalize binding to a value which might be
       * responsible for the blind alley
       */
      glsPenalizeBinding(otherBinding, 1.0);
      noPenalties++;
    }

    if (noPenalties == 0) {
      /* cdgPrintf(CDG_WARNING, "WARNING: nothing penalized in blind alley, penalizing this\n");*/
      glsPenalizeBinding(binding, 1.0);
    }

#if 0
      cdgExecHook(HOOK_GLSINTERACTION, net, "blind alley action");
#endif

    /* check devaluation */
    if (net->tag & GlsPenalized) {
      glsOptimizeNet(net);
      net->tag &= ~GlsPenalized;
    }

    /* 
     * we only need to break lexem deletion of the chosen value
     * if we want to go on with this value
     *
     * if we what to use this value we have to reenable modifiers
     * and modifiees
     */

#if 0
    
    for (l = ((LevelValue)value)->modifiers; l != NULL; l = listNext(l)) {
      ln = (LexemNode)(listElement(l));
      if (bvElement(state->isDeletedLexem, ln->no)) {
        bvSetElement(state->isDeletedLexem, FALSE, ln->no);
      }
    }
    for (l = ((LevelValue)value)->modifiees; l != NULL; l = listNext(l)) {
      ln = (LexemNode)(listElement(l));
      if (bvElement(state->isDeletedLexem, ln->no)) {
	bvSetElement(state->isDeletedLexem, FALSE, ln->no);
      }
    }

#else


    return FALSE;
#endif
  }

  return TRUE;
}

/* ---------------------------------------------------------------------------
 * compute the costs of a search state.
 *
 * @param state the search state which we evaluate
 */
void glsEvalState(GlsState state)
{
//#define DEBUG_GLSEVALSTATE
  GlsBinding binding1, binding2;
  GlsValue value1, value2;
  GlsValuePair valuePair;
  List oldConflicts, remainingConflicts;
  List l;
  Boolean isValid = TRUE;

#ifdef DEBUG_GLSEVALSTATE
  cdgPrintf(CDG_DEBUG, "DEBUG: glsEvalState() called\n");
#endif

  /* init state costs */
  state->costs = 0.0;
  state->augCosts = 0.0;
  state->score = 1.0;
  state->noSoftViolations = 0;
  state->noHardViolations = 0;
  listForEachDelete(state->violations, cvDelete);
  state->violations = NULL;

  /* reset all binding costs */
  for (binding1 = state->firstBinding; 
       binding1 != NULL; binding1 = binding1->next) {
    binding1->costs = 0.0;
    binding1->augCosts = 0.0;
  }

  /* save conflicts of pairs which didn't change */
  oldConflicts = remainingConflicts = NULL;
  for (l = state->conflictPairs; l != NULL; l = listNext(l)) {
    valuePair = (GlsValuePair) (listElement(l));

    /* get the involved bindings */
    binding1 = (GlsBinding) 
      vectorElement(state->bindings, valuePair->node1Index);
    binding2 = (GlsBinding) 
      vectorElement(state->bindings, valuePair->node2Index);

    if (!(binding1->tag & GlsChanged) &&
	!(binding2->tag & GlsChanged)) {
      remainingConflicts = listPrependElement(remainingConflicts, valuePair);

      /* add unchanged binary costs */
      binding1->costs += valuePair->costs / 2;
      binding2->costs += valuePair->costs / 2;
      binding1->augCosts += valuePair->costs / 2;
      binding2->augCosts += valuePair->costs / 2;
    } else {
      oldConflicts = listPrependElement(oldConflicts, valuePair);
    }
  }

  /* assert remaining conflictPairs */
  listDelete(state->conflictPairs);
  state->conflictPairs = remainingConflicts;

  /* kick old conflicts */
  if (glsCurrentPool->useCache) {
    listDelete(oldConflicts);
  } else {
    listForEachDelete(oldConflicts, freeGlsValuePair);
  }
  
  /* compute costs of changed bindings */
  for (binding1 = state->firstBinding;
       binding1 != NULL; binding1 = binding1->next) {

    value1 = binding1->value;
    if (value1 == NULL) 
      continue;

    /* add unary costs */
    binding1->costs += value1->costs;
    binding1->augCosts += value1->augCosts;

    /* add binary costs */
    for (binding2 = binding1->next; 
	 binding2 != NULL; binding2 = binding2->next) {

      value2 = binding2->value;
      if (value2 == NULL) 
	continue;

      /* skip unchanged pairs */
      if (!(binding1->tag & GlsChanged) && 
	  !(binding2->tag & GlsChanged)) 
	continue;

      valuePair = glsEvalBinary(value1, value2);

      /* if these values conflict ... */
      if (valuePair->costs > 0.0) {

	/* register reason of conflict */
	state->conflictPairs = listPrependElement(state->conflictPairs, valuePair);

	/* distribute costs to both values */
	binding1->costs += valuePair->costs / 2;
       	binding2->costs += valuePair->costs / 2;  
	binding1->augCosts += valuePair->costs / 2;
       	binding2->augCosts += valuePair->costs / 2;  
      } else {
	if (!glsCurrentPool->useCache)
	  freeGlsValuePair(valuePair);
      }
    }
  }

  /* compute non local constraint violations */
  {
#define DISTRIBUTE 1
    static BadnessStruct badnessStruct;
    Badness badness = &badnessStruct;
#if DISTRIBUTE
    List violatedLevels = NULL;
    Number distCosts, totalCosts = 0;
#endif
    Number costs;
    int nrBindings;

    /* trigger thorough evaluation by a dummy badness */
    badness->no = 0;
    badness->hard = 0;
    badness->soft = 1.0;

#ifdef DEBUG_GLSEVALSTATE
    {
      int i;
      for (i = 0; i < vectorSize(state->context); i++) {
	LevelValue lv = vectorElement(state->context, i);
	if (lv == NULL)
	  continue;
	cdgPrintf(CDG_DEBUG, "DEBUG: context %03d : ", i);
	lvPrint(CDG_DEBUG, lv, NULL);
	cdgPrintf(CDG_DEBUG, "\n");
      }
    }
#endif
    evalInContext(state->context, state->context, badness, &(state->violations));
    state->score = badness->soft;
    state->noSoftViolations = badness->no - badness->hard;
    state->noHardViolations = badness->hard;

    /* collect all violated levels and compute costs */
    for (l = state->violations; l; l = listNext(l)) {
      ConstraintViolation cv = listElement(l);

      costs = glsPenaltyToCosts(state->net, cv->penalty);

#if 0
      cdgPrintf(CDG_DEBUG, "DEBUG: non-local violation '%s' at %03d",
	  cv->constraint->id, cv->nodeBindingIndex1);
      if (cv->nodeBindingIndex2 >= 0) {
	cdgPrintf(CDG_DEBUG, " %03d\n", cv->nodeBindingIndex2);
      } else {
	cdgPrintf(CDG_DEBUG, "\n");
      }
#endif

#if 1
      /* extra costs to bindings the violation points to */
      binding1 = vectorElement(state->bindings, cv->nodeBindingIndex1);
      if (cv->nodeBindingIndex2 >= 0) {
	binding2 = vectorElement(state->bindings, cv->nodeBindingIndex2);
	binding1->costs += costs / 2;
	binding1->augCosts += costs / 2;
	binding2->costs += costs / 2;
	binding2->augCosts += costs / 2;
      } else {
	binding1->costs += costs;
	binding1->augCosts += costs;
      }
#endif

#if DISTRIBUTE
      totalCosts += costs;
      violatedLevels = listAddUniqueElement(violatedLevels, cv->lv1->level);
      if (cv->lv2)
	violatedLevels = listAddUniqueElement(violatedLevels, cv->lv2->level);
#endif
    }

    /* count involved bindings */
    nrBindings = 0;
#if DISTRIBUTE
    for (binding1 = state->firstBinding; binding1 != NULL;
     	binding1 = binding1->next) {
      if (binding1->value && 
	  listContains(violatedLevels, ((LevelValue)binding1->value)->level)) {
	nrBindings++;
      }
    }
    distCosts = (totalCosts / nrBindings) * 1; /* TODO: magic factor */ 
#ifdef DEBUG_GLSEVALSTATE
    cdgPrintf(CDG_DEBUG, "DEBUG: distributing non-local costs %g to %d bindings, %g each\n",
	totalCosts, nrBindings, distCosts);
#endif
    
    /* distribute costs to all values of violated levels */
    for (binding1 = state->firstBinding; binding1 != NULL;
     	binding1 = binding1->next) {
      if (binding1->value && 
	  listContains(violatedLevels, ((LevelValue)binding1->value)->level)) {
	binding1->costs += distCosts; 
	binding1->augCosts += distCosts; 
      }
    }

    /* inform search state about non local violations */
    listDelete(violatedLevels);
#endif
  }

  /* collect costs in all bindings */
  for (binding1 = state->firstBinding; binding1 != NULL;
       binding1 = binding1->next) {

    value1 = binding1->value;
    if (value1 == NULL) {
      isValid = FALSE;
      continue;
    }

    state->costs += binding1->costs;
    state->augCosts += binding1->augCosts;
    state->score *= ((LevelValue)value1)->score;
    state->noSoftViolations += value1->noSoftViolations;
  }

  /* collect score in all value pairs */
  for (l = state->conflictPairs; l != NULL; l = listNext(l)) {
    valuePair = (GlsValuePair) (listElement(l));
    state->score *= valuePair->score;
    state->noSoftViolations += valuePair->noSoftViolations;
    state->noHardViolations += valuePair->noHardViolations;
  }

  /* settle */
  if (isValid) {
    /* track maximum augmented costs so far for this net */
    if (state->maxCostsSoFar < state->augCosts) {
      state->maxCostsSoFar = state->augCosts;
    }
  } else {
    cdgPrintf(CDG_DEBUG, "DEBUG: evaluating invalid state\n");
    state->costs = MY_MAXDOUBLE;
    state->augCosts = MY_MAXDOUBLE;
    state->score = 0.0;
  }
#ifdef DEBUG_GLSEVALSTATE
  cdgPrintf(CDG_DEBUG, "DEBUG: glsEvalState() done\n");
#endif
}

/* ---------------------------------------------------------------------------
 * glsCompareStates:
 *
 *      compare the quality of two states. the quality is not computed 
 *      by a defined numerical function. moreover the quality functin is given
 *      by the strict order of visited search states defined by this function
 * 
 *      returns > 0 if state1 is better than state 2
 *              = 0 if state1 and state2 have the same quality
 *              < 0 if state1 is worse than state 2
 *
 */
int glsCompareStates(GlsState state1, GlsState state2)
{

  /*
   * CAUTION: costs comparison can only be used if two states of
   *          the same net are compared; this is due to possible
   *          different utility functions in the nets which lead
   *          to different absolute costs
   */
  
  /* compare states by costs or ... */
  if (glsCurrentPool->compareFunction == GLS_CMP_COSTS
      && state1->net == state2->net) {
      if (state1->costs < state2->costs)
	return 1;
      if (state1->costs > state2->costs)
	return -1;
  } 
  
  /* ... by badness */
  else {
      if(state1->noHardViolations < state2->noHardViolations) 
	return 1;
      if(state1->noHardViolations > state2->noHardViolations) 
	return -1;

      if((state1->score - state2->score) > 1e-10)
	return 1;
      if((state2->score - state1->score) > 1e-10) 
	return -1;

      if(state1->noSoftViolations < state2->noSoftViolations) 
	return 1;
      if(state1->noSoftViolations > state2->noSoftViolations) 
	return -1;
  }

  /* same quality */
  return 0;
}

/* ---------------------------------------------------------------------------
 * glsGetPruning:
 *	get the pruning costs depending on the net phase
 */
Number glsGetPruning(GlsNet net)
{
  switch (net->phase) {
    case GlsInit:
      return net->initialUnaryPruning;
    case GlsProlong:
      return net->prolongUnaryPruning;
    case GlsTerminate:
      return net->terminationUnaryPruning;
    default:  /** never reach */
      cdgPrintf(CDG_ERROR, "ERROR: unknown gls phase\n");
      abort();
  }
}

/* ---------------------------------------------------------------------------
 * glsPruneValue
 *	Removes a given value from its domain depending on the pruning costs.
 *	The Value is not pruned if
 *	- it is still good enuf
 *	- it is already deleted
 *	- less than a certain ammount of values would be left (5 by default)
 *	returns TRUE if the value has been pruned.
 */
Boolean glsPruneValue(GlsValue value)
{
  GlsDomain domain = value->domain;
  GlsNet net = domain->net;
  Number unaryPruning = glsGetPruning(net);

  /* don't prune values if ... */
  if (glsCurrentPool->mode == GlsDynamic ||
      unaryPruning == 0.0 ||
      (unaryPruning - value->augCosts > 1e-10) ||
      value->tag & GlsDeleted ||
      domain->noValues <= 5) { /* TODO: magic value */
    return FALSE;
  } 

#if 0
  if (domain->annoValue == value) {
    cdgPrintf(CDG_DEBUG,
	      "DEBUG: bad luck ... you just deleted the anno value "
	      "%03d in domain %03d with costs %.2f\n", value->index,
	      domain->index, value->augCosts);
  } else {
    cdgPrintf(CDG_DEBUG,
	      "DEBUG: pruning value %03d in domain %03d with costs %.2g\n",
	      value->index, domain->index, value->augCosts);
  }
#endif
  glsRemoveValue(value);
  glsUnRegisterModifiers(value);
  
#if 0
  glsOptimizeNet(net);
#endif

  return TRUE;
}

/* ---------------------------------------------------------------------------
 * glsPruneNet
 *	cut away all values which are worse than the
 *	given score
 *
 * returns the number of pruned values
 */
int glsPruneNet(GlsNet net)
{
  int noPrunedValues = 0;
  GlsValue value, nextValue;

  /* real pruning due to softland experiences */
  if (net->bestState->noHardViolations == 0) {
    GlsDomain domain;
    GlsBinding binding1, binding2;
    GlsState bestState = net->bestState;

    /* loop thru all bindings */
    for (binding1 = net->bestState->firstBinding,
	 binding2 = net->state->firstBinding; binding1 != NULL;
	 binding1 = binding1->next,
	 binding2 = binding2->next) {

      /* skip frozen stuff */
      if (binding1->tag & GlsFrozen)
	continue;

      domain = binding1->domain;

      /* loop thru domain */
      for (value = domain->firstValue; value != NULL; value = nextValue) {
	nextValue = value->next;

	/* find worse part of the domain */
	if (bestState->score - ((LevelValue)value)->score > 1e-10) {

	  /* don't prune values which are in use */
	  if (value == binding1->value || value == binding2->value) 
	    continue;
#if 0
	  cdgPrintf(CDG_DEBUG, "DEBUG: glsPruneNet() - pruning value %03d in domain %03d\n",
	      value->index, domain->index);
#endif

	  noPrunedValues++;
	  glsRemoveValue(value);
	  glsUnRegisterModifiers(value);
	} else {
	  value = value->next;
	}
      }
    }
  }

  /* extra pruning strategy */
  if (glsGetPruning(net) != 0.0) {
    GlsBinding binding;

    /* prune every domain */
    for (binding = net->state->firstBinding; binding; binding = binding->next) {
      if (!binding || binding->tag & GlsFrozen)
	continue;

      /* remove too costy values */
      for (value = binding->domain->firstValue; value; value = nextValue) {
	nextValue = value->next;
	if (glsPruneValue(value))
	  noPrunedValues++;
      }
    }
  }

  /* there might be stuff to optimize */
  if (noPrunedValues) {
    glsOptimizeNet(net);
  }

  return noPrunedValues;
}

/* ---------------------------------------------------------------------------
 * glsRegisterModifiers
 *	increase the reference counters of all used modifier lexemes of 
 *	the given value
 */
void glsRegisterModifiers(GlsValue value)
{
  LexemNode ln;
  List l;
  Vector noModifiersOfLexem = value->domain->net->noModifiersOfLexem;
  int count;

  for (l = ((LevelValue)value)->modifiers; l != NULL; l = listNext(l)) {
    ln = (LexemNode)(listElement(l));
    count = (int)vectorElement(noModifiersOfLexem, ln->no) + 1;
    vectorSetElement(noModifiersOfLexem, (void *)count, ln->no);
  }
}

/* ---------------------------------------------------------------------------
 * glsUnRegisterModifiers
 *	decrease the reference counters of all used modifier lexemes of 
 *	the given value
 */
void glsUnRegisterModifiers(GlsValue value)
{
  int count;
  LexemNode ln;
  List l;
  Vector noModifiersOfLexem = value->domain->net->noModifiersOfLexem;
  ByteVector isDeletedLexem = value->domain->net->lexemgraph->isDeletedNode;

  for (l = ((LevelValue)value)->modifiers; l != NULL; l = listNext(l)) {
    ln = (LexemNode)(listElement(l));
	
    count = (int)vectorElement(noModifiersOfLexem, ln->no);
    if (count > 0) {
      if (count == 1) {
#if 0
	cdgPrintf(CDG_DEBUG, "DEBUG: deleting lexem '%s'\n",
	    ln->lexem->description);
#endif
	bvSetElement(isDeletedLexem, TRUE, ln->no);
      }
      vectorSetElement(noModifiersOfLexem, (void *)(count-1), ln->no);
    } 
#if 0
    if (count < 0) {
      cdgPrintf(CDG_WARNING, "WARNING: refcounter of lexem '%s' is negative\n",
	  ln->lexem->description);
    }
#endif
    
  }
}

/* ---------------------------------------------------------------------------
 * glsPrunePool
 *	- cut away all values which are worse than the
 *	  best search state found so far
 *	- reduce domains by a given percentage
 */
void glsPrunePool(void)
{
  GlsNet net;
  GlsBinding binding;

  if (glsCurrentPool->mode == GlsDynamic)
    return;

#if 0
  cdgPrintf(CDG_DEBUG, "DEBUG: pruning pool with score %4.3e\n",
      score);
#endif

  /* prune every net in the pool */
  for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {

    if (net->tag & GlsFrozen || !net->values)
      continue;

    glsPruneNet(net);
  }

  /* check if a net binds a deleted value */
  for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {

    if (net->tag & GlsFrozen || !net->values)
      continue;

    for (binding = net->state->firstBinding;
	 binding != NULL; binding = binding->next) {
      if (binding->value && binding->value->tag & GlsDeleted) {
#if 1
	cdgPrintf(CDG_DEBUG, "DEBUG: binding %03d uses a deleted value %03d\n",
	    binding->index, binding->value->index);

	if (binding->tag & GlsFrozen) {
	  cdgPrintf(CDG_ERROR, "ERROR: altered a frozen binding\n");
	}
#endif
	binding->value = NULL;
      }
    }
  }
}

/* ---------------------------------------------------------------------------
 * glsOptimizeNet
 *
 * returns TRUE if progress has been made
 */
Boolean glsOptimizeNet(GlsNet net) 
{
  int i;
  LexemNode lna;
#if 0
  int oldNoFrozen = net->state->noFrozenBindings;
#endif
  GlsDomain domain;
  Boolean progress = TRUE;
  GlsValue value, nextValue;
  Boolean found;
  GlsBinding binding;
  int j;
  LexemNode lnb;
  LexemGraph lg = net->lexemgraph;
  ByteVector isDeletedLexem;

  if (glsCurrentPool->mode == GlsDynamic)
    return FALSE;
  
  isDeletedLexem = bvNew(bvSize(lg->isDeletedNode));

#if 0
  cdgPrintf(CDG_DEBUG, "DEBUG: optimizing net %d\n", net->index);
#endif

  /* 
   * steps to be performed here:
   *
   * (1) a value can be deleted if all of its modifiers or modifiees are deleted
   * (3) check for an empty domain and possibly freeze the net
   * (4) domains with a single value freeze their binding
   * (5) a lexem can be deleted when it renders a binding unasignable
   * (6) check wether an active state is bound to a recently deleted value (FIXME)
   */

  while(progress) {
    progress = FALSE;

    /* optimize all domains of a net */
    for (binding = net->state->firstBinding; binding; binding = binding->next) {
      if(!binding)
	continue;

      domain = binding->domain;
      
      /* (1) */

      /* loop thru every domain */
      for (value = domain->firstValue; value != NULL; value = nextValue) {
	nextValue = value->next;

	/* remove invalid values from the domain */
	if (glsAreDeletedLexemes(lg->isDeletedNode, ((LevelValue)value)->modifiers) ||
	    (((LevelValue)value)->modifiees != NULL &&
	     glsAreDeletedLexemes(lg->isDeletedNode, ((LevelValue)value)->modifiees))) {
#if 0
	cdgPrintf(CDG_DEBUG, "DEBUG: glsOptimizeNet() - pruning value %03d in domain %03d\n", 
	    value->index, domain->index);
#endif
	  glsRemoveValue(value);
	  glsUnRegisterModifiers(value);
	  progress = TRUE;
	}
      }

      /* (3) */
      if (domain->noValues == 0) {
	cdgPrintf(CDG_WARNING, "WARNING: illegal net %d, freezing\n",
	    net->index);

	/* freeze net */
	glsFreezeNet(net);
	break;
      } 

      /* (4) */
      if (domain->noValues == 1) {
	if (glsFreezeDomain(domain)) /* this incooperates new deleted values */
	  progress = TRUE;
      }
    }

    /* early exit */
    if (net->tag & GlsFrozen)
      break;

    /* (5) */
    for (i = vectorSize(lg->nodes)-1; i>=0; i--) {
      lna = (LexemNode)vectorElement(lg->nodes, i);

      /* skip lexemes which are alreay deleted in every net */
      if (bvElement(lg->isDeletedNode, lna->no))
	continue;

#if 0
	cdgPrintf(CDG_DEBUG, "DEBUG: selecting lexem '%s' used %d times\n",
	    lna->lexem->description, (int)vectorElement(net->noModifiersOfLexem, lna->no));
#endif


      /* get a copy of the original lexem information */
      bvCopy(isDeletedLexem, lg->isDeletedNode);

      /* 
       * simulate selection of lexem lna,
       * erase lexemes incompatible to lna,
       * look if that gives us an empty domain
       */
      for (j = vectorSize(lg->nodes)-1; j >=0; j--) {
	lnb = (LexemNode)vectorElement(lg->nodes, j);
	if (bvElement(isDeletedLexem, lnb->no))
	  continue;

	if (!lgCompatibleNodes(lg, lna, lnb)) {
#if 0
	  cdgPrintf(CDG_DEBUG, "DEBUG: trying to delete lexem '%s'\n",
	      lnb->lexem->description);
	  
#endif
	  bvSetElement(isDeletedLexem, TRUE, lnb->no);
	}
      }

      /* loop thru bindings */
      for (binding = net->state->firstBinding; binding; binding = binding->next) {
	if(!binding)
	  continue;

	domain = binding->domain;

	/* find a valid value */
	found = FALSE;
	for (value = domain->firstValue; value; value = value->next) {
	  if (!glsAreDeletedLexemes(isDeletedLexem, ((LevelValue)value)->modifiers) 
	      &&
	      !glsAreDeletedLexemes(isDeletedLexem, ((LevelValue)value)->modifiees)) {
	    found = TRUE;
	    break;
	  }
	}

	/* 
	 * if we've found a domain without any valid values then we delete 
	 * the tested lexem in the maximum net
	 */
	if (!found) {
#if 0
	  if (glsCurrentPool->debugLevel >= GLS_DEBUG1) {
	    cdgPrintf(CDG_DEBUG, "DEBUG: deleting lexem '%s', no valid value in domain %d\n",
		lna->lexem->description, domain->index);
	  }
#endif

          bvSetElement(lg->isDeletedNode, TRUE, lna->no);

	  progress = TRUE;
	  break; /* we can break if existence is proven */
	}
      }
    }
  }

  /* (6) */
  for (binding = net->state->firstBinding; 
      binding != NULL; binding = binding->next) {

    if (!binding || 
	binding->tag & GlsFrozen || 
	!binding->value)
      continue;

    if (binding->value->tag & GlsDeleted) {
#if 0
	cdgPrintf(CDG_DEBUG, "DEBUG: binding %03d uses a deleted value %d in net %d (current state)\n",
	    binding->index, binding->value->index, net->index);
	glsPrintValue(CDG_DEBUG, NULL, binding->value);
	cdgPrintf(CDG_DEBUG, "\n");
#endif
      binding->value = NULL;
    }
  }

  /* DELETEME : the following is allowed */
#if 0
  for (binding = net->bestState->firstBinding; 
      binding != NULL; binding = binding->next) {

    if (binding->tag & GlsFrozen) {
      continue;
    }

    if (binding->value != NULL &&
	binding->value->tag & GlsDeleted) {
#if 0
	cdgPrintf(CDG_DEBUG, "DEBUG: binding %03d bound to a deleted value %d in net %d (best state)\n",
	    binding->index, binding->value->index, net->index);
	glsPrintValue(CDG_DEBUG, NULL, binding->value);
	cdgPrintf(CDG_DEBUG, "\n");
#endif
      binding->value = NULL;
    }
  }
#endif

  /* talk about deletions */
#if 0
  if (glsCurrentPool->debugLevel >= GLS_DEBUG4) {
    if (net->state->noFrozenBindings > oldNoFrozen) {
      cdgPrintf(CDG_DEBUG, 
	  "DEBUG: %d frozen binding(s), %d left in net %d\n", 
	  net->state->noFrozenBindings,
	  vectorSize(net->state->bindings) - net->state->noFrozenBindings,
	  net->index);
    }
  }
#endif

  bvDelete(isDeletedLexem);
  return progress;
}

/* ---------------------------------------------------------------------------
 * glsOptimizePool
 */
void glsOptimizePool(void)
{
  GlsNet net;

  /* optimize all nets in the pool */
  for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {
    if (!(net->tag & GlsFrozen)) {
      glsOptimizeNet(net);
    }
  }
}

/* ---------------------------------------------------------------------------
 * glsFreezeDomain
 *
 *	freeze the given domain, so that no more binding changes are possible
 */
int glsFreezeDomain(GlsDomain domain)
{
  GlsNet net = domain->net;
  int oldNoValidValues = net->noValidValues;
  GlsBinding binding1 = NULL, binding2 = NULL;
  GlsValue value, nextValue;

  /* freeze binding in the best net first */
  if (net->bestState) {
    binding1 = vectorElement(net->bestState->bindings, domain->index);
    glsFreezeBinding(binding1);
  }

  if (net->state) {
    binding2 = vectorElement(net->state->bindings, domain->index);
    glsFreezeBinding(binding2);
  }

  /* remove all unused values */
  for (value = domain->firstValue; value && domain->noValues > 1; value = nextValue) {
    nextValue = value->next;
    if ((binding1 && value == binding1->value) || 
	(binding2 && value == binding2->value)) {
#if 0
      cdgPrintf(CDG_DEBUG, "DEBUG: keeping value ");
      glsPrintValue(CDG_DEBUG, NULL, value);
      cdgPrintf(CDG_DEBUG, "\n");
#endif
      continue;
    }
#if 0
    cdgPrintf(CDG_DEBUG, "DEBUG: removing value ");
    glsPrintValue(CDG_DEBUG, NULL, value);
    cdgPrintf(CDG_DEBUG, "\n");
#endif
    glsRemoveValue(value);
  }

  return oldNoValidValues - net->noValidValues;
}

/* ---------------------------------------------------------------------------
 * glsComputeUtility
 *	recalculate utility information
 */
void glsComputeUtility(GlsNet net)
{
  GlsBinding binding;
  GlsState state  = net->state;
  GlsState bestState = net->bestState;

  /* reset maximum utility */
  state->maxUtility = -MY_MAXDOUBLE;

  /* recompute utility of all unfrozen bindings */
  for (binding = state->firstBinding;
       binding != NULL; binding = binding->next) {
    
    /* skip unwanted loop items */ 
    if (binding->tag & GlsFrozen)
      continue;
    
    /* compute utility */
    if (binding->value == NULL) {
      binding->utility = MY_MAXDOUBLE; /* highest utility */
    } else {
      /* TRYME: use costs or augCosts */
#if 0
      binding->utility = binding->augCosts / (1.0 + binding->value->penalty);
#else
      binding->utility = binding->costs / (1.0 + binding->value->penalty);
#endif

      /* get maximum utility */
      if (binding->utility > state->maxUtility) {
	state->maxUtility = binding->utility;
      } 
    }
  }

  /* recompute the state utilities */
  {
#if 0
    Number currentTime = (Number)glsCurrentPool->lastRecord->time;
    Number costsFactor= log(net->maxCosts) / net->maxCosts;
    Number timeFactor = log(net->maxTime) / (Number)net->maxTime;

    state->utility = 
      exp(-state->costs * costsFactor - currentTime * timeFactor);
    bestState->utility = 
      exp(-bestState->costs * costsFactor - currentTime * timeFactor);
#endif
#if 0
    state->utility = state->costs / net->noPenalties;
    bestState->utility = bestState->costs / net->noPenalties;
#endif

#if 0
    state->utility = (1.0 - state->score) / currentTime;
    bestState->utility = (1.0 - bestState->score) / currentTime;
#endif

#if 0
    state->utility = state->costs / currentTime;
    bestState->utility = bestState->costs / currentTime;
#endif
  }

  {
    GlsStatRecord solutionRecord = glsGetSolutionRecord();
    GlsStatRecord currentRecord = net->statRecord;
    int effort;

    if (!solutionRecord)
      solutionRecord = net->statRecord->prev;
    if (!solutionRecord)
      return;
    
    effort = currentRecord->cycle - solutionRecord->cycle;
    state->utility = effort / state->costs;
    bestState->utility = effort / bestState->costs;
  }

}

/* ---------------------------------------------------------------------------
 * glsLearn
 *
 *	- learn the sorting function of values
 *	- build repair jobs.
 *	
 */
void glsLearn(GlsState state)
{
  GlsNet net = state->net;
  GlsBinding binding;
  Number cmpUtility;

  glsComputeUtility(net);
  cmpUtility = state->maxUtility * (1.0 - net->toleranceFraction / 100.0);
  
  /* remove previous jobs */
  glsClearQueue(state->allJobs);

  /* clear penalizing information in the bindings */
  for (binding = state->firstBinding; binding != NULL;
       binding = binding->next) {
    binding->tag &= ~GlsPenalized;
  }

  /* 
   * penalize values
   */
  for (binding = state->firstBinding;
       binding != NULL; binding = binding->next) {

    /* skip 
     * - frozen bindings or 
     * - bindings which arent worth it or
     * - already penalized values */
    if ((binding->tag & GlsFrozen) ||
	(binding->tag & GlsPenalized))
      continue;

    if (binding->utility < cmpUtility) {
      continue;
    }

#if 0
    if (glsCurrentPool->debugLevel >= GLS_DEBUG5) {
      cdgPrintf(CDG_DEBUG, "DEBUG: queuing starter job\n");
    }
#endif

    glsQueueJob(state->allJobs, GlsUnaryJob, binding);

#if 0
    /* skip penalization if the current state has improved */
    if (net->tag & GlsImproved)
      continue;
#endif

    glsPenalizeBinding(binding, 1.0);
  }

  /* check devaluation */
  if (net->tag & GlsPenalized) {
    glsOptimizeNet(net);
    net->tag &= ~GlsPenalized;
  }
}

/* ---------------------------------------------------------------------------
 * glsPenalizeBinding
 *	increase penalty of the value used by a given binding,
 *	resort the value in its domain due to its new augmented costs
 */
void glsPenalizeBinding(GlsBinding binding, Number penalty)
{
  GlsValue value = binding->value, nextValue;
  GlsState state = binding->state;
  GlsNet net = state->net;

  if (value == NULL)
    return;

  value->penalty += penalty;
  net->noPenalties += penalty;
  value->augCosts = glsAugCostsOfValue(state, value);
  binding->tag |= GlsPenalized;

#ifdef DEBUG_GLSPENALIZE_BINDING
  if (glsCurrentPool->debugLevel >= GLS_DEBUG2) {
    cdgPrintf(CDG_DEBUG, "PENALIZE: %03d %03d penalty=%g aug costs=%g",
	      binding->index, value->index, value->penalty,
	      value->augCosts);
    glsPrintValue(CDG_DEBUG, NULL, value);
    cdgPrintf(CDG_DEBUG, "\n");
  }
#endif

  /* try to get rid of this value */
  if (!glsPruneValue(value)) {

    /* if not, resort value in its domain */

    nextValue = value->next;
    glsRemoveValue(value);
    glsInsertValueAt(nextValue, value);

  }
}

/* ---------------------------------------------------------------------------
 * glsAugCostsOfValue
 */
INLINE Number glsAugCostsOfValue(GlsState state, GlsValue value)
{
  GlsNet net = state->net;

  return (value->costs 
      + value->penalty * net->lambda1 
      - value->reinforcement * net->lambda2);
}

/* ---------------------------------------------------------------------------
 * store best state so far into constraint net
 * defered parse registration until the end of gls 
 */
void glsRegisterSolution(void)
{
  Parse parse;
  GlsStatRecord thisRecord;
  GlsNet bestNet = glsCurrentPool->bestNet;
  GlsState bestState = bestNet->state;

#if 0
  cdgPrintf(CDG_DEBUG, "DEBUG: registering solution\n");
#endif

  /* defer solution construction til normal mode */
  if (glsCurrentPool->mode == GlsDynamic) {
    return;
  }
  
  if (bestNet == NULL) {
    cdgPrintf(CDG_WARNING, "WARNING: no solution\n");
    return;
  }

  thisRecord = bestNet->statRecord;
  
  /* transform state to parse */
  glsEvalState(bestState);
  parse = glsToParse(bestState);

  if (parse == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: cannot build a parse!\n");
    if (glsCurrentPool->debugLevel >= GLS_DEBUG1)
      cdgExecHook(HOOK_GLSINTERACTION, glsCurrentPool->bestNet, "");
    return;
  } 

#if 0
  /* TRY: freeze bindings with low augmented costs */
  if (bestNet->phase == GlsTerminate) {
    GlsBinding binding;

    for (binding = bestState->firstBinding; binding; binding = binding->next) {
      if (binding->augCosts <= 0.0) {
	glsFreezeDomain(binding->domain);
      }
    }
    if (!glsIsValidState(bestState)) {
      cdgPrintf(CDG_ERROR, "ERROR: best state should not be invalid\n");
      cdgExecHook(HOOK_GLSINTERACTION, glsCurrentPool->bestNet, "");
    } else if (!glsIsValidState(bestNet->state)) {
      cdgPrintf(CDG_ERROR, "ERROR: current state should not be invalid\n");
      cdgExecHook(HOOK_GLSINTERACTION, glsCurrentPool->bestNet, "");
    }
  }
#endif

  /* add link of parse to current statistics record */
  thisRecord->parse = parse;
  bvCopy(thisRecord->isDeletedLexem, bestState->isDeletedLexem);


#if 0
  {
    Vector lvs = parse->LVs;

    parse->LVs = NULL;
    parseComputeStructure(parse, lvs, bestState->isDeletedLexem);
    parseDecorate(parse, glsCurrentPool->constraintNet->lexemgraph, NULL);
    vectorDelete(lvs);
    parseEval(parse);
    if (fabs(parse->score - bestState->score) > 1e-10) {
      parsePrint(parse);
      cdgPrintf(CDG_ERROR, 
	  "\nERROR: wrong score %4.3e in net %d, should be %4.3e, difference %4.3e",
	  bestState->score,
	  bestState->net->index,
	  parse->score,
	  parse->score - bestState->score);
      cdgExecHook(HOOK_GLSINTERACTION, glsCurrentPool->bestNet, "");
    } 
  }
#endif
}

/* ---------------------------------------------------------------------------
 * glsToParse
 * 
 * returns the bindings in parse format
 *	or NULL if the bindings are incomplete or malformed
 */
Parse glsToParse(GlsState state) 
{
  Parse parse;
  GlsNet net = state->net;
  ConstraintNet constraintNet = glsCurrentPool->constraintNet;
  GlsBinding binding;
  
  /* adjust lexem-deletions */
  bvCopy(state->isDeletedLexem, net->lexemgraph->isDeletedNode);

  /* loop thru all bindings and check if they are ok */
  for (binding = state->firstBinding; binding != NULL;
       binding = binding->next) {

    /* security check */
    if (!glsIsValidValue(state, binding->value)) {
      cdgPrintf(CDG_WARNING, "WARNING: invalid assignment %03d no parse for this binding\n", 
	  binding->index);
      return NULL;
    }

    glsDeleteIncompatibles(state->isDeletedLexem, binding->value);
  }

  /* build a parse */  
  parse = parseNew(NULL);
  parse->latticeId = constraintNet->lexemgraph->lattice->id;
  parse->searchStrategy = strRegister("gls");
  parse->comment = NULL;
  parse->solutionNo = 1;
  parse->noOfSolutions = 1;
  parse->score = (state->noHardViolations > 0) ? 0.0: state->score;
  parse->optimal = FALSE; /* we don't know that here */
  parse->LVs = vectorNew(100);
  for (binding = state->firstBinding; binding != NULL;
       binding = binding->next) {
    vectorAddElement(parse->LVs, binding->value);
  }
#if 0
  cdgPrintf(CDG_DEBUG, "DEBUG: constructed parse %s\n", parse->id);
#endif

  /* defere decoration and evaluation rest til the end of gls */

  return parse;
}

/* ---------------------------------------------------------------------------
 * glsFreezeNet
 *	mark a net frozen and inactive
 */
INLINE void glsFreezeNet(GlsNet net) 
{
#if 0
  cdgPrintf(CDG_DEBUG, "DEBUG: freezing net %d\n", net->index);
#endif

  net->tag |= GlsFrozen; /* gcc optimizes this for us */
  net->tag &= ~GlsActive;
  glsCurrentPool->noFrozenNets++;

}

/* ---------------------------------------------------------------------------
 * unfreeze a net and active it.
 * @param net the net we are talking about
 */
INLINE void glsUnfreezeNet(GlsNet net) 
{
#if 0
  cdgPrintf(CDG_DEBUG, "DEBUG: freezing net %d\n", net->index);
#endif
  if (!(net->tag & GlsFrozen))
    return;

  net->tag = GlsDefault;
  glsCurrentPool->noFrozenNets--;
}

/* ---------------------------------------------------------------------------
 * check the phase of a net
 *
 * This function is called in order to mark the phase the search net is
 * currently in. A net 
 * - finishes the init phase when ... (TODO: better heuristics )
 * - finishes the prolongation if we find a first soft solution
 * @param net the considered GlsNet
 */
void glsCheckPhase(GlsNet net)
{
  GlsStatRecord currentRecord = net->statRecord;

  switch(net->phase) {
    case GlsInit: 
      if (net->bestState->noHardViolations == 0) {
	net->phase = GlsTerminate;
	cdgPrintf(CDG_INFO, "INFO: net %d finished initialization, no prolongation needed\n",
	    net->index);
      } else if (currentRecord->prevSolution != NULL) {
	if(currentRecord->maxUtility > 1e-10 
	    && currentRecord->maxUtility == currentRecord->prevSolution->maxUtility) {
	  net->phase = GlsProlong;
	  cdgPrintf(CDG_INFO, "INFO: net %d finished initialization (%4.3e)\n",
	      net->index, currentRecord->maxUtility);
	} 
      } 
#if 0
      else {
	cdgPrintf(CDG_DEBUG, "DEBUG: no prev solution of net %d\n", net->index);
      }
#endif
      break;
    case GlsProlong:
      if (net->bestState->noHardViolations == 0) {
	net->phase = GlsTerminate;
	cdgPrintf(CDG_INFO, "INFO: net %d finished prolongation\n",
	    net->index);
      }
      break;
    case GlsTerminate:
      break;
    default: /* never reach */
      cdgPrintf(CDG_ERROR, "ERROR: programming error in glsCheckPhase\n");
      abort();
  }
}

/* ---------------------------------------------------------------------------
 * glsScheduleNet:
 *	- select the next active net.
 *	- test contracts
 *	- returns the new net or NULL if the gls can be finished.
 */
GlsNet glsScheduleNet(void)
{
  GlsNet net, minNet;
  List minNets, l;
  Boolean found;
#if 0
  GlsNet prevNet = glsCurrentPool->currentNet;
#endif
  unsigned long laps;
  GlsStatRecord currentRecord;

  /* check for maximum quality */
  currentRecord = glsCurrentPool->currentNet->statRecord;
  if (currentRecord->score == 1.0) {
    cdgPrintf(CDG_INFO, "INFO: net %d found maximum quality\n", 
	glsCurrentPool->currentNet->index);
    glsCurrentPool->currentNet = NULL;
    return NULL;
  }
  
  /* pool time expired */
  if (timerExpired) {
    cdgPrintf(CDG_INFO, " \nINFO: time expired\n");
    return NULL; /* game over */
  }

  if (glsCurrentPool->maxTime > 0) {
    laps = glsCurrentPool->lastRecord->time;
  } else {
    laps = 0;
  }

  /* 
   * tag the working set of networks:
   *    - initialy the networks are nonactive and nonfrozen.
   *    - we set all networks to active which operate within the given
   *      contract boundaries.
   *    - if a network exceeds its contract it will be frozen and deactivated
   *    - CAUTION: the utility is computed later, not here
   */
  for (net = glsCurrentPool->firstNet; net != NULL; net = net->next) {

    /* frozen nets never get active again */
    if (net->tag & GlsFrozen)
      continue;


    /* network time expired */
    if (net->maxTime > 0 && net->maxTime < laps) {
      cdgPrintf(CDG_INFO, " \nINFO: time expired for net %d\n", net->index);
      glsFreezeNet(net);
      continue;
    }

    /* maximum cycles used */
    else if (glsCurrentPool->cycle > net->toCycle) {
      cdgPrintf(CDG_INFO, " \nINFO: net %d reached maximum number of %lu cycles\n",
		net->index, net->toCycle);
      glsFreezeNet(net);
      continue;
    }

    /* maximum investment */
    else if (net->maxCosts > 0 && net->state->augCosts > net->maxCosts) {
      cdgPrintf(CDG_INFO, " \nINFO: further usage of net %d too expensive\n",
		net->index);
      glsFreezeNet(net);
      continue;
    }

    /* tag all networks active which match the current cycle */
    if (net->fromCycle <= glsCurrentPool->cycle && net->toCycle > glsCurrentPool->cycle && 
	!(net->tag & GlsActive)) {

	  net->tag |= GlsActive;
#if 0
	  cdgPrintf(CDG_INFO, 
	      "INFO: started net %d with %d values and %d active bindings \n", 
	      net->index,
	      net->noValidValues,
	      vectorSize(net->state->bindings) - net->state->noFrozenBindings);
#endif
    }
  }

  /* if all networks are frozen then game over */
  if (glsCurrentPool->noFrozenNets > glsCurrentPool->lastNet->index) {
    return NULL;
  }

  /* 
   * test end of slice:
   *    - a slice is finished if the underlying net expired
   *    - a slice can have a limited time and limited number of cycles
   *    - if a net expired but the contract of another net doesn't hold jet
   *      we search the earliest contract to come and jump into it.
   */

  if ((glsCurrentPool->currentNet->tag & GlsFrozen) ||
      (glsCurrentPool->sliceCycles >= glsCurrentPool->currentNet->priority)) {
#if 0
    cdgPrintf(CDG_DEBUG, "DEBUG: suspending net %d\n", 
	glsCurrentPool->currentNet->index);
#endif

    /* find a new net */
    found = FALSE;
    minNet = NULL;		       /* example net from minNets */
    minNets = NULL;
    for (net = glsCurrentPool->currentNet->next;; net = net->next) {

      /* turn arround at the end of the list of networks */
      if (net == NULL)
	net = glsCurrentPool->firstNet;

      /* if a full loop was made without finding another net then break */
      if (net == glsCurrentPool->currentNet)
	break;

      /* skip frozen networks */
      if (net->tag & GlsFrozen)
	continue;

      /* another active net was found */
      if (net->tag & GlsActive) {
	found = TRUE;
	break;
      }
      /* register minimal contracts */
      if (minNet != NULL) {
	if (minNet->fromCycle > net->fromCycle) {
	  listDelete(minNets);
	  minNets = listPrependElement(NULL, net);
	  minNet = net;
	} else if (minNet->fromCycle == net->fromCycle) {
	  minNets = listPrependElement(minNets, net);
	}
      } else {
	minNets = listPrependElement(NULL, net);
	minNet = net;
      }
    }

    if (found) {
      glsCurrentPool->currentNet = net;
      glsCurrentPool->sliceCycles = 1;
    } else { 
      /* is the current slice stil usable ... */
      /* ... or must we search the minimal contracts? */
      if (glsCurrentPool->currentNet->tag & GlsFrozen) {
#if 0
      	/* never reach this point because we test the number of frozen nets */
	if (minNet == NULL) {
	  cdgPrintf(CDG_ERROR,
      	      "ERROR: no minimal contract found, but still unfrozen nets available???\n");
  	  abort();
    	}
#endif

      	/* force all minimal contracts to begin now */
	glsCurrentPool->currentNet = minNet;
	glsCurrentPool->sliceCycles = 1;
	for (l = minNets; l != NULL; l = listNext(l)) {
	  minNet = (GlsNet) (listElement(l));
	  minNet->fromCycle = glsCurrentPool->cycle;
	  minNet->tag |= GlsActive;
	  cdgPrintf(CDG_INFO, "INFO: jumpstarting net %d\n", minNet->index);
	}
	listDelete(minNets);
      }
    }
  } else {
    /* the slice is stil usefull */
    glsCurrentPool->sliceCycles++;
  }

  /* process the current net from here on */
  net = glsCurrentPool->currentNet;
  currentRecord = net->statRecord;

  /* check utility: 
   *	- the pool has now set the currentNet.
   *	- if the bestStateSoFar changed then adapt it.
   *	- if the utility dropped beyond the allowed minimum 
   *	  then freeze the net and call glsScheduleNet once again
   */

  if (glsCurrentPool->cooperativeFlag && (net->tag & GlsChanged))
    glsAdaptToBestNet(net);

  if (net->phase != GlsInit && net->maxUtility > 0) {
    if (net->bestState->utility > net->maxUtility) {
      cdgPrintf(CDG_INFO, " \nINFO: utility %4.3e exeeded for net %d\n",
		net->bestState->utility, net->index);

      /* freeze the net */
      glsFreezeNet(net);

      /* schedule some other net, not this one */
      return glsScheduleNet();
    }
  }

  glsCurrentPool->cycle++;

#if 0
  cdgPrintf(CDG_DEBUG, "DEBUG: previous net %d, next net %d\n",
      prevNet->index, glsCurrentPool->currentNet->index);
#endif

  return net;
}

/* ---------------------------------------------------------------------------
 * glsAdaptToBestNet:
 *	set all items of the given state to match the best state so far
 */
void glsAdaptToBestNet(GlsNet targetNet) 
{
  GlsState targetState = targetNet->bestState;
  GlsNet sourceNet = glsCurrentPool->bestNet;
  GlsState sourceState = sourceNet->bestState;
  GlsBinding targetBinding, sourceBinding;
  GlsValue value;
  LexemGraph lg = targetNet->lexemgraph;

  /* unmark GlsChange of this targetNet */
  targetNet->tag &= ~GlsChanged;

  /* 
   * we never need to adapt to ourself, though this point should never
   * be reached, because it wouldn't have been set to GlsChanged; but 
   * in order to minimize code dependencies we do one more check here
   */
  if (sourceNet == targetNet) {
    return;
  }

  /* check if every value we will adapt to is ok here */
  for (sourceBinding = sourceState->firstBinding;
       sourceBinding != NULL; sourceBinding = sourceBinding->next) {
    value = (GlsValue)
      vectorElement(targetNet->values, sourceBinding->value->index);

    /* if not then let it be */
    if (value->tag & GlsDeleted) {
      cdgPrintf(CDG_DEBUG, "DEBUG: can't learn from net %d, value %d is deleted in net %d\n",
	  sourceNet->index, value->index, targetNet->index);
      return;
    }
  }

#if 0
  cdgPrintf(CDG_DEBUG, "DEBUG: net %d learns from net %d\n",
      targetNet->index, sourceNet->index);
#endif

  /* reset the lexeminformation */
  bvCopy(targetState->isDeletedLexem, lg->isDeletedNode);

  /* clear the conflict pairs */
  if (glsCurrentPool->useCache) {
    listDelete(targetState->conflictPairs);
  } else {
    listForEachDelete(targetState->conflictPairs, freeGlsValuePair);
  }
  targetState->conflictPairs = NULL;

  /* adapt bindings */
  for (sourceBinding = sourceState->firstBinding,
       targetBinding = targetState->firstBinding;
       sourceBinding != NULL;
       sourceBinding = sourceBinding->next,
       targetBinding = targetBinding->next) {

    /* things which are frozen here stay untouched */
    if (targetBinding->tag & GlsFrozen)
      continue;
    
    value = (GlsValue)
      vectorElement(targetNet->values, sourceBinding->value->index);
    glsAssignValue(targetBinding, value);

    /* learn penalties for adapted better bindings */
    if (targetNet->phase != GlsInit) {
      targetBinding->value->penalty = 0.0; /* TODO : make this a parameter */
      targetBinding->value->reinforcement += 1.0;
      targetBinding->value->augCosts = 
	glsAugCostsOfValue(targetState, targetBinding->value);
      glsResortValue(targetBinding->value);
    }
  }

  /* continue search at the adapted state */
  glsEvalState(targetState);
}

/* ---------------------------------------------------------------------------
 * glsGetPropeller
 *	returns consecutive characters in order to animate progress
 */
char glsGetPropeller(void)
{
  static char propeller = '-';

  switch(propeller) {
    case '-': propeller = '\\'; break;
    case '\\': propeller = '|'; break;
    case '|': propeller = '/'; break;
    case '/': propeller = '-'; break;
  }

  return propeller;
}

/* ---------------------------------------------------------------------------
 * glsCompleteStatistics
 */
void glsCompleteStatistics(GlsNet net)
{
  GlsStatRecord currentRecord = net->statRecord;
  GlsState state = net->state;
  GlsState bestState = net->bestState;

  currentRecord->isSolution = net->tag & GlsImproved;
  currentRecord->score = state->noHardViolations?0.0:state->score;
  currentRecord->costs = state->costs;
  currentRecord->augCosts = state->augCosts;
  currentRecord->noHardViolations = state->noHardViolations;
  currentRecord->noSoftViolations = state->noSoftViolations;
  currentRecord->time = timerElapsed(glsCurrentPool->profileTimer);
  currentRecord->noPenalties = net->noPenalties;
  currentRecord->utility = bestState->utility;
  if (currentRecord->maxDetour < net->currentDetour)
    currentRecord->maxDetour = net->currentDetour;

  /* count correct edges */
  if (glsCurrentPool->annoParse) {
    int noCorrect = 0;
    int noTotal = 0;
    GlsBinding binding;
    for (binding = state->firstBinding; binding; binding = binding->next) {
      noTotal++;
      if (binding->value == binding->domain->annoValue) {
	noCorrect++;
      }
    }
    currentRecord->errorRate = ((float)noTotal-noCorrect) / ((float)noTotal);
    if (currentRecord->errorRate < currentRecord->minErrorRate) {
      currentRecord->minErrorRate = currentRecord->errorRate;
    }
  } else {
    currentRecord->errorRate = -1.0; /* the unknown error rate */
    currentRecord->minErrorRate = -1.0; /* the unknown error rate */
  }

  /* don't count utilities in init phase */
  if (currentRecord->minUtility > bestState->utility) {
    currentRecord->minUtility = bestState->utility;
  } 

  if (currentRecord->maxUtility < bestState->utility) {
    currentRecord->maxUtility = bestState->utility;
  }

  if (currentRecord->maxCosts < state->maxCostsSoFar) {
    currentRecord->maxCosts = state->maxCostsSoFar;
  }
}

/* ---------------------------------------------------------------------------
 * get the current solution record.
 * @returns the information describing the best parse found so far
 */
INLINE GlsStatRecord glsGetSolutionRecord(void)
{
  GlsStatRecord record = glsCurrentPool->lastRecord;
  
  if (!record)
    return NULL;
  else 
    return (record->isSolution)?record:record->prevSolution;
}

/* ---------------------------------------------------------------------------
 * glsGetSoftRecord:
 *	return the first statistics record without any hard violations
 */
GlsStatRecord glsGetSoftRecord(void)
{
  GlsStatRecord record;

  for (record = glsCurrentPool->firstRecord; record != NULL; record = record->next) {
    if (record->noHardViolations == 0) {
      return record;
    }
  }

  return NULL;
}

/* --------------------------------------------------------------------
 * glsParseFloatParameter:
 *	get float parameter for all nets
 */
int glsParseFloatParameter(int pos, char **args)
{
  GlsNet net, lastNet;
  int argIndex = 0;

  if (args == NULL) 
    return CMDGLS_EXPECTED_FLOAT;

  /* loop thru all int values of this parameter */
  for (net = glsCurrentPool->firstNet; 
      args[argIndex] && args[argIndex][0] != '-';
      argIndex++, net = net->next) {

    /* build a new net if we have more parameters */
    if (net == NULL) 
      net = newGlsNet();
      
    if (!sscanf(args[argIndex], "%lf", (Number *) ((Pointer)net+pos))) 
      return CMDGLS_CANNOT_PARSE_FLOAT;
  }

  /* propagate last parameters to remaining networks */
  lastNet = net;
  for (; net != NULL; net = net->next) {
    *(Number *)((Pointer)net + pos) = 
	*(Number *)((Pointer)lastNet + pos);
  }

  return argIndex;
}

/* --------------------------------------------------------------------
 * glsParseIntParameter:
 *	get integer parameter for all nets
 */
int glsParseIntParameter(int pos, char **args)
{
  GlsNet net, lastNet;
  int argIndex = 0;

  if (args == NULL) 
    return CMDGLS_EXPECTED_INT;

  /* loop thru all int values of this parameter */
  for (net = glsCurrentPool->firstNet; 
      args[argIndex] && args[argIndex][0] != '-';
      argIndex++, net = net->next) {

    /* build a new net if we have more parameters */
    if (net == NULL) 
      net = newGlsNet();
      
#if 0
    cdgPrintf(CDG_DEBUG, "DEBUG: parsing parameter %d='%s'\n", 
	argIndex, args[argIndex]);
#endif
    if (!sscanf(args[argIndex], "%lu", (unsigned long *) ((Pointer)net+pos))) 
      return CMDGLS_CANNOT_PARSE_INT;
  }

  /* propagate last parameters to remaining networks */
  lastNet = net;
  for (; net != NULL; net = net->next) {
    *(unsigned long *)((Pointer)net + pos) = 
	*(unsigned long *)((Pointer)lastNet + pos);
  }

  return argIndex;
}

/* ---------------------------------------------------------------------- 
 * implementation of the gls scripting command.
 * This function has the duty to parse commandline parameters, 
 * build a gls pool and call gls.
 * \param no the number of arguments in \a args
 * \param args an array of gls arguments
 * \returns -1 if the command failed, 1 on success
 */
int glsCommand(int no, char **args)
{
  unsigned long oldHkVerbosity;
  int error = CMDGLS_NO_ERROR;
  int argIndex;
  ConstraintNet constraintnet;
  String errorMsg = NULL;
  GlsNet net, aNet;
  Boolean oldScUseCache;
  FILE *statisticsFile;
#define PARAMOFFSET(parameter) ((Pointer)&(aNet->parameter) - (Pointer)aNet)
#define GLS_HELP_STR \
  "Syntax: gls [<netId>] [<options>]*\n" \
  "      <netId>                           : name of constraint net\n" \
  "      -c, --cutoff [<x>]+               : maximum costs of one hard violation\n"  \
  "      -ca, --cache <b>                  : switch on or off value pair caching\n" \
  "      -co, --costs [<x>]+               : maximum allowed augmented costs\n" \
  "      -coop, --cooperate                : networks learn from each other\n" \
  "      -comp, --compete                  : networks don't learn from each other\n" \
  "      -cmp, --compare <name>            : compare function 'badness' or 'costs'\n" \
  "      -prio, --priority [<n>]+          : cycles spent on a network per timeslice\n" \
  "      -d, --debug <n>                   : debug level\n" \
  "      -dt, --detour [<n>]+              : maximum allowed cycles to spend on worse bindings\n" \
  "      -dyn, --dynamic                   : run in dynamic mode (similar to frobbing dynamic)\n" \
  "      -from, --from-cycle [<n>]+        : start computing at the specified cycle\n" \
  "      -i, --interrupt [<n>]             : interrupt gls after given number of cycles\n" \
  "      -init, --init-mode <init-mode>    : how to build the initial search state\n" \
  "      -iup, --initial-unary-pruning     : pruning costs in init phase\n" \
  "      -g1, -g, --guide1 [<x>]+          : strength of penalization\n" \
  "      -g2, --guide2 [<x>]+              : strength of reinforcement\n" \
  "      -pup, --prolong-unary-pruning     : pruning costs in prolongation phase\n" \
  "      -n, --normalize [<x>]+            : normalization factor\n" \
  "      -s, --statistics <filename>       : generate statistics and print them into a file\n" \
  "      -t, --time [<n>]+                 : maximum milliseconds available to solve the problem\n" \
  "      -tup, --termination-unary-pruning : pruning costs in termination phase\n" \
  "      -to, --to-cycle [<n>]+            : stop computing at the specified cycle\n" \
  "      -tol, --tolerance [<x>]+          : tolerated utility of a binding to be repaired\n" \
  "      -u, --utility [<x>]+              : maximal useful termination\n" \
  "      -up, --unary-pruning [<x>]+       : pruning costs applied in any phase\n" \
  "      -v, --version                     : show version of gls\n" \
  "      -?, -h, --help                    : this text\n" \
  "\n" \
  "      <n>                               : integer\n" \
  "      <x>                               : float\n" \
  "      <b>                               : boolean (0 or 1, true or false, on or off)\n" \
  "      <init-mode>                       : one of the following values to initialize the first state\n" \
  "                                          'normal' - use an initial repair cycle\n"\
  "                                          'root' - use best root bindings\n"\
  "                                          'last' - use a parse attached to the constraintnet\n" \
  "                                          'best' - use an annotation\n" \
  "      [<x>]+                            : iterated parameters; parameter at position i is\n" \
  "                                          assigned to net i; if there are more parameters than nets\n" \
  "                                          new nets are created with default parameters set to\n" \
  "                                          the most previously created net in the pool\n" \
  "Examples:\n" \
  "      gls -g 3                          : guiding strength\n" \
  "      gls -i 1000                       : drop into interactive mode after 1000 iterations\n" \
  "      gls -t 60000 -c 500               : terminate after a minute or after exceeded costs of 500 \n" \
  "      gls -n 0.6 -g 3 3.2               : two search nets \n" \
  "                                          (1) normalized=0.6, guide=3.0\n" \
  "                                          (2) normalized=0.6, guide=3.2\n" \
  "      gls -g 0.7 3 -co 200 1000 \n" \
  "          -from 0 99999                 : two search nets, the second starts when the first finished\n" \
  "                                          exceeding its max costs of 300\n" \
  "\n"
#define GLS_VERS_STR \
  "\nINFO: guided local search\n" \
  "      $Revision: 1.85 $\n" \
  "      Michael Schulz <micha@nats.informatik.uni-hamburg.de>\n" \
  "\n"

  /* security check */
  if (args == NULL) {
    cdgPrintf(CDG_ERROR, "ERROR: argument is NULL\n");
    return -1;
  }

  newGlsPool();
  aNet = newGlsNet();

  /* default values */
  oldHkVerbosity = hkVerbosity;
  cdgCtrlCAllowed = TRUE;
  cdgCtrlCTrapped = FALSE;

  /* collect parameters */
  for (argIndex = 0; argIndex < no; argIndex += error?error:1) {

    error = CMDGLS_NO_ERROR;

    if (args[argIndex] == NULL) {
      error = CMDGLS_NOT_ENUF_PARAMETERS;
    }

    /* debug level */
    else if (strcmp(args[argIndex], "-d") == 0 ||
	strcmp(args[argIndex], "--debug") == 0) {

      if (++argIndex >= no) {
	error = CMDGLS_EXPECTED_INT;
      }

      else if (sscanf(args[argIndex], "%d", &(glsCurrentPool->debugLevel))) {
	if (glsCurrentPool->debugLevel > 0) {
	  hkVerbosity |= CDG_DEBUG;
	} else if (glsCurrentPool->debugLevel == 0) {
	  hkVerbosity &= ~CDG_DEBUG;
	}
      } else {
	error = CMDGLS_CANNOT_PARSE_INT;
      }
    }

    /* caching */
    else if (strcmp(args[argIndex], "-ca") == 0 ||
	     strcmp(args[argIndex], "--cache") == 0) {
      if (++argIndex >= no) {
	error = CMDGLS_EXPECTED_BOOLEAN;
      } else {
	if (strcmp(args[argIndex], "0") == 0 ||
	    strcmp(args[argIndex], "off") == 0 ||
	    strcmp(args[argIndex], "false") == 0) {
	  glsCurrentPool->useCache = FALSE;
	} else if (strcmp(args[argIndex], "1") == 0 ||
	    strcmp(args[argIndex], "on") == 0 ||
	    strcmp(args[argIndex], "true") == 0) {
	  glsCurrentPool->useCache = TRUE;
	} else {
	  error = CMDGLS_EXPECTED_BOOLEAN;
	}
      }
    }

    /* interruption */
    else if (strcmp(args[argIndex], "-i") == 0 ||
	     strcmp(args[argIndex], "--interrupt") == 0) {

      glsCurrentPool->interrupt = 0;
      if (++argIndex < no && 
	  !sscanf(args[argIndex], "%d", &(glsCurrentPool->interrupt))) {
	glsCurrentPool->interrupt = 0;
	argIndex--;
      }
    }

    /* cooperation or competition */
    else if (strcmp(args[argIndex], "-coop") == 0 ||
	     strcmp(args[argIndex], "--cooperate") == 0) {

      glsCurrentPool->cooperativeFlag = TRUE;
    }
    else if (strcmp(args[argIndex], "-comp") == 0 ||
	     strcmp(args[argIndex], "--compete") == 0) {

      glsCurrentPool->cooperativeFlag = FALSE;
    }

    /* priority */
    else if (strcmp(args[argIndex], "-prio") == 0 ||
	strcmp(args[argIndex], "--priority") == 0) {
      error = glsParseIntParameter(PARAMOFFSET(priority), &args[++argIndex]);
    }

    /* cutoff */
    else if (strcmp(args[argIndex], "-c") == 0 ||
	       strcmp(args[argIndex], "--cutoff") == 0) {
      error = glsParseFloatParameter(PARAMOFFSET(cutoff), &args[++argIndex]);
    }

    /* maximum detour */
    else if (strcmp(args[argIndex], "-dt") == 0 ||
	strcmp(args[argIndex], "--detour") == 0) {
      error = glsParseIntParameter(PARAMOFFSET(maxDetour), &args[++argIndex]);
    }

    /* guiding strength of bad penalties */
    else if (strcmp(args[argIndex], "-g") == 0 || /* backwards compatibility */
	strcmp(args[argIndex], "-g1") == 0 ||
	strcmp(args[argIndex], "--guide1") == 0 ||
	strcmp(args[argIndex], "--lambda") == 0 ||
	strcmp(args[argIndex], "--lambda1") == 0 ||
	strcmp(args[argIndex], "--guide") == 0) {
      error = glsParseFloatParameter(PARAMOFFSET(lambda1), &args[++argIndex]);
    }

    /* guiding strength of good penalties */
    else if (strcmp(args[argIndex], "-g2") == 0 ||
	strcmp(args[argIndex], "--lambda2") == 0 ||
	strcmp(args[argIndex], "--reinforce") == 0 ||
	strcmp(args[argIndex], "--guide2") == 0) {
      error = glsParseFloatParameter(PARAMOFFSET(lambda2), &args[++argIndex]);
    }

    /* dynamic mode */
    else if (strcmp(args[argIndex], "-dyn") == 0 ||
	strcmp(args[argIndex], "--dynamic") == 0) {
      glsCurrentPool->mode = GlsDynamic;
    }

    /* normalization factor */
    else if (strcmp(args[argIndex], "-n") == 0 ||
	strcmp(args[argIndex], "--normalize") == 0) {
      error = glsParseFloatParameter(PARAMOFFSET(beta), &args[++argIndex]);
    }

    /* time resources */
    else if (strcmp(args[argIndex], "-t") == 0 ||
	strcmp(args[argIndex], "--time") == 0) {
      error = glsParseIntParameter(PARAMOFFSET(maxTime), &args[++argIndex]);
    }

    /* cycle resources */
    else if (strcmp(args[argIndex], "-from") == 0 ||
	strcmp(args[argIndex], "--from-cylce") == 0) {
      error = glsParseIntParameter(PARAMOFFSET(fromCycle), &args[++argIndex]);
    }
    else if (strcmp(args[argIndex], "-to") == 0 ||
	strcmp(args[argIndex], "--to-cylce") == 0) {
      error = glsParseIntParameter(PARAMOFFSET(toCycle), &args[++argIndex]);
    }

    /* costs resources */
    else if (strcmp(args[argIndex], "-co") == 0 ||
	strcmp(args[argIndex], "--costs") == 0) {
      error = glsParseFloatParameter(PARAMOFFSET(maxCosts), &args[++argIndex]);
    }

    /* tolerance */
    else if (strcmp(args[argIndex], "-tol") == 0 ||
	       strcmp(args[argIndex], "--tolerance") == 0) {
      error = glsParseFloatParameter(PARAMOFFSET(toleranceFraction), &args[++argIndex]);

      for (net = glsCurrentPool->firstNet; net && error >= 0; net = net->next) {
	if (net->toleranceFraction > 100.0) {
	  error = CMDGLS_DEFAULT_ERROR;
	  errorMsg = "ERROR: tolerance fraction should be between 0.0 and 100.0\n";
	}
      }
    }

    /* unary pruning */
    else if (strcmp(args[argIndex], "-up") == 0 ||
	       strcmp(args[argIndex], "--unary-pruning") == 0) {
      error = glsParseFloatParameter(PARAMOFFSET(initialUnaryPruning), &args[++argIndex]);

      for (net = glsCurrentPool->firstNet; net && error >= 0; net = net->next) {
	net->prolongUnaryPruning = net->initialUnaryPruning;
	net->terminationUnaryPruning = net->initialUnaryPruning;
      }
    }

    /* initial unary pruning */
    else if (strcmp(args[argIndex], "-iup") == 0 ||
	       strcmp(args[argIndex], "--initial-unary-pruning") == 0) {
      error = glsParseFloatParameter(PARAMOFFSET(initialUnaryPruning), &args[++argIndex]);
    }

    /* prolong unary pruning */
    else if (strcmp(args[argIndex], "-pup") == 0 ||
	       strcmp(args[argIndex], "--prolong-unary-pruning") == 0) {
      error = glsParseFloatParameter(PARAMOFFSET(prolongUnaryPruning), &args[++argIndex]);
    }

    /* termination unary pruning */
    else if (strcmp(args[argIndex], "-tup") == 0 ||
	       strcmp(args[argIndex], "--termination-unary-pruning") == 0) {
      error = glsParseFloatParameter(PARAMOFFSET(terminationUnaryPruning), &args[++argIndex]);
    }

    /* maximal utility */
    else if (strcmp(args[argIndex], "-u") == 0 ||
	       strcmp(args[argIndex], "--utility") == 0) {
      error = glsParseFloatParameter(PARAMOFFSET(maxUtility), &args[++argIndex]);
    }

    /* compare function */
    else if (strcmp(args[argIndex], "-cmp") == 0 ||
	     strcmp(args[argIndex], "--compare") == 0) {

      if (++argIndex >= no) {
	error = CMDGLS_EXPECTED_CMP;
      } else {
	if (strcmp(args[argIndex], "badness") == 0) {
	  glsCurrentPool->compareFunction = GLS_CMP_BADNESS;
	} else if (strcmp(args[argIndex], "costs") == 0) {
	  glsCurrentPool->compareFunction = GLS_CMP_COSTS;
	} else {
	  error = CMDGLS_UNKNOWN_CMP;
	}
      }
    }

    /* init mode */
    else if (strcmp(args[argIndex], "-init") == 0 ||
	     strcmp(args[argIndex], "--init-mode") == 0) {
      if (++argIndex >= no) {
	error = CMDGLS_EXPECTED_INITMODE;
      } else {
	if (strcmp(args[argIndex], "best") == 0) {
	  glsCurrentPool->initMode = GLS_INIT_BEST;
	} else if (strcmp(args[argIndex], "normal") == 0) {
	  glsCurrentPool->initMode = GLS_INIT_NORMAL;
	} else if (strcmp(args[argIndex], "root") == 0) {
	  glsCurrentPool->initMode = GLS_INIT_ROOT;
	} else if (strcmp(args[argIndex], "last") == 0) {
	  glsCurrentPool->initMode = GLS_INIT_LAST;
	} else {
	  error = CMDGLS_UNKNOWN_INITMODE;
	}

      }
    }

    /* statistics */
    else if (strcmp(args[argIndex], "-s") == 0 ||
	     strcmp(args[argIndex], "--statistics") == 0) {

      if (++argIndex >= no) {
	error = CMDGLS_EXPECTED_FILENAME;
      } else {
	glsCurrentPool->statisticsFileName = strRegister(args[argIndex]);

	statisticsFile = fopen(glsCurrentPool->statisticsFileName, "w");
	if (!statisticsFile) {
	  error = CMDGLS_FILEERROR;
	  fclose(statisticsFile);
	}
	fclose(statisticsFile);
      }
    }

    /* help */
    else if (strcmp(args[argIndex], "-?") == 0 ||
	     strcmp(args[argIndex], "-h") == 0 ||
	     strcmp(args[argIndex], "--help") == 0) {

      error = CMDGLS_HELP;
    } 

    /* verison */
    else if (strcmp(args[argIndex], "-v") == 0 ||
	     strcmp(args[argIndex], "--version") == 0) {

      error = CMDGLS_VERSION;
    }

    /* constraintNet */
    else if (args[argIndex][0] != '-') {
      constraintnet = cnFindNet(args[argIndex]);
      if (!constraintnet) {
	error = CMDGLS_UNKNOWN_NET;
      } else {
	glsCurrentPool->constraintNet = constraintnet;
      }
    }

    /* unknown flag */
    else {
      error = CMDGLS_UNKNOWN_OPTION;
    }

    if (error < 0)
      break;
  }
  
  /* get default constraintNet */
  if (error >= 0 && glsCurrentPool->constraintNet == NULL) {
    if (cnMostRecentlyCreatedNet == NULL) {
      error = CMDGLS_NO_NET;
    } else {
      glsCurrentPool->constraintNet = cnMostRecentlyCreatedNet;
    }
  }


  /* handle error messages */
  if (error < 0) {
    switch (error) {
      case CMDGLS_NO_ERROR:
	break;
      case CMDGLS_NOT_ENUF_PARAMETERS:
	cdgPrintf(CDG_ERROR, "ERROR: not enbuf parameters\n");
	break;
      case CMDGLS_CANNOT_PARSE_FLOAT:
	cdgPrintf(CDG_ERROR, "ERROR: can't parse value '%s', expected float\n", 
	    args[argIndex]);
	break;
      case CMDGLS_CANNOT_PARSE_INT:
	cdgPrintf(CDG_ERROR, "ERROR: can't parse value '%s', expected integer\n", 
	    args[argIndex]);
	break;
      case CMDGLS_EXPECTED_INT:
	cdgPrintf(CDG_ERROR, "ERROR: expected an integer\n");
	break;
      case CMDGLS_EXPECTED_FLOAT:
	cdgPrintf(CDG_ERROR, "ERROR: expected a float\n");
	break;
      case CMDGLS_EXPECTED_BOOLEAN:
	cdgPrintf(CDG_ERROR, "ERROR: expected a boolean (0 or 1, true of false, on or off)\n");
	break;
      case CMDGLS_CANNOT_PARSE_BOOLEAN:
	cdgPrintf(CDG_ERROR, "ERROR: can't parse '%s', expected boolean\n", 
	    args[argIndex]);
	break;
      case CMDGLS_UNKNOWN_NET:
	cdgPrintf(CDG_ERROR, "ERROR: constraint net '%s' unknown\n", 
	    args[argIndex]);
	break;
      case CMDGLS_NO_NET:
	cdgPrintf(CDG_ERROR, "ERROR: no net specified\n");
	break;
      case CMDGLS_UNKNOWN_CMP:
	cdgPrintf(CDG_ERROR, "ERROR: unknown compare function '%s', should be 'badness' or 'costs'\n", 
	    args[argIndex]);
	break;
      case CMDGLS_UNKNOWN_INITMODE:
	cdgPrintf(CDG_ERROR, "ERROR: unknown init mode '%s', should be 'normal', 'best' or 'root'\n", 
	    args[argIndex]);
	break;
      case CMDGLS_EXPECTED_CMP:
	cdgPrintf(CDG_ERROR, "ERROR: expected compare function 'badness' or 'costs'\n", 
	    args[argIndex]);
	break;
      case CMDGLS_UNKNOWN_OPTION:
	cdgPrintf(CDG_ERROR, "ERROR: can't parse option '%s'\n", 
	    args[argIndex]);
	break;
      case CMDGLS_EXPECTED_FILENAME:
	cdgPrintf(CDG_ERROR, "ERROR: expected a filename\n");
	break;
      case CMDGLS_EXPECTED_INITMODE:
	cdgPrintf(CDG_ERROR, "ERROR: expected init mode 'best' or 'root'\n");
	break;
      case CMDGLS_FILEERROR:
	cdgPrintf(CDG_ERROR, "ERROR: can't open file '%s' for writing\n",
	    glsCurrentPool->statisticsFileName);
	break;
      case CMDGLS_UNKNOWN_GLSMODE:
	cdgPrintf(CDG_ERROR, "ERROR: unknown gls mode %s\n", args[argIndex]);
	break;
      case CMDGLS_HELP:
	cdgPrintf(CDG_INFO, GLS_VERS_STR);
	cdgPrintf(CDG_INFO, GLS_HELP_STR);
	break;
      case CMDGLS_VERSION:
	cdgPrintf(CDG_INFO, GLS_VERS_STR);
	break;
      default:
	cdgPrintf(CDG_ERROR, "%s", errorMsg);
	break;
    }
  }

  /* if we had problems parsing the commandline we exit now */
  if (error < 0) {
    freeGlsPool();
    return -1;
  }

  /* we do our own caching */
  oldScUseCache = scUseCache;
  scUseCache = FALSE;

#if 0
  /* always renew a constraint net before glsing */
  cdgPrintf(CDG_INFO, "INFO: renewing net '%s'\n", glsCurrentPool->constraintNet->id);
  if (!cnRenew(glsCurrentPool->constraintNet)) {
    cdgPrintf(CDG_ERROR, "ERROR: cannot renew net\n");
    abort();
  }
#endif

  /* do it */
  if (!gls()) {
    cdgPrintf(CDG_WARNING, "WARNING: no solution found\n");
  }

  /* free the used gls pool */
  freeGlsPool();

  /* restore some settings we changed during gls */
  cdgCtrlCTrapped = FALSE;
  cdgCtrlCAllowed = FALSE;
  hkVerbosity = oldHkVerbosity;
  scUseCache = oldScUseCache;

  return 1;
}

/* ---------------------------------------------------------------------------
 * glsNormal:
 *	operate in normal mode. @see GlsMode
 */
void glsNormal(void)
{
  GlsNet net;
  GlsState state;
  int lastMessage = 0;
  Boolean hasImproved, displayProgress, foundMinErrorRate;
  GlsStatRecord currentRecord = NULL;

  if (glsCurrentPool->currentNet == NULL) {
    cdgPrintf(CDG_WARNING, "WARNING: no initial net found\n");
    return;
  } else if (glsCurrentPool->bestNet == NULL) {
    glsCurrentPool->currentNet = glsCurrentPool->firstNet;
  } else {
    glsCurrentPool->currentNet = glsCurrentPool->bestNet;
  }

#if 0
  cdgExecHook(HOOK_GLSINTERACTION, glsCurrentPool->currentNet, "got init state");
#endif


  for (net = glsCurrentPool->currentNet; net; net = glsScheduleNet()) {

    currentRecord = net->statRecord;
    state = net->state;
    foundMinErrorRate = (currentRecord->prev &&
			 (currentRecord->prev->minErrorRate -
			 currentRecord->minErrorRate) > 1e-10);

    /* prepair a new statistics record */
    if (net->statRecord->isSolution || foundMinErrorRate || glsCurrentPool->statisticsFileName != NULL) {
      currentRecord = newGlsStatRecord(net);
    } else {
      currentRecord->cycle = glsCurrentPool->cycle;
    }

    /* interrupt mainloop */
    if (cdgCtrlCTrapped || glsCurrentPool->interrupt == 0) {
      glsPrintState(CDG_DEBUG, state);
      cdgExecHook(HOOK_GLSINTERACTION, state->net, 
	  "interrupt in mainloop");
    }

    /* break if we still have a ctrl-c */
    if (cdgCtrlCTrapped)
      break;

    /* prepare next searchstep */
    glsLearn(state);

    /* test cylce mode */
    if (glsCurrentPool->interrupt > 0)
      glsCurrentPool->interrupt--;

    /* try to improve the actual state */
    hasImproved = glsImprove(state);

    /* collect the rest of the usefull information */
    glsCompleteStatistics(net);

    /* check the search net phase */
    if (hasImproved)
      glsCheckPhase(net);


    /* show information about the current cycle */
    displayProgress = (hasImproved || lastMessage == 100 || foundMinErrorRate);
    if (displayProgress) {
      cdgPrintf(displayProgress? CDG_INFO:CDG_PROGRESS, 
		"%c %d : %5lu : %5lums : util=%4.3e/%4.3e : score=%2d/%02d/%4.3e : costs=%6.2f/%6.2f/%6.2f",
		hasImproved?'*':(foundMinErrorRate?' ':glsGetPropeller()),
		net->index,
		glsCurrentPool->cycle,
		currentRecord->time,
		currentRecord->utility,
		currentRecord->maxUtility,
		currentRecord->noHardViolations,
		currentRecord->noSoftViolations,
		state->score,
		currentRecord->costs,
		currentRecord->augCosts,
		currentRecord->maxCosts);
      if (glsCurrentPool->annoParse) {
	cdgPrintf(displayProgress? CDG_INFO:CDG_PROGRESS, " : errors=%.2f/%.2f",
		currentRecord->errorRate, currentRecord->minErrorRate);
      }
      cdgPrintf(displayProgress? CDG_INFO:CDG_PROGRESS, "    %c", 
		(hasImproved || foundMinErrorRate)? '\n':'\r');
      lastMessage = 0;
      cdgFlush();
    } else {
      lastMessage++;
    }

  } 
}

/* ---------------------------------------------------------------------------
 * glsDynamic:
 *	subdivide the problem along the interpunctation.
 */
void glsDynamic(void)
{
#define DEBUG_GLSDYNAMIC
  List spans; /** a list of lists of timepoints, that is:
	          - a list of spans
		  - a span is a list of timepoints
		  - a timepoint is an index in the vector timepoints */
  List span;
  List domains;
  List l, ll;
  Vector timepoints; /** vector of lists of domains */
  GlsNet net = glsCurrentPool->firstNet;
  GlsDomain domain;
  String word = NULL;
  int i, n;
  int index = 0;
  GlsBinding binding;

  /* inline functions ------------------------------------------------------- */

  /* ------------------------------------------------------------------------
   * prepare a search state for a span.
   * @param state the current state to be manipulated
   * @param span the span which to adapt the state to
   */
  void prepareState(GlsState state, List span) {
    GlsBinding binding, prevBinding;
    int i, index = 0;

    /* null the complete context */
    for (i = 0; i < vectorSize(state->context); i++) {
      vectorSetElement(state->context, NULL, i);
    }

    /* insert spanned and already frozen bindings */
    prevBinding =  NULL;
    state->firstBinding = NULL;
    for (i = 0; i < vectorSize(state->bindings); i++) {
      binding = vectorElement(state->bindings, i);
      if (!binding)
	continue;
      binding->next = NULL;
      index = binding->domain->constraintNode->gn->arc->from;
      if (listContains(span, (Pointer)index)) {
	if (prevBinding == NULL) {
	  state->firstBinding = binding;
	} else {
	  prevBinding->next = binding;
	}
	prevBinding = binding;
      }
    }

    /* reset statistics */
    state->utility = 0.0;
    state->maxUtility = 100;
    state->maxCostsSoFar = 0.0;
  }

  /* ------------------------------------------------------------------------
   * undo the span preparation.
   * @param state the current search state
   */
  void restoreState(GlsState state)
  {
    GlsBinding binding, prevBinding;
    int i;

    prevBinding = NULL;
    for (i = 0; i < vectorSize(state->bindings); i++) {
      binding = vectorElement(state->bindings, i);
      if (!binding)
	continue;
      binding->next = NULL;
      if (prevBinding == NULL) {
	state->firstBinding = binding;
      } else {
	prevBinding->next = binding;
      }
      prevBinding = binding;
    }

    /* restore statistics */
    state->utility = 0.0;
    state->maxUtility = 100;
    state->maxCostsSoFar = 0.0;
  }

  /* ------------------------------------------------------------------------
   * prepare a network for a span.
   * @param net the current network to be manipulated
   * @param span the span which to adapt the state to
   */
  void prepareNet(GlsNet net, List span) 
  {
    GlsValue value;
    LevelValue lv;
    GlsBinding binding;
    Boolean isValid;

    prepareState(net->state, span);
    prepareState(net->bestState, span);
    net->phase = GlsInit;

    /* prepare all domains */
    for (binding = net->state->firstBinding; binding;
	 binding = binding->next) {
      for (value = binding->domain->firstValue; value; value = value->next) {
	lv = (LevelValue)value;
	if (!listContains(span, (Pointer)(lv->modifier->arc->from)) ||
	    (spec(lv->modifiee) &&
	     !listContains(span, (Pointer)(lv->modifiee->arc->from)))
	    ) {
	  value->tag &= ~GlsActive; /* deactivate for this span */
	} else {
	  value->tag |= GlsActive;
	}
      }
    }

    /* chose the init mode */
    switch (glsCurrentPool->initMode) {
      case GLS_INIT_NORMAL:
	isValid = glsInitState(net->state);
	break;
      case GLS_INIT_ROOT: 
	isValid = glsInitStateRoot(net->state); 
	break;
      case GLS_INIT_BEST:
	isValid = glsInitStateBest(net->state);
	break;
      case GLS_INIT_LAST:
	isValid = glsInitStateLast(net->state);
	break;
      default: /* never reach */
	cdgPrintf(CDG_ERROR, "illegal code reached\n");
	abort();
	break;
    }
  }

  /* ------------------------------------------------------------------------
   * undo the span preparation.
   * @param state the current search state
   * @see prepareState
   */
  void restoreNet(GlsNet net)
  {
    GlsBinding binding;
    GlsValue value;

    restoreState(net->state);
    restoreState(net->bestState);
    net->phase = GlsInit;

    /* acivate all values */
    for (binding = net->state->firstBinding; binding;
	 binding = binding->next) {
      for (value = binding->domain->firstValue; value; value = value->next) {
	value->tag |= GlsActive; 
      }
    }

    glsUnfreezeNet(net);
  }

  /* ------------------------------------------------------------------------
   * prepare the pool for the a span.
   */
  void preparePool(List span)
  {
    GlsStatRecord record, oldRecord;

    /* free all statistic records */
    for (record = glsCurrentPool->firstRecord; record; ) {
      oldRecord = record;
      record = record->next;
      freeGlsStatRecord(oldRecord);
    }
    glsCurrentPool->firstRecord = NULL;
    glsCurrentPool->lastRecord = NULL;

    /* prepare each net */
    for (net = glsCurrentPool->firstNet; net; net = net->next) {
      net->statRecord = NULL;
      newGlsStatRecord(net);
      prepareNet(net, span);
    }
  }

  /* ------------------------------------------------------------------------
   * undo the span preparation
   */
  void restorePool()
  {
    /* - retract the best net 
     * - kick all old conflictPairs 
     * - reactivate possibly frozen stuff 
     * - mark a first net for the next round
     */
    glsCurrentPool->bestNet = NULL;
    for (net = glsCurrentPool->firstNet; net; net = net->next) {
      glsUnfreezeNet(net);
      if (glsCurrentPool->useCache) {
	listDelete(net->state->conflictPairs);
      } else {
	listForEachDelete(net->state->conflictPairs, freeGlsValuePair);
      }
      net->state->conflictPairs = NULL;
    }
    glsCurrentPool->currentNet = glsCurrentPool->firstNet;
  }


  /* ------------------------------------------------------------------------ */
  /* ------------------------------------------------------------------------ */
  
  /* collect timepoints */
  timepoints = vectorNew(listSize(net->lexemgraph->lattice->arcs));
  vectorSetElement(timepoints, NULL, vectorCapacity(timepoints)-1);
  for (i = 0; i < vectorSize(net->domains); i++) {
    domain = vectorElement(net->domains, i);
    if(!domain)
      continue;
    index = domain->constraintNode->gn->arc->from;
    domains = vectorElement(timepoints, index);
    domains = listPrependElement(domains, domain);
    vectorSetElement(timepoints, domains, index);
  }

  /* get spans */
  spans = NULL;
  span = NULL;
  for (i = 0; i < vectorSize(timepoints); i++) {
    domains = vectorElement(timepoints, i);
    
    if (!domains)
      continue;

    domain = listElement(domains); /** get a pivot */
    word = domain->constraintNode->gn->arc->word;

    /* detect clause boundary */
    if (/*listSize(span) >= 6 &&*/
	((strlen(word) == 1 && strspn(word, ",:;-()!?")) || 
	strcmp(word, "oder") == 0 ||
	strcmp(word, "und") == 0)) { 
      /** NOTE: "." (dot) is not included, cus this sucks in something like
	*	"Der 3 . Mrz passt mir schlecht !" 
	*/
      span = NULL;
    }

    if (span == NULL) {
      span = listAppendElement(NULL, (Pointer)i);
      spans = listAppendElement(spans, span);
    } else {
      span = listAppendElement(span, (Pointer)i);
    } 
  }

  /* show me your subdivisions */
  if (listSize(spans) < 2) {
    cdgPrintf(CDG_INFO, "INFO: no subdivisions found ... falling back to normal mode\n");
    glsCurrentPool->mode = GlsNormal;
  } else {

    /* show subdivision */
    cdgPrintf(CDG_INFO, "INFO: using %d subdivisions:\n", listSize(spans));
    for (i = 1, l = spans; l; i++, l = listNext(l)) {
      span = listElement(l);
      cdgPrintf(CDG_INFO, "(%2d) ", i);
      for (ll = span; ll; ll = listNext(ll)) {
	index = (int)listElement(ll);
	domain = listElement(vectorElement(timepoints, index));
	cdgPrintf(CDG_INFO, "%s ", domain->constraintNode->gn->arc->word);
      }
      cdgPrintf(CDG_INFO, "\n");
    }

    /**
     * solve subdivisions: loop over all spans 
     */
    for (i = 1, l = spans; l; i++, l = listNext(l)) {
      span = listElement(l);

      preparePool(span);

      /* be verbose about what happens next */
      cdgPrintf(CDG_INFO, "INFO: starting subdivision (%d)\n", i);

      /** doit */
      n = glsCurrentPool->cycle;
      glsNormal();

      cdgCtrlCTrapped = FALSE;

#if 0
      if (glsCurrentPool->debugLevel > GLS_DEBUG1) {
	cdgExecHook(HOOK_GLSINTERACTION, glsCurrentPool->firstNet, 
	    "INFO: interaction after subdivision");
	glsCurrentPool->interrupt = -1;
      }
#endif
#if 0
      glsPrintState(CDG_DEBUG, glsCurrentPool->bestNet->bestState);
#endif
      restorePool();

    } /* end of spans calculation */

    /* free all statistic records */
    {
      GlsStatRecord record, oldRecord;

      for (record = glsCurrentPool->firstRecord; record; ) {
	oldRecord = record;
	record = record->next;
	freeGlsStatRecord(oldRecord);
      }
      glsCurrentPool->firstRecord = NULL;
      glsCurrentPool->lastRecord = NULL;
    }

    /* restore all search states and domains */
    for (net = glsCurrentPool->firstNet; net; net = net->next) {
      restoreNet(net);
      net->statRecord = NULL;
      newGlsStatRecord(net);
    }

    /*
     * recombination
     * - combine all parial solutions 
     * - handle mismatching lexem usage and find a best valid substitute
     * - activate promising bindings
     * - start a last search
     */

    /* switch to normal mode */
    glsCurrentPool->mode = GlsNormal;

    /* get the best net and the best state */
    for (net = glsCurrentPool->firstNet; net; net = net->next) {
      GlsState state = net->state;
      Boolean isValid;
      glsRestoreBestState(net);

      /* reassign new state */
      isValid = TRUE;
      for (binding = state->firstBinding; binding; binding = binding->next) {
	GlsValue value;
	if (!binding || binding->tag & GlsFrozen)
	  continue;
	value = binding->value;
	binding->value = NULL;
	if (glsIsValidValue(state, value)) {
	  glsAssignValue(binding, value);
	} else {
	  glsQueueJob(state->allJobs, GlsUnaryJob, binding);
	  isValid = FALSE;
	}
      }

      /* carry out one local search */
      glsActivateInvalids(state);
      while(glsRequeueJob(state->allJobs, state->interimJobs));
      if (!rbIsEmpty(state->allJobs)) {
	glsImprove(state);
      } else {
	/* found a solution? */
	glsEvalState(state);
	if (!glsCurrentPool->bestNet || 
	    glsCompareStates(state, glsCurrentPool->bestNet->bestState) > 0) {

	  /* set the improve flag */
	  net->tag |= GlsImproved;
      
	  /* register net with the best state so far */
	  glsRegisterBestStateSoFar(state);
	  glsRegisterBestNetSoFar(net);
	  glsRegisterSolution();
	  glsPrunePool();
	}
      }

      /* freeze unpromising bindings */
      glsComputeUtility(net);
      glsCompleteStatistics(net);
#if 1
      glsPrintState(CDG_DEBUG, net->bestState);
#endif
      for (binding = state->firstBinding; binding; binding = binding->next) {
	if (!binding || binding->tag & GlsFrozen)
	  continue;
	if (binding->costs > (net->cutoff * 0.8) || /* TODO : magic value */
	    (((LevelValue)binding->value)->modifiee == NULL && 
	     binding->costs > (net->cutoff * 0.5))) {  /* TODO : magic value */
#if 0
	  cdgPrintf(CDG_DEBUG, "DEBUG: keeping domain %03d\n",
	      binding->domain->index);
#endif
	} else {
#if 0
	  cdgPrintf(CDG_DEBUG, "DEBUG: freezing binding ");
	  glsPrintBinding(CDG_DEBUG, state->isDeletedLexem, binding);
#endif
	  glsFreezeDomain(binding->domain);
	}
      }
      glsOptimizeNet(net);
    }

#ifdef DEBUG_GLSDYNAMIC
    if (glsCurrentPool->debugLevel > GLS_DEBUG1) {
      cdgExecHook(HOOK_GLSINTERACTION, glsCurrentPool->firstNet, 
	  "INFO: before overall search");
      glsCurrentPool->interrupt = -1;
    }
#endif
  }

  glsNormal();
  
  /* deallocation thingy */
  vectorDelete(timepoints);
  listForEachDelete(spans, listDelete);
}

/* ---------------------------------------------------------------------------
 * gls: 
 *	returns true if solution found, else false
 */
Boolean gls(void)
{
  Boolean success;
  GlsStatRecord solutionRecord;
  Parse parse;

  /* gls does not implement branching lattices */
  if (latticeBranches(glsCurrentPool->constraintNet->lexemgraph->lattice)) {
    cdgPrintf(CDG_ERROR, 
	"ERROR: non trivial wordgraphs cannot be analysed by this method\n");
    return FALSE;
  }

  glsPrintSettings(CDG_INFO);

  /* start the profile timer */
  timerStart(glsCurrentPool->profileTimer);

  /* initialise the pool */
  success = glsInitPool();
  if (!success) {
    cdgPrintf(CDG_WARNING, "WARNING: abording after trying to inilialize\n");
    return FALSE;
  }

  cdgPrintf(CDG_PROFILE, "PROFILE: it took %lums to establish the pool\n",
      glsCurrentPool->lastRecord->time);

  /* call gls mode */
  switch (glsCurrentPool->mode) {
    case GlsNormal:
      glsNormal();
      break;
    case GlsDynamic:
      glsDynamic();
      break;
    default: /* never reach */
      cdgPrintf(CDG_ERROR, "ERROR: programming error\n");
      abort();
  }
  
  /* be verbose about user interrupts */
  if (cdgCtrlCTrapped) {
    cdgPrintf(CDG_INFO, "INFO: user interrupt after %d cycles\n\n",
	      glsCurrentPool->cycle);
  } 

  /* print out result */
  if (hkVerbosity & CDG_SEARCHRESULT && glsCurrentPool->bestNet) {
    glsEvalState(glsCurrentPool->bestNet->bestState);
    glsPrintState(CDG_INFO, glsCurrentPool->bestNet->bestState);
  }

  /* finish the registered parses */
  {
    GlsStatRecord record;
    Vector lvs;

    for (record = glsCurrentPool->firstRecord; record != NULL;
	 record = record->next) {
      parse = record->parse;

      if (!parse)
	continue;

      lvs = parse->LVs;
      parse->LVs = NULL;
      parseComputeStructure(parse, lvs, record->isDeletedLexem);
      parseDecorate(parse, glsCurrentPool->constraintNet->lexemgraph, NULL);
      vectorDelete(lvs);
      parseEval(parse);
      parseRegister(parse);
      cdgExecHook(HOOK_PARTIALRESULT, parse);
    }
  }

  /* register the best parse */
  solutionRecord = glsGetSolutionRecord();
  if (solutionRecord && solutionRecord->parse) {
    parse = solutionRecord->parse;
    parse->optimal = TRUE;
    glsCurrentPool->constraintNet->parses = listPrependElement(NULL, parse);
    parseMostRecentlyCreatedParse = parse;
    cdgPrintf(CDG_INFO, "INFO: best parse found is '%s' with score %4.3e\n",
	parse->id, parse->score);
  }

  glsPrintStatistics(CDG_INFO);
  glsWriteStatistics();

#if 1
  if (glsCurrentPool->debugLevel > 1) {
    cdgExecHook(HOOK_GLSINTERACTION, glsCurrentPool->bestNet, "INFO: end of gls");
  }
#endif

  return TRUE;
}

/* ---------------------------------------------------------------------- 
 * initialize the gls module
 */
void glsInitialize()
{
}

/* ---------------------------------------------------------------------- */
/* -- ENDOFFILE --------------------------------------------------------- */
/** @} */
